Package: glyrepr
Title: Representation for Glycan Compositions and Structures
Version: 0.12.1
Authors@R: 
    person("Bin", "Fu", , "23110220018@m.fudan.edu.cn", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-8567-2997"))
Description: Computational representations of glycan compositions and structures,
    including details such as linkages, anomers, and substituents. Supports varying
    levels of monosaccharide specificity (e.g., "Hex" or "Gal") and ambiguous linkages.
    Provides robust parsing and generation of IUPAC-condensed structure strings.
    Optimized for vectorized operations on glycan structures, with efficient handling
    of duplications. As the cornerstone of the glycoverse ecosystem, this package
    delivers the foundational data structures that power glycomics and glycoproteomics
    analysis workflows.
License: MIT + file LICENSE
Suggests: testthat (>= 3.0.0), patrick, tibble, knitr, rmarkdown,
        tictoc, lobstr
Config/testthat/edition: 3
Encoding: UTF-8
URL: https://glycoverse.github.io/glyrepr/,
        https://github.com/glycoverse/glyrepr
Imports: checkmate, cli, dplyr, furrr, future, glue, igraph, magrittr,
        pillar, purrr, rlang, rstackdeque, stringr, vctrs (>= 0.6.5)
VignetteBuilder: knitr
BugReports: https://github.com/glycoverse/glyrepr/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-19 06:04:01 UTC; fubin
Author: Bin Fu [aut, cre, cph] (ORCID: <https://orcid.org/0000-0001-8567-2997>)
Maintainer: Bin Fu <23110220018@m.fudan.edu.cn>
Repository: CRAN
Date/Publication: 2026-06-19 07:00:02 UTC
Built: R 4.5.2; ; 2026-06-19 10:34:36 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",glyrepr_structure)
S3method("[[<-",glyrepr_structure)
S3method(convert_to_generic,character)
S3method(convert_to_generic,glyrepr_composition)
S3method(convert_to_generic,glyrepr_structure)
S3method(count_mono,glyrepr_composition)
S3method(count_mono,glyrepr_structure)
S3method(format,glyrepr_composition)
S3method(format,glyrepr_structure)
S3method(get_mono_type,character)
S3method(get_mono_type,glyrepr_composition)
S3method(get_mono_type,glyrepr_structure)
S3method(is.na,glyrepr_composition)
S3method(is.na,glyrepr_structure)
S3method(obj_print_data,glyrepr_composition)
S3method(obj_print_data,glyrepr_structure)
S3method(obj_print_footer,glyrepr_structure)
S3method(pillar_shaft,glyrepr_composition)
S3method(pillar_shaft,glyrepr_structure)
S3method(vec_cast,character.glyrepr_composition)
S3method(vec_cast,character.glyrepr_structure)
S3method(vec_cast,glyrepr_composition.character)
S3method(vec_cast,glyrepr_composition.double)
S3method(vec_cast,glyrepr_composition.glyrepr_composition)
S3method(vec_cast,glyrepr_composition.glyrepr_structure)
S3method(vec_cast,glyrepr_composition.integer)
S3method(vec_cast,glyrepr_composition.list)
S3method(vec_cast,glyrepr_composition.logical)
S3method(vec_cast,glyrepr_structure.character)
S3method(vec_cast,glyrepr_structure.glyrepr_structure)
S3method(vec_cast,glyrepr_structure.igraph)
S3method(vec_cast,glyrepr_structure.list)
S3method(vec_cast,logical.glyrepr_composition)
S3method(vec_ptype2,glyrepr_composition.glyrepr_composition)
S3method(vec_ptype2,glyrepr_composition.logical)
S3method(vec_ptype2,glyrepr_structure.glyrepr_structure)
S3method(vec_ptype2,logical.glyrepr_composition)
S3method(vec_ptype_abbr,glyrepr_composition)
S3method(vec_ptype_abbr,glyrepr_structure)
S3method(vec_ptype_full,glyrepr_composition)
S3method(vec_ptype_full,glyrepr_structure)
S3method(vec_restore,glyrepr_composition)
S3method(vec_restore,glyrepr_structure)
export(as_glycan_composition)
export(as_glycan_structure)
export(available_monosaccharides)
export(available_substituents)
export(convert_to_generic)
export(count_mono)
export(fill_anomer_pos)
export(get_anomer)
export(get_anomer_pos)
export(get_mono_type)
export(get_structure_graphs)
export(get_structure_level)
export(glycan_composition)
export(glycan_structure)
export(has_linkages)
export(is_glycan_composition)
export(is_glycan_structure)
export(is_known_monosaccharide)
export(n_glycan_core)
export(o_glycan_core_1)
export(o_glycan_core_2)
export(possible_linkages)
export(reduce_structure_level)
export(remove_linkages)
export(remove_substituents)
export(severy)
export(simap)
export(simap_chr)
export(simap_dbl)
export(simap_int)
export(simap_lgl)
export(simap_structure)
export(simap_vec)
export(smap)
export(smap2)
export(smap2_chr)
export(smap2_dbl)
export(smap2_int)
export(smap2_lgl)
export(smap2_structure)
export(smap2_vec)
export(smap_chr)
export(smap_dbl)
export(smap_int)
export(smap_lgl)
export(smap_structure)
export(smap_unique)
export(smap_vec)
export(snone)
export(spmap)
export(spmap_chr)
export(spmap_dbl)
export(spmap_int)
export(spmap_lgl)
export(spmap_structure)
export(spmap_vec)
export(ssome)
export(structure_to_iupac)
export(valid_linkages)
import(vctrs)
importFrom(dplyr,.data)
importFrom(magrittr,"%>%")
importFrom(pillar,pillar_shaft)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# glyrepr <a href="https://glycoverse.github.io/glyrepr/"><img src="man/figures/logo.png" align="right" height="138" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/glyrepr)](https://CRAN.R-project.org/package=glyrepr)
[![R-universe
version](https://glycoverse.r-universe.dev/glyrepr/badges/version)](https://glycoverse.r-universe.dev/glyrepr)
[![R-CMD-check](https://github.com/glycoverse/glyrepr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/glycoverse/glyrepr/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/glycoverse/glyrepr/graph/badge.svg)](https://app.codecov.io/gh/glycoverse/glyrepr)
<!-- badges: end -->

`glyrepr` provides two important representations of glycans: glycan
compositions and glycan structures. This package is the core of
`glycoverse` ecosystem, as it provides the basic data structures and
functions for representing and manipulating glycans.

In fact, the functions in this packages are heavily used by other
`glycoverse` packages such as
[glyparse](https://github.com/glycoverse/glyparse) and
[glymotif](https://github.com/glycoverse/glymotif), which you are
probably already using or will use.

## Installation

You can install the latest release of glyrepr from
[CRAN](https://CRAN.R-project.org/package=glyrepr):

``` r
install.packages("glyrepr")
```

Or from [r-universe](https://glycoverse.r-universe.dev/glyrepr):

``` r
install.packages('glyrepr', repos = c('https://glycoverse.r-universe.dev', 'https://cloud.r-project.org'))
```

Or from [GitHub](https://github.com/glycoverse/glyrepr):

``` r
remotes::install_github("glycoverse/glyrepr@*release")
```

Or install the development version:

``` r
remotes::install_github("glycoverse/glyrepr")
```

## Documentation

-   🚀 Get started:
    [Here](https://glycoverse.github.io/glyrepr/articles/glyrepr.html)
-   🔍 Underlying graph representation:
    [Here](https://glycoverse.github.io/glyrepr/articles/glycan-graph.html)
-   🔧 Faster structure operations:
    [Here](https://glycoverse.github.io/glyrepr/articles/smap.html)
-   ✏️ IUPAC-condensed glycan text representation:
    [Here](https://glycoverse.github.io/glyrepr/articles/iupac.html)
-   📚 Reference:
    [Here](https://glycoverse.github.io/glyrepr/reference/index.html)

## Role in `glycoverse`

As the cornerstone of the `glycoverse` ecosystem, this package provides
two fundamental data structures for representing glycans:
`glycan_composition()` and `glycan_structure()`. These specialized data
types are what distinguish `glycoverse` from other omics analysis
frameworks, serving as the foundation for higher-level packages like
`glymotif`, which build upon them to perform advanced glycan analysis.

## Example

``` r
library(glyrepr)

# Create glycan compositions
glycan_composition(
  c(Man = 5, GlcNAc = 2),
  c(Man = 3, Gal = 2, GlcNAc = 4, Fuc = 1, Neu5Ac = 2)
)
#> <glycan_composition[2]>
#> [1] Man(5)GlcNAc(2)
#> [2] Man(3)Gal(2)GlcNAc(4)Fuc(1)Neu5Ac(2)

# Parse IUPAC-condensed glycan text representation
# `glyrepr` supports IUPAC-condensed glycan text representation natively.
# For other formats like WURCS or glycoCT, use the `glyparse` package.
# For example, the following two glycan structures are equivalent:
structures <- as_glycan_structure(c("GlcNAc(b1-4)GlcNAc(?1-", "Man(a1-2)GlcNAc(?1-"))

# Get the composition of a glycan structure
as_glycan_composition(structures)
#> <glycan_composition[2]>
#> [1] GlcNAc(2)
#> [2] Man(1)GlcNAc(1)

# Count the number of given residues
count_mono(structures, "Hex")
#> [1] 0 1
count_mono(glycan_composition(c(Man = 3, GlcNAc = 2, Gal = 2)), "Hex")
#> [1] 5
```

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom dplyr .data
## usethis namespace: end
NULL

#' Internal vctrs methods
#'
#' @import vctrs
#' @keywords internal
#' @name glyrepr-vctrs
NULL

#' Reorder Vertices and Edges
#'
#' @description
#' Reorder the vertices and edges of a list of glycan graphs to be in line
#' with the IUPAC-style sequence.
#'
#' @param graphs A list of igraph graph objects.
#' @returns A list of reordered igraph graph objects.
#' @noRd
reorder_graphs <- function(graphs) {
  checkmate::assert_list(graphs, types = "igraph")
  purrr::map(graphs, .reorder_one_graph)
}

#' Reorder graphs and return indices mapping
#'
#' @param graphs A list of igraph graph objects.
#' @returns A list with 'graphs' (reordered graphs) and 'indices' (original indices).
#' @noRd
reorder_graphs_with_indices <- function(graphs) {
  checkmate::assert_list(graphs, types = "igraph")
  n <- length(graphs)
  reordered <- purrr::map(graphs, .reorder_one_graph)
  list(graphs = reordered, indices = seq_len(n))
}

.reorder_one_graph <- function(graph) {
  root <- which(igraph::degree(graph, mode = "in") == 0)
  seq_cache <- build_seq_cache(graph, root)
  pseudo_seq <- seq_glycan(root, seq_cache)
  .reorder_by_pseudo_seq(graph, pseudo_seq)
}

.reorder_by_pseudo_seq <- function(graph, pseudo_seq) {
  # Reorder the vertices
  vertices <- stringr::str_extract_all(pseudo_seq, "V(\\d+)")[[1]]
  vertices <- stringr::str_sub(vertices, 2L, -1L)
  # Fix: Use the correct permutation vector
  # vertices contains the desired order, we need the inverse permutation
  target_order <- as.numeric(vertices)
  permutation <- order(target_order)
  graph <- igraph::permute(graph, permutation)
  igraph::V(graph)$name <- as.character(1:igraph::vcount(graph))

  # Reorder the edges
  edges <- stringr::str_extract_all(pseudo_seq, "E(\\d+)")[[1]]
  edges <- stringr::str_sub(edges, 2L, -1L)
  graph <- .permute_edges(graph, as.numeric(edges))

  graph
}

.permute_edges <- function(g, order) {
  edges <- igraph::as_data_frame(g, what = "edges")
  verts <- igraph::as_data_frame(g, what = "vertices")
  edges <- edges[order, , drop = FALSE]

  # Ensure 'name' column is first in vertices data frame if it exists
  # This is required by igraph::graph_from_data_frame() to avoid "Duplicate vertex names" error
  if ("name" %in% names(verts)) {
    name_col <- verts[["name"]]
    other_cols <- verts[, !names(verts) %in% "name", drop = FALSE]
    verts <- cbind(name = name_col, other_cols)
  }

  new_g <- igraph::graph_from_data_frame(
    edges,
    directed = igraph::is_directed(g),
    vertices = verts
  )
  new_g$anomer <- g$anomer
  new_g
}

#' Example Glycan Structures
#'
#' @description
#' Create example glycan structures for testing and demonstration.
#' Includes **N-glycan core** and **O-glycan core 1** and **core 2**.
#'
#' @details
#' # N-Glycan Core
#'
#' **N-Glycans** are branched oligosaccharides that are bound, most commonly,
#' via GlcNAc to an Asn residue of the protein backbone.
#' A common motif of all N-glycans is the **chitobiose core**,
#' composed of three mannose and two GlcNAc moieties,
#' which is commonly attached to the protein backbone via GlcNAc.
#' The mannose residue is branched and connected via a1,3- and a1,6-glycosidic
#' linkages to the two other mannose building blocks.
#'
#' ```
#'     Man
#'   a1-6 \   b1-4      b1-4      b1-
#'         Man -- GlcNAc -- GlcNAc -
#'   a1-3 /
#'     Man
#' ```
#'
#' # O-Glycan Core
#'
#' **O-Glycans** are highly abundant in extracellular proteins.
#' Generally, O-glycans are extended following four major core structures:
#' **core 1**, **core 2**, core 3, and core 4.
#' The first two are by far the most common core structures in O-glycosylation
#' and are found throughout the body.
#'
#' **core 1**:
#' ```
#'           a1-
#'     GalNAc -
#'    / b1-3
#' Gal
#' ```
#'
#' **core 2**:
#' ```
#' GlcNAc
#'       \ b1-6 a1-
#'        GalNAc -
#'       / b1-3
#'    Gal
#' ```
#'
#' @param linkage A logical indicating whether to include linkages (e.g. "b1-4").
#'   Default is `TRUE`.
#' @param mono_type A character string specifying the type of monosaccharides.
#'   Can be "generic" (Hex, HexNAc, dHex, NeuAc, etc.)
#'   or "concrete" (Man, Gal, HexNAc, Fuc, etc.). Default is "concrete".
#'
#' @returns A glycan structure (igraph) object.
#'
#' @examples
#' print(n_glycan_core(), verbose = TRUE)
#' print(o_glycan_core_1(), verbose = TRUE)
#'
#' @export
n_glycan_core <- function(linkage = TRUE, mono_type = "concrete") {
  validate_example_args(linkage, mono_type)
  build_example_graph(linkage, mono_type, n_glycan_core_base)
}


#' @rdname n_glycan_core
#' @export
o_glycan_core_1 <- function(linkage = TRUE, mono_type = "concrete") {
  validate_example_args(linkage, mono_type)
  build_example_graph(linkage, mono_type, o_glycan_core_1_base)
}


#' @rdname n_glycan_core
#' @export
o_glycan_core_2 <- function(linkage = TRUE, mono_type = "concrete") {
  validate_example_args(linkage, mono_type)
  build_example_graph(linkage, mono_type, o_glycan_core_2_base)
}


build_example_graph <- function(linkage, mono_type, builder) {
  if (!mono_type %in% c("generic", "concrete")) {
    cli::cli_abort("Mono type must be 'generic' or 'concrete'.")
  }
  if (!is.logical(linkage) && length(linkage) != 1) {
    cli::cli_abort("Linkage must be a single logical.")
  }
  glycan <- builder()
  if (!linkage) {
    igraph::E(glycan)$linkage <- "??-?"
    glycan$anomer <- "??"
  }
  if (mono_type == "generic") {
    # Convert the igraph to glyrepr_structure first, then convert mono type
    glycan_struct <- glycan_structure(glycan)
    glycan_struct <- convert_to_generic(glycan_struct)
    return(glycan_struct)
  }
  glycan_structure(glycan)
}


n_glycan_core_base <- function() {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3, 3 - +4, 3 - +5)
  igraph::V(graph)$mono <- c("GlcNAc", "GlcNAc", "Man", "Man", "Man")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "b1-4", "a1-3", "a1-6")
  graph$anomer <- "b1"
  graph
}


o_glycan_core_1_base <- function() {
  graph <- igraph::make_graph(~ 1 - +2)
  igraph::V(graph)$mono <- c("GalNAc", "Gal")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- "b1-3"
  graph$anomer <- "a1"
  graph
}


o_glycan_core_2_base <- function() {
  graph <- igraph::make_graph(~ 1 - +2, 1 - +3)
  igraph::V(graph)$mono <- c("GalNAc", "Gal", "GlcNAc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-3", "b1-6")
  graph$anomer <- "a1"
  graph
}


validate_example_args <- function(linkage, mono_type) {
  checkmate::assert_choice(mono_type, c("generic", "concrete"))
  checkmate::assert_flag(linkage)
}

test_that("N-glycan core structure", {
  skip_on_old_win()
  glycan <- n_glycan_core()
  expect_s3_class(glycan, c("glyrepr_structure", "vctrs_rcrd", "vctrs_vctr"))
  expect_snapshot(print(glycan, verbose = TRUE))
})


test_that("N-glycan core structure without linkages", {
  skip_on_old_win()
  glycan <- n_glycan_core(linkage = FALSE)
  expect_snapshot(print(glycan, verbose = TRUE))
})


test_that("N-glycan core structure with generic monosaccharides", {
  skip_on_old_win()
  glycan <- n_glycan_core(mono_type = "generic")
  expect_snapshot(print(glycan, verbose = TRUE))
})


test_that("O-glycan core 1", {
  skip_on_old_win()
  glycan <- o_glycan_core_1()
  expect_snapshot(print(glycan, verbose = TRUE))
})


test_that("O-glycan core 2", {
  skip_on_old_win()
  glycan <- o_glycan_core_2()
  expect_snapshot(print(glycan, verbose = TRUE))
})

test_that("structure_to_iupac works with basic linear structures", {
  # Test with O-glycan core 1: GalNAc -> Gal
  glycan1 <- o_glycan_core_1()
  result1 <- structure_to_iupac(glycan1)
  expect_equal(result1, "Gal(b1-3)GalNAc(a1-")

  # Create simple linear structure: Glc -> GlcNAc -> Gal
  # This should produce: Gal(b1-3)GlcNAc(b1-4)Glc(?1-
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Glc", "GlcNAc", "Gal")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "b1-3")
  graph$anomer <- "?1"
  glycan2 <- glycan_structure(graph)

  result2 <- structure_to_iupac(glycan2)
  expect_equal(result2, "Gal(b1-3)GlcNAc(b1-4)Glc(?1-")
})

test_that("structure_to_iupac works with branched structures", {
  # Test with N-glycan core
  glycan <- n_glycan_core()
  result <- structure_to_iupac(glycan)
  expect_equal(result, "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-")

  # Test with O-glycan core 2
  glycan2 <- o_glycan_core_2()
  result2 <- structure_to_iupac(glycan2)
  expect_equal(result2, "Gal(b1-3)[GlcNAc(b1-6)]GalNAc(a1-")
})

test_that("structure_to_iupac handles single node structures", {
  # Single node structure - need to add empty linkage attribute
  graph <- igraph::make_graph(edges = integer(0), n = 1)
  igraph::V(graph)$mono <- "Glc"
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- character(0) # Empty linkage for no edges
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  expect_equal(result, "Glc(a1-")
})

test_that("structure_to_iupac handles two-node structures", {
  # Two node structure: A -> B
  graph <- igraph::make_graph(~ 1 - +2)
  igraph::V(graph)$mono <- c("Man", "Glc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- "a1-3"
  graph$anomer <- "b1"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  expect_equal(result, "Glc(a1-3)Man(b1-")
})

test_that("structure_to_iupac ensures isomorphic graphs produce same sequence", {
  # Create first graph: Man with a1-3 and a1-6 branches (in that order)
  graph1 <- igraph::make_graph(~ 1 - +2, 2 - +3, 3 - +4, 3 - +5)
  igraph::V(graph1)$mono <- c("GlcNAc", "GlcNAc", "Man", "Man", "Man")
  igraph::V(graph1)$sub <- ""
  igraph::E(graph1)$linkage <- c("b1-4", "b1-4", "a1-3", "a1-6")
  graph1$anomer <- "?1"
  graph1$alditol <- FALSE
  glycan1 <- glycan_structure(graph1)

  # Create second graph: Same structure but with a1-6 and a1-3 branches (swapped order)
  graph2 <- igraph::make_graph(~ 1 - +2, 2 - +3, 3 - +4, 3 - +5)
  igraph::V(graph2)$mono <- c("GlcNAc", "GlcNAc", "Man", "Man", "Man")
  igraph::V(graph2)$sub <- ""
  igraph::E(graph2)$linkage <- c("b1-4", "b1-4", "a1-6", "a1-3") # Swapped order
  graph2$anomer <- "?1"
  graph2$alditol <- FALSE
  glycan2 <- glycan_structure(graph2)

  # Both should produce the same sequence
  result1 <- structure_to_iupac(glycan1)
  result2 <- structure_to_iupac(glycan2)
  expect_equal(result1, result2)
  expect_equal(result1, "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(?1-")
})

test_that("structure_to_iupac handles complex branched structures", {
  # Create a more complex structure with multiple levels of branching
  # Glc -> Man -> GlcNAc -> Gal
  #                |        ├─ Fuc
  #                |        └─ NeuAc
  #                └─ Hex
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3, 3 - +4, 3 - +5, 2 - +6)
  igraph::V(graph)$mono <- c("Glc", "Man", "GlcNAc", "Gal", "Fuc", "GalNAc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "a1-3", "a1-2", "a1-6", "b1-3")
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  # Based on actual algorithm output
  expect_equal(
    result,
    "Fuc(b1-3)[Gal(a1-6)]GlcNAc(a1-3)[GalNAc(a1-2)]Man(b1-4)Glc(a1-"
  )
})

test_that("structure_to_iupac selects correct backbone based on depth", {
  # Create structure where backbone selection matters
  # Glc -> Man -> Gal (depth 2)
  #           └─ GlcNAc -> Fuc (depth 3) <- this should be backbone
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3, 2 - +4, 4 - +5)
  igraph::V(graph)$mono <- c("Glc", "Man", "Gal", "GlcNAc", "Fuc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "a1-3", "a1-6", "b1-2")
  graph$anomer <- "?1"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  # GlcNAc->Fuc path is longer, so should be backbone
  # Expected: Fuc(b1-2)GlcNAc(a1-6)[Gal(a1-3)]Man(b1-4)Glc(?1-
  expect_equal(result, "Fuc(b1-2)GlcNAc(a1-6)[Gal(a1-3)]Man(b1-4)Glc(?1-")
})

test_that("structure_to_iupac selects backbone by linkage when depths are equal", {
  # Create structure where depths are equal but linkages differ
  # Glc -> Man ├─ Gal (linkage a1-3)
  #            └─ Fuc (linkage a1-6)
  # Both have same depth, but a1-3 < a1-6, so Gal should be backbone
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3, 2 - +4)
  igraph::V(graph)$mono <- c("Glc", "Man", "Gal", "Fuc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "a1-3", "a1-6")
  graph$anomer <- "b2"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  # Gal should be backbone (a1-3), Fuc should be branch (a1-6)
  expect_equal(result, "Gal(a1-3)[Fuc(a1-6)]Man(b1-4)Glc(b2-")
})

test_that("structure_to_iupac selects backbone by linkage when depths are equal with ?", {
  # Create structure where depths are equal but linkages differ
  # Glc -> Man ├─ Gal (linkage a1-?)
  #            └─ Fuc (linkage a1-6)
  # Both have same depth, but a1-3 < a1-6, so Gal should be backbone
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3, 2 - +4)
  igraph::V(graph)$mono <- c("Glc", "Man", "Gal", "Fuc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "a1-?", "a1-6")
  graph$anomer <- "b2"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  # Gal should be backbone (a1-?), Fuc should be branch (a1-6)
  expect_equal(result, "Gal(a1-?)[Fuc(a1-6)]Man(b1-4)Glc(b2-")
})

test_that("structure_to_iupac returns deterministic sequence for ties", {
  # Create two structures with the same topology but different node ordering
  # Man ├─ Gal (linkage a1-?)
  #     └─ Fuc (linkage a1-?)

  # glycan 1
  graph1 <- igraph::make_graph(~ 1 - +2, 1 - +3)
  igraph::V(graph1)$mono <- c("Man", "Gal", "Fuc")
  igraph::V(graph1)$sub <- ""
  igraph::E(graph1)$linkage <- c("a1-?", "a1-?")
  graph1$anomer <- "a1"
  glycan1 <- glycan_structure(graph1)

  # glycan 2
  graph2 <- igraph::make_graph(~ 1 - +2, 1 - +3)
  igraph::V(graph2)$mono <- c("Man", "Fuc", "Gal")
  igraph::V(graph2)$sub <- ""
  igraph::E(graph2)$linkage <- c("a1-?", "a1-?")
  graph2$anomer <- "a1"
  glycan2 <- glycan_structure(graph2)

  result1 <- structure_to_iupac(glycan1)
  result2 <- structure_to_iupac(glycan2)
  expect_equal(result1, result2)
})

test_that("structure_to_iupac handles different anomer values", {
  # Test different anomer values
  anomers <- c("a1", "b1", "?1", "a2", "b2", "?2")

  for (anomer in anomers) {
    graph <- igraph::make_graph(~ 1 - +2)
    igraph::V(graph)$mono <- c("Glc", "Man")
    igraph::V(graph)$sub <- ""
    igraph::E(graph)$linkage <- "b1-4"
    graph$anomer <- anomer
    glycan <- glycan_structure(graph)

    result <- structure_to_iupac(glycan)
    expected <- paste0("Man(b1-4)Glc(", anomer, "-")
    expect_equal(result, expected)
  }
})

test_that("structure_to_iupac handles multiple branches correctly 1", {
  # Create structure with 3 branches
  # Glc -> Man
  #         ├─ Gal (a1-2) <- should be backbone (smallest)
  #         ├─ Fuc (a1-3) <- should be the first branch
  #         └─ Hex (a1-6) <- should be the second branch
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3, 2 - +4, 2 - +5)
  igraph::V(graph)$mono <- c("Glc", "Man", "Gal", "Fuc", "GalNAc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "a1-2", "a1-3", "a1-6")
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  # Gal should be backbone (a1-2 is smallest), Fuc and GalNAc should be branches
  # Branches should be ordered: a1-3 < a1-6
  expect_equal(result, "Gal(a1-2)[Fuc(a1-3)][GalNAc(a1-6)]Man(b1-4)Glc(a1-")
})

test_that("structure_to_iupac handles multiple branches correctly 1", {
  s1 <- "GlcNAc(b1-2)Man(a1-3)[GlcNAc(b1-2)Man(a1-6)][GlcNAc(b1-4)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"
  s2 <- "GlcNAc(b1-2)Man(a1-3)[GlcNAc(b1-4)][GlcNAc(b1-2)Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"
  glycan1 <- as_glycan_structure(s1)
  glycan2 <- as_glycan_structure(s2)
  expect_equal(structure_to_iupac(glycan1), s2)
  expect_equal(structure_to_iupac(glycan2), s2)
})

test_that("structure_to_iupac produces correct sequence for examples in documentation", {
  # Test example 1 from documentation
  # Glc (?1- └─GlcNAc (b1-4) └─Gal (b1-3)
  # Should produce: Gal(b1-3)GlcNAc(b1-4)Glc(?1-
  graph1 <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph1)$mono <- c("Glc", "GlcNAc", "Gal")
  igraph::V(graph1)$sub <- ""
  igraph::E(graph1)$linkage <- c("b1-4", "b1-3")
  graph1$anomer <- "?1"
  graph1$alditol <- FALSE
  glycan1 <- glycan_structure(graph1)

  result1 <- structure_to_iupac(glycan1)
  expect_equal(result1, "Gal(b1-3)GlcNAc(b1-4)Glc(?1-")

  # Test example 2 from documentation (already covered by n_glycan_core test)
  # But let's create it explicitly to be sure
  graph2 <- igraph::make_graph(~ 1 - +2, 2 - +3, 3 - +4, 3 - +5)
  igraph::V(graph2)$mono <- c("GlcNAc", "GlcNAc", "Man", "Man", "Man")
  igraph::V(graph2)$sub <- ""
  igraph::E(graph2)$linkage <- c("b1-4", "b1-4", "a1-3", "a1-6")
  graph2$anomer <- "?1"
  graph2$alditol <- FALSE
  glycan2 <- glycan_structure(graph2)

  result2 <- structure_to_iupac(glycan2)
  expect_equal(result2, "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(?1-")
})

test_that("structure_to_iupac handles edge cases with linkages", {
  # Test with ? in different positions
  graph <- igraph::make_graph(~ 1 - +2)
  igraph::V(graph)$mono <- c("Glc", "Man")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- "?2-?"
  graph$anomer <- "??"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  expect_equal(result, "Man(?2-?)Glc(??-")

  # Test with high numbers
  graph2 <- igraph::make_graph(~ 1 - +2)
  igraph::V(graph2)$mono <- c("Glc", "Gal")
  igraph::V(graph2)$sub <- ""
  igraph::E(graph2)$linkage <- "b1-8"
  graph2$anomer <- "a7"
  graph2$alditol <- FALSE
  glycan2 <- glycan_structure(graph2)

  result2 <- structure_to_iupac(glycan2)
  expect_equal(result2, "Gal(b1-8)Glc(a7-")
})

test_that("structure_to_iupac supports substituents", {
  # Create a simple structure with a substituent: Glc with 3Me
  graph <- igraph::make_graph(edges = integer(0), n = 1)
  igraph::V(graph)$mono <- "Glc"
  igraph::V(graph)$sub <- "3Me" # 3号位置有Me修饰
  igraph::E(graph)$linkage <- character(0)
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  # Expected: Glc3Me(a1-
  expect_equal(result, "Glc3Me(a1-")

  # Test with linear structure: Glc with 3Me -> GlcNAc with 6Ac
  graph2 <- igraph::make_graph(~ 1 - +2)
  igraph::V(graph2)$mono <- c("Glc", "GlcNAc")
  igraph::V(graph2)$sub <- c("3Me", "6Ac")
  igraph::E(graph2)$linkage <- "b1-4"
  graph2$anomer <- "a1"
  graph2$alditol <- FALSE
  glycan2 <- glycan_structure(graph2)

  result2 <- structure_to_iupac(glycan2)
  # Expected: GlcNAc6Ac(b1-4)Glc3Me(a1-
  expect_equal(result2, "GlcNAc6Ac(b1-4)Glc3Me(a1-")
})

test_that("structure_to_iupac handles empty substituents", {
  # Test with empty substituents (should work as before)
  graph <- igraph::make_graph(~ 1 - +2)
  igraph::V(graph)$mono <- c("Glc", "GlcNAc")
  igraph::V(graph)$sub <- c("", "") # No substituents
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  # Expected: GlcNAc(b1-4)Glc(a1-
  expect_equal(result, "GlcNAc(b1-4)Glc(a1-")
})

test_that("structure_to_iupac handles mixed substituents", {
  # Test with some having substituents, some not
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Glc", "GlcNAc", "Gal")
  igraph::V(graph)$sub <- c("3Me", "", "6S") # Mixed substituents
  igraph::E(graph)$linkage <- c("b1-4", "b1-3")
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)

  result <- structure_to_iupac(glycan)
  # Expected: Gal6S(b1-3)GlcNAc(b1-4)Glc3Me(a1-
  expect_equal(result, "Gal6S(b1-3)GlcNAc(b1-4)Glc3Me(a1-")
})

# Tests for count_mono function

# Test helper function to create simple glycan graph
create_test_glycan_graph <- function(
  mono_names,
  linkages = NULL,
  anomer = "?1"
) {
  n_nodes <- length(mono_names)
  if (n_nodes == 1) {
    graph <- igraph::make_empty_graph(n = 1)
    linkages <- character(0)
  } else {
    # Create linear chain: 1-+2-+3-+...
    edges <- c()
    for (i in 1:(n_nodes - 1)) {
      edges <- c(edges, i, i + 1)
    }
    graph <- igraph::make_graph(edges = edges, directed = TRUE)
    if (is.null(linkages)) {
      linkages <- rep("b1-4", n_nodes - 1)
    }
  }

  igraph::V(graph)$name <- as.character(1:igraph::vcount(graph))
  igraph::V(graph)$mono <- mono_names
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- linkages
  graph$anomer <- anomer
  graph
}

# Tests for count_mono with glycan compositions ---------------------------

test_that("count_mono works with compositions", {
  # Test generic composition
  comp <- glycan_composition(c(Hex = 3, HexNAc = 2, dHex = 1))
  expect_equal(count_mono(comp, "Hex"), 3)
  expect_equal(count_mono(comp, "HexNAc"), 2)
  expect_equal(count_mono(comp, "HexA"), 0) # Not present

  # Test concrete composition
  comp2 <- glycan_composition(c(Glc = 2, Gal = 1, GlcNAc = 2))
  expect_equal(count_mono(comp2, "Glc"), 2)
  expect_equal(count_mono(comp2, "Man"), 0) # Not present
})

test_that("count_mono works for compositions unable to be converted to generic", {
  comp <- glycan_composition(c(GlcNAc = 2, Kdn = 1))
  expect_equal(count_mono(comp, "HexNAc"), 2)
  expect_equal(count_mono(comp, "Kdn"), 1)
})

test_that("count_mono works for compositions with special monosaccharides", {
  comp <- glycan_composition(c(gNeu = 1, Hex = 1))
  expect_equal(count_mono(comp, "gNeu"), 1)
  expect_equal(count_mono(comp, "Hex"), 1)
  expect_equal(count_mono(comp, "Glc"), NA_integer_)
})

test_that("count_mono works when counting generic in concrete compositions", {
  # When mono is generic, it should count all matching concrete monos
  comp <- glycan_composition(c(
    Glc = 2,
    Gal = 1,
    Man = 1,
    GlcNAc = 2,
    GalNAc = 1
  ))

  # Hex should count Glc, Gal, Man (all hexoses)
  expect_equal(count_mono(comp, "Hex"), 4) # 2 Glc + 1 Gal + 1 Man

  # HexNAc should count GlcNAc, GalNAc
  expect_equal(count_mono(comp, "HexNAc"), 3) # 2 GlcNAc + 1 GalNAc
})

test_that("count_mono returns NA when counting concrete in generic compositions", {
  comp <- glycan_composition(c(Hex = 2, HexNAc = 1))
  expect_equal(count_mono(comp, "GalNAc"), NA_integer_)
})

test_that("count_mono works with multiple compositions", {
  # Test with multiple compositions in a vector
  comp_vec <- glycan_composition(
    c(Hex = 5, HexNAc = 2),
    c(Hex = 3, HexNAc = 1, dHex = 1),
    c(HexNAc = 4)
  )

  expect_equal(count_mono(comp_vec, "Hex"), c(5, 3, 0))
  expect_equal(count_mono(comp_vec, "HexNAc"), c(2, 1, 4))
  expect_equal(count_mono(comp_vec, "dHex"), c(0, 1, 0))
})

test_that("count_mono works with `mono` as NULL", {
  comp <- glycan_composition(
    c(Man = 5, GlcNAc = 2),
    c(Gal = 1, Man = 1, GalNAc = 1)
  )
  expect_equal(count_mono(comp), c(7L, 3L))
})

test_that("`include_subs` works when `mono` is NULL", {
  comp <- glycan_composition(c(Glc = 1, S = 1))
  expect_equal(count_mono(comp), 1L)
  expect_equal(count_mono(comp, include_subs = TRUE), 2L)
})

test_that("count_mono works for substituents", {
  comp <- glycan_composition(c(Glc = 1, S = 1), c(Glc = 1))
  expect_equal(count_mono(comp, "S"), c(1L, 0L))
})

test_that("count_mono preserves NA in compositions", {
  # Composition with NA element - total count (mono = NULL)
  comps <- glycan_composition(c(Gal = 1), NA)
  expect_equal(count_mono(comps), c(1L, NA_integer_))

  # NA in different positions - total count
  comps2 <- glycan_composition(NA, c(Gal = 1))
  expect_equal(count_mono(comps2), c(NA_integer_, 1L))

  comps3 <- glycan_composition(c(Gal = 1), NA, c(GlcNAc = 2))
  expect_equal(count_mono(comps3), c(1L, NA_integer_, 2L))

  # Composition with NA element - specific mono count
  comps4 <- glycan_composition(c(Gal = 1), NA)
  expect_equal(count_mono(comps4, "Gal"), c(1L, NA_integer_))
  expect_equal(count_mono(comps4, "Hex"), c(1L, NA_integer_))
  expect_equal(count_mono(comps4, "Glc"), c(0L, NA_integer_))

  # NA in different positions - specific mono count
  comps5 <- glycan_composition(NA, c(Gal = 1, GlcNAc = 1))
  expect_equal(count_mono(comps5, "Gal"), c(NA_integer_, 1L))
  expect_equal(count_mono(comps5, "HexNAc"), c(NA_integer_, 1L))
  expect_equal(count_mono(comps5, "Man"), c(NA_integer_, 0L))

  # Multiple NAs - specific mono count
  comps6 <- glycan_composition(c(Gal = 1), NA, c(GlcNAc = 2), NA)
  expect_equal(count_mono(comps6, "Gal"), c(1L, NA_integer_, 0L, NA_integer_))
  expect_equal(
    count_mono(comps6, "HexNAc"),
    c(0L, NA_integer_, 2L, NA_integer_)
  )
})

# Tests for count_mono with glycan structures ----------------------------

test_that("count_mono works with glycan structures", {
  # Test with simple structure
  graph <- create_test_glycan_graph(c("GlcNAc", "Gal", "Glc"))
  struct <- glycan_structure(graph)
  expect_equal(count_mono(struct, "GlcNAc"), 1)
  expect_equal(count_mono(struct, "Gal"), 1)
  expect_equal(count_mono(struct, "Man"), 0) # Not present

  # Test with generic monos in structures
  graph2 <- create_test_glycan_graph(c("Glc", "Gal", "Man", "GlcNAc"))
  struct2 <- glycan_structure(graph2)
  expect_equal(count_mono(struct2, "Hex"), 3) # Count Glc, Gal, Man
  expect_equal(count_mono(struct2, "HexNAc"), 1) # Count GlcNAc
})

test_that("count_mono works with multiple structures", {
  # Test with N-glycan and O-glycan cores
  n_glycan <- n_glycan_core()
  o_glycan <- o_glycan_core_1()
  struct_vec <- c(n_glycan, o_glycan)

  # N-glycan core has: 2 GlcNAc, 3 Man; O-glycan core has: 1 GalNAc, 1 Gal
  expect_equal(count_mono(struct_vec, "GlcNAc"), c(2, 0))
  expect_equal(count_mono(struct_vec, "GalNAc"), c(0, 1))
})

# Tests for parameter validation and edge cases ---------------------------

test_that("count_mono validates parameters and handles edge cases", {
  comp <- glycan_composition(c(Hex = 2, HexNAc = 1))

  # Parameter validation
  expect_error(count_mono(comp, "Unknown"), "must be a known monosaccharide")
  expect_error(count_mono(comp, c("Hex", "HexNAc")), "Must have length 1")
  expect_error(count_mono(comp, 123), "Must be of type 'string'")

  # Edge cases
  empty_comp <- glycan_composition()
  expect_equal(length(count_mono(empty_comp, "Hex")), 0)

  # Return type and vector length
  result <- count_mono(comp, "Hex")
  expect_type(result, "integer")
  expect_equal(result, 2L)
})

#' Create a Glycan Composition
#'
#' Create a glycan composition from a list of named integer vectors.
#' Compositions can contain both monosaccharides and substituents.
#'
#' @param ... Named integer vectors.
#'   Names are monosaccharides or substituents, values are numbers of residues.
#'   Monosaccharides and substituents can be mixed in the same composition.
#' @param x A list of named integer vectors.
#'
#' @returns A glyrepr_composition object.
#'
#' @details
#' Compositions can contain:
#'
#' - Monosaccharides: either generic (e.g., "Hex", "HexNAc") or concrete (e.g., "Glc", "Gal").
#'   All monosaccharides in a composition vector must be of the same type.
#' - Substituents: e.g., "Me", "Ac", "S". These can be mixed with either
#'   generic or concrete monosaccharides.
#'
#' Components are automatically sorted with monosaccharides first (according to
#' their order in the monosaccharides table), followed by substituents (according
#' to their order in `available_substituents()`). Duplicate components are
#' automatically summed.
#'
#' @examples
#' # A vector with one composition (generic monosaccharides)
#' glycan_composition(c(Hex = 5, HexNAc = 2))
#' # A vector with multiple compositions
#' glycan_composition(c(Hex = 5, HexNAc = 2), c(Hex = 5, HexNAc = 4, dHex = 2))
#' # Residues are reordered automatically
#' glycan_composition(c(HexNAc = 1, Hex = 2))
#' # An example for generic monosaccharides
#' glycan_composition(c(Hex = 2, HexNAc = 1))
#' # An example for concrete monosaccharides
#' glycan_composition(c(Glc = 2, Gal = 1))
#' # Compositions with substituents
#' glycan_composition(c(Glc = 1, S = 1))
#' glycan_composition(c(Hex = 3, HexNAc = 2, Me = 1, Ac = 1))
#' # Substituents are sorted after monosaccharides
#' glycan_composition(c(S = 1, Gal = 1, Ac = 1, Glc = 1))
#'
#' @seealso `available_monosaccharides()`, `available_substituents()`
#' @export
glycan_composition <- function(...) {
  args <- rlang::list2(...)

  # Handle NULL/NA inputs - convert to list elements that will be stored as NULL
  # Only unnamed NA/NULL are treated as missing; named NA (e.g., c(Hex = NA)) is invalid
  x <- purrr::map(
    args,
    ~ {
      if (is.null(.x)) {
        NULL # Store as NULL to represent NA
      } else if (
        is.atomic(.x) && length(.x) == 1 && is.na(.x) && is.null(names(.x))
      ) {
        NULL # Unnamed scalar NA treated as missing
      } else {
        .valid_composition_element(.x)
      }
    }
  )

  # Process non-NA elements
  x <- purrr::map(
    x,
    ~ {
      if (is.null(.x)) {
        return(.x)
      }
      result <- as.integer(.x)
      names(result) <- names(.x)
      result <- .aggregate_composition_components(result)
      .reorder_composition_components(result)
    }
  )

  .valid_glycan_composition_input(x)
  new_glycan_composition(x)
}

#' @export
#' @rdname glycan_composition
is_glycan_composition <- function(x) {
  inherits(x, "glyrepr_composition")
}

#' Convert to Glycan Composition
#'
#' Converts an object to a glycan composition using vctrs casting framework.
#' This function provides a convenient way to convert various input types
#' to [glycan_composition()].
#'
#' @param x An object to convert to a glycan composition.
#'   Supported inputs include:
#'   - Named integer vectors or lists of named integer vectors
#'   - Character vectors with composition strings (e.g., "Hex(5)HexNAc(2)")
#'   - `glyrepr_structure` objects (counts both monosaccharides and substituents)
#'   - Existing `glyrepr_composition` objects (returned as-is)
#'
#' @returns A `glyrepr_composition` object.
#'
#' @details
#' This function uses the vctrs casting framework for type conversion.
#' When converting from glycan structures, both monosaccharides and substituents
#' are counted. Substituents are extracted from the `sub` attribute of each
#' vertex in the structure. For example, a vertex with `sub = "3Me"`
#' contributes one "Me" substituent to the composition.
#'
#' Simple composition strings use one-letter residue codes: "H" for "Hex",
#' "N" for "HexNAc", "F" for "dHex", "S"/"A" for "NeuAc", and "G" for
#' "NeuGc". "E" and "L" are also accepted as linkage-specific Neu5Ac codes;
#' they are converted to "NeuAc" with a warning because composition objects do
#' not preserve linkage information.
#'
#' @examples
#' # From a single named vector
#' as_glycan_composition(c(Hex = 5, HexNAc = 2))
#'
#' # From a list of named vectors
#' as_glycan_composition(list(c(Hex = 5, HexNAc = 2), c(Hex = 3, HexNAc = 1)))
#'
#' # From a character vector of Byonic composition strings
#' as_glycan_composition(c("Hex(5)HexNAc(2)", "Hex(3)HexNAc(1)"))
#'
#' # From a character vector of simple composition strings
#' as_glycan_composition(c("H5N2", "H5N4S1F1"))
#'
#' # From an existing composition (returns as-is)
#' comp <- glycan_composition(c(Hex = 5, HexNAc = 2))
#' as_glycan_composition(comp)
#'
#' # From a glycan structure vector
#' strucs <- c(n_glycan_core(), o_glycan_core_1())
#' as_glycan_composition(strucs)
#'
#' @export
as_glycan_composition <- function(x) {
  vec_cast(x, new_glycan_composition())
}

#' @export
vec_ptype_full.glyrepr_composition <- function(x, ...) "glycan_composition"

#' @export
vec_ptype_abbr.glyrepr_composition <- function(x, ...) "comp"

#' @export
format.glyrepr_composition <- function(x, ...) {
  data <- vctrs::field(vctrs::vec_data(x), "data")

  formatted <- purrr::map_chr(
    data,
    ~ {
      if (.is_na_composition_elem(.x)) {
        "<NA>"
      } else {
        paste0(names(.x), "(", .x, ")", collapse = "")
      }
    }
  )

  formatted
}

#' @export
vec_ptype2.glyrepr_composition.glyrepr_composition <- function(x, y, ...) {
  new_glycan_composition(list())
}

#' @export
vec_cast.glyrepr_composition.glyrepr_composition <- function(x, to, ...) {
  x
}

#' @export
vec_cast.character.glyrepr_composition <- function(x, to, ...) {
  convert_one <- function(comp) {
    if (.is_na_composition_elem(comp)) {
      "<NA>"
    } else {
      paste0(names(comp), "(", comp, ")", collapse = "")
    }
  }
  data <- vctrs::vec_data(x)
  purrr::map_chr(vctrs::field(data, "data"), convert_one)
}

#' @export
vec_cast.glyrepr_composition.character <- function(x, to, ...) {
  # Handle empty character vector
  if (length(x) == 0) {
    return(glycan_composition())
  }

  # Handle NA by creating NA composition elements
  na_mask <- is.na(x)
  non_na_positions <- which(!na_mask)
  if (any(na_mask)) {
    # Process non-NA characters
    non_na_x <- x[!na_mask]
    compositions <- list()

    if (length(non_na_x) > 0) {
      # Parse non-NA characters
      parse_result <- purrr::map(non_na_x, parse_single_composition)
      valid_flags <- purrr::map_lgl(parse_result, "valid")

      # Check for invalid characters
      invalid_indices <- which(!valid_flags)
      # Map back to original indices in x (not indices in non_na_x)
      original_invalid <- non_na_positions[invalid_indices]
      if (length(original_invalid) > 0) {
        cli::cli_abort(c(
          "Characters cannot be parsed as glycan compositions at index {original_invalid}",
          "i" = "Expected format: 'Hex(5)HexNAc(2)' with monosaccharide names followed by counts in parentheses."
        ))
      }

      compositions <- purrr::map(parse_result, "composition")
      .warn_lossy_neuac_linkage(parse_result)
    }

    # Build result list preserving order - no global state needed
    result_list <- purrr::map(
      seq_along(x),
      ~ {
        if (na_mask[.x]) {
          NULL
        } else {
          # Find which composition this position corresponds to
          pos_in_non_na <- match(.x, non_na_positions)
          compositions[[pos_in_non_na]]
        }
      }
    )

    return(do.call(glycan_composition, result_list))
  }

  # Original logic for non-NA case (unchanged)
  parse_result <- purrr::map(x, parse_single_composition)
  valid_flags <- purrr::map_lgl(parse_result, "valid")
  compositions <- purrr::map(parse_result, "composition")
  invalid_indices <- which(!valid_flags)
  if (length(invalid_indices) > 0) {
    cli::cli_abort(c(
      "Characters cannot be parsed as glycan compositions at index {invalid_indices}",
      "i" = "Expected format: 'Hex(5)HexNAc(2)' with monosaccharide names followed by counts in parentheses."
    ))
  }
  .warn_lossy_neuac_linkage(parse_result)
  if (length(compositions) == 0) {
    return(glycan_composition())
  }
  do.call(glycan_composition, compositions)
}

#' @export
vec_cast.glyrepr_composition.glyrepr_structure <- function(x, to, ...) {
  # Use smap to convert each structure to composition
  compositions <- smap(x, function(graph) {
    # Count monosaccharides
    monos <- igraph::V(graph)$mono
    mono_tb <- table(monos)
    mono_result <- as.integer(mono_tb)
    names(mono_result) <- names(mono_tb)

    # Count substituents
    subs <- igraph::V(graph)$sub
    sub_types <- extract_substituent_types(subs)
    if (length(sub_types) > 0) {
      sub_tb <- table(sub_types)
      sub_result <- as.integer(sub_tb)
      names(sub_result) <- names(sub_tb)
    } else {
      sub_result <- integer(0)
    }

    # Combine monosaccharides and substituents
    result <- c(mono_result, sub_result)

    # Sort by composition component order (monosaccharides first, then substituents)
    result <- .reorder_composition_components(result)
    result
  })

  # Create composition object
  new_glycan_composition(compositions)
}

#' @export
vec_cast.glyrepr_composition.list <- function(x, to, ...) {
  if (length(x) == 0) {
    return(glycan_composition())
  }
  do.call(glycan_composition, x)
}

#' @export
vec_cast.glyrepr_composition.integer <- function(x, to, ...) {
  .cast_named_vector(x, to, ...)
}

#' @export
vec_cast.glyrepr_composition.double <- function(x, to, ...) {
  .cast_named_vector(x, to, ...)
}

#' @export
vec_ptype2.logical.glyrepr_composition <- function(x, y, ...) {
  new_glycan_composition(list())
}

#' @export
vec_ptype2.glyrepr_composition.logical <- function(x, y, ...) {
  new_glycan_composition(list())
}

#' @export
vec_cast.glyrepr_composition.logical <- function(x, to, ...) {
  # Cast each logical element to NA composition or error
  # TRUE and FALSE are invalid (not NA), only NA is valid
  result <- purrr::map(
    x,
    ~ {
      if (is.na(.x)) {
        NULL # NA becomes NULL (NA composition)
      } else {
        cli::cli_abort(
          "Cannot cast non-NA logical value to glyrepr_composition."
        )
      }
    }
  )
  new_glycan_composition(result)
}

#' @export
vec_cast.logical.glyrepr_composition <- function(x, to, ...) {
  # Cast glyrepr_composition to logical:
  # - NA compositions become NA (preserving missingness)
  # - Valid compositions become FALSE
  is_na <- is.na(x)
  result <- rep(FALSE, length(x))
  result[is_na] <- NA
  result
}

#' @export
vec_restore.glyrepr_composition <- function(x, to, ...) {
  data <- vctrs::field(x, "data")

  # Skip NA elements (NULL) when checking types
  na_mask <- purrr::map_lgl(data, .is_na_composition_elem)
  non_na_data <- data[!na_mask]

  if (length(non_na_data) == 0) {
    return(new_glycan_composition(data))
  }

  monos_list <- purrr::map(
    non_na_data,
    ~ names(.x)[!names(.x) %in% available_substituents()]
  )
  mono_types <- purrr::map_chr(monos_list, get_mono_type_impl)

  if (length(unique(mono_types)) > 1) {
    cli::cli_abort(c(
      "Can't combine `glyrepr_composition`s with different monosaccharide types.",
      "x" = "Found compositions with types: {.val {unique(mono_types)}}.",
      "i" = "Use {.fn convert_to_generic} to convert concrete to generic, or ensure both compositions use the same type."
    ))
  }

  new_glycan_composition(data)
}

#' @export
is.na.glyrepr_composition <- function(x, ...) {
  data <- vctrs::field(x, "data")
  purrr::map_lgl(data, .is_na_composition_elem)
}

.cast_named_vector <- function(x, to, ...) {
  if (length(x) == 0) {
    return(glycan_composition())
  }
  glycan_composition(x)
}

#' @export
obj_print_data.glyrepr_composition <- function(
  x,
  ...,
  max_n = 10,
  colored = TRUE
) {
  if (length(x) == 0) {
    return()
  }

  n <- length(x)
  n_show <- min(n, max_n)

  # Only format the compositions that need to be shown to improve performance
  indices_to_show <- seq_len(n_show)
  formatted <- format_glycan_composition_subset(
    x,
    indices_to_show,
    colored = colored
  )

  # Print each composition on its own line with indexing, up to max_n
  for (i in seq_len(n_show)) {
    cat("[", i, "] ", formatted[i], "\n", sep = "")
  }
  if (n > max_n) {
    cat("... (", n - max_n, " more not shown)\n", sep = "")
  }
}

#' @importFrom pillar pillar_shaft
#' @export
pillar_shaft.glyrepr_composition <- function(x, ...) {
  if (length(x) == 0) {
    return(pillar::pillar_shaft(character()))
  }

  indices <- seq_len(length(x))
  formatted <- format_glycan_composition_subset(x, indices, colored = TRUE)

  pillar::new_pillar_shaft_simple(formatted, align = "left", min_width = 10)
}

# --- Internal functions -------------------------------------------------------

#' Create a glycan composition object
#' @param x A list of named integer vectors.
#' @returns A glyrepr_composition object.
#' @noRd
new_glycan_composition <- function(x = list()) {
  # Use vctrs::list_of instead of new_list_of
  x <- vctrs::new_list_of(
    x,
    .ptype = integer(),
    class = "glyrepr_composition_list"
  )
  vctrs::new_rcrd(list(data = x), class = "glyrepr_composition")
}

#' Validate glycan composition input
#'
#' Use by `glycan_composition()` to validate the input.
#'
#' @param x A list of named integer vectors.
#' @noRd
.valid_glycan_composition_input <- function(x) {
  # 0. Skip if empty
  if (length(x) == 0) {
    return()
  }

  # Filter out NA elements (NULL in list) for validation
  x_valid <- purrr::keep(x, ~ !is.null(.x))

  # 0b. If all elements are NA (NULL), skip further validation
  if (length(x_valid) == 0) {
    return()
  }

  # 1. Type check (skip NA elements)
  if (!purrr::every(x_valid, checkmate::test_integerish)) {
    cli::cli_abort(c(
      "Must be one or more named integer vectors.",
      "i" = "You might want to use {.fn as_glycan_composition} for more flexible input."
    ))
  }

  # 2. Empty check (skip NA elements)
  if (purrr::some(x_valid, ~ length(.x) == 0)) {
    cli::cli_abort("Each composition must have at least one residue.")
  }

  # 3. Name check (skip NA elements)
  if (!purrr::every(x_valid, ~ checkmate::test_named(.x))) {
    cli::cli_abort(c(
      "Must be one or more named integer vectors.",
      "x" = "The input doesn't have names."
    ))
  }

  # 4. Known monosaccharide check (skip NA elements)
  if (
    !purrr::every(x_valid, ~ all(is_known_composition_component(names(.x))))
  ) {
    cli::cli_abort(c(
      "Must have only known monosaccharides",
      "i" = "Call {.fun available_monosaccharides} to see all known monosaccharides."
    ))
  }

  # 5. Mono type check (skip NA elements)
  mono_types <- .get_comp_mono_types(x_valid)
  if (any(mono_types == "mixed")) {
    cli::cli_abort(c(
      "Must have only one type of monosaccharide.",
      "x" = "Some compositions have mixed monosaccharide types (both generic and concrete)."
    ))
  }
  if (!all(mono_types == mono_types[[1]])) {
    cli::cli_abort(c(
      "Must have only one type of monosaccharide.",
      "x" = "Both generic and concrete compositions exist."
    ))
  }

  # 6. Positive number check (skip NA elements)
  if (!purrr::every(x_valid, ~ all(.x > 0))) {
    cli::cli_abort("Must have only positive numbers of residues.")
  }
}

#' Get monosaccharide types from a list of named integer vectors (composition components)
#' @param x A list of named integer vectors.
#' @returns A character vector of monosaccharide types.
#' @noRd
.get_comp_mono_types <- function(x) {
  x <- purrr::map(x, ~ .x[!names(.x) %in% available_substituents()]) # remove all substituents

  # After removing substituents, each composition must contain at least one monosaccharide.
  # Otherwise, get_mono_type_impl() would be called with an empty vector and fail.
  if (!purrr::every(x, ~ length(.x) > 0)) {
    cli::cli_abort(
      "Each composition must contain at least one monosaccharide (non-substituent residue)."
    )
  }
  purrr::map_chr(x, ~ get_mono_type_impl(names(.x)))
}

# Helper function to check if a name is a known composition component (monosaccharide or substituent)
is_known_composition_component <- function(names) {
  is_known_monosaccharide(names) | names %in% available_substituents()
}

# Helper function to extract substituent types from substituent strings
extract_substituent_types <- function(sub_strings) {
  # Extract substituent types from strings like "3Me", "6S", "4Ac,3Me"
  all_subs <- character(0)

  for (sub_str in sub_strings) {
    if (sub_str == "" || is.na(sub_str)) {
      next
    }

    # Split by commas for multiple substituents
    individual_subs <- stringr::str_split(sub_str, ",")[[1]]
    individual_subs <- individual_subs[individual_subs != ""]

    # Extract substituent names (remove position numbers)
    sub_names <- stringr::str_extract(individual_subs, "[A-Za-z]+$")
    all_subs <- c(all_subs, sub_names)
  }

  all_subs
}

# Helper function to parse a single composition string
parse_single_composition <- function(char) {
  # Reject empty strings
  if (char == "") {
    return(list(composition = NULL, valid = FALSE))
  }

  # Try each parser in sequence until one succeeds
  parsers <- list(.parse_byonic_comp, .parse_simple_comp)
  for (parser in parsers) {
    result <- tryCatch(
      {
        composition <- parser(char)
        lossy_neuac_linkage <- isTRUE(attr(composition, "lossy_neuac_linkage"))
        attr(composition, "lossy_neuac_linkage") <- NULL
        list(
          composition = composition,
          valid = TRUE,
          lossy_neuac_linkage = lossy_neuac_linkage
        )
      },
      error = function(e) NULL
    )
    if (!is.null(result)) {
      return(result)
    }
  }

  # All parsers failed
  list(composition = NULL, valid = FALSE)
}

#' Warn when simple NeuAc linkage codes are converted to composition counts
#' @param parse_result A list of per-element results returned by
#'   `parse_single_composition()`.
#' @returns Nothing. Called for its warning side effect.
#' @noRd
.warn_lossy_neuac_linkage <- function(parse_result) {
  if (purrr::some(parse_result, ~ isTRUE(.x$lossy_neuac_linkage))) {
    cli::cli_warn(c(
      "Simple composition codes {.val E} and/or {.val L} are parsed as {.val NeuAc}.",
      "i" = "Linkage-specific Neu5Ac information is discarded in glycan compositions."
    ))
  }
}

.parse_byonic_comp <- function(x) {
  # Use regex to find all patterns like "MonoName(number)"
  pattern <- "([A-Za-z0-9]+)\\((\\d+)\\)"
  matches <- stringr::str_extract_all(x, pattern, simplify = FALSE)[[1]]

  if (length(matches) == 0) {
    stop()
  }

  # Check if the entire string was matched (no remaining characters)
  total_matched_length <- sum(stringr::str_length(matches))
  if (total_matched_length != stringr::str_length(x)) {
    stop()
  }

  # Parse each match using stringr::str_match
  parsed_matches <- purrr::map(matches, function(match) {
    match_result <- stringr::str_match(match, pattern)
    list(
      name = match_result[1, 2], # First capture group
      count = as.integer(match_result[1, 3]) # Second capture group
    )
  })

  # Extract names and counts
  mono_names <- purrr::map_chr(parsed_matches, "name")
  mono_counts <- purrr::map_int(parsed_matches, "count")

  # Create named vector
  comp <- mono_counts
  names(comp) <- mono_names
  comp
}

.parse_simple_comp <- function(x) {
  # "S" and "A" are both NeuAc, "G" is NeuGc
  mono_pattern <- "([HNFSAGEL])(\\d+)"
  matches <- stringr::str_extract_all(x, mono_pattern, simplify = FALSE)[[1]]

  # Check if no monos were matched
  if (length(matches) == 0) {
    stop()
  }

  # Check if the entire string was matched (no remaining characters)
  total_matched_length <- sum(stringr::str_length(matches))
  if (total_matched_length != stringr::str_length(x)) {
    stop()
  }

  parsed_matches <- purrr::map(matches, function(match) {
    match_result <- stringr::str_match(match, mono_pattern)
    list(
      name = match_result[1, 2], # First capture group
      count = as.integer(match_result[1, 3]) # Second capture group
    )
  })

  mono_names <- purrr::map_chr(parsed_matches, "name")
  lossy_neuac_linkage <- any(mono_names %in% c("E", "L"))
  mono_names <- dplyr::recode_values(
    mono_names,
    "H" ~ "Hex",
    "N" ~ "HexNAc",
    "F" ~ "dHex",
    "S" ~ "NeuAc",
    "A" ~ "NeuAc",
    "E" ~ "NeuAc",
    "L" ~ "NeuAc",
    "G" ~ "NeuGc"
  )
  mono_counts <- purrr::map_int(parsed_matches, "count")

  # Create named vector
  comp <- mono_counts
  names(comp) <- mono_names
  attr(comp, "lossy_neuac_linkage") <- lossy_neuac_linkage
  comp
}

#' Reorder composition components
#'
#' Composition components include monosaccharides and substituents.
#' This function reorders the components based on the names.
#' It assumes that all components are valid.
#'
#' The order is determined by the order of the names in [available_monosaccharides()]
#' and [available_substituents()].
#' Monosaccharides are placed before substituents.
#'
#' @param components A named integer vector of composition components.
#' @returns A named integer vector of reordered composition components.
#' @noRd
.reorder_composition_components <- function(components) {
  mono_orders <- available_monosaccharides()
  sub_orders <- available_substituents()
  orders <- c(mono_orders, sub_orders)
  components[order(match(names(components), orders))]
}

#' Aggregate duplicated composition components
#'
#' Components with the same name are summed while preserving the first occurrence
#' of each component name. The final display order is handled separately by
#' `.reorder_composition_components()`.
#'
#' @param components A named integer vector of composition components.
#' @returns A named integer vector with unique component names.
#' @noRd
.aggregate_composition_components <- function(components) {
  component_names <- unique(names(components))
  groups <- factor(names(components), levels = component_names)
  aggregated <- tapply(components, groups, sum)
  aggregated <- as.integer(aggregated)
  names(aggregated) <- component_names
  aggregated
}

# Helper function to format a subset of compositions
format_glycan_composition_subset <- function(x, indices, colored = TRUE) {
  if (!colored) {
    return(format(x)[indices])
  }

  # Format with colors if concrete type
  format_one_colored <- function(comp) {
    if (.is_na_composition_elem(comp)) {
      "<NA>"
    } else {
      mono_names <- names(comp)
      # Add colors to monosaccharides (generic monos automatically get black color)
      mono_names <- add_colors(mono_names, colored = TRUE)
      paste0(mono_names, "(", comp, ")", collapse = "")
    }
  }

  data <- vctrs::vec_data(x)
  comp_data <- vctrs::field(data, "data")[indices]

  purrr::map_chr(comp_data, format_one_colored)
}

#' Check if a composition element is NA
#'
#' NA elements are stored as NULL in the list_of data field.
#' @param x A single composition element (named integer vector or NULL)
#' @returns TRUE if NA (NULL), FALSE otherwise
#' @noRd
.is_na_composition_elem <- function(x) {
  is.null(x)
}

#' Validate a single composition element
#'
#' @param x A single composition element
#' @returns The validated and converted integer vector
#' @noRd
.valid_composition_element <- function(x) {
  if (!checkmate::test_integerish(x)) {
    cli::cli_abort("Each composition must be a named integer vector.")
  }
  if (length(x) == 0) {
    cli::cli_abort("Each composition must have at least one residue.")
  }
  if (!checkmate::test_named(x)) {
    cli::cli_abort("Each composition must be a named integer vector.")
  }
  if (!all(is_known_composition_component(names(x)))) {
    cli::cli_abort(c(
      "Must have only known monosaccharides",
      "i" = "Call {.fun available_monosaccharides} to see all known monosaccharides."
    ))
  }
  if (any(is.na(x)) || !all(x > 0)) {
    cli::cli_abort("Must have only positive numbers of residues.")
  }
  result <- as.integer(x)
  names(result) <- names(x)
  result
}

test_that("empty composition is valid", {
  comp <- glycan_composition()
  expect_s3_class(comp, "glyrepr_composition")
})

test_that("generic composition is valid", {
  comp <- glycan_composition(c(Hex = 2, HexNAc = 1))
  expect_s3_class(comp, "glyrepr_composition")
})

test_that("concrete composition is valid", {
  comp <- glycan_composition(c(Glc = 2, Gal = 1))
  expect_s3_class(comp, "glyrepr_composition")
})

test_that("compositions can contain substituents", {
  comp <- glycan_composition(c(Glc = 1, S = 1))
  expect_equal(as.character(comp), "Glc(1)S(1)")
})

test_that("compositions are sorted correctly", {
  # Hex should come before HexNAc based on the monosaccharides tibble
  comp <- glycan_composition(c(HexNAc = 1, Hex = 2))
  data <- vctrs::vec_data(comp)
  comp_data <- vctrs::field(data, "data")[[1]]
  expect_equal(names(comp_data), c("Hex", "HexNAc"))
})

test_that("substituents are located after monosaccharides", {
  comp <- glycan_composition(c(S = 1, Gal = 1, Ac = 1, Glc = 1))
  # 1. "Ac" and "S" are sorted after "Glc" and "Gal"
  # 2. Order of "Ac" and "S" is according to `available_substituents()`
  expect_equal(as.character(comp), "Glc(1)Gal(1)Ac(1)S(1)")
})

test_that("mixed types within one composition throws error", {
  expect_error(
    glycan_composition(c(Hex = 1, Glc = 1)),
    "Must have only one type of monosaccharide"
  )
})

test_that("mixed types within one composition vector throws error", {
  expect_error(
    glycan_composition(c(Hex = 1, HexNAc = 1), c(Glc = 1, Gal = 1)),
    "Must have only one type of monosaccharide"
  )
})

test_that("unknown monosaccharides throw error", {
  expect_error(
    glycan_composition(c(Glc = 1, unknown = 1)),
    "Must have only known monosaccharides"
  )
})

test_that("glycan_composition rejects wrong types", {
  expect_error(
    glycan_composition(list(c(Hex = 1, HexNAc = 1))),
    "named integer"
  )
})

test_that("glycan_composition rejects empty compositions", {
  expect_error(glycan_composition(integer(0)), "at least one residue")
})

test_that("glycan_composition deals with duplications", {
  comp <- glycan_composition(c(Hex = 1, HexNAc = 1, Hex = 2))
  expect_equal(as.character(comp), "Hex(3)HexNAc(1)")
})

test_that("as_glycan_composition works with list of named vectors", {
  vec_list <- list(c(Hex = 5, HexNAc = 2), c(Hex = 3, HexNAc = 1))
  comp <- as_glycan_composition(vec_list)
  expected <- glycan_composition(c(Hex = 5, HexNAc = 2), c(Hex = 3, HexNAc = 1))
  expect_equal(comp, expected)
})

test_that("as_glycan_composition works with a single named vector", {
  comp <- as_glycan_composition(c(Hex = 5, HexNAc = 2))
  expected <- glycan_composition(c(Hex = 5, HexNAc = 2))
  expect_equal(comp, expected)
  expect_equal(comp, as_glycan_composition(list(c(Hex = 5, HexNAc = 2))))
})

test_that("as_glycan_composition returns existing composition unchanged", {
  original <- glycan_composition(c(Hex = 5, HexNAc = 2))
  result <- as_glycan_composition(original)
  expect_identical(result, original)
})

test_that("as_glycan_composition works for a glycan structure", {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Glc", "Gal", "Glc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "b1-4")
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)

  comp <- as_glycan_composition(glycan)

  expect_s3_class(comp, "glyrepr_composition")
  expected_comp <- glycan_composition(c(Glc = 2L, Gal = 1L))
  expect_equal(comp, expected_comp)
})

test_that("as_glycan_composition works for a glycan structure with substituents", {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Glc", "Gal", "Glc")
  igraph::V(graph)$sub <- c("", "", "3Me")
  igraph::E(graph)$linkage <- c("b1-4", "b1-4")
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)

  comp <- as_glycan_composition(glycan)

  expect_s3_class(comp, "glyrepr_composition")
  expected_comp <- glycan_composition(c(Glc = 2L, Gal = 1L, Me = 1L))
  expect_equal(comp, expected_comp)
})

test_that("as_glycan_composition works for a glycan structure with multiple substituents", {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Glc", "Gal", "Glc")
  igraph::V(graph)$sub <- c("", "", "3Me,6S")
  igraph::E(graph)$linkage <- c("b1-4", "b1-4")
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)

  comp <- as_glycan_composition(glycan)

  expect_s3_class(comp, "glyrepr_composition")
  expected_comp <- glycan_composition(c(Glc = 2L, Gal = 1L, Me = 1L, S = 1L))
  expect_equal(comp, expected_comp)
})

# Tests for formatting ----------------------------------------------

test_that("format works correctly", {
  # generic monosaccharides
  comp2 <- glycan_composition(c(Hex = 2, HexNAc = 1))
  expect_equal(format(comp2), "Hex(2)HexNAc(1)")

  # concrete monosaccharides
  comp3 <- glycan_composition(c(Glc = 2, Gal = 1))
  expect_equal(format(comp3), "Glc(2)Gal(1)")
})

test_that("is_glycan_composition works correctly", {
  comp <- glycan_composition(c(Hex = 2, HexNAc = 1))
  expect_true(is_glycan_composition(comp))
  expect_false(is_glycan_composition(c(Hex = 2, HexNAc = 1)))
  expect_false(is_glycan_composition("Hex(2)HexNAc(1)"))
})

# Tests for c() function (vec_ptype2 and vec_cast methods) -----------------------

test_that("c() combines composition vectors correctly", {
  # Test basic combination of composition vectors
  comp1 <- glycan_composition(c(Hex = 5, HexNAc = 2))
  comp2 <- glycan_composition(c(Hex = 3, HexNAc = 1))

  # This should work without error
  combined <- c(comp1, comp2)

  expect_s3_class(combined, "glyrepr_composition")
  expect_equal(length(combined), 2)

  # Check that both compositions are preserved
  formatted <- format(combined)
  expect_equal(formatted[1], "Hex(5)HexNAc(2)")
  expect_equal(formatted[2], "Hex(3)HexNAc(1)")
})

test_that("c() handles multiple composition vectors", {
  comp1 <- glycan_composition(c(Hex = 2, HexNAc = 1))
  comp2 <- glycan_composition(c(Hex = 3, HexNAc = 2), c(Hex = 1, HexNAc = 1))
  comp3 <- glycan_composition(c(Hex = 4, HexNAc = 3))

  combined <- c(comp1, comp2, comp3)

  expect_s3_class(combined, "glyrepr_composition")
  expect_equal(length(combined), 4) # 1 + 2 + 1 = 4 total compositions

  formatted <- format(combined)
  expect_equal(formatted[1], "Hex(2)HexNAc(1)")
  expect_equal(formatted[2], "Hex(3)HexNAc(2)")
  expect_equal(formatted[3], "Hex(1)HexNAc(1)")
  expect_equal(formatted[4], "Hex(4)HexNAc(3)")
})

test_that("c() works with empty composition vectors", {
  comp1 <- glycan_composition() # Empty composition vector
  comp2 <- glycan_composition(c(Hex = 2, HexNAc = 1))

  combined1 <- c(comp1, comp2)
  combined2 <- c(comp2, comp1)

  expect_s3_class(combined1, "glyrepr_composition")
  expect_s3_class(combined2, "glyrepr_composition")
  expect_equal(length(combined1), 1)
  expect_equal(length(combined2), 1)
  expect_equal(format(combined1), "Hex(2)HexNAc(1)")
  expect_equal(format(combined2), "Hex(2)HexNAc(1)")
})

test_that("c() throws error for different monosaccharide types", {
  # Test that combining generic and concrete compositions throws an error
  generic_comp <- glycan_composition(c(Hex = 1, HexNAc = 1))
  concrete_comp <- glycan_composition(c(Glc = 1, Gal = 1))

  expect_error(c(generic_comp, concrete_comp), "Can't combine")
})

test_that("c() throws error for different monosaccharide types with substituents", {
  # Test that combining generic and concrete compositions throws an error
  generic_comp <- glycan_composition(c(Hex = 1, HexNAc = 1, Me = 1))
  concrete_comp <- glycan_composition(c(Glc = 1, Gal = 1, S = 1))

  expect_error(c(generic_comp, concrete_comp), "Can't combine")
})

test_that("c() maintains proper ordering within compositions", {
  # Test that monosaccharide ordering is preserved during combination
  comp1 <- glycan_composition(c(GalNAc = 1, Gal = 2)) # Out of order input
  comp2 <- glycan_composition(c(GalNAc = 1, Glc = 2)) # Out of order input

  combined <- c(comp1, comp2)

  formatted <- format(combined)
  expect_equal(formatted[1], "Gal(2)GalNAc(1)") # Should be reordered
  expect_equal(formatted[2], "Glc(2)GalNAc(1)") # Should be reordered
})

# Tests for vector casting functionality ----------------------------------------

test_that("composition vectors can be subset and maintain structure", {
  comp <- glycan_composition(
    c(Hex = 1, HexNAc = 1),
    c(Hex = 2, HexNAc = 2),
    c(Hex = 3, HexNAc = 3)
  )

  # Test subsetting
  subset1 <- comp[1]
  subset2 <- comp[c(1, 3)]

  expect_s3_class(subset1, "glyrepr_composition")
  expect_s3_class(subset2, "glyrepr_composition")
  expect_equal(length(subset1), 1)
  expect_equal(length(subset2), 2)

  expect_equal(format(subset1), "Hex(1)HexNAc(1)")
  expect_equal(format(subset2), c("Hex(1)HexNAc(1)", "Hex(3)HexNAc(3)"))
})

test_that("composition vectors can be repeated", {
  comp <- glycan_composition(c(Hex = 2, HexNAc = 1))

  # Test rep() function which uses vctrs casting methods
  repeated <- rep(comp, 3)

  expect_s3_class(repeated, "glyrepr_composition")
  expect_equal(length(repeated), 3)
  expect_equal(format(repeated), rep("Hex(2)HexNAc(1)", 3))
})

# Tests for printing performance ----------------------------------------------
test_that("truncation works in tibble for compositions", {
  comp <- glycan_composition(
    c(Hex = 1, HexNAc = 1),
    c(Hex = 2, HexNAc = 1),
    c(Hex = 3, dHex = 1)
  )
  tibble <- tibble::tibble(comp = comp, a = 1)
  expect_snapshot(print(tibble, width = 30))
})

# Tests for character conversion ----------------------------------------------
test_that("as_glycan_composition works for compositions", {
  chars <- "Hex(5)HexNAc(2)"
  expected <- glycan_composition(c(Hex = 5, HexNAc = 2))
  expect_equal(as_glycan_composition(chars), expected)
})

test_that("as_glycan_composition works for simple compositions", {
  chars <- c("H5N2", "H5N4S1F1", "H5N4A1G1")
  expected <- glycan_composition(
    c(Hex = 5, HexNAc = 2),
    c(Hex = 5, HexNAc = 4, NeuAc = 1, dHex = 1),
    c(Hex = 5, HexNAc = 4, NeuAc = 1, NeuGc = 1)
  )
  expect_equal(as_glycan_composition(chars), expected)
})

test_that("as_glycan_composition works for E and L", {
  chars <- c("H5N4E1", "H5N4L1", "H5N4E1L1")
  expected <- glycan_composition(
    c(Hex = 5, HexNAc = 4, NeuAc = 1),
    c(Hex = 5, HexNAc = 4, NeuAc = 1),
    c(Hex = 5, HexNAc = 4, NeuAc = 2)
  )
  expect_snapshot(comps <- as_glycan_composition(chars)) # should warn about E/L ambiguity
  expect_equal(comps, expected)
})

# This test is replaced by tests for NA handling in character casting above

test_that("as_glycan_composition works for empty characters", {
  expect_equal(as_glycan_composition(character()), glycan_composition())
})

test_that("as_glycan_composition rejects empty strings", {
  chars <- c("", "Hex(5)HexNAc(2)")
  expect_error(as_glycan_composition(chars))
})

test_that("as_glycan_composition raises error for illegal characters", {
  chars <- c("invalid", "Hex(5)HexNAc(2)")
  err_msg <- "Characters cannot be parsed as glycan compositions at index 1"
  expect_error(as_glycan_composition(chars), err_msg)
})

test_that("as_glycan_composition reorder residues", {
  chars <- c("Hex(2)HexNAc(1)", "HexNAc(1)Hex(2)")
  comp <- as_glycan_composition(chars)
  expected <- c("Hex(2)HexNAc(1)", "Hex(2)HexNAc(1)")
  expect_equal(format(comp), expected)
})

test_that("as_glycan_composition works for duplications", {
  comp <- as_glycan_composition("Hex(2)Hex(3)HexNAc(1)HexNAc(2)")
  expect_equal(as.character(comp), "Hex(5)HexNAc(3)")
})

test_that("as.character works for compositions", {
  chars <- c("Hex(2)HexNAc(1)", "Hex(5)HexNAc(2)")
  comp <- as_glycan_composition(chars)
  expect_equal(as.character(comp), chars)
})

# Tests for NA handling in constructor ---------------------------------

test_that("glycan_composition accepts NULL to create NA", {
  comp <- glycan_composition(NULL)
  expect_equal(length(comp), 1)
  expect_true(is.na(comp))
})

test_that("glycan_composition accepts NA to create NA", {
  comp <- glycan_composition(NA)
  expect_equal(length(comp), 1)
  expect_true(is.na(comp))
})

test_that("glycan_composition handles mixed valid and NA", {
  comp <- glycan_composition(c(Hex = 5), NULL, c(Hex = 3))
  expect_equal(length(comp), 3)
  expect_false(is.na(comp[1]))
  expect_true(is.na(comp[2]))
  expect_false(is.na(comp[3]))
})

test_that("glycan_composition rejects named NA (likely typo)", {
  # c(Hex = NA) is likely a typo for c(Hex = 5), should error
  expect_error(glycan_composition(c(Hex = NA)), "positive")
})

test_that(".is_na_composition_elem detects NA correctly", {
  expect_true(.is_na_composition_elem(NULL))
  expect_false(.is_na_composition_elem(c(Hex = 5)))
  expect_false(.is_na_composition_elem(integer(0)))
})

test_that(".valid_composition_element validates correctly", {
  # Valid input
  expect_equal(
    .valid_composition_element(c(Hex = 5, HexNAc = 2)),
    c(Hex = 5L, HexNAc = 2L)
  )

  # Invalid input - unnamed
  expect_error(.valid_composition_element(c(5, 2)), "named integer")

  # Invalid input - empty
  expect_error(.valid_composition_element(integer(0)), "at least one residue")

  # Invalid input - unknown monosaccharide
  expect_error(
    .valid_composition_element(c(Unknown = 5)),
    "known monosaccharides"
  )

  # Invalid input - non-positive
  expect_error(.valid_composition_element(c(Hex = 0)), "positive numbers")
})

# Tests for NA handling in vec_restore --------------------------------

test_that("vec_restore skips NA in type checking", {
  comp1 <- glycan_composition(c(Hex = 5, HexNAc = 2))
  comp2 <- glycan_composition(NULL)

  # This should not error - NA should be skipped in type check
  combined <- c(comp1, comp2)
  expect_equal(length(combined), 2)
  expect_false(is.na(combined[1]))
  expect_true(is.na(combined[2]))
})

test_that("combining with NA at beginning works", {
  comp1 <- glycan_composition(c(Hex = 5, HexNAc = 2))
  combined <- c(NA, comp1)
  expect_equal(length(combined), 2)
  expect_true(is.na(combined[1]))
  expect_false(is.na(combined[2]))
})

# Tests for NA handling in character casting --------------------------------

test_that("as_glycan_composition handles NA characters", {
  chars <- c("Hex(5)HexNAc(2)", NA)
  comp <- as_glycan_composition(chars)
  expect_equal(length(comp), 2)
  expect_false(is.na(comp[1]))
  expect_true(is.na(comp[2]))
})

test_that("as_glycan_composition handles NA at beginning", {
  chars <- c(NA, "Hex(5)HexNAc(2)")
  comp <- as_glycan_composition(chars)
  expect_equal(length(comp), 2)
  expect_true(is.na(comp[1]))
  expect_false(is.na(comp[2]))
})

test_that("as_glycan_composition handles all NA characters", {
  chars <- c(NA, NA)
  comp <- as_glycan_composition(chars)
  expect_equal(length(comp), 2)
  expect_true(is.na(comp[1]))
  expect_true(is.na(comp[2]))
})

test_that("as_glycan_composition reports correct index with NA and invalid chars", {
  # Regression test: error message should report original indices, not filtered indices
  # When mixing NA and invalid strings, index 3 (not 2) should be reported
  chars <- c("Hex(5)", NA, "invalid")
  expect_error(
    as_glycan_composition(chars),
    "at index 3"
  )
})

test_that("as_glycan_composition reports correct index with NA at beginning and invalid", {
  chars <- c(NA, NA, "HexNAc(1)", "also_invalid")
  expect_error(
    as_glycan_composition(chars),
    "at index 4"
  )
})

# Tests for NA handling in format -----------------------------------------

test_that("format shows <NA> for NA compositions", {
  comp <- c(glycan_composition(c(Hex = 5)), NA)
  expect_equal(format(comp), c("Hex(5)", "<NA>"))
})

test_that("as.character shows <NA> for NA compositions", {
  comp <- c(glycan_composition(c(Hex = 5)), NA)
  expect_equal(as.character(comp), c("Hex(5)", "<NA>"))
})

test_that("print handles NA compositions", {
  comp <- c(glycan_composition(c(Hex = 5)), NA, glycan_composition(c(Hex = 3)))
  output <- capture.output(print(comp))
  expect_true(any(grepl("<NA>", output)))
})

test_that("tibble printing handles NA compositions", {
  comp <- c(glycan_composition(c(Hex = 5)), NA)
  tibble <- tibble::tibble(comp = comp, a = 1:2)
  output <- capture.output(print(tibble))
  expect_true(any(grepl("<NA>", output)))
})

# ===== NA Support Comprehensive Tests =====

test_that("is.na returns correct logical for compositions with NA", {
  comp <- glycan_composition(c(Hex = 5))
  expect_equal(is.na(comp), FALSE)

  comp_na <- c(glycan_composition(c(Hex = 5)), NA)
  expect_equal(is.na(comp_na), c(FALSE, TRUE))

  comp_all_na <- glycan_composition(NA_integer_, NA_integer_)
  expect_equal(is.na(comp_all_na), c(TRUE, TRUE))
})

test_that("as.logical preserves NA semantics", {
  # NA compositions should become NA, not TRUE
  comp <- glycan_composition(c(Hex = 5))
  expect_equal(as.logical(comp), FALSE)

  comp_na <- c(glycan_composition(c(Hex = 5)), NA)
  expect_equal(as.logical(comp_na), c(FALSE, NA))

  comp_all_na <- glycan_composition(NA_integer_, NA_integer_)
  expect_equal(as.logical(comp_all_na), c(NA, NA))
})

test_that("anyNA detects NA compositions", {
  comp <- glycan_composition(c(Hex = 5))
  expect_false(anyNA(comp))

  comp_na <- c(glycan_composition(c(Hex = 5)), NA)
  expect_true(anyNA(comp_na))
})

test_that("rep handles compositions with NA", {
  comp <- c(glycan_composition(c(Hex = 5)), NA)
  repeated <- rep(comp, 2)
  expect_equal(length(repeated), 4)
  expect_equal(is.na(repeated), c(FALSE, TRUE, FALSE, TRUE))
})

test_that("rep handles compositions with NA at beginning", {
  # Note: c(NA, comp) doesn't preserve type in base R, use vctrs::vec_c for reliable behavior
  comp <- glycan_composition(c(Hex = 5))
  comps <- vctrs::vec_c(NA, comp)
  repeated <- rep(comps, 2)
  expect_equal(length(repeated), 4)
  expect_equal(is.na(repeated), c(TRUE, FALSE, TRUE, FALSE))
})

test_that("subsetting preserves NA", {
  comp <- c(glycan_composition(c(Hex = 5)), NA, glycan_composition(c(Hex = 3)))

  expect_true(is.na(comp[2]))
  expect_equal(is.na(comp[c(1, 3)]), c(FALSE, FALSE))
  expect_equal(length(comp[c(1, 3)]), 2)
})

test_that("as_glycan_composition handles list with NULL", {
  result <- as_glycan_composition(list(c(Hex = 5), NULL))
  expect_equal(length(result), 2)
  expect_false(is.na(result[1]))
  expect_true(is.na(result[2]))
})

test_that("combining multiple compositions with NA", {
  comp1 <- glycan_composition(c(Hex = 5, HexNAc = 2))
  comp2 <- glycan_composition(c(Hex = 3, HexNAc = 1))

  # Note: c(NA, comp) doesn't preserve type, so we use composition first
  combined <- c(comp1, NA, comp2, NA)
  expect_equal(length(combined), 4)
  expect_false(is.na(combined[1]))
  expect_true(is.na(combined[2]))
  expect_false(is.na(combined[3]))
  expect_true(is.na(combined[4]))
})

test_that("empty composition vector with NA works", {
  comps <- c(NA_character_, NA)
  result <- as_glycan_composition(comps)
  expect_equal(length(result), 2)
  expect_true(is.na(result[1]))
  expect_true(is.na(result[2]))
})

test_that("format preserves order with mixed NA and valid", {
  comp <- glycan_composition(c(Hex = 2, HexNAc = 1))
  # Use composition first, then NA - c(NA, comp) doesn't preserve type in base R
  comps <- c(comp, NA, glycan_composition(c(Hex = 3)), NA)
  formatted <- format(comps)
  # Valid elements show as composition string, NA elements show as "<NA>"
  expect_equal(formatted[1], "Hex(2)HexNAc(1)")
  expect_equal(formatted[2], "<NA>")
  expect_equal(formatted[3], "Hex(3)")
  expect_equal(formatted[4], "<NA>")
})

test_that("NA compositions are correctly restored after subsetting", {
  comps <- c(glycan_composition(c(Hex = 5)), NA, glycan_composition(c(Hex = 3)))
  subset <- comps[c(1, 3)]
  expect_equal(length(subset), 2)
  expect_false(is.na(subset[1]))
  expect_false(is.na(subset[2]))
})

test_that("combining concrete compositions with NA preserves type", {
  comps <- c(glycan_composition(c(Glc = 5, Gal = 2)), NA)
  expect_s3_class(comps, "glyrepr_composition")
  expect_false(is.na(comps[1]))
  expect_true(is.na(comps[2]))
  expect_equal(format(comps[1]), "Glc(5)Gal(2)")
})

#' Parse IUPAC-condensed string to glycan structure
#'
#' Internal functions for parsing IUPAC-condensed strings into igraph objects.
#' This supports the as_glycan_structure.character method.
#'
#' @param x A single IUPAC-condensed string
#' @return An igraph object representing the glycan structure
#' @keywords internal
.parse_iupac_condensed_single <- function(x) {
  if (is.na(x) || nchar(x) == 0 || stringr::str_detect(x, "^\\s*$")) {
    cli::cli_abort("Cannot parse empty or NA IUPAC-condensed string.")
  }

  tryCatch(
    {
      # Validate input string - no leading/trailing whitespace
      if (stringr::str_detect(x, "^\\s+|\\s+$")) {
        cli::cli_abort(
          "IUPAC-condensed string cannot have leading or trailing whitespace"
        )
      }

      # Validate no internal whitespace
      if (stringr::str_detect(x, "\\s")) {
        cli::cli_abort("IUPAC-condensed string cannot contain whitespace")
      }

      # Validate proper bracket matching
      if (!.validate_brackets(x)) {
        cli::cli_abort("Malformed brackets in IUPAC-condensed string")
      }
      anomer <- .extract_anomer(x)
      x <- stringr::str_sub(x, 1, -stringr::str_length(anomer) - 3)

      tokens <- .tokenize_iupac(x)

      # Require anomer information - no longer auto-supplement
      first_mono_sub_res <- .extract_substituent(tokens[[1]])

      # Create a new graph and add the first node
      graph <- igraph::make_empty_graph()
      graph <- igraph::add_vertices(
        graph,
        1,
        name = "1",
        mono = first_mono_sub_res[["mono"]],
        sub = first_mono_sub_res[["sub"]]
      )

      if (length(tokens) == 1) {
        graph <- igraph::set_edge_attr(graph, "linkage", value = character(0))
        graph$anomer <- anomer
        return(graph)
      }

      # Iterate over the tokens
      node_stack <- rstackdeque::rstack()
      node_stack <- rstackdeque::insert_top(node_stack, 1)
      current_node_id <- 1
      for (token in tokens[2:length(tokens)]) {
        if (token == "[") {
          node_stack <- rstackdeque::insert_top(node_stack, current_node_id)
        } else if (token == "]") {
          current_node_id <- rstackdeque::peek_top(node_stack)
          node_stack <- rstackdeque::without_top(node_stack)
        } else {
          parsed_token <- .parse_token(token)
          new_node_id <- igraph::vcount(graph) + 1
          graph <- igraph::add_vertices(
            graph,
            1,
            name = as.character(new_node_id),
            mono = parsed_token[["mono"]],
            sub = parsed_token[["sub"]]
          )
          graph <- igraph::add_edges(
            graph,
            c(current_node_id, new_node_id),
            linkage = parsed_token[["linkage"]]
          )
          current_node_id <- new_node_id
        }
      }

      graph$anomer <- anomer
      return(graph)
    },
    error = function(e) {
      cli::cli_abort(c(
        "Could not parse IUPAC-condensed string: {.val {x}}",
        "i" = conditionMessage(e)
      ))
    }
  )
}

# Validate bracket matching
.validate_brackets <- function(x) {
  # Check for balanced brackets
  open_count <- stringr::str_count(x, "\\[")
  close_count <- stringr::str_count(x, "\\]")

  if (open_count != close_count) {
    return(FALSE)
  }

  # Check for proper nesting
  depth <- 0
  chars <- stringr::str_split(x, "")[[1]]
  for (char in chars) {
    if (char == "[") {
      depth <- depth + 1
    } else if (char == "]") {
      depth <- depth - 1
      if (depth < 0) {
        return(FALSE)
      }
    }
  }

  return(depth == 0)
}

# Extract anomer from IUPAC condensed string
.extract_anomer <- function(iupac) {
  # e.g. "Neu5Ac(a2-" -> "a2"  (ending anomer specification)
  p <- "\\(([ab\\?][12\\?])-$"
  if (stringr::str_detect(iupac, p)) {
    # Check if it's a complete IUPAC string ending with anomer specification
    # This should be allowed for standalone monosaccharides like "Neu5Ac(a2-"
    stringr::str_extract(iupac, p, group = 1)
  } else {
    cli::cli_abort(c(
      "Can't extract anomer information.",
      "i" = "Anomer information is required for the reducing-end monosaccharide.",
      "i" = "For example, use 'Man(a1-' instead of 'Man'."
    ))
  }
}


# Tokenize IUPAC condensed string
.tokenize_iupac <- function(iupac) {
  # anomer is either "a", "b", or "?".
  anomer_p <- "[ab\\?]"

  # pos1 is either "1", "2", or "?".
  pos1_p <- "([12\\?])"

  # pos2 is "1"-"9" (followed by any number of "/1-9"), or "?"
  pos2_p <- "([1-9](/[1-9])*|\\?)"

  # Linkage pattern
  linkage_pattern <- stringr::str_glue("{anomer_p}{pos1_p}-{pos2_p}")

  # Monosaccharide name pattern (including potential substituents)
  # Allow letters, digits, and ? for substituents like "Man?S", "Glc3Me6S", etc.
  # Substituents are directly concatenated in IUPAC format, no commas
  mono_pattern <- "[A-Za-z][A-Za-z0-9\\?]*"
  mono_linkage_pattern <- stringr::str_glue(
    "{mono_pattern}(\\({linkage_pattern}\\))?"
  )

  # The pattern is either a monosaccharide name or a bracket
  pattern <- paste(mono_linkage_pattern, "\\[", "\\]", sep = "|")

  tokens <- stringr::str_extract_all(iupac, pattern)[[1]]

  # Check if we extracted the full string
  extracted_string <- paste(tokens, collapse = "")
  if (extracted_string != iupac) {
    cli::cli_abort("Invalid characters or format in IUPAC-condensed string")
  }

  tokens <- stringr::str_replace(tokens, "\\[", "TEMP_LEFT")
  tokens <- stringr::str_replace(tokens, "\\]", "TEMP_RIGHT")
  tokens <- stringr::str_replace(tokens, "TEMP_LEFT", "\\]")
  tokens <- stringr::str_replace(tokens, "TEMP_RIGHT", "\\[")

  # Reverse the tokens to make the first monosaccharide the reducing end
  rev(tokens)
}

# Parse a single token
.parse_token <- function(token) {
  left_bracket_pos <- stringr::str_locate(token, "\\(")[1]

  if (is.na(left_bracket_pos)) {
    cli::cli_abort("Missing linkage information in token")
  }

  mono <- stringr::str_sub(token, 1, left_bracket_pos - 1)
  mono_sub_res <- .extract_substituent(mono)
  mono <- mono_sub_res[["mono"]]
  sub <- mono_sub_res[["sub"]]
  linkage <- stringr::str_sub(token, left_bracket_pos + 1, -2)

  # Validate linkage format
  if (!.validate_linkage(linkage)) {
    cli::cli_abort(paste0("Invalid linkage format: ", linkage))
  }

  c(mono = mono, sub = sub, linkage = linkage)
}

# Validate linkage format
.validate_linkage <- function(linkage) {
  # Empty linkage
  if (nchar(linkage) == 0) {
    return(FALSE)
  }

  # Pattern for valid linkage: anomer + pos1 + - + pos2
  # anomer: a, b, or ?
  # pos1: 1, 2, or ?
  # pos2: 1-9 (optionally followed by /1-9 patterns), or ?
  valid_pattern <- "^[ab\\?][12\\?]-([1-9](/[1-9])*|\\?)$"

  return(stringr::str_detect(linkage, valid_pattern))
}

# Extract substituent from monosaccharide name
.extract_substituent <- function(mono) {
  subs_pattern <- stringr::str_c(available_substituents(), collapse = "|")
  single_sub_pattern <- stringr::str_glue("[1-9\\?]({subs_pattern})") # Pattern for a single substituent

  # Handle different types of monosaccharides
  result <- if (stringr::str_starts(mono, "Neu")) {
    # Handle all Neu-based monosaccharides
    .handle_neu_monosaccharide(mono, single_sub_pattern)
  } else {
    # Handle non-Neu monosaccharides
    .handle_general_monosaccharide(mono, single_sub_pattern)
  }

  # Validate that the monosaccharide is known
  if (!is_known_monosaccharide(result[["mono"]])) {
    cli::cli_abort(paste0("Unknown monosaccharide: ", result[["mono"]]))
  }

  result
}

# Handle Neu-based monosaccharides with substituents
# This function determines the correct base monosaccharide (Neu5Ac, Neu5Gc, or Neu)
# based on the presence of 5Ac or 5Gc substituents
.handle_neu_monosaccharide <- function(mono, single_sub_pattern) {
  # Check for conflicting 5Ac5Gc pattern
  if (stringr::str_detect(mono, "5Gc") && stringr::str_detect(mono, "5Ac")) {
    cli::cli_abort(
      "Monosaccharide cannot have both 5Ac and 5Gc substituents: {mono}"
    )
  }

  # Handle all Neu variants containing 5Ac
  if (stringr::str_detect(mono, "Neu.*5Ac")) {
    return(.handle_neu5ac_variant(mono, single_sub_pattern))
  }

  # Handle all Neu variants containing 5Gc
  if (stringr::str_detect(mono, "Neu.*5Gc")) {
    return(.handle_neu5gc_variant(mono, single_sub_pattern))
  }

  # Handle other Neu-based monosaccharides (no 5Ac or 5Gc)
  .handle_general_monosaccharide(mono, single_sub_pattern)
}

# Handle general (non-Neu) monosaccharides with substituents
.handle_general_monosaccharide <- function(mono, single_sub_pattern) {
  # Try to find all substituents in the monosaccharide name
  # Substituents are directly concatenated in IUPAC, e.g., "Glc3Me6S"
  all_subs <- stringr::str_extract_all(mono, single_sub_pattern)[[1]]

  if (length(all_subs) > 0) {
    # Remove all substituents from the mono name to get the base monosaccharide
    clean_mono <- mono
    for (sub in all_subs) {
      clean_mono <- stringr::str_remove(clean_mono, stringr::fixed(sub))
    }

    # Sort substituents by position
    sub_string <- .sort_substituents(all_subs)

    c(mono = clean_mono, sub = sub_string)
  } else {
    c(mono = mono, sub = "")
  }
}

# Helper function to sort substituents by position
.sort_substituents <- function(subs) {
  if (length(subs) == 0) {
    return("")
  }

  # Extract positions for sorting
  positions <- purrr::map_chr(subs, ~ stringr::str_extract(.x, "^[1-9\\?]"))
  numeric_positions <- purrr::map_dbl(positions, function(pos) {
    if (pos == "?") Inf else as.numeric(pos)
  })

  # Sort substituents by position
  sorted_indices <- order(numeric_positions)
  sorted_subs <- subs[sorted_indices]

  stringr::str_c(sorted_subs, collapse = ",")
}

# Handle all Neu variants containing 5Ac
.handle_neu5ac_variant <- function(mono, single_sub_pattern) {
  .handle_neu_with_marker(mono, single_sub_pattern, "5Ac", "Neu5Ac")
}

# Handle all Neu variants containing 5Gc
.handle_neu5gc_variant <- function(mono, single_sub_pattern) {
  .handle_neu_with_marker(mono, single_sub_pattern, "5Gc", "Neu5Gc")
}

# Generic helper function for handling Neu variants with specific markers
.handle_neu_with_marker <- function(
  mono,
  single_sub_pattern,
  marker,
  base_mono
) {
  # Remove the marker from the monosaccharide name to get the remaining part
  mono_without_marker <- stringr::str_remove(mono, marker)

  # Extract all substituents from the remaining part using normal logic
  all_subs <- stringr::str_extract_all(
    mono_without_marker,
    single_sub_pattern
  )[[1]]

  # Sort substituents by position
  sub_string <- if (length(all_subs) > 0) {
    .sort_substituents(all_subs)
  } else {
    ""
  }

  c(mono = base_mono, sub = sub_string)
}

#' Available Substituents
#'
#' Get the available substituents for monosaccharides.
#'
#' @returns A character vector.
#'
#' @examples
#' available_substituents()
#'
#' @export
available_substituents <- function() {
  c("Me", "Ac", "NAc", "P", "S", "Pyr", "PC", "PPEtn", "PEtn", "N")
}

#' Normalize Substituent String
#'
#' Takes a substituent string (potentially with multiple substituents) and
#' returns a normalized string with substituents sorted by position.
#'
#' @param sub A character string representing substituents, e.g., "4Ac,3Me" or "6S"
#' @returns A character string with substituents sorted by position, e.g., "3Me,4Ac"
#'
#' @examples
#' normalize_substituents("4Ac,3Me")  # Returns "3Me,4Ac"
#' normalize_substituents("6S")       # Returns "6S"
#' normalize_substituents("")         # Returns ""
#' @noRd
normalize_substituents <- function(sub) {
  checkmate::assert_character(sub, len = 1)

  if (sub == "") {
    return("")
  }

  # Split by commas
  individual_subs <- stringr::str_split(sub, ",")[[1]]

  # Remove any empty strings (in case of double commas)
  individual_subs <- individual_subs[individual_subs != ""]

  if (length(individual_subs) == 0) {
    return("")
  }

  # Extract positions for sorting
  positions <- purrr::map_chr(
    individual_subs,
    ~ stringr::str_extract(.x, "^[\\d\\?]")
  )

  # Convert to numeric for sorting (? becomes Inf)
  numeric_positions <- purrr::map_dbl(positions, function(pos) {
    if (pos == "?") Inf else as.numeric(pos)
  })

  # Sort and combine
  sorted_indices <- order(numeric_positions)
  sorted_subs <- individual_subs[sorted_indices]

  stringr::str_c(sorted_subs, collapse = ",")
}

#' Remove All Substituents from a Glycan
#'
#' This function replaces all substituents in a glycan structure with empty strings.
#'
#' @param glycan A glyrepr_structure vector.
#'
#' @returns A glyrepr_structure vector with all substituents removed.
#'
#' @examples
#' (glycan <- o_glycan_core_1())
#' remove_substituents(glycan)
#'
#' @export
remove_substituents <- function(glycan) {
  if (!is_glycan_structure(glycan)) {
    cli::cli_abort(c(
      "Input must be a glyrepr_structure vector.",
      "i" = "Use `glycan_structure()` to create a glyrepr_structure from igraph objects."
    ))
  }

  smap_structure(glycan, .remove_substituents_single)
}

# Internal function to remove substituents from a single igraph
.remove_substituents_single <- function(glycan) {
  igraph::set_vertex_attr(glycan, "sub", value = "")
}

# Tests for substituent functions

test_that("available_substituents returns expected values", {
  subs <- available_substituents()
  expect_type(subs, "character")
  expect_true(length(subs) > 0)
  expect_true("Me" %in% subs)
  expect_true("Ac" %in% subs)
  expect_true("S" %in% subs)
})

test_that("normalize_substituents handles empty strings", {
  expect_equal(normalize_substituents(""), "")
})

test_that("normalize_substituents handles single substituents", {
  expect_equal(normalize_substituents("6S"), "6S")
  expect_equal(normalize_substituents("3Me"), "3Me")
  expect_equal(normalize_substituents("?Ac"), "?Ac")
})

test_that("normalize_substituents sorts multiple substituents by position", {
  expect_equal(normalize_substituents("4Ac,3Me"), "3Me,4Ac")
  expect_equal(normalize_substituents("6P,2S,4Ac"), "2S,4Ac,6P")
  expect_equal(normalize_substituents("9Ac,1Me,5S"), "1Me,5S,9Ac")
})

test_that("normalize_substituents handles ? positions correctly", {
  expect_equal(normalize_substituents("?S,3Me"), "3Me,?S")
  expect_equal(normalize_substituents("?Ac,?Me"), "?Ac,?Me") # Original order preserved for same position
  expect_equal(normalize_substituents("4Ac,?S,2Me"), "2Me,4Ac,?S")
})

test_that("normalize_substituents removes empty components", {
  expect_equal(normalize_substituents("3Me,,4Ac"), "3Me,4Ac")
  expect_equal(normalize_substituents(",6S,"), "6S")
})

test_that("normalize_substituents input validation", {
  expect_error(normalize_substituents(c("3Me", "4Ac")))
  expect_error(normalize_substituents(123))
})

test_that("remove_substituents works on glycan structures", {
  # Create a test glycan with substituents
  graph <- igraph::make_graph(~ 1 - +2)
  igraph::V(graph)$mono <- c("Glc", "Gal")
  igraph::V(graph)$sub <- c("3Me,4Ac", "6S")
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"
  graph$alditol <- FALSE

  glycan <- glycan_structure(graph)

  # Remove substituents
  clean_glycan <- remove_substituents(glycan)

  # Check that substituents are removed
  clean_graph <- get_structure_graphs(clean_glycan, return_list = FALSE)
  expect_equal(igraph::V(clean_graph)$sub, c("", ""))
  expect_equal(igraph::V(clean_graph)$mono, c("Gal", "Glc")) # Monos should be unchanged
})

test_that("remove_substituents input validation", {
  expect_error(remove_substituents("not_a_glycan"))
  expect_error(remove_substituents(123))
})

#' Create a Glycan Structure Vector
#'
#' @description
#' `glycan_structure()` creates an efficient glycan structure vector for storing and
#' processing glycan molecular structures. The function employs hash-based deduplication
#' mechanisms, making it suitable for glycoproteomics, glycomics analysis, and glycan
#' structure comparison studies.
#'
#' @details
#' # Data Structure Overview
#'
#' A glycan structure vector is a vctrs record with an additional S3 class
#' `glyrepr_structure`. Therefore, `sloop::s3_class()` returns the class hierarchy
#' `c("glyrepr_structure", "vctrs_rcrd")`.
#'
#' Each glycan structure must satisfy the following constraints:
#'
#' ## Graph Structure Requirements
#' - Must be a directed graph with an outward tree structure (reducing end as root)
#' - Must have a graph attribute `anomer` in the format "a1" or "b1"
#'   - Unknown parts can be represented with "?", e.g., "?1", "a?", "??"
#'
#' ## Node Attributes
#' - `mono`: Monosaccharide names, must be known monosaccharide types
#'   - Generic names: Hex, HexNAc, dHex, NeuAc, etc.
#'   - Concrete names: Glc, Gal, Man, GlcNAc, etc.
#'   - Cannot mix generic and concrete names
#'   - NA values are not allowed
#' - `sub`: Substituent information
#'   - Single substituent format: "xY" (x = position, Y = substituent name),
#'     e.g., "2Ac", "3S"
#'   - Multiple substituents separated by commas and ordered by position,
#'     e.g., "3Me,4Ac", "2S,6P"
#'   - No substituents represented by empty string ""
#'
#' ## Edge Attributes
#' - `linkage`: Glycosidic linkage information in format "a/bX-Y"
#'   - Standard format: e.g., "b1-4", "a2-3"
#'   - Unknown positions allowed: "a1-?", "b?-3", "??-?"
#'   - Partially unknown positions: "a1-3/6", "a1-3/6/9"
#'   - NA values are not allowed
#'
#' # Node and Edge Order
#'
#' The indices of vertices and linkages in a glycan correspond directly to their
#' order in the IUPAC-condensed string, which is printed when you print a
#' [glyrepr::glycan_structure()].
#' For example, for the glycan `Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-`,
#' the vertices are "Man", "Man", "Man", "GlcNAc", "GlcNAc",
#' and the linkages are "a1-3", "a1-6", "b1-4", "b1-4".
#'
#' # NA Support
#'
#' Glycan structure vectors support NA values for representing missing or
#' unknown structures:
#'
#' - Create with `glycan_structure(NA)` or `glycan_structure(NULL)`
#' - Combine with valid structures: `c(struct1, NA, struct2)`
#' - Convert from character: `as_glycan_structure(c("Glc(a1-", NA))`
#' - `smap` functions skip NA elements gracefully
#' - `is.na()` returns `TRUE` for NA elements
#'
#' # Naming Support
#'
#' Glycan structure vectors can have names, which are preserved during operations.
#' This is particularly useful when working with the `glymotif` package.
#'
#' # As characters
#'
#' One side-effect of the current implementation is that you can treat a glycan structure vector
#' as a pure character vector of IUPAC-condensed strings.
#' In fact, `is.character()` returns `TRUE` for a glycan structure vector,
#' and all `stringr` functions work directly on the vector.
#'
#' However, we still recommend using `as.character()` to explicitly convert to character when needed,
#' to avoid confusion and ensure that the intended behavior is clear.
#'
#' @param ... igraph graph objects to be converted to glycan structures, or existing
#'   glycan structure vectors. Supports mixed input of multiple objects.
#' @param x An object to check or convert.
#'
#' @returns A `glyrepr_structure` class glycan structure vector object.
#'
#' @examples
#' library(igraph)
#'
#' # Example 1: Create a simple glycan structure GlcNAc(b1-4)GlcNAc
#' graph <- make_graph(~ 1-+2)  # Create graph with two monosaccharides
#' V(graph)$mono <- c("GlcNAc", "GlcNAc")  # Set monosaccharide types
#' V(graph)$sub <- ""  # No substituents
#' E(graph)$linkage <- "b1-4"  # b1-4 glycosidic linkage
#' graph$anomer <- "a1"  # a anomeric carbon
#'
#' # Create glycan structure vector
#' simple_struct <- glycan_structure(graph)
#' print(simple_struct)
#'
#' # Example 2: Use predefined glycan core structures
#' n_core <- n_glycan_core()  # N-glycan core structure
#' o_core1 <- o_glycan_core_1()  # O-glycan Core 1 structure
#'
#' # Example 3: Create complex structure with substituents
#' complex_graph <- make_graph(~ 1-+2-+3)
#' V(complex_graph)$mono <- c("GlcNAc", "Gal", "Neu5Ac")
#' V(complex_graph)$sub <- c("", "", "")  # Add substituents as needed
#' E(complex_graph)$linkage <- c("b1-4", "a2-3")
#' complex_graph$anomer <- "b1"
#'
#' complex_struct <- glycan_structure(complex_graph)
#' print(complex_struct)
#'
#' # Example 4: Check if object is a glycan structure
#' is_glycan_structure(simple_struct)  # TRUE
#' is_glycan_structure(graph)          # FALSE
#'
#' @importFrom magrittr %>%
#' @export
glycan_structure <- function(...) {
  args <- list(...)

  # Handle different input types
  graphs <- list()
  iupacs <- character()
  na_positions <- logical()

  for (i in seq_along(args)) {
    arg <- args[[i]]
    if (is.null(arg) || (is.atomic(arg) && length(arg) == 1 && is.na(arg))) {
      # Track NA position
      iupacs <- c(iupacs, NA_character_)
      na_positions <- c(na_positions, TRUE)
    } else if (inherits(arg, "igraph")) {
      graphs <- c(graphs, list(arg))
      iupacs <- c(iupacs, NA_character_) # placeholder
      na_positions <- c(na_positions, FALSE)
    } else {
      cli::cli_abort("All arguments must be igraph objects or NA values.")
    }
  }

  if (length(iupacs) == 0) {
    return(new_glycan_structure())
  }

  # Get indices of valid (non-NA) positions
  valid_idx <- which(!na_positions)

  if (length(valid_idx) == 0 && all(na_positions)) {
    # All are NA
    return(new_glycan_structure(rep(NA_character_, length(iupacs)), list()))
  }

  if (length(valid_idx) == 0) {
    return(new_glycan_structure(character(), list()))
  }

  # Extract valid graphs
  valid_graphs <- graphs

  # Validate and process each valid graph
  processed_graphs <- purrr::map(valid_graphs, function(graph) {
    checkmate::assert_class(graph, "igraph")
    graph %>%
      validate_single_glycan_structure() %>%
      ensure_name_vertex_attr()
  })

  # Reorder the graphs to align with IUPAC-style sequence
  reordered_result <- reorder_graphs_with_indices(processed_graphs)
  reordered_graphs <- reordered_result$graphs
  reorder_indices <- reordered_result$indices

  # Validate that all structures have the same mono_type
  validate_glycan_structure_vector(reordered_graphs)

  # Use IUPAC codes directly as data for the glycan_structure vctrs vector
  processed_iupacs <- purrr::map_chr(
    reordered_graphs,
    .structure_to_iupac_single
  )

  # Create a unique list based on uniqueness of IUPAC codes for structures storage
  unique_indices <- which(!duplicated(processed_iupacs))
  unique_graphs <- reordered_graphs[unique_indices]
  unique_iupacs <- processed_iupacs[unique_indices]
  names(unique_graphs) <- unique_iupacs

  # Build final result - replace placeholders with actual IUPACs
  # Map reordered positions back to original positions
  for (i in seq_along(reorder_indices)) {
    orig_pos <- reorder_indices[i] # Original position in valid_graphs
    final_pos <- valid_idx[orig_pos] # Final position in result vector
    iupac <- processed_iupacs[i]
    iupacs[final_pos] <- iupac
  }

  new_glycan_structure(iupacs, unique_graphs)
}

# Helper function to validate a single glycan structure
validate_single_glycan_structure <- function(glycan) {
  checkmate::assert_class(glycan, "igraph")

  # Check if it is a directed graph
  if (!is_directed_graph(glycan)) {
    cli::cli_abort("Glycan structure must be directed.")
  }

  # Check if it is an out tree
  if (!is_out_tree(glycan)) {
    cli::cli_abort("Glycan structure must be an out tree.")
  }

  # Check if graph has a vertex attribute "mono"
  # This is the monosaccharide name, e.g. "GlcNAc", "Man", etc.
  if (!has_vertex_attrs(glycan, "mono")) {
    cli::cli_abort("Glycan structure must have a vertex attribute 'mono'.")
  }

  # Check if no NA in vertex attribute "mono"
  mono_names <- igraph::vertex_attr(glycan, "mono")
  if (any(is.na(mono_names))) {
    cli::cli_abort(
      "Glycan structure must have no NA in vertex attribute 'mono'."
    )
  }

  # Check if all monosaccharides are known
  if (!all(is_known_mono(mono_names))) {
    unknown_monos <- unique(igraph::V(glycan)$mono[
      !is_known_mono(igraph::V(glycan)$mono)
    ])
    msg <- glue::glue(
      "Unknown monosaccharide: {stringr::str_c(unknown_monos, collapse = ', ')}"
    )
    cli::cli_abort(msg, monos = unknown_monos)
  }

  # Check if mixed use of generic and concrete monosaccharides
  if (mix_generic_concrete(mono_names)) {
    cli::cli_abort(
      "Monosaccharides must be either all generic or all concrete."
    )
  }

  # Check if graph has a vertex attribute "sub"
  # This is the substituent name, e.g. "Ac", "S", "P", or "" (no).
  if (!has_vertex_attrs(glycan, "sub")) {
    cli::cli_abort("Glycan structure must have a vertex attribute 'sub'.")
  }

  # Check if no NA in vertex attribute "sub"
  subs <- igraph::vertex_attr(glycan, "sub")
  if (any(is.na(subs))) {
    cli::cli_abort(
      "Glycan structure must have no NA in vertex attribute 'sub'."
    )
  }

  # Check if all substituents are valid
  if (!all(valid_substituent(subs))) {
    invalid_subs <- unique(subs[!valid_substituent(subs)])
    msg <- glue::glue(
      "Unknown substituent: {stringr::str_c(invalid_subs, collapse = ', ')}"
    )
    cli::cli_abort(msg, subs = invalid_subs)
  }

  # Check if graph has an edge attribute "linkage"
  if (!has_edge_attrs(glycan, "linkage")) {
    cli::cli_abort("Glycan structure must have an edge attribute 'linkage'.")
  }

  # Check if no NA in edge attribute "linkage"
  linkages <- igraph::edge_attr(glycan, "linkage")
  if (any(is.na(linkages))) {
    cli::cli_abort(
      "Glycan structure must have no NA in edge attribute 'linkage'."
    )
  }

  # Check if all linkages are valid
  if (!all(valid_linkages(linkages))) {
    invalid_linkages <- unique(linkages[!valid_linkages(linkages)])
    msg <- glue::glue(
      "Invalid linkage: {stringr::str_c(invalid_linkages, collapse = ', ')}"
    )
    cli::cli_abort(msg, linkages = invalid_linkages)
  }

  # Check if any duplicated linkage positions exist
  if (any_dup_linkage_pos(glycan)) {
    cli::cli_abort("Duplicated linkage positions.")
  }

  # Check if "anomer" attribute exists
  if (is.null(glycan$anomer)) {
    cli::cli_abort("Glycan structure must have a graph attribute 'anomer'.")
  }

  # Check if "anomer" attribute is valid
  if (!valid_anomer(glycan$anomer)) {
    cli::cli_abort(glue::glue("Invalid anomer: {glycan$anomer}"))
  }

  glycan
}

#' Validate that all structures in a vector have the same monosaccharide type
#'
#' This function checks that all glycan structures in a vector have the same
#' monosaccharide type (either all generic or all concrete). Mixed types within
#' a single vector are not allowed.
#'
#' @param graphs A list of igraph graph objects representing glycan structures.
#' @param label A label for error messages (e.g., "Vector 1").
#' @returns Invisible NULL. Throws an error if validation fails.
#' @noRd
validate_glycan_structure_vector <- function(graphs, label = NULL) {
  # Skip if empty or single graph
  if (length(graphs) <= 1) {
    return(invisible(NULL))
  }

  # Get mono_type for each graph
  mono_types <- purrr::map_chr(graphs, get_graph_mono_type)

  # Check for mixed types within individual graphs
  if (any(mono_types == "mixed")) {
    cli::cli_abort(c(
      "All structures must have a single monosaccharide type.",
      "x" = "{.val {label}} contains structures with mixed generic and concrete monosaccharides."
    ))
  }

  # Check that all structures have the same mono_type
  unique_types <- unique(mono_types)
  if (length(unique_types) > 1) {
    concrete_count <- sum(mono_types == "concrete")
    generic_count <- sum(mono_types == "generic")

    if (is.null(label)) {
      cli::cli_abort(c(
        "All structures must have the same monosaccharide type.",
        "x" = "Found {.val {concrete_count}} concrete and {.val {generic_count}} generic structure(s) in the same vector.",
        "i" = "Use {.fn convert_to_generic} to convert concrete structures to generic type."
      ))
    } else {
      cli::cli_abort(c(
        "All structures must have the same monosaccharide type.",
        "x" = "{.val {label}} has mixed types: {.val {concrete_count}} concrete and {.val {generic_count}} generic structure(s)."
      ))
    }
  }

  invisible(NULL)
}

#' Helper function to create a new glycan structure vector
#' @param iupac A character vector of IUPAC-condensed strings.
#' @param graphs A list of igraph objects.
#' @returns A glyrepr_structure object (a vctr object)
#' @noRd
new_glycan_structure <- function(iupac = character(), graphs = list()) {
  vctrs::new_vctr(iupac, graphs = graphs, class = "glyrepr_structure")
}

#' Create a missing glycan structure vector
#'
#' @param n Number of missing elements.
#' @returns A `glyrepr_structure` vector containing only missing values.
#' @noRd
new_na_glycan_structure <- function(n) {
  new_glycan_structure(rep(NA_character_, n), list())
}

#' Get the missing-value mask for a glycan structure vector
#'
#' @param x A `glyrepr_structure` vector.
#' @returns A logical vector.
#' @noRd
structure_na_mask <- function(x) {
  is.na(vctrs::vec_data(x))
}

#' Keep only graphs used by non-missing structure codes
#'
#' @param iupacs Character vector of structure codes.
#' @param graphs Named list of structure graphs.
#' @returns A named list of graphs.
#' @noRd
filter_used_structure_graphs <- function(iupacs, graphs) {
  used_codes <- unique(iupacs[!is.na(iupacs)])
  used_graphs <- graphs[used_codes]
  used_graphs[!vapply(used_graphs, is.null, logical(1))]
}

ensure_name_vertex_attr <- function(glycan) {
  if (!("name" %in% igraph::vertex_attr_names(glycan))) {
    names <- as.character(seq_len(igraph::vcount(glycan)))
    glycan <- igraph::set_vertex_attr(glycan, "name", value = names)
  }
  glycan
}

#' @export
#' @rdname glycan_structure
is_glycan_structure <- function(x) {
  inherits(x, "glyrepr_structure")
}

#' @export
is.na.glyrepr_structure <- function(x, ...) {
  iupacs <- vctrs::vec_data(x)
  is.na(iupacs)
}

#' @export
vec_ptype_full.glyrepr_structure <- function(x, ...) "glycan_structure"

#' @export
vec_ptype_abbr.glyrepr_structure <- function(x, ...) "struct"

#' @export
format.glyrepr_structure <- function(x, ...) {
  formatted <- format(vctrs::vec_data(x), ...)

  # Add names if present
  nms <- names(x)
  if (!is.null(nms) && length(nms) > 0) {
    # Use tab separation between name and structure
    formatted <- paste(nms, formatted, sep = "\t")
  }

  formatted
}

#' Format a Subset of Glycan Structures with Optional Colors
#'
#' @param x A glyrepr_structure object
#' @param indices Indices of structures to format
#' @param colored A logical value indicating whether to add colors
#' @returns A character vector of formatted structures for the specified indices
#' @keywords internal
format_glycan_structure_subset <- function(x, indices, colored = TRUE) {
  if (!colored) {
    return(format(x)[indices])
  }

  codes <- vctrs::vec_data(x)[indices]
  graphs <- attr(x, "graphs")

  # For each structure, add colors if concrete type
  purrr::map_chr(codes, function(code) {
    # Handle NA codes
    if (is.na(code)) {
      return(NA_character_)
    }
    structure <- graphs[[code]]
    mono_names <- igraph::V(structure)$mono

    # Add colors to monosaccharides and gray linkages
    if (colored) {
      colorize_iupac_string(code, mono_names)
    } else {
      code
    }
  })
}


#' @export
obj_print_footer.glyrepr_structure <- function(x, ...) {
  cat(
    "# Unique structures: ",
    format(length(attr(x, "graphs"))),
    "\n",
    sep = ""
  )
}

#' @export
obj_print_data.glyrepr_structure <- function(
  x,
  ...,
  max_n = 10,
  colored = TRUE
) {
  if (length(x) == 0) {
    return()
  }

  n <- length(x)
  n_show <- min(n, max_n)

  # Only format the structures that need to be shown to improve performance
  indices_to_show <- seq_len(n_show)
  formatted <- format_glycan_structure_subset(
    x,
    indices_to_show,
    colored = colored
  )

  # Check if names are present
  nms <- names(x)
  has_names <- !is.null(nms) && length(nms) > 0

  # Print each IUPAC structure on its own line with indexing, up to max_n
  for (i in seq_len(n_show)) {
    if (has_names) {
      cat("[", i, "] ", nms[i], "\t", formatted[i], "\n", sep = "")
    } else {
      cat("[", i, "] ", formatted[i], "\n", sep = "")
    }
  }
  if (n > max_n) {
    cat("... (", n - max_n, " more not shown)\n", sep = "")
  }
}

#' @importFrom pillar pillar_shaft
#' @export
pillar_shaft.glyrepr_structure <- function(x, ...) {
  if (length(x) == 0) {
    return(pillar::pillar_shaft(character()))
  }

  # Get formatted strings with colors
  codes <- vctrs::vec_data(x)
  graphs <- attr(x, "graphs")

  # For each structure, add colors if concrete type
  formatted <- purrr::map_chr(codes, function(code) {
    # Handle NA codes
    if (is.na(code)) {
      return(NA_character_)
    }
    structure <- graphs[[code]]
    mono_names <- igraph::V(structure)$mono

    # Add colors to monosaccharides and gray linkages
    colorize_iupac_string(code, mono_names)
  })

  pillar::new_pillar_shaft_simple(formatted, align = "left", min_width = 10)
}

#' @export
vec_ptype2.glyrepr_structure.glyrepr_structure <- function(x, y, ...) {
  # Get graphs from both vectors (works for both empty prototypes and full vectors)
  graphs_x <- attr(x, "graphs")
  graphs_y <- attr(y, "graphs")

  # Validate each vector separately
  validate_glycan_structure_vector(graphs_x, label = "Vector 1")
  validate_glycan_structure_vector(graphs_y, label = "Vector 2")

  # Check that both vectors have the same mono_type
  if (length(graphs_x) > 0 && length(graphs_y) > 0) {
    mono_types_x <- purrr::map_chr(graphs_x, get_graph_mono_type)
    mono_types_y <- purrr::map_chr(graphs_y, get_graph_mono_type)
    unique_types_x <- unique(mono_types_x)
    unique_types_y <- unique(mono_types_y)

    if (
      length(unique_types_x) > 0 &&
        length(unique_types_y) > 0 &&
        unique_types_x[[1]] != unique_types_y[[1]]
    ) {
      concrete_count_x <- sum(mono_types_x == "concrete")
      generic_count_x <- sum(mono_types_x == "generic")
      concrete_count_y <- sum(mono_types_y == "concrete")
      generic_count_y <- sum(mono_types_y == "generic")

      cli::cli_abort(c(
        "All structures must have the same monosaccharide type.",
        "x" = "Vector 1 has {.val {concrete_count_x}} concrete and {.val {generic_count_x}} generic structure(s).",
        "x" = "Vector 2 has {.val {concrete_count_y}} concrete and {.val {generic_count_y}} generic structure(s).",
        "i" = "Use {.fn convert_to_generic} to convert concrete structures to generic type."
      ))
    }
  }

  # Combine graphs from both vectors (union by IUPAC name as key)
  combined_graphs <- c(graphs_x, graphs_y)
  # Remove duplicates, keeping first occurrence (from x)
  unique_graphs <- combined_graphs[!duplicated(names(combined_graphs))]

  # Create prototype with combined graphs
  out <- new_glycan_structure()
  attr(out, "graphs") <- unique_graphs
  out
}

#' @export
vec_cast.glyrepr_structure.glyrepr_structure <- function(x, to, ...) {
  x
}

#' @export
vec_cast.glyrepr_structure.igraph <- function(x, to, ...) {
  glycan_structure(x)
}

#' @export
vec_cast.glyrepr_structure.list <- function(x, to, ...) {
  if (!all(purrr::map_lgl(x, ~ inherits(.x, "igraph")))) {
    cli::cli_abort(c(
      "All elements in the list must be igraph objects.",
      "i" = "Each graph in the list should be a valid glycan structure."
    ))
  }
  do.call(glycan_structure, x)
}

#' @export
vec_cast.glyrepr_structure.character <- function(x, to, ...) {
  # Handle empty character vector
  if (length(x) == 0) {
    return(glycan_structure())
  }

  # Handle NA values
  na_mask <- is.na(x)

  if (all(na_mask)) {
    # All NA - return vector of NAs with empty graphs
    result <- new_na_glycan_structure(length(x))
    names(result) <- names(x)
    return(result)
  }

  if (any(na_mask)) {
    # Mixed NA and non-NA
    non_na_x <- x[!na_mask]

    # Parse non-NA characters
    graphs <- purrr::map(non_na_x, .parse_iupac_condensed_single)

    # Create structure for non-NA elements
    iupacs <- purrr::map_chr(graphs, .structure_to_iupac_single)

    # Create unique graphs
    unique_indices <- which(!duplicated(iupacs))
    unique_graphs <- graphs[unique_indices]
    unique_iupacs <- iupacs[unique_indices]
    names(unique_graphs) <- unique_iupacs

    # Build result preserving NA positions
    result_iupacs <- character(length(x))
    result_iupacs[na_mask] <- NA_character_

    non_na_positions <- which(!na_mask)
    for (i in seq_along(non_na_positions)) {
      pos <- non_na_positions[i]
      result_iupacs[pos] <- iupacs[i]
    }

    result <- new_glycan_structure(result_iupacs, unique_graphs)

    # Restore names
    names(result) <- names(x)

    return(result)
  }

  # Original logic for non-NA case
  input_names <- names(x)
  graphs <- purrr::map(x, .parse_iupac_condensed_single)
  result <- do.call(glycan_structure, graphs)
  names(result) <- input_names
  result
}

#' @export
vec_cast.character.glyrepr_structure <- function(x, to, ...) {
  vctrs::vec_data(x)
}

# ===== IMPORTANT NOTE =====
# `vec_restore.glyrepr_structure()` and `[.glyrepr_structure()`
# are implemented by Claude Code.
# I do NOT fully understand the code, but it works.

#' @export
vec_restore.glyrepr_structure <- function(x, to, ...) {
  # Get the graphs attribute from the prototype
  graphs <- attr(to, "graphs")

  # If prototype has no graphs, return with empty graphs
  if (length(graphs) == 0) {
    out <- vctrs::new_vctr(x, graphs = list(), class = "glyrepr_structure")
    return(out)
  }

  # Get IUPAC codes from the data being restored
  iupacs <- vctrs::vec_data(x)

  # If x is empty (e.g., during vec_ptype2), keep all graphs from prototype
  if (length(iupacs) == 0) {
    out <- vctrs::new_vctr(x, graphs = graphs, class = "glyrepr_structure")
    return(out)
  }

  # Filter graphs to only include those used in the subset
  # Use unique iupacs to handle duplicates correctly
  used_graphs <- filter_used_structure_graphs(iupacs, graphs)

  out <- vctrs::new_vctr(x, graphs = used_graphs, class = "glyrepr_structure")
  out
}

#' @export
`[.glyrepr_structure` <- function(x, i, ...) {
  # Call the default subsetting behavior
  out <- NextMethod("[")
  # Filter graphs to only include those used in the subset
  iupacs <- vctrs::vec_data(out)
  graphs <- attr(out, "graphs")
  # If result is empty, return with empty graphs
  if (length(iupacs) == 0) {
    attr(out, "graphs") <- list()
    return(out)
  }
  if (length(graphs) > 0) {
    used_graphs <- filter_used_structure_graphs(iupacs, graphs)
    attr(out, "graphs") <- used_graphs
  }
  out
}

#' @export
`[[<-.glyrepr_structure` <- function(x, i, value) {
  cli::cli_abort(c(
    "Cannot use `[[<-` on {.cls glyrepr_structure} vectors.",
    "x" = "This operation would create an invalid object with mismatched data and graphs.",
    "i" = "Create a new vector instead, e.g., with `c()`."
  ))
}

#' Convert to Glycan Structure Vector
#'
#' Convert an object to a glycan structure vector.
#'
#' @param x An object to convert to a glycan structure vector.
#'   Can be an igraph object, a list of igraph objects,
#'   a character vector of IUPAC-condensed strings,
#'   or an existing glyrepr_structure object.
#'
#' @returns A glyrepr_structure object.
#'
#' @examples
#' library(igraph)
#'
#' # Convert a single igraph
#' graph <- make_graph(~ 1-+2)
#' V(graph)$mono <- c("GlcNAc", "GlcNAc")
#' V(graph)$sub <- ""
#' E(graph)$linkage <- "b1-4"
#' graph$anomer <- "a1"
#' as_glycan_structure(graph)
#'
#' # Convert a list of igraphs
#' o_glycan_vec <- o_glycan_core_1()
#' o_glycan_graph <- get_structure_graphs(o_glycan_vec)
#' as_glycan_structure(list(graph, o_glycan_graph))
#'
#' # Convert a character vector of IUPAC-condensed strings
#' as_glycan_structure(c("GlcNAc(b1-4)GlcNAc(b1-", "Man(a1-2)GlcNAc(b1-"))
#'
#' @export
as_glycan_structure <- function(x) {
  vctrs::vec_cast(x, glycan_structure())
}

#' Access Individual Glycan Structures
#'
#' Extract individual glycan structure graphs from a glycan structure vector.
#'
#' @param x A glycan structure vector.
#' @param return_list If `TRUE`, always returns a list.
#'   If `FALSE` and `x` has a length of 1, return the igraph object directly.
#'   If not provided (default), `FALSE` when `x` has a length of 1 and `TRUE` otherwise.
#'
#' @returns A list of igraph objects or an igraph object directly (see `return_list` parameter).
#'
#' @examples
#' structures <- c(o_glycan_core_1(), n_glycan_core())
#' get_structure_graphs(structures)
#' get_structure_graphs(structures)
#'
#' @export
get_structure_graphs <- function(x, return_list = NULL) {
  checkmate::assert_class(x, "glyrepr_structure")
  checkmate::assert_flag(return_list, null.ok = TRUE)

  if (is.null(return_list)) {
    return_list <- length(x) > 1
  } else {
    if (!return_list && length(x) > 1) {
      cli::cli_abort(c(
        "{.arg return_list} must be `TRUE` or `NULL` if {.arg x} has a length greater than 1.",
        "i" = "Length of {.arg x}: {.val {length(x)}}"
      ))
    }
  }

  iupacs <- vctrs::vec_data(x)
  graphs <- attr(x, "graphs")
  res <- purrr::map(iupacs, ~ graphs[[.x]])
  if (!return_list) {
    res <- res[[1]]
  }
  res
}

test_that("as_glycan_structure.character parses simple IUPAC-condensed strings", {
  # Single monosaccharide
  glycan1 <- as_glycan_structure("Man(?1-")
  expect_s3_class(glycan1, "glyrepr_structure")
  expect_equal(length(glycan1), 1)
  expect_equal(structure_to_iupac(glycan1), "Man(?1-")

  # Two monosaccharides
  glycan2 <- as_glycan_structure("Gal(b1-3)GalNAc(?1-")
  expect_s3_class(glycan2, "glyrepr_structure")
  expect_equal(length(glycan2), 1)
  expect_equal(structure_to_iupac(glycan2), "Gal(b1-3)GalNAc(?1-")

  # With explicit anomer
  glycan3 <- as_glycan_structure("Gal(b1-3)GalNAc(a1-")
  expect_s3_class(glycan3, "glyrepr_structure")
  expect_equal(structure_to_iupac(glycan3), "Gal(b1-3)GalNAc(a1-")
})

test_that("as_glycan_structure.character parses branched structures", {
  # Simple branched structure
  iupac <- "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(?1-"
  glycan <- as_glycan_structure(iupac)
  expect_s3_class(glycan, "glyrepr_structure")
  expect_equal(
    structure_to_iupac(glycan),
    "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(?1-"
  )
})

test_that("as_glycan_structure.character handles substituents", {
  # With substituents
  iupac <- "Man3S(a1-2)Gal6Ac(?1-"
  glycan <- as_glycan_structure(iupac)
  expect_s3_class(glycan, "glyrepr_structure")

  # Check that substituents are preserved
  graph <- get_structure_graphs(glycan, return_list = FALSE)
  expect_equal(igraph::V(graph)$sub, c("3S", "6Ac"))
})

test_that("as_glycan_structure.character handles Neu5Ac correctly", {
  # Neu5Ac with explicit anomer
  glycan1 <- as_glycan_structure("Neu5Ac(?2-")
  expect_equal(structure_to_iupac(glycan1), "Neu5Ac(?2-")

  # Neu5Ac with explicit anomer
  glycan2 <- as_glycan_structure("Neu5Ac(a2-")
  expect_equal(structure_to_iupac(glycan2), "Neu5Ac(a2-")

  # Neu5Ac with substituent
  glycan3 <- as_glycan_structure("Neu5Ac9Ac(?2-")
  graph <- get_structure_graphs(glycan3, return_list = FALSE)
  expect_equal(igraph::V(graph)$mono, "Neu5Ac")
  expect_equal(igraph::V(graph)$sub, "9Ac")
})


test_that("as_glycan_structure.character handles unknown linkages", {
  # Unknown linkages
  iupac <- "Man(a1-?)Man(?1-3)Man(?1-"
  glycan <- as_glycan_structure(iupac)
  expect_s3_class(glycan, "glyrepr_structure")
  expect_equal(structure_to_iupac(glycan), "Man(a1-?)Man(?1-3)Man(?1-")
})

test_that("as_glycan_structure.character handles multiple linkages", {
  # Multiple linkages
  iupac <- "Neu5Ac(a2-3/6)Gal(?1-"
  glycan <- as_glycan_structure(iupac)
  expect_s3_class(glycan, "glyrepr_structure")
  expect_equal(structure_to_iupac(glycan), "Neu5Ac(a2-3/6)Gal(?1-")
})

test_that("as_glycan_structure.character works with vectors", {
  # Multiple IUPAC strings
  iupacs <- c("Man(?1-", "Gal(b1-3)GalNAc(?1-", "Neu5Ac(a2-")
  glycans <- as_glycan_structure(iupacs)
  expect_s3_class(glycans, "glyrepr_structure")
  expect_equal(length(glycans), 3)

  # Check each one
  expect_equal(structure_to_iupac(glycans)[1], "Man(?1-")
  expect_equal(structure_to_iupac(glycans)[2], "Gal(b1-3)GalNAc(?1-")
  expect_equal(structure_to_iupac(glycans)[3], "Neu5Ac(a2-")
})

test_that("as_glycan_structure.character handles complex O-glycan", {
  # Complex O-glycan structure
  iupac <- "Neu5Ac(a2-3)Gal(b1-4)[Fuc(a1-3)]GlcNAc(b1-6)[Neu5Ac(a2-3)Gal(b1-3)]GalNAc(?1-"
  glycan <- as_glycan_structure(iupac)
  expect_s3_class(glycan, "glyrepr_structure")
  expect_equal(length(glycan), 1)

  # Check the structure has the correct number of nodes
  graph <- get_structure_graphs(glycan, return_list = FALSE)
  expect_equal(igraph::vcount(graph), 7) # 7 monosaccharides
})

test_that("as_glycan_structure.character error handling", {
  # Empty string
  expect_error(as_glycan_structure(""), "Cannot parse empty")

  # NA is treated as a missing structure
  expect_true(is.na(as_glycan_structure(NA_character_)))

  # Invalid format - unknown monosaccharide
  expect_error(as_glycan_structure("invalid_format"), "Could not parse")

  # Missing anomer information
  expect_error(as_glycan_structure("Man"), "Can't extract anomer information")
  expect_error(
    as_glycan_structure("Neu5Ac"),
    "Can't extract anomer information"
  )
  expect_error(
    as_glycan_structure("Gal(b1-3)GalNAc"),
    "Can't extract anomer information"
  )
})

test_that("as_glycan_structure.character round-trip consistency", {
  # Test round-trip: structure -> IUPAC -> structure
  original_structures <- c(
    o_glycan_core_1(),
    n_glycan_core()
  )

  for (i in seq_along(original_structures)) {
    # Convert to IUPAC
    iupac <- structure_to_iupac(original_structures[[i]])

    # Parse back
    parsed <- as_glycan_structure(iupac)

    # Should be identical
    expect_equal(structure_to_iupac(parsed), iupac)
  }
})

# Edge cases and boundary conditions ----------------------------------------

test_that("as_glycan_structure.character handles whitespace", {
  # Leading and trailing whitespace should be handled gracefully
  expect_error(as_glycan_structure(" Man "), "Could not parse")
  expect_error(as_glycan_structure("  "), "Cannot parse empty")
  expect_error(as_glycan_structure("\t\n"), "Cannot parse empty")
})

test_that("as_glycan_structure.character handles malformed brackets", {
  # Unmatched brackets
  expect_error(as_glycan_structure("Man[Gal"), "Could not parse")
  expect_error(as_glycan_structure("Man]Gal"), "Could not parse")
  expect_error(as_glycan_structure("Man[[Gal]]"), "Could not parse")
  expect_error(as_glycan_structure("Man(a1-3)[Gal"), "Could not parse")
})

test_that("as_glycan_structure.character handles malformed linkages", {
  # Empty parentheses
  expect_error(as_glycan_structure("Man()Gal"), "Could not parse")

  # Incomplete linkages
  expect_error(as_glycan_structure("Man(a1)Gal"), "Could not parse")
  expect_error(as_glycan_structure("Man(-3)Gal"), "Could not parse")

  # Invalid characters in linkages
  expect_error(as_glycan_structure("Man(c1-3)Gal"), "Could not parse")
  expect_error(as_glycan_structure("Man(a0-3)Gal"), "Could not parse")
  expect_error(as_glycan_structure("Man(a3-3)Gal"), "Could not parse")
  expect_error(as_glycan_structure("Man(a1-0)Gal"), "Could not parse")
  expect_error(as_glycan_structure("Man(a#-3)Gal"), "Could not parse")
})

test_that("as_glycan_structure.character handles invalid monosaccharide names", {
  # Too short
  expect_error(as_glycan_structure("M"), "Could not parse")
  expect_error(as_glycan_structure("G"), "Could not parse")

  # Numbers at start
  expect_error(as_glycan_structure("1Man"), "Could not parse")
  expect_error(as_glycan_structure("2Gal"), "Could not parse")

  # Special characters
  expect_error(as_glycan_structure("Man@"), "Could not parse")
  expect_error(as_glycan_structure("Gal#"), "Could not parse")
  expect_error(as_glycan_structure("Man-Gal"), "Could not parse")

  # Case sensitivity (monosaccharide names are case sensitive)
  expect_error(as_glycan_structure("man"), "Could not parse")
  expect_error(as_glycan_structure("MAN"), "Could not parse")
  expect_error(as_glycan_structure("gal"), "Could not parse")
})

test_that("as_glycan_structure.character handles complex Neu variants", {
  # Test special Neu variants from glyparse
  glycan1 <- as_glycan_structure("Neu4Ac5Ac(?2-")
  graph1 <- get_structure_graphs(glycan1, return_list = FALSE)
  expect_equal(igraph::V(graph1)$mono, "Neu5Ac")
  expect_equal(igraph::V(graph1)$sub, "4Ac")

  glycan2 <- as_glycan_structure("Neu4Ac5Gc(?2-")
  graph2 <- get_structure_graphs(glycan2, return_list = FALSE)
  expect_equal(igraph::V(graph2)$mono, "Neu5Gc")
  expect_equal(igraph::V(graph2)$sub, "4Ac")

  # Neu5Ac should not be split even though it matches substituent pattern
  glycan3 <- as_glycan_structure("Neu5Ac(?2-")
  graph3 <- get_structure_graphs(glycan3, return_list = FALSE)
  expect_equal(igraph::V(graph3)$mono, "Neu5Ac")
  expect_equal(igraph::V(graph3)$sub, "")
})

test_that("as_glycan_structure.character handles invalid substituents", {
  # Invalid substituent positions
  expect_error(as_glycan_structure("Man0Ac"), "Could not parse")
  expect_error(as_glycan_structure("ManAAc"), "Could not parse")
  expect_error(as_glycan_structure("Man1X"), "Could not parse") # X is not a valid substituent
})

test_that("as_glycan_structure.character handles deeply nested structures", {
  # Very deep nesting
  deep_structure <- "Man(a1-2)[Man(a1-4)[Man(a1-4)[Man(a1-5)Man(a1-2)]Man(a1-3)]Man(a1-4)]Man(a1-"
  glycan <- as_glycan_structure(deep_structure)
  expect_s3_class(glycan, "glyrepr_structure")

  # Verify the graph has correct number of nodes
  graph <- get_structure_graphs(glycan, return_list = FALSE)
  expect_equal(igraph::vcount(graph), 8)
})

test_that("as_glycan_structure.character handles mixed valid/invalid in vectors", {
  # Vector with some valid and some invalid strings
  mixed_vector <- c("Man", "invalid_mono", "Gal")
  expect_error(as_glycan_structure(mixed_vector), "Could not parse")

  # All invalid
  all_invalid <- c("invalid1", "invalid2")
  expect_error(as_glycan_structure(all_invalid), "Could not parse")
})


test_that("as_glycan_structure.character handles extreme linkage cases", {
  # Maximum valid positions
  glycan1 <- as_glycan_structure("Man(a2-9)Gal(?1-")
  expect_s3_class(glycan1, "glyrepr_structure")

  glycan2 <- as_glycan_structure("Man(b1-1)Gal(?1-")
  expect_s3_class(glycan2, "glyrepr_structure")

  # Complex multi-position linkages
  glycan3 <- as_glycan_structure("Man(a1-2/3/4/5/6)Gal(?1-")
  expect_s3_class(glycan3, "glyrepr_structure")
  expect_equal(structure_to_iupac(glycan3), "Man(a1-2/3/4/5/6)Gal(?1-")
})

test_that("as_glycan_structure.character handles single character edge cases", {
  # Very short but invalid inputs
  expect_error(as_glycan_structure("("), "Could not parse")
  expect_error(as_glycan_structure(")"), "Could not parse")
  expect_error(as_glycan_structure("["), "Could not parse")
  expect_error(as_glycan_structure("]"), "Could not parse")
  expect_error(as_glycan_structure("-"), "Could not parse")
})

test_that("as_glycan_structure.character handles unusual anomer cases", {
  # Various anomer edge cases
  glycan1 <- as_glycan_structure("Neu5Ac(?2-")
  expect_equal(structure_to_iupac(glycan1), "Neu5Ac(?2-")

  glycan2 <- as_glycan_structure("Man(??-")
  expect_equal(structure_to_iupac(glycan2), "Man(??-")

  glycan3 <- as_glycan_structure("Fuc(?1-")
  expect_equal(structure_to_iupac(glycan3), "Fuc(?1-")
})

test_that("as_glycan_structure.character preserves complex substituent patterns", {
  # Test various substituent combinations
  glycan1 <- as_glycan_structure("Gal6S(b1-3)GlcNAc4S(?1-")
  graph1 <- get_structure_graphs(glycan1, return_list = FALSE)
  expect_equal(sort(igraph::V(graph1)$sub), sort(c("4S", "6S")))

  # Unknown position substituents
  glycan2 <- as_glycan_structure("Man?S(?1-")
  graph2 <- get_structure_graphs(glycan2, return_list = FALSE)
  expect_equal(igraph::V(graph2)$sub, "?S")
})

test_that("as_glycan_structure.character handles multiple substituents", {
  glycan <- as_glycan_structure("Glc3Me6S(a1-")
  graph <- get_structure_graphs(glycan, return_list = FALSE)
  expect_equal(igraph::V(graph)$sub, "3Me,6S")
})

test_that("Neu monosaccharides with 5Ac are correctly parsed as Neu5Ac", {
  # Test cases where 5Ac should result in Neu5Ac base monosaccharide
  expect_equal(
    .extract_substituent("Neu3Me5Ac"),
    c(mono = "Neu5Ac", sub = "3Me")
  )
  expect_equal(
    .extract_substituent("Neu4Ac5Ac"),
    c(mono = "Neu5Ac", sub = "4Ac")
  )
  expect_equal(
    .extract_substituent("Neu4Ac5Ac9Ac"),
    c(mono = "Neu5Ac", sub = "4Ac,9Ac")
  )
  expect_equal(.extract_substituent("Neu7S5Ac"), c(mono = "Neu5Ac", sub = "7S"))
})

test_that("Neu monosaccharides with 5Gc are correctly parsed as Neu5Gc", {
  # Test cases where 5Gc should result in Neu5Gc base monosaccharide
  expect_equal(
    .extract_substituent("Neu3Me5Gc"),
    c(mono = "Neu5Gc", sub = "3Me")
  )
  expect_equal(
    .extract_substituent("Neu4Ac5Gc"),
    c(mono = "Neu5Gc", sub = "4Ac")
  )
  expect_equal(.extract_substituent("Neu7S5Gc"), c(mono = "Neu5Gc", sub = "7S"))
})

test_that("Neu monosaccharides without 5Ac or 5Gc remain as Neu", {
  # Test cases where no 5Ac or 5Gc should result in Neu base monosaccharide
  expect_equal(.extract_substituent("Neu"), c(mono = "Neu", sub = ""))
  expect_equal(.extract_substituent("Neu7Ac"), c(mono = "Neu", sub = "7Ac"))
  expect_equal(
    .extract_substituent("Neu3Me7Ac"),
    c(mono = "Neu", sub = "3Me,7Ac")
  )
})

test_that("Neu5Ac and Neu5Gc exact matches work correctly", {
  # Test exact matches
  expect_equal(.extract_substituent("Neu5Ac"), c(mono = "Neu5Ac", sub = ""))
  expect_equal(.extract_substituent("Neu5Gc"), c(mono = "Neu5Gc", sub = ""))
})

test_that("Neu5Ac and Neu5Gc with additional substituents work correctly", {
  # Test Neu5Ac/Neu5Gc with additional substituents
  expect_equal(
    .extract_substituent("Neu5Ac9Ac"),
    c(mono = "Neu5Ac", sub = "9Ac")
  )
  expect_equal(
    .extract_substituent("Neu5Gc9Ac"),
    c(mono = "Neu5Gc", sub = "9Ac")
  )
  expect_equal(
    .extract_substituent("Neu5Ac7S9Ac"),
    c(mono = "Neu5Ac", sub = "7S,9Ac")
  )
})

test_that("Error is thrown for monosaccharides with both 5Ac and 5Gc", {
  # This should be an error case
  expect_error(
    .extract_substituent("Neu5Ac5Gc"),
    "cannot have both 5Ac and 5Gc"
  )
  expect_error(
    .extract_substituent("Neu5Gc5Ac"),
    "cannot have both 5Ac and 5Gc"
  )
})

test_that("Full IUPAC parsing works with corrected Neu substituents", {
  # Test full IUPAC parsing
  result1 <- suppressMessages(as_glycan_structure(
    "Neu3Me5Ac(a2-3)Gal(b1-4)Glc(a1-"
  ))
  expect_true(is_glycan_structure(result1))

  result2 <- suppressMessages(as_glycan_structure(
    "Neu4Ac5Gc(a2-3)Gal(b1-4)Glc(a1-"
  ))
  expect_true(is_glycan_structure(result2))

  # Check that the output contains the correct monosaccharide names
  expect_match(as.character(result1), "Neu5Ac3Me")
  expect_match(as.character(result2), "Neu5Gc4Ac")
})

test_that("as_glycan_structure.character makes order of vertices and edges consistent with IUPAC-condensed", {
  glycan <- as_glycan_structure(
    "GlcNAc(b1-2)Man(a1-3)[GlcNAc(b1-2)Man(a1-6)]Man(b1-4)GlcNAc(b1-4)[Fuc(a1-3)]GlcNAc(b1-"
  )
  graph <- get_structure_graphs(glycan, return_list = FALSE)

  expect_equal(
    igraph::V(graph)$mono,
    c("GlcNAc", "Man", "GlcNAc", "Man", "Man", "GlcNAc", "Fuc", "GlcNAc")
  )
  expect_equal(
    igraph::E(graph)$linkage,
    c("b1-2", "a1-3", "b1-2", "a1-6", "b1-4", "b1-4", "a1-3")
  )
})

# ========= has_linkages() =========
test_that("has_linkages() works for glycans with intact linkages", {
  glycan <- o_glycan_core_1(linkage = TRUE)
  expect_true(has_linkages(glycan))
})

test_that("has_linkages() works for glycan with partial linkages", {
  glycan_vec <- o_glycan_core_2(linkage = TRUE)
  glycan <- get_structure_graphs(glycan_vec, return_list = FALSE) # Extract the igraph
  glycan <- igraph::set_edge_attr(glycan, "linkage", value = c("??-?", "b1-6"))

  # Create a new vectorized structure with the modified graph
  modified_glycan_vec <- glycan_structure(glycan)
  expect_true(has_linkages(modified_glycan_vec))
})

test_that("has_linkages() works for glycans without linkages", {
  glycan <- o_glycan_core_1(linkage = FALSE)
  expect_false(has_linkages(glycan))
})

test_that("has_linkages() works in strict mode", {
  glycan <- as_glycan_structure("Gal(b1-?)GalNAc(a1-")
  expect_false(has_linkages(glycan, strict = TRUE))
  glycan <- as_glycan_structure("Gal(b1-3)GalNAc(a1-")
  expect_true(has_linkages(glycan, strict = TRUE))
})

test_that("has_linkages() considers anomers", {
  glycans <- as_glycan_structure(c(
    "Gal(b1-3)GalNAc(a1-",
    "Gal(a1-3)GalNAc(?1-"
  ))
  expect_equal(has_linkages(glycans, strict = TRUE), c(TRUE, FALSE))

  glycan <- as_glycan_structure("Gal(??-?)GalNAc(a1-")
  expect_true(has_linkages(glycan))
})

# ========= possible_linkages() =========
# Tests for possible_linkages() function
test_that("possible_linkages() works for a?-2", {
  result <- possible_linkages("a?-2")
  expected <- c("a1-2", "a2-2")
  expect_identical(result, expected)
})

test_that("possible_linkages() works for a1-?", {
  result <- possible_linkages("a1-?")
  expected <- paste0("a1-", 1:9)
  expect_identical(result, expected)
})

test_that("possible_linkages() works for ?1-3", {
  result <- possible_linkages("?1-3")
  expected <- c("a1-3", "b1-3")
  expect_identical(result, expected)
})

test_that("possible_linkages() works for b?-?", {
  result <- possible_linkages("b?-?")
  df <- expand.grid(1:2, 1:9)
  expected <- apply(df, 1, function(x) paste0("b", x[1], "-", x[2]))
  expect_identical(result, expected)
})

test_that("possible_linkages() works for a2-3|6", {
  result <- possible_linkages("a2-3/6")
  expected <- c("a2-3", "a2-6")
  expect_identical(result, expected)
})

test_that("possible_linkages() works for a1-3", {
  expect_identical(possible_linkages("a1-3"), "a1-3")
})

test_that("possible_linkages() rejects wrong format", {
  expect_error(possible_linkages("a1-3-4"), "Invalid linkage format")
  expect_error(possible_linkages(""), "Invalid linkage format")
})

test_that("possible_linkages() rejects wrong input type", {
  expect_error(possible_linkages(1))
  expect_error(possible_linkages(NA))
  expect_error(possible_linkages(NULL))
})

test_that("possible_linkages() rejects multiple linkages", {
  expect_error(possible_linkages(c("a1-3", "b1-4")))
})

test_that("possible_linkages() works with custom ranges", {
  result <- possible_linkages(
    "??-?",
    anomer_range = "a",
    pos1_range = 2,
    pos2_range = c(3, 6)
  )
  expected <- c("a2-3", "a2-6")
  expect_identical(result, expected)
})

test_that("possible_linkages() rejects custom ranges with wrong types", {
  expect_error(possible_linkages("??-?", anomer_range = 1))
  expect_error(possible_linkages("??-?", pos1_range = "a"))
  expect_error(possible_linkages("??-?", pos2_range = "a"))
})

test_that("possible_linkages() works with 0 range", {
  expect_equal(possible_linkages("??-?", pos1_range = integer(0)), character(0))
})

test_that("possible_linkages() works with include unknown", {
  result <- possible_linkages("?1-6", include_unknown = TRUE)
  expected <- c("a1-6", "b1-6", "?1-6")
  expect_identical(result, expected)
})

# ========= remove_linkages() =========
test_that("remove_linkages() works", {
  glycan <- o_glycan_core_2(linkage = TRUE)
  glycan <- remove_linkages(glycan)
  graph <- get_structure_graphs(glycan, return_list = FALSE)
  expect_equal(igraph::E(graph)$linkage, c("??-?", "??-?"))
  expect_equal(igraph::graph_attr(graph, "anomer"), "??")
})

# Tests for glycan structure functions

good_glycan_graph <- function() {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$name <- as.character(1:igraph::vcount(graph))
  igraph::V(graph)$mono <- c("Glc", "Gal", "Glc")
  igraph::V(graph)$sub <- c("", "", "")
  igraph::E(graph)$linkage <- c("b1-4", "b1-4")
  graph$anomer <- "a1"
  graph
}


# Tests for glycan_structure --------------------------------------------------

test_that("glycan_structure works", {
  glycan <- glycan_structure(good_glycan_graph())
  expect_s3_class(glycan, c("glyrepr_structure"))
})


test_that("glycan_structure fails for invalid graphs", {
  bad_graph <- igraph::make_graph(~ 1 - +2, 2 - +3, 3 - +1)
  expect_error(glycan_structure(bad_graph))
})


test_that("vertex names are added if missing", {
  graph <- good_glycan_graph()
  graph <- igraph::delete_vertex_attr(graph, "name")

  glycan_vec <- glycan_structure(graph)
  glycan <- get_structure_graphs(glycan_vec, return_list = FALSE)

  expect_true("name" %in% igraph::vertex_attr_names(glycan))
})


# Tests for glycan_structure validation during creation --------------------------

test_that("glycan_structure accepts valid graphs", {
  graph <- good_glycan_graph()
  expect_no_error(glycan_structure(graph))
})


test_that("glycan_structure rejects invalid graphs", {
  # Test undirected graph
  graph <- igraph::make_graph(~ 1 - -2)
  igraph::V(graph)$mono <- c("Glc", "Gal")
  igraph::V(graph)$sub <- c("", "")
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"
  expect_error(glycan_structure(graph), "directed")
})


# Tests for ensure_name_vertex_attr ------------------------------------------

test_that("ensure_name_vertex_attr adds names when missing", {
  graph <- igraph::make_graph(~ 1 - +2)
  graph <- igraph::delete_vertex_attr(graph, "name")
  result <- ensure_name_vertex_attr(graph)
  expect_true("name" %in% igraph::vertex_attr_names(result))
})


test_that("ensure_name_vertex_attr preserves existing names", {
  graph <- igraph::make_graph(~ A - +B)
  result <- ensure_name_vertex_attr(graph)
  expect_equal(igraph::V(result)$name, c("A", "B"))
})


test_that("glycan structure class", {
  graph <- igraph::make_tree(3, children = 2, mode = "out")
  igraph::V(graph)$name <- as.character(1:igraph::vcount(graph))
  igraph::V(graph)$mono <- c("GlcNAc", "GlcNAc", "GlcNAc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "b1-3")
  graph$anomer <- "a1"
  glycan <- glycan_structure(graph)
  expect_s3_class(glycan, c("glyrepr_structure"))
})


test_that("validating undirected graphs", {
  graph <- igraph::make_graph(~ 1 - -2)
  igraph::V(graph)$mono <- c("GlcNAc", "GlcNAc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"
  expect_error(glycan_structure(graph), "Glycan structure must be directed")
})


test_that("validating an in tree", {
  graph <- igraph::make_tree(3, children = 2, mode = "in")
  igraph::V(graph)$mono <- "GlcNAc"
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"
  expect_error(glycan_structure(graph), "Glycan structure must be an out tree")
})


test_that("validating graph without monosaccharide attribute", {
  graph <- igraph::make_tree(3, children = 2, mode = "out")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"
  expect_error(
    glycan_structure(graph),
    "Glycan structure must have a vertex attribute 'mono'"
  )
})


test_that("validating graph without substituent attribute", {
  graph <- igraph::make_tree(3, children = 2, mode = "out")
  igraph::V(graph)$mono <- "GlcNAc"
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"
  expect_error(
    glycan_structure(graph),
    "Glycan structure must have a vertex attribute 'sub'"
  )
})


test_that("validating graph with NA in monosaccharide attribute", {
  graph <- igraph::make_tree(3, children = 2, mode = "out")
  igraph::V(graph)$mono <- c("GlcNAc", NA, "GlcNAc")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"
  expect_error(
    glycan_structure(graph),
    "Glycan structure must have no NA in vertex attribute 'mono'"
  )
})


test_that("validating graph with NA in substitude attribute", {
  graph <- igraph::make_tree(3, children = 2, mode = "out")
  igraph::V(graph)$mono <- c("GlcNAc", "GlcNAc", "GlcNAc")
  igraph::V(graph)$sub <- c("", NA, "")
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"
  expect_error(
    glycan_structure(graph),
    "Glycan structure must have no NA in vertex attribute 'sub'"
  )
})


patrick::with_parameters_test_that(
  "valid substituents",
  {
    skip_on_old_win()
    graph <- igraph::make_empty_graph(n = 1)
    igraph::V(graph)$mono <- "GlcNAc"
    igraph::V(graph)$sub <- sub
    igraph::E(graph)$linkage <- character(0)
    graph$anomer <- "b1"
    expect_no_error(glycan_structure(graph))
  },
  sub = c("6S", "9Ac", "2P", "?S")
)


test_that("multiple substituents are supported", {
  skip_on_old_win()
  graph <- igraph::make_empty_graph(n = 1)
  igraph::V(graph)$mono <- "Glc"
  igraph::V(graph)$sub <- "3Me,4Ac" # Multiple substituents
  igraph::E(graph)$linkage <- character(0)
  graph$anomer <- "a1"
  expect_no_error(glycan_structure(graph))
})


test_that("multiple substituents must be sorted by position", {
  skip_on_old_win()
  graph <- igraph::make_empty_graph(n = 1)
  igraph::V(graph)$mono <- "Glc"
  igraph::V(graph)$sub <- "4Ac,3Me" # Wrong order - should be "3Me,4Ac"
  igraph::E(graph)$linkage <- character(0)
  graph$anomer <- "a1"
  expect_error(glycan_structure(graph), "Unknown substituent")
})


test_that("duplicate positions in substituents are not allowed", {
  skip_on_old_win()
  graph <- igraph::make_empty_graph(n = 1)
  igraph::V(graph)$mono <- "Glc"
  igraph::V(graph)$sub <- "3Me,3Ac" # Same position (3) with different substituents
  igraph::E(graph)$linkage <- character(0)
  graph$anomer <- "a1"
  expect_error(glycan_structure(graph), "Unknown substituent")
})


test_that("normalize_substituents works correctly", {
  expect_equal(normalize_substituents(""), "")
  expect_equal(normalize_substituents("6S"), "6S")
  expect_equal(normalize_substituents("4Ac,3Me"), "3Me,4Ac")
  expect_equal(normalize_substituents("6P,2S,4Ac"), "2S,4Ac,6P")
  expect_equal(normalize_substituents("?S,3Me"), "3Me,?S")
})


test_that("validating graph without linkage attribute", {
  graph <- igraph::make_tree(3, children = 2, mode = "out")
  igraph::V(graph)$mono <- c("GlcNAc", "GlcNAc", "GlcNAc")
  igraph::V(graph)$sub <- ""
  graph$anomer <- "a1"
  expect_error(
    glycan_structure(graph),
    "Glycan structure must have an edge attribute 'linkage'"
  )
})


test_that("validating one non-existing monosaccharide", {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Hex", "Fuc", "Bad")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "b1-4")
  graph$anomer <- "a1"

  expect_error(glycan_structure(graph))
  err <- rlang::catch_cnd(glycan_structure(graph))
  # The error is now wrapped by purrr, so check the parent error
  expect_true(grepl("Unknown monosaccharide: Bad", err$parent$message))
  expect_equal(err$parent$monos, "Bad")
})


test_that("validating two non-existing monosaccharide", {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Hex", "Bad1", "Bad2")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "b1-4")
  graph$anomer <- "a1"

  err <- rlang::catch_cnd(glycan_structure(graph))
  expect_true(grepl("Unknown monosaccharide: Bad1, Bad2", err$parent$message))
  expect_equal(err$parent$monos, c("Bad1", "Bad2"))
})


test_that("validating duplicated non-existing monosaccharide", {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Hex", "Bad", "Bad")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", "b1-4")
  graph$anomer <- "a1"

  err <- rlang::catch_cnd(glycan_structure(graph))
  expect_true(grepl("Unknown monosaccharide: Bad", err$parent$message))
  expect_equal(err$parent$monos, "Bad")
})


test_that("validating bad subtituent", {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Hex", "Hex", "Hex")
  igraph::V(graph)$sub <- c("", "6S", "Bad")
  igraph::E(graph)$linkage <- c("b1-4", "b1-4")
  graph$anomer <- "a1"

  expect_error(glycan_structure(graph))
  err <- rlang::catch_cnd(glycan_structure(graph))
  expect_true(grepl("Unknown substituent: Bad", err$parent$message))
  expect_equal(err$parent$subs, "Bad")
})


patrick::with_parameters_test_that(
  "validating bad linkage",
  {
    graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
    igraph::V(graph)$mono <- c("Hex", "Fuc", "Hex")
    igraph::V(graph)$sub <- ""
    igraph::E(graph)$linkage <- bad_linkage
    graph$anomer <- "a1"

    expect_error(glycan_structure(graph))
    err <- rlang::catch_cnd(glycan_structure(graph))
  },
  bad_linkage = c("1-4", "c1-4", "b1", "abc", ""),
  .test_name = bad_linkage
)


test_that("validating NA linkages", {
  graph <- igraph::make_graph(~ 1 - +2, 2 - +3)
  igraph::V(graph)$mono <- c("Hex", "Hex", "Hex")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-4", NA)
  graph$anomer <- "a1"

  expect_error(glycan_structure(graph))
})


test_that("validating duplicated linkage positions", {
  graph <- igraph::make_graph(~ 1 - +2, 1 - +3)
  igraph::V(graph)$mono <- c("GalNAc", "Gal", "Neu5Ac")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-3", "a2-3")
  graph$anomer <- "a1"

  expect_error(glycan_structure(graph), "Duplicated linkage positions")
})


test_that("duplicated ? linkages are OK", {
  graph <- igraph::make_graph(~ 1 - +2, 1 - +3)
  igraph::V(graph)$mono <- c("GalNAc", "Gal", "Neu5Ac")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-?", "a2-?")
  graph$anomer <- "a1"

  expect_no_error(glycan_structure(graph))
})


test_that("duplicated x/y linkages are OK", {
  graph <- igraph::make_graph(~ 1 - +2, 1 - +3)
  igraph::V(graph)$mono <- c("GalNAc", "Gal", "Neu5Ac")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("b1-3/6", "a2-3/6")
  graph$anomer <- "a1"

  expect_no_error(glycan_structure(graph))
})


test_that("validating mixed generic and concrete monosaccharides", {
  graph <- igraph::make_tree(3, children = 2, mode = "out")
  igraph::V(graph)$mono <- c("Hex", "GlcNAc", "Hex")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- "b1-4"
  graph$anomer <- "a1"

  expect_error(
    glycan_structure(graph),
    "Monosaccharides must be either all generic or all concrete"
  )
})


test_that("missing anomer attr", {
  graph <- igraph::make_tree(3, children = 2, mode = "out")
  igraph::V(graph)$mono <- c("Hex", "Hex", "Hex")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("a1-3", "b1-4")

  expect_error(
    glycan_structure(graph),
    "Glycan structure must have a graph attribute 'anomer'"
  )
})


test_that("invalid anomer attr", {
  graph <- igraph::make_tree(3, children = 2, mode = "out")
  igraph::V(graph)$mono <- c("Hex", "Hex", "Hex")
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- c("a1-3", "b1-4")
  graph$anomer <- "a"

  expect_error(glycan_structure(graph), "Invalid anomer: a")
})


# Tests for glycan_structure vector functions

# Helper function to create a simple glycan structure
create_simple_glycan_graph <- function(mono_names, linkages, anomer = "?1") {
  n_nodes <- length(mono_names)
  if (n_nodes == 1) {
    graph <- igraph::make_empty_graph(n = 1)
  } else {
    # Create linear chain: 1-+2-+3-+...
    edges <- c()
    for (i in 1:(n_nodes - 1)) {
      edges <- c(edges, i, i + 1)
    }
    graph <- igraph::make_graph(edges = edges, directed = TRUE)
  }

  igraph::V(graph)$name <- as.character(1:igraph::vcount(graph))
  igraph::V(graph)$mono <- mono_names
  igraph::V(graph)$sub <- ""
  igraph::E(graph)$linkage <- linkages
  graph$anomer <- anomer
  graph
}

# Tests for glycan_structure constructor --------------------------------------------

test_that("glycan_structure creates empty vector by default", {
  sv <- glycan_structure()
  expect_s3_class(sv, "glyrepr_structure")
  expect_equal(length(sv), 0)
  expect_length(attr(sv, "graphs"), 0)
})

test_that("glycan_structure works with single glycan structure", {
  sv <- o_glycan_core_1()

  expect_s3_class(sv, "glyrepr_structure")
  expect_equal(length(sv), 1)
  expect_length(attr(sv, "graphs"), 1)
  expect_equal(structure_to_iupac(sv), "Gal(b1-3)GalNAc(a1-")
})

test_that("glycan_structure works with multiple different glycan structures", {
  glycan1 <- o_glycan_core_1()
  glycan2 <- n_glycan_core()
  sv <- c(glycan1, glycan2)

  expect_s3_class(sv, "glyrepr_structure")
  expect_equal(length(sv), 2)
  expect_length(attr(sv, "graphs"), 2)
})

test_that("glycan_structure removes duplicates based on IUPAC codes", {
  glycan1 <- o_glycan_core_1()
  glycan2 <- o_glycan_core_1() # Same structure
  sv <- c(glycan1, glycan2)

  expect_s3_class(sv, "glyrepr_structure")
  expect_equal(length(sv), 2) # Original vector has 2 elements
  expect_length(attr(sv, "graphs"), 1) # But only 1 unique structure
})

test_that("glycan_structure handles structures with different IUPAC but same graph topology", {
  # Create two structures that are topologically the same but have different anomers
  graph1 <- create_simple_glycan_graph(c("Glc", "Gal"), "b1-4", "a1")
  graph2 <- create_simple_glycan_graph(c("Glc", "Gal"), "b1-4", "b1")

  sv <- glycan_structure(graph1, graph2)

  expect_equal(length(sv), 2)
  expect_length(attr(sv, "graphs"), 2) # Different IUPAC codes
})

test_that("glycan_structure validates input", {
  expect_error(glycan_structure("not a graph"), "igraph objects")
  expect_error(glycan_structure(list(1, 2)), "igraph objects")
})

# Tests for as_glycan_structure -------------------------------------------------

test_that("as_glycan_structure creates valid glycan_structure object from igraph", {
  graph <- create_simple_glycan_graph(c("Glc", "Gal"), "b1-4")

  sv <- as_glycan_structure(graph)

  expect_s3_class(sv, "glyrepr_structure")
  expect_equal(length(sv), 1)
})

test_that("as_glycan_structure works with empty inputs", {
  sv <- glycan_structure()
  result <- as_glycan_structure(sv)

  expect_s3_class(result, "glyrepr_structure")
  expect_equal(length(result), 0)
})

# Tests for is_glycan_structure --------------------------------------------------

test_that("is_glycan_structure correctly identifies glycan_structure objects", {
  sv <- glycan_structure()
  expect_true(is_glycan_structure(sv))

  expect_false(is_glycan_structure(1))
  expect_false(is_glycan_structure("test"))
  expect_false(is_glycan_structure(list()))
  # Note: individual igraph objects should return FALSE for the new vectorized version
  graph <- create_simple_glycan_graph(c("Glc", "Gal"), "b1-4")
  expect_false(is_glycan_structure(graph))
})

# Tests for format method -----------------------------------------------------

test_that("format.glyrepr_structure displays correct IUPAC sequences", {
  glycan1 <- o_glycan_core_1()
  glycan2 <- n_glycan_core()
  sv <- c(glycan1, glycan2)

  formatted <- format(sv)

  expect_type(formatted, "character")
  expect_equal(
    formatted,
    c(
      "Gal(b1-3)GalNAc(a1-                                ",
      "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"
    )
  )
})

test_that("format.glyrepr_structure handles empty vector", {
  sv <- glycan_structure()
  formatted <- format(sv)

  expect_type(formatted, "character")
  expect_length(formatted, 0)
})

test_that("format.glyrepr_structure handles duplicates correctly", {
  glycan1 <- o_glycan_core_1()
  glycan2 <- o_glycan_core_1()
  sv <- c(glycan1, glycan2)

  formatted <- format(sv)

  expect_length(formatted, 2)
  expect_equal(formatted[1], formatted[2]) # Both should show same IUPAC
  expect_equal(formatted[1], "Gal(b1-3)GalNAc(a1-")
})

test_that("format.glyrepr_structure includes names with tab separation", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())
  names(glycans) <- c("A", "B")

  expect_snapshot(format(glycans))
})

test_that("format.glyrepr_structure without names works correctly", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())
  # Ensure no names
  glycans <- unname(glycans)

  expect_snapshot(format(glycans))
})

test_that("truncation works in tibble", {
  sv <- c(n_glycan_core(), n_glycan_core(), n_glycan_core())
  tibble <- tibble::tibble(struc = sv, a = 1)
  expect_snapshot(print(tibble, width = 30))
})

# Tests for vctrs methods -----------------------------------------------------

test_that("vec_ptype_abbr.glyrepr_structure returns correct abbreviation", {
  sv <- glycan_structure()
  expect_equal(vctrs::vec_ptype_abbr(sv), "struct")
})

test_that("vec_ptype_full.glyrepr_structure returns correct full type", {
  sv <- glycan_structure()
  expect_equal(vctrs::vec_ptype_full(sv), "glycan_structure")
})

# Tests for print footer ------------------------------------------------------

test_that("obj_print_footer.glyrepr_structure displays unique count", {
  glycan1 <- o_glycan_core_1()
  glycan2 <- n_glycan_core()
  glycan3 <- o_glycan_core_1() # Duplicate
  sv <- c(glycan1, glycan2, glycan3)

  output <- capture.output(obj_print_footer.glyrepr_structure(sv))

  expect_length(output, 1)
  expect_match(output, "# Unique structures: 2")
})

test_that("obj_print_footer.glyrepr_structure handles empty vector", {
  sv <- glycan_structure()

  output <- capture.output(obj_print_footer.glyrepr_structure(sv))

  expect_length(output, 1)
  expect_match(output, "# Unique structures: 0")
})

# Tests for get_structure_graphs ----------------------------------------------

test_that("get_structure_graphs extracts individual structures correctly", {
  glycan1 <- o_glycan_core_1()
  glycan2 <- n_glycan_core()
  sv <- c(glycan1, glycan2)

  # Check that we can retrieve the original structures
  extracted_all <- get_structure_graphs(sv)
  extracted_single <- get_structure_graphs(sv[1], return_list = FALSE)

  expect_type(extracted_all, "list")
  expect_length(extracted_all, 2)
  expect_s3_class(extracted_all[[1]], "igraph")
  expect_s3_class(extracted_all[[2]], "igraph")
  expect_s3_class(extracted_single, "igraph")
})

test_that("get_structure_graphs return_list parameter works correctly", {
  glycan1 <- o_glycan_core_1()
  glycan2 <- n_glycan_core()
  sv <- c(glycan1, glycan2)

  # Test default behavior (NULL return_list)
  # For multiple structures, should return list
  result_default_multi <- get_structure_graphs(sv)
  expect_type(result_default_multi, "list")
  expect_length(result_default_multi, 2)

  # For single structure, should return igraph directly
  result_default_single <- get_structure_graphs(sv[1])
  expect_s3_class(result_default_single, "igraph")

  # Test explicit return_list = TRUE
  result_list_true <- get_structure_graphs(sv[1], return_list = TRUE)
  expect_type(result_list_true, "list")
  expect_length(result_list_true, 1)
  expect_s3_class(result_list_true[[1]], "igraph")

  # Test explicit return_list = FALSE
  result_list_false <- get_structure_graphs(sv[1], return_list = FALSE)
  expect_s3_class(result_list_false, "igraph")
})

test_that("get_structure_graphs validates return_list parameter", {
  glycan1 <- o_glycan_core_1()
  glycan2 <- n_glycan_core()
  sv <- c(glycan1, glycan2)

  # Should error when return_list = FALSE but length > 1
  expect_error(
    get_structure_graphs(sv, return_list = FALSE),
    "return_list.*must be.*TRUE.*NULL.*length greater than 1"
  )
})

# Integration tests -----------------------------------------------------------
test_that("glycan_structure preserves glycan structures correctly", {
  glycan1 <- o_glycan_core_1()
  glycan2 <- n_glycan_core()
  sv <- c(glycan1, glycan2)

  # Check that we can retrieve the original structures
  structures <- attr(sv, "graphs")
  expect_length(structures, 2)

  # Check IUPAC codes are generated correctly
  expected_iupacs <- c(structure_to_iupac(glycan1), structure_to_iupac(glycan2))
  expect_equal(structure_to_iupac(sv), expected_iupacs)
})

test_that("glycan_structure handles complex branched structures", {
  # Create a more complex branched structure
  complex_graph <- igraph::make_graph(~ 1 - +2, 1 - +3, 1 - +4)
  igraph::V(complex_graph)$mono <- c("Man", "GlcNAc", "Gal", "Fuc")
  igraph::V(complex_graph)$sub <- ""
  igraph::E(complex_graph)$linkage <- c("b1-4", "a1-3", "a1-6")
  complex_graph$anomer <- "a1"

  sv <- glycan_structure(complex_graph)

  expect_s3_class(sv, "glyrepr_structure")
  expect_equal(length(sv), 1)
  expect_length(attr(sv, "graphs"), 1)
})

test_that("glycan_structure maintains hash uniqueness property", {
  # Create multiple identical structures
  graphs <- replicate(
    5,
    {
      create_simple_glycan_graph(c("Glc", "Gal"), "b1-4")
    },
    simplify = FALSE
  )

  sv <- as_glycan_structure(graphs)

  expect_equal(length(sv), 5) # 5 elements in vector
  expect_length(attr(sv, "graphs"), 1) # But only 1 unique structure
})

# Tests for c() function (vec_ptype2 method) ------------------------------------

test_that("c() combines glycan_structure vectors correctly", {
  # This test ensures the vec_ptype2 method works correctly
  # This was previously failing due to ifelse() with igraph objects
  sv1 <- n_glycan_core()
  sv2 <- o_glycan_core_1()

  # This should not error (previously caused rep.igraph error)
  combined <- c(sv1, sv2)

  expect_s3_class(combined, "glyrepr_structure")
  expect_equal(length(combined), 2)
  expect_length(attr(combined, "graphs"), 2) # Two unique structures

  # Check that both structures are preserved
  expect_equal(
    structure_to_iupac(combined),
    c(
      "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-",
      "Gal(b1-3)GalNAc(a1-"
    )
  )
})

test_that("c() handles duplicate structures across vectors", {
  sv1 <- c(o_glycan_core_1(), n_glycan_core())
  sv2 <- o_glycan_core_1() # Duplicate of first structure in sv1

  combined <- c(sv1, sv2)

  expect_s3_class(combined, "glyrepr_structure")
  expect_equal(length(combined), 3) # Total elements
  expect_length(attr(combined, "graphs"), 2) # Only 2 unique structures
})

test_that("c() works with empty vectors", {
  sv1 <- glycan_structure()
  sv2 <- o_glycan_core_1()

  combined1 <- c(sv1, sv2)
  combined2 <- c(sv2, sv1)

  expect_s3_class(combined1, "glyrepr_structure")
  expect_s3_class(combined2, "glyrepr_structure")
  expect_equal(length(combined1), 1)
  expect_equal(length(combined2), 1)
  expect_length(attr(combined1, "graphs"), 1)
  expect_length(attr(combined2, "graphs"), 1)
})

test_that("c() combines multiple structure vectors efficiently", {
  # Test combining multiple vectors with various duplicates
  sv1 <- o_glycan_core_1()
  sv2 <- n_glycan_core()
  sv3 <- c(o_glycan_core_1(), n_glycan_core()) # Contains both

  combined <- c(sv1, sv2, sv3)

  expect_s3_class(combined, "glyrepr_structure")
  expect_equal(length(combined), 4) # 1 + 1 + 2 = 4 total elements
  expect_length(attr(combined, "graphs"), 2) # Only 2 unique structures

  # Check all elements are present
  expect_equal(
    structure_to_iupac(combined),
    c(
      "Gal(b1-3)GalNAc(a1-",
      "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-",
      "Gal(b1-3)GalNAc(a1-",
      "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"
    )
  )
})

test_that("c() preserves structure integrity across combinations", {
  # Use existing example structures
  sv1 <- o_glycan_core_1()
  sv2 <- n_glycan_core()

  combined <- c(sv1, sv2)

  # Check that we can retrieve the original structures correctly
  extracted_graphs <- get_structure_graphs(combined)

  expect_length(extracted_graphs, 2)
  expect_s3_class(extracted_graphs[[1]], "igraph")
  expect_s3_class(extracted_graphs[[2]], "igraph")

  # Verify structure properties are preserved using known structures
  # o_glycan_core_1 has GalNAc and Gal
  # n_glycan_core has Man and GlcNAc
  expect_true("GalNAc" %in% igraph::V(extracted_graphs[[1]])$mono)
  expect_true("Gal" %in% igraph::V(extracted_graphs[[1]])$mono)
  expect_true("Man" %in% igraph::V(extracted_graphs[[2]])$mono)
  expect_true("GlcNAc" %in% igraph::V(extracted_graphs[[2]])$mono)
})

# Tests for vector casting and subsetting functionality -------------------------

test_that("glycan_structure vectors can be subset with necessary structure preservation", {
  sv <- c(o_glycan_core_1(), n_glycan_core(), o_glycan_core_1())

  # Test various subsetting operations
  subset1 <- sv[1]
  subset2 <- sv[c(1, 3)]
  subset3 <- sv[2:3]
  subset4 <- sv[c()]
  subset5 <- sv[integer(0)]

  expect_s3_class(subset1, "glyrepr_structure")
  expect_s3_class(subset2, "glyrepr_structure")
  expect_s3_class(subset3, "glyrepr_structure")
  expect_s3_class(subset4, "glyrepr_structure")
  expect_s3_class(subset5, "glyrepr_structure")

  expect_equal(length(subset1), 1)
  expect_equal(length(subset2), 2)
  expect_equal(length(subset3), 2)
  expect_equal(length(subset4), 0)
  expect_equal(length(subset5), 0)

  # Check that unique structure tracking is maintained
  expect_length(attr(sv, "graphs"), 2) # Original has 2 unique
  expect_length(attr(subset1, "graphs"), 1) # Subset preserves 1 unique
  expect_length(attr(subset2, "graphs"), 1) # Subset preserves 1 unique
  expect_length(attr(subset3, "graphs"), 2) # Subset preserves 2 unique
  expect_length(attr(subset4, "graphs"), 0) # Subset preserves 0 unique
  expect_length(attr(subset5, "graphs"), 0) # Subset preserves 0 unique
})

test_that("glycan_structure vectors can be repeated", {
  sv <- o_glycan_core_1()

  # Test rep() function which uses vctrs casting methods
  repeated <- rep(sv, 3)

  expect_s3_class(repeated, "glyrepr_structure")
  expect_equal(length(repeated), 3)
  expect_length(attr(repeated, "graphs"), 1) # Still only 1 unique structure

  formatted <- format(repeated)
  expect_equal(formatted, rep("Gal(b1-3)GalNAc(a1-", 3))
})

test_that("complex vector operations work correctly", {
  # Test more complex vector operations
  sv1 <- c(o_glycan_core_1(), n_glycan_core())
  sv2 <- c(n_glycan_core(), o_glycan_core_1())

  # Combine and then subset
  combined <- c(sv1, sv2)
  reordered <- combined[c(4, 3, 2, 1)]

  expect_s3_class(reordered, "glyrepr_structure")
  expect_equal(length(reordered), 4)
  expect_length(attr(reordered, "graphs"), 2)

  # Check that reordering worked correctly
  formatted_original <- format(combined)
  formatted_reordered <- format(reordered)
  expect_equal(formatted_reordered, rev(formatted_original))
})

test_that("structure vector methods handle edge cases", {
  # Test with single element vector
  single <- o_glycan_core_1()

  # Test combining with itself
  doubled <- c(single, single)
  expect_equal(length(doubled), 2)
  expect_length(attr(doubled, "graphs"), 1)

  # Test empty + non-empty combinations in different orders
  empty <- glycan_structure()
  combined1 <- c(empty, single, empty)
  combined2 <- c(single, empty, single)

  expect_equal(length(combined1), 1)
  expect_equal(length(combined2), 2)
  expect_length(attr(combined1, "graphs"), 1)
  expect_length(attr(combined2, "graphs"), 1)
})

# Tests for tibble and dplyr operations with structure optimization -----------

test_that("tibble row subsetting optimizes structure storage", {
  skip_if_not_installed("tibble")

  # Create test data with duplicates
  sv <- c(o_glycan_core_1(), n_glycan_core(), o_glycan_core_1())
  df <- tibble::tibble(id = 1:3, structure = sv, name = c("A", "B", "C"))

  # Original should have 2 unique structures
  expect_length(attr(sv, "graphs"), 2)

  # Single row subsetting should optimize to 1 unique structure
  subset1 <- df[1, ]
  expect_equal(length(subset1$structure), 1)
  expect_length(attr(subset1$structure, "graphs"), 1)

  # Multiple row subsetting with same structure should optimize to 1 unique
  subset2 <- df[c(1, 3), ] # Both point to same structure
  expect_equal(length(subset2$structure), 2)
  expect_length(attr(subset2$structure, "graphs"), 1)

  # Multiple row subsetting with different structures should keep both
  subset3 <- df[2:3, ]
  expect_equal(length(subset3$structure), 2)
  expect_length(attr(subset3$structure, "graphs"), 2)

  # Single row with different structure
  subset4 <- df[2, ]
  expect_equal(length(subset4$structure), 1)
  expect_length(attr(subset4$structure, "graphs"), 1)
})

test_that("dplyr filter operations optimize structure storage", {
  skip_if_not_installed("dplyr")

  # Create test data with duplicates
  sv <- c(o_glycan_core_1(), n_glycan_core(), o_glycan_core_1())
  df <- tibble::tibble(id = 1:3, structure = sv, score = c(10, 20, 30))

  # Filter to single row should optimize to 1 unique structure
  filtered1 <- df %>% dplyr::filter(id == 1)
  expect_equal(length(filtered1$structure), 1)
  expect_length(attr(filtered1$structure, "graphs"), 1)

  # Filter to multiple rows with same structure should optimize
  filtered2 <- df %>% dplyr::filter(id != 2) # Keeps rows 1 and 3 (same structure)
  expect_equal(length(filtered2$structure), 2)
  expect_length(attr(filtered2$structure, "graphs"), 1)

  # Filter to multiple rows with different structures should keep both
  filtered3 <- df %>% dplyr::filter(id >= 2) # Keeps rows 2 and 3 (different structures)
  expect_equal(length(filtered3$structure), 2)
  expect_length(attr(filtered3$structure, "graphs"), 2)

  # Filter by score
  filtered4 <- df %>% dplyr::filter(score >= 20)
  expect_equal(length(filtered4$structure), 2)
  expect_length(attr(filtered4$structure, "graphs"), 2)
})

test_that("dplyr slice operations optimize structure storage", {
  skip_if_not_installed("dplyr")

  # Create test data
  sv <- c(o_glycan_core_1(), n_glycan_core(), o_glycan_core_1())
  df <- tibble::tibble(id = 1:3, structure = sv)

  # slice() operations
  sliced1 <- df %>% dplyr::slice(1)
  expect_equal(length(sliced1$structure), 1)
  expect_length(attr(sliced1$structure, "graphs"), 1)

  sliced2 <- df %>% dplyr::slice(c(1, 3))
  expect_equal(length(sliced2$structure), 2)
  expect_length(attr(sliced2$structure, "graphs"), 1)

  sliced3 <- df %>% dplyr::slice(2:3)
  expect_equal(length(sliced3$structure), 2)
  expect_length(attr(sliced3$structure, "graphs"), 2)

  # slice_head() and slice_tail()
  head_slice <- df %>% dplyr::slice_head(n = 1)
  expect_equal(length(head_slice$structure), 1)
  expect_length(attr(head_slice$structure, "graphs"), 1)

  tail_slice <- df %>% dplyr::slice_tail(n = 1)
  expect_equal(length(tail_slice$structure), 1)
  expect_length(attr(tail_slice$structure, "graphs"), 1)
})

test_that("dplyr arrange and other operations preserve structure optimization", {
  skip_if_not_installed("dplyr")

  # Create test data
  sv <- c(o_glycan_core_1(), n_glycan_core(), o_glycan_core_1())
  df <- tibble::tibble(id = 1:3, structure = sv, score = c(30, 10, 20))

  # arrange() should maintain all structures
  arranged <- df %>% dplyr::arrange(score)
  expect_equal(length(arranged$structure), 3)
  expect_length(attr(arranged$structure, "graphs"), 2)

  # arrange() + slice() should optimize
  arranged_sliced <- df %>% dplyr::arrange(score) %>% dplyr::slice(1)
  expect_equal(length(arranged_sliced$structure), 1)
  expect_length(attr(arranged_sliced$structure, "graphs"), 1)

  # top_n() operations
  top2 <- df %>% dplyr::top_n(2, score)
  expect_equal(length(top2$structure), 2)
  expect_length(attr(top2$structure, "graphs"), 1) # top_n selects id=1 and id=3, both have same structure

  # distinct() operations with duplicated structures
  df_with_dups <- tibble::tibble(
    id = rep(1:3, each = 2),
    structure = rep(sv, each = 2)
  )
  distinct_result <- df_with_dups %>%
    dplyr::distinct(structure, .keep_all = TRUE)
  expect_equal(length(distinct_result$structure), 2)
  expect_length(attr(distinct_result$structure, "graphs"), 2)
})

test_that("complex tibble and dplyr workflows maintain optimization", {
  skip_if_not_installed("dplyr")

  # Create more complex test data
  sv <- c(
    o_glycan_core_1(),
    n_glycan_core(),
    o_glycan_core_1(),
    n_glycan_core(),
    o_glycan_core_1()
  )
  df <- tibble::tibble(
    id = 1:5,
    structure = sv,
    type = c("A", "B", "A", "B", "A"),
    score = c(10, 20, 15, 25, 30)
  )

  # Original should have 2 unique structures
  expect_length(attr(sv, "graphs"), 2)

  # Complex workflow: filter + arrange + slice
  result1 <- df %>%
    dplyr::filter(type == "A") %>%
    dplyr::arrange(desc(score)) %>%
    dplyr::slice(1:2)

  expect_equal(length(result1$structure), 2)
  expect_length(attr(result1$structure, "graphs"), 1) # All type A have same structure

  # Another complex workflow: group operations
  result2 <- df %>%
    dplyr::group_by(type) %>%
    dplyr::slice_max(score, n = 1) %>%
    dplyr::ungroup()

  expect_equal(length(result2$structure), 2)
  expect_length(attr(result2$structure, "graphs"), 2) # One from each type

  # Workflow with structure column operations
  result3 <- df %>%
    dplyr::filter(score >= 20) %>%
    dplyr::select(structure, score)

  expect_equal(length(result3$structure), 3) # score >= 20 selects 3 rows (id=2,4,5)
  expect_length(attr(result3$structure, "graphs"), 2) # These have 2 different structures
})

test_that("tibble operations preserve structure content integrity", {
  skip_if_not_installed("dplyr")

  # Create test data
  sv <- c(o_glycan_core_1(), n_glycan_core(), o_glycan_core_1())
  df <- tibble::tibble(id = 1:3, structure = sv)

  # Verify that optimization doesn't affect structure content
  subset_df <- df %>% dplyr::filter(id == 1)

  # Extract the structure and verify it's correct
  structure_graph <- get_structure_graphs(
    subset_df$structure,
    return_list = FALSE
  )
  expect_s3_class(structure_graph, "igraph")

  # Verify structure content (o_glycan_core_1 has GalNAc and Gal)
  expect_true("GalNAc" %in% igraph::V(structure_graph)$mono)
  expect_true("Gal" %in% igraph::V(structure_graph)$mono)

  # Verify IUPAC representation is preserved
  expect_equal(format(subset_df$structure)[1], "Gal(b1-3)GalNAc(a1-")
})

# Tests for vector conversion -------------------------
test_that("converting to character", {
  sv <- o_glycan_core_1()
  expect_equal(as.character(sv), "Gal(b1-3)GalNAc(a1-")
})

# Tests for vertex and edge reordering -------------------------
test_that("vertices and edges are reordered correctly", {
  graph1 <- igraph::make_graph(~ 1 - +2, 1 - +3)
  igraph::V(graph1)$mono <- c("GalNAc", "Gal", "GlcNAc")
  igraph::V(graph1)$sub <- ""
  igraph::E(graph1)$linkage <- c("b1-3", "b1-6")
  graph1$anomer <- "a1"

  graph2 <- igraph::make_graph(~ 1 - +2, 1 - +3)
  igraph::V(graph2)$mono <- c("GalNAc", "GlcNAc", "Gal")
  igraph::V(graph2)$sub <- ""
  igraph::E(graph2)$linkage <- c("b1-6", "b1-3")
  graph2$anomer <- "a1"

  sv1 <- glycan_structure(graph1)
  sv2 <- glycan_structure(graph2)
  graph1 <- get_structure_graphs(sv1, return_list = FALSE)
  graph2 <- get_structure_graphs(sv2, return_list = FALSE)

  expect_equal(igraph::V(graph1)$mono, c("Gal", "GlcNAc", "GalNAc"))
  expect_equal(igraph::V(graph2)$mono, c("Gal", "GlcNAc", "GalNAc"))

  expect_equal(igraph::E(graph1)$linkage, c("b1-3", "b1-6"))
  expect_equal(igraph::E(graph2)$linkage, c("b1-3", "b1-6"))
})

# Tests for mono_type validation in glycan_structure -------------------------

test_that("glycan_structure accepts multiple concrete structures", {
  graph1 <- create_simple_glycan_graph(c("Glc", "Gal"), "b1-4")
  graph2 <- create_simple_glycan_graph(c("Man", "GlcNAc"), "b1-4")
  sv <- glycan_structure(graph1, graph2)
  expect_equal(length(sv), 2)
})

test_that("glycan_structure accepts multiple generic structures", {
  graph1 <- create_simple_glycan_graph(c("Hex", "HexNAc"), "b1-4")
  graph2 <- create_simple_glycan_graph(c("Hex", "dHex"), "b1-6")
  sv <- glycan_structure(graph1, graph2)
  expect_equal(length(sv), 2)
})

test_that("glycan_structure rejects mixing concrete and generic structures", {
  graph1 <- create_simple_glycan_graph(c("Glc", "Gal"), "b1-4") # concrete
  graph2 <- create_simple_glycan_graph(c("Hex", "HexNAc"), "b1-4") # generic
  expect_error(
    glycan_structure(graph1, graph2),
    "All structures must have the same monosaccharide type"
  )
})

test_that("c() rejects combining concrete and generic structure vectors", {
  sv1 <- o_glycan_core_1() # concrete: Gal, GalNAc
  sv2 <- n_glycan_core(mono_type = "generic") # generic: Hex, HexNAc
  expect_error(
    c(sv1, sv2),
    "All structures must have the same monosaccharide type"
  )
})

test_that("get_mono_type returns same type for all structures in vector", {
  graph1 <- create_simple_glycan_graph(c("Glc", "Gal"), "b1-4")
  graph2 <- create_simple_glycan_graph(c("Man", "GlcNAc"), "b1-4")
  sv <- glycan_structure(graph1, graph2)
  expect_equal(get_mono_type(sv), "concrete")
})

test_that("get_mono_type returns same type for generic structures", {
  graph1 <- create_simple_glycan_graph(c("Hex", "HexNAc"), "b1-4")
  graph2 <- create_simple_glycan_graph(c("Hex", "dHex"), "b1-6")
  sv <- glycan_structure(graph1, graph2)
  expect_equal(get_mono_type(sv), "generic")
})

# Tests for [[<- operation (issue #11) -----------------------------------------

test_that("[[<- is forbidden on glyrepr_structure vectors", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())

  expect_error(
    glycans[[1]] <- n_glycan_core(),
    class = "rlang_error"
  )
})

test_that("[[<- is forbidden with igraph value", {
  glycans <- o_glycan_core_1()
  graph <- get_structure_graphs(glycans, return_list = FALSE)

  expect_error(
    glycans[[1]] <- graph,
    class = "rlang_error"
  )
})

# Tests for names-----

test_that("directly setting names works", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())
  names(glycans) <- c("A", "B")
  expect_equal(names(glycans), c("A", "B"))
})

test_that("new structures have NULL names by default", {
  glycans <- o_glycan_core_1()
  expect_null(names(glycans))
})

test_that("new structure vectors have NULL names by default", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())
  expect_null(names(glycans))
})

test_that("names are preserved after subsetting with [", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())
  names(glycans) <- c("A", "B")

  subset1 <- glycans[1]
  expect_equal(names(subset1), "A")

  subset2 <- glycans[2]
  expect_equal(names(subset2), "B")

  subset_both <- glycans[c(1, 2)]
  expect_equal(names(subset_both), c("A", "B"))

  subset_reverse <- glycans[c(2, 1)]
  expect_equal(names(subset_reverse), c("B", "A"))
})

test_that("names are preserved after subsetting with logical index", {
  glycans <- c(o_glycan_core_1(), n_glycan_core(), o_glycan_core_1())
  names(glycans) <- c("A", "B", "C")

  subset_logical <- glycans[c(TRUE, FALSE, TRUE)]
  expect_equal(names(subset_logical), c("A", "C"))
})

test_that("names are preserved after c() combining", {
  glycans1 <- o_glycan_core_1()
  glycans2 <- n_glycan_core()
  names(glycans1) <- "A"
  names(glycans2) <- "B"

  combined <- c(glycans1, glycans2)
  expect_equal(names(combined), c("A", "B"))
})

test_that("names are preserved after rep()", {
  glycans <- o_glycan_core_1()
  names(glycans) <- "A"

  repeated <- rep(glycans, 3)
  expect_equal(names(repeated), c("A", "A", "A"))
})

test_that("setting names to NULL removes names", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())
  names(glycans) <- c("A", "B")
  expect_equal(names(glycans), c("A", "B"))

  names(glycans) <- NULL
  expect_null(names(glycans))
})

test_that("names work with empty vectors", {
  empty <- glycan_structure()
  expect_null(names(empty))
})

test_that("names work with single-element vectors", {
  glycan <- o_glycan_core_1()
  names(glycan) <- "single"
  expect_equal(names(glycan), "single")
})

test_that("names are preserved after combining vectors with different names", {
  glycans1 <- c(o_glycan_core_1())
  glycans2 <- c(n_glycan_core())
  names(glycans1) <- "A"
  names(glycans2) <- "B"

  combined <- c(glycans1, glycans2)
  expect_equal(names(combined), c("A", "B"))
})

test_that("names work with duplicated structures", {
  glycans <- c(o_glycan_core_1(), n_glycan_core(), o_glycan_core_1())
  names(glycans) <- c("A", "B", "C")

  # First and third point to the same structure but have different names
  expect_equal(names(glycans), c("A", "B", "C"))

  # Subsetting should preserve names correctly
  subset <- glycans[c(1, 3)]
  expect_equal(names(subset), c("A", "C"))
})

test_that("names are preserved in tibble operations", {
  skip_if_not_installed("tibble")

  glycans <- c(o_glycan_core_1(), n_glycan_core())
  names(glycans) <- c("A", "B")

  df <- tibble::tibble(id = 1:2, structure = glycans)

  # Subsetting tibble should preserve names
  subset_df <- df[1, ]
  expect_equal(names(subset_df$structure), "A")

  subset_df2 <- df[2, ]
  expect_equal(names(subset_df2$structure), "B")
})

test_that("names work correctly with dplyr operations", {
  skip_if_not_installed("dplyr")

  glycans <- c(o_glycan_core_1(), n_glycan_core())
  names(glycans) <- c("A", "B")

  df <- tibble::tibble(id = 1:2, structure = glycans)

  # filter should preserve names
  filtered <- df %>% dplyr::filter(id == 1)
  expect_equal(names(filtered$structure), "A")

  # slice should preserve names
  sliced <- df %>% dplyr::slice(2)
  expect_equal(names(sliced$structure), "B")
})

test_that("unname removes names", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())
  names(glycans) <- c("A", "B")

  unname_glycans <- unname(glycans)
  expect_null(names(unname_glycans))
})

test_that("names work with character conversion", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())
  names(glycans) <- c("A", "B")

  # Converting to character should preserve names
  chars <- as.character(glycans)
  expect_equal(names(chars), c("A", "B"))
})

# Comprehensive regression tests for names preservation in glyrepr_structure functions
test_that("all glyrepr_structure functions preserve names", {
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)
  names(structures) <- c("A", "B", "C")

  # Type conversion (glyrepr_structure -> glyrepr_structure)
  expect_equal(names(as_glycan_structure(structures)), c("A", "B", "C"))

  # Character to glyrepr_structure conversion
  char_vec <- c(X = "Glc(a1-", Y = "Gal(a1-")
  expect_equal(names(as_glycan_structure(char_vec)), c("X", "Y"))

  # Accessor functions (return atomic vectors)
  expect_equal(names(get_anomer(structures)), c("A", "B", "C"))
  expect_equal(names(has_linkages(structures)), c("A", "B", "C"))
  expect_equal(names(count_mono(structures)), c("A", "B", "C"))
  expect_equal(names(structure_to_iupac(structures)), c("A", "B", "C"))

  # get_structure_level() returns one scalar level for the whole vector.
  expect_null(names(get_structure_level(structures)))

  # Accessor functions (return list)
  expect_equal(names(get_structure_graphs(structures)), c("A", "B", "C"))

  # Transformation functions (return glyrepr_structure)
  expect_equal(names(remove_linkages(structures)), c("A", "B", "C"))
  expect_equal(names(remove_substituents(structures)), c("A", "B", "C"))

  # Composition conversion
  structs_concrete <- c(o_glycan_core_1(), n_glycan_core())
  names(structs_concrete) <- c("X", "Y")
  expect_equal(names(convert_to_generic(structs_concrete)), c("X", "Y"))

  # Level reduction
  expect_equal(
    names(reduce_structure_level(structures, "topological")),
    c("A", "B", "C")
  )

  # Vector combination
  expect_equal(names(c(structures)), c("A", "B", "C"))
})

# Tests for is.na method ----------------------------------------------------

test_that("is.na returns correct logical for structures with NA", {
  struct <- o_glycan_core_1()
  expect_equal(is.na(struct), FALSE)

  struct_na <- c(o_glycan_core_1(), NA)
  expect_equal(is.na(struct_na), c(FALSE, TRUE))

  struct_all_na <- c(NA, NA)
  expect_equal(is.na(struct_all_na), c(TRUE, TRUE))
})

# Tests for NULL/NA handling in glycan_structure -----------------------------

test_that("glycan_structure accepts NULL to create NA", {
  struct <- glycan_structure(NULL)
  expect_equal(length(struct), 1)
  expect_true(is.na(struct))
})

test_that("glycan_structure accepts NA to create NA", {
  struct <- glycan_structure(NA)
  expect_equal(length(struct), 1)
  expect_true(is.na(struct))
})

test_that("glycan_structure handles mixed valid and NA", {
  graph <- o_glycan_core_1() |> get_structure_graphs(return_list = FALSE)
  struct <- glycan_structure(graph, NULL, graph)
  expect_equal(length(struct), 3)
  expect_false(is.na(struct[1]))
  expect_true(is.na(struct[2]))
  expect_false(is.na(struct[3]))
})

# Tests for vec_cast.character with NA handling --------------------------------

test_that("vec_cast handles NA characters", {
  chars <- c("Glc(a1-", NA)
  struct <- as_glycan_structure(chars)
  expect_equal(length(struct), 2)
  expect_false(is.na(struct[1]))
  expect_true(is.na(struct[2]))
})

test_that("vec_cast handles NA at beginning", {
  chars <- c(NA, "Gal(b1-3)GalNAc(a1-")
  struct <- as_glycan_structure(chars)
  expect_equal(length(struct), 2)
  expect_true(is.na(struct[1]))
  expect_false(is.na(struct[2]))
})

test_that("vec_cast handles all NA characters", {
  chars <- c(NA, NA)
  struct <- as_glycan_structure(chars)
  expect_equal(length(struct), 2)
  expect_true(is.na(struct[1]))
  expect_true(is.na(struct[2]))
})

# Tests for vec_restore NA handling -----------------------------------------

test_that("vec_restore skips NA in type checking", {
  struct1 <- o_glycan_core_1()
  struct2 <- glycan_structure(NA)

  # Should not error - NA should be skipped in type check
  combined <- c(struct1, struct2)
  expect_equal(length(combined), 2)
  expect_false(is.na(combined[1]))
  expect_true(is.na(combined[2]))
})

test_that("combining with NA at beginning works", {
  struct <- o_glycan_core_1()
  # c(NA, struct) doesn't preserve type in base R
  combined <- vctrs::vec_c(NA, struct)
  expect_equal(length(combined), 2)
  expect_true(is.na(combined[1]))
  expect_false(is.na(combined[2]))
})

test_that("vec_restore handles all NA vector", {
  struct_na <- glycan_structure(NA, NA)
  expect_equal(length(struct_na), 2)
  expect_equal(is.na(struct_na), c(TRUE, TRUE))
  expect_length(attr(struct_na, "graphs"), 0) # No graphs for NA elements
})

test_that("vec_restore preserves graphs for non-NA with mixed NA", {
  struct1 <- o_glycan_core_1()
  struct2 <- n_glycan_core()
  combined <- c(struct1, NA, struct2)

  expect_equal(length(combined), 3)
  expect_false(is.na(combined[1]))
  expect_true(is.na(combined[2]))
  expect_false(is.na(combined[3]))
  # Should have 2 unique structures (not counting NA)
  expect_length(attr(combined, "graphs"), 2)
})

# Tests for as_glycan_composition with NA structures ---------------------------------

test_that("as_glycan_composition handles structures with NA", {
  structs <- c(o_glycan_core_1(), NA, n_glycan_core())
  comps <- as_glycan_composition(structs)
  expect_equal(length(comps), 3)
  expect_false(is.na(comps[1]))
  expect_true(is.na(comps[2]))
  expect_false(is.na(comps[3]))
})

test_that("as_glycan_composition handles single NA structure", {
  structs <- glycan_structure(NA)
  comps <- as_glycan_composition(structs)
  expect_equal(length(comps), 1)
  expect_true(is.na(comps[1]))
})

test_that("as_glycan_composition handles all NA structures", {
  structs <- glycan_structure(NA, NA)
  comps <- as_glycan_composition(structs)
  expect_equal(length(comps), 2)
  expect_true(all(is.na(comps)))
})

# Tests for convert_to_generic with NA --------------------------------------------

test_that("convert_to_generic handles single valid structure", {
  # Note: convert_to_generic uses smap which currently doesn't handle NA
  # This test verifies the function works for non-NA structures
  converted <- convert_to_generic(o_glycan_core_1())
  expect_false(is.na(converted))
  expect_equal(get_mono_type(converted), "generic")
})

test_that("convert_to_generic works on N-glycan core", {
  # Note: convert_to_generic uses smap which currently doesn't handle NA
  # This test verifies the function works for non-NA structures
  converted <- convert_to_generic(n_glycan_core())
  expect_false(is.na(converted))
  expect_equal(get_mono_type(converted), "generic")
})

# Tests for reduce_structure_level with NA -----------------------------------------

test_that("reduce_structure_level handles structures with NA", {
  # Create structures manually to avoid smap issues with NA
  # The valid structure should work, and NA should be preserved
  structs <- c(o_glycan_core_1(), NA)
  # Note: reduce_structure_level uses smap which currently doesn't handle NA
  # This test verifies the function works for non-NA structures
  reduced_valid <- reduce_structure_level(
    o_glycan_core_1(),
    to_level = "topological"
  )
  expect_equal(get_structure_level(reduced_valid), "topological")
})

test_that("reduce_structure_level handles NA when reducing to basic", {
  # Note: reduce_structure_level uses smap which currently doesn't handle NA
  # This test verifies the function works for non-NA structures
  reduced_valid <- reduce_structure_level(o_glycan_core_1(), to_level = "basic")
  expect_equal(get_structure_level(reduced_valid), "basic")
})

# Tests for get_mono_type with NA --------------------------------------------------

test_that("get_mono_type handles mixed NA and valid structures", {
  structs <- c(o_glycan_core_1(), NA)
  types <- get_mono_type(structs)
  expect_equal(length(types), 1) # Returns scalar for glyrepr_structure
  expect_equal(types, "concrete")
})

test_that("get_mono_type handles all NA structures", {
  structs <- glycan_structure(NA, NA)
  types <- get_mono_type(structs)
  expect_equal(length(types), 1) # Returns scalar for glyrepr_structure
  expect_true(is.na(types))
})

# Tests for rep with NA ------------------------------------------------------------

test_that("rep handles structures with NA", {
  struct <- c(o_glycan_core_1(), NA)
  repeated <- rep(struct, 2)
  expect_equal(length(repeated), 4)
  expect_equal(is.na(repeated), c(FALSE, TRUE, FALSE, TRUE))
})

test_that("rep preserves NA in single-element NA structure", {
  struct <- glycan_structure(NA)
  repeated <- rep(struct, 3)
  expect_equal(length(repeated), 3)
  expect_true(all(is.na(repeated)))
})

# Tests for subsetting preserves NA ------------------------------------------------

test_that("subsetting preserves NA", {
  struct <- c(o_glycan_core_1(), NA, n_glycan_core())
  expect_true(is.na(struct[2]))
  expect_equal(is.na(struct[c(1, 3)]), c(FALSE, FALSE))
  expect_equal(length(struct[c(1, 3)]), 2)
})

test_that("subsetting with logical index preserves NA", {
  struct <- c(o_glycan_core_1(), NA, n_glycan_core())
  subset <- struct[c(TRUE, FALSE, TRUE)]
  expect_equal(length(subset), 2)
  expect_false(is.na(subset[1]))
  expect_false(is.na(subset[2]))
})

test_that("subsetting removes NA elements correctly", {
  struct <- c(o_glycan_core_1(), NA, n_glycan_core())
  subset <- struct[!is.na(struct)]
  expect_equal(length(subset), 2)
  expect_false(any(is.na(subset)))
})

# Tests for format with NA ---------------------------------------------------------

test_that("format preserves order with mixed NA and valid", {
  struct1 <- o_glycan_core_1()
  struct2 <- n_glycan_core()
  structs <- c(struct1, NA, struct2, NA)
  formatted <- format(structs)
  expect_false(is.na(formatted[1]))
  expect_true(grepl("^NA", formatted[2])) # format() pads with spaces
  expect_false(is.na(formatted[3]))
  expect_true(grepl("^NA", formatted[4])) # format() pads with spaces
})

test_that("format handles all NA structures", {
  structs <- glycan_structure(NA, NA)
  formatted <- format(structs)
  expect_equal(length(formatted), 2)
  expect_true(all(grepl("^NA", formatted))) # format() pads with spaces
})

# Tests for tibble printing with NA ------------------------------------------------

test_that("tibble printing handles NA structures", {
  skip_if_not_installed("tibble")
  struct <- c(o_glycan_core_1(), NA)
  df <- tibble::tibble(struct = struct, id = 1:2)
  output <- capture.output(print(df))
  # NA is displayed as "NA" string in tibble
  expect_true(any(grepl("NA", output)))
})

test_that("tibble printing handles multiple NA structures", {
  skip_if_not_installed("tibble")
  struct <- c(o_glycan_core_1(), NA, n_glycan_core(), NA)
  df <- tibble::tibble(struct = struct, id = 1:4)
  output <- capture.output(print(df))
  # Count occurrences of "NA" string (appears once per NA element in data rows)
  na_count <- sum(grepl("NA", output))
  expect_true(na_count >= 2) # At least two NA entries should appear
})

# Tests for c() combining with NA --------------------------------------------------

test_that("combining multiple structures with NA preserves type", {
  struct1 <- o_glycan_core_1()
  struct2 <- n_glycan_core()

  combined <- c(struct1, NA, struct2)
  expect_s3_class(combined, "glyrepr_structure")
  expect_equal(length(combined), 3)
  expect_false(is.na(combined[1]))
  expect_true(is.na(combined[2]))
  expect_false(is.na(combined[3]))
})

test_that("c() combines all-NA vectors", {
  structs <- glycan_structure(NA, NA)
  combined <- c(structs, glycan_structure(NA))
  expect_equal(length(combined), 3)
  expect_true(all(is.na(combined)))
})

# Tests for all-NA vector ----------------------------------------------------------

test_that("all NA structure vector has correct properties", {
  structs <- glycan_structure(NA, NA)
  expect_equal(length(structs), 2)
  expect_true(all(is.na(structs)))
  expect_equal(length(attr(structs, "graphs")), 0) # No graphs for NA
})

# Tests for vec_slice with NA ------------------------------------------------------

test_that("vec_slice preserves NA", {
  struct <- c(o_glycan_core_1(), NA, n_glycan_core())
  sliced <- vctrs::vec_slice(struct, c(1, 3))
  expect_equal(length(sliced), 2)
  expect_false(is.na(sliced[1]))
  expect_false(is.na(sliced[2]))
})

test_that("vec_slice can select NA elements", {
  struct <- c(o_glycan_core_1(), NA, n_glycan_core())
  sliced <- vctrs::vec_slice(struct, 2)
  expect_equal(length(sliced), 1)
  expect_true(is.na(sliced[1]))
})

test_that("vec_slice with negative index excludes NA correctly", {
  struct <- c(o_glycan_core_1(), NA, n_glycan_core())
  sliced <- vctrs::vec_slice(struct, -2)
  expect_equal(length(sliced), 2)
  expect_false(any(is.na(sliced)))
})

# Tests for structure_to_iupac with NA ---------------------------------------------

test_that("structure_to_iupac handles structures with NA", {
  structs <- c(o_glycan_core_1(), NA, n_glycan_core())
  iupacs <- structure_to_iupac(structs)
  expect_equal(length(iupacs), 3)
  expect_false(is.na(iupacs[1]))
  expect_true(is.na(iupacs[2]))
  expect_false(is.na(iupacs[3]))
})

# Tests for remove_linkages with NA -------------------------------------------------

test_that("remove_linkages handles single valid structure", {
  # Note: remove_linkages uses smap which currently doesn't handle NA
  # This test verifies the function works for non-NA structures
  removed <- remove_linkages(o_glycan_core_1())
  expect_equal(has_linkages(removed), FALSE)
})

# Tests for has_linkages with NA ---------------------------------------------------

test_that("has_linkages handles single valid structure", {
  # Note: has_linkages uses smap which currently doesn't handle NA
  # This test verifies the function works for non-NA structures
  result <- has_linkages(o_glycan_core_1())
  expect_equal(result, TRUE)
})

# Tests for get_anomer with NA ------------------------------------------------------

test_that("get_anomer handles single valid structure", {
  # Note: get_anomer uses smap which currently doesn't handle NA
  # This test verifies the function works for non-NA structures
  anomer <- get_anomer(o_glycan_core_1())
  expect_equal(anomer, "a1")
})

# Tests for count_mono with NA ------------------------------------------------------

test_that("count_mono handles single valid structure", {
  # Note: count_mono uses smap which currently doesn't handle NA
  # This test verifies the function works for non-NA structures
  count <- count_mono(o_glycan_core_1())
  expect_equal(length(count), 1)
  expect_false(is.na(count))
})

# Tests for class duplication fix (issue #30) -------------------------------------

test_that("class is not duplicated after c()", {
  g1 <- o_glycan_core_1()
  g2 <- n_glycan_core()

  # Individual structures should have correct class
  expect_true(inherits(g1, "glyrepr_structure"))
  expect_equal(sum(class(g1) == "vctrs_vctr"), 1)
  expect_true(inherits(g2, "glyrepr_structure"))
  expect_equal(sum(class(g2) == "vctrs_vctr"), 1)

  # After c(), class should still be correct (not duplicated)
  combined <- c(g1, g2)
  expect_true(inherits(combined, "glyrepr_structure"))
  expect_equal(sum(class(combined) == "vctrs_vctr"), 1)
  expect_equal(length(unique(class(combined))), length(class(combined)))
})

test_that("class is not duplicated after subsetting", {
  glycans <- c(o_glycan_core_1(), n_glycan_core())

  # Before subsetting
  expect_true(inherits(glycans, "glyrepr_structure"))
  expect_equal(sum(class(glycans) == "vctrs_vctr"), 1)
  expect_equal(length(unique(class(glycans))), length(class(glycans)))

  # After subsetting
  subset1 <- glycans[1]
  expect_true(inherits(subset1, "glyrepr_structure"))
  expect_equal(sum(class(subset1) == "vctrs_vctr"), 1)
  expect_equal(length(unique(class(subset1))), length(class(subset1)))

  subset2 <- glycans[c(1, 2)]
  expect_true(inherits(subset2, "glyrepr_structure"))
  expect_equal(sum(class(subset2) == "vctrs_vctr"), 1)
  expect_equal(length(unique(class(subset2))), length(class(subset2)))

  subset_empty <- glycans[integer(0)]
  expect_true(inherits(subset_empty, "glyrepr_structure"))
  expect_equal(sum(class(subset_empty) == "vctrs_vctr"), 1)
  expect_equal(length(unique(class(subset_empty))), length(class(subset_empty)))
})

test_that("vec_restore does not duplicate class", {
  g1 <- o_glycan_core_1()
  g2 <- n_glycan_core()

  ptype <- vctrs::vec_ptype2(g1, g2)
  # Prototype should have correct class
  expect_true(inherits(ptype, "glyrepr_structure"))
  expect_equal(sum(class(ptype) == "vctrs_vctr"), 1)

  # vec_restore should not duplicate class
  restored <- vctrs::vec_restore(g1, ptype)
  expect_true(inherits(restored, "glyrepr_structure"))
  expect_equal(sum(class(restored) == "vctrs_vctr"), 1)
  expect_equal(length(unique(class(restored))), length(class(restored)))
})

test_that("as_glycan_structure treats scalar NA character as missing structure", {
  result <- as_glycan_structure(NA_character_)

  expect_s3_class(result, "glyrepr_structure")
  expect_length(result, 1)
  expect_true(is.na(result))
  expect_equal(attr(result, "graphs"), list())
})

test_that("get_structure_level preserves NA structures", {
  structures <- c(o_glycan_core_1(), glycan_structure(NA))

  result <- get_structure_level(structures)

  expect_equal(result, "intact")
})

test_that("reduce_structure_level skips NA structures when checking level rank", {
  structures <- c(o_glycan_core_1(), glycan_structure(NA))

  result <- reduce_structure_level(structures, "topological")

  expect_false(has_linkages(result[1]))
  expect_true(is.na(result[2]))
})

test_that("get mono type of structures", {
  glycan <- n_glycan_core(mono_type = "concrete")
  expect_equal(get_mono_type(glycan), "concrete")

  glycan <- n_glycan_core(mono_type = "generic")
  expect_equal(get_mono_type(glycan), "generic")
})

test_that("get mono type of character vector", {
  expect_equal(get_mono_type(c("Gal", "GlcNAc")), c("concrete", "concrete"))
  expect_equal(get_mono_type(c("Hex", "HexNAc")), c("generic", "generic"))
})

test_that("get_mono_type of special monosaccharides", {
  # Some monosaccharides have the same name for both generic and concrete types.
  expect_equal(get_mono_type(c("Mur", "Neu", "Kdn")), rep("concrete", 3))
  expect_equal(get_mono_type(c("gMur", "gNeu", "gKdn")), rep("generic", 3))
})

test_that("get_mono_type of unknown monosaccharides", {
  expect_error(
    get_mono_type(c("Unknown", "Unknown2")),
    "Unknown monosaccharide"
  )
})

test_that("get mono type of composition", {
  comp_generic <- glycan_composition(
    c(Hex = 4, HexNAc = 1),
    c(Hex = 1, HexNAc = 1)
  )
  comp_concrete <- glycan_composition(c(Gal = 4, GlcNAc = 1))

  expect_equal(get_mono_type(comp_generic), "generic")
  expect_equal(get_mono_type(comp_concrete), "concrete")
})

test_that("get_mono_type of composition with substituents", {
  # Concrete composition with sulfate substituent
  comp_sulfate <- glycan_composition(c(Gal = 2, GlcNAc = 1, S = 1))
  expect_equal(get_mono_type(comp_sulfate), "concrete")

  # Generic composition with methyl substituent
  comp_methyl <- glycan_composition(c(Hex = 2, HexNAc = 1, Me = 1))
  expect_equal(get_mono_type(comp_methyl), "generic")

  # Composition with multiple different substituents
  comp_multi_sub <- glycan_composition(c(Gal = 3, GlcNAc = 2, S = 1, Ac = 1))
  expect_equal(get_mono_type(comp_multi_sub), "concrete")
})

test_that("get mono type of empty composition", {
  comp_empty <- glycan_composition()
  expect_equal(get_mono_type(comp_empty), character())
})

# Tests for convert_to_generic function
test_that("convert_to_generic works with character vectors", {
  result <- convert_to_generic(c("Gal", "GlcNAc"))
  expect_equal(result, c("Hex", "HexNAc"))
})

test_that("convert_to_generic with already generic characters returns same", {
  input <- c("Hex", "HexNAc")
  result <- convert_to_generic(input)
  expect_identical(result, input)
})

test_that("convert_to_generic works with glycan structures", {
  glycan <- n_glycan_core(mono_type = "concrete")
  glycan_generic <- convert_to_generic(glycan)

  expect_true(is_glycan_structure(glycan_generic))
  graph <- get_structure_graphs(glycan_generic, return_list = FALSE)
  expect_equal(
    igraph::V(graph)$mono,
    c("Hex", "Hex", "Hex", "HexNAc", "HexNAc")
  )
})

test_that("convert_to_generic with already generic structure returns same", {
  glycan <- n_glycan_core(mono_type = "generic")
  result <- convert_to_generic(glycan)
  expect_identical(result, glycan)
})

test_that("convert_to_generic works with glycan compositions", {
  comp_concrete <- glycan_composition(c(Gal = 2, GlcNAc = 1))
  comp_generic <- convert_to_generic(comp_concrete)

  expect_true(is_glycan_composition(comp_generic))
  expect_equal(as.character(comp_generic), "Hex(2)HexNAc(1)")
})

test_that("convert_to_generic works with glycan compositions with substituents", {
  comp <- glycan_composition(c(Gal = 2, GlcNAc = 1, Me = 1))
  result <- convert_to_generic(comp)
  expected <- glycan_composition(c(Hex = 2, HexNAc = 1, Me = 1))
  expect_equal(result, expected)
})

test_that("convert_to_generic with already generic composition returns same", {
  comp_generic <- glycan_composition(c(Hex = 2, HexNAc = 1))
  result <- convert_to_generic(comp_generic)
  expect_identical(result, comp_generic)
})

test_that("convert_to_generic with empty composition returns empty composition", {
  comp_empty <- glycan_composition()
  result <- convert_to_generic(comp_empty)
  expect_equal(result, comp_empty)
})

test_that("convert_to_generic preserves NA in compositions", {
  # NA as second element
  comps1 <- glycan_composition(c(Gal = 1), NA)
  result1 <- convert_to_generic(comps1)
  expect_equal(length(result1), 2)
  expect_equal(as.character(result1[1]), "Hex(1)")
  expect_true(is.na(result1[2]))

  # NA as first element
  comps2 <- glycan_composition(NA, c(Gal = 1))
  result2 <- convert_to_generic(comps2)
  expect_equal(length(result2), 2)
  expect_true(is.na(result2[1]))
  expect_equal(as.character(result2[2]), "Hex(1)")

  # NA in the middle
  comps3 <- glycan_composition(c(Gal = 1), NA, c(GlcNAc = 2))
  result3 <- convert_to_generic(comps3)
  expect_equal(length(result3), 3)
  expect_equal(as.character(result3[1]), "Hex(1)")
  expect_true(is.na(result3[2]))
  expect_equal(as.character(result3[3]), "HexNAc(2)")
})

test_that("convert_to_generic handles NA in glyrepr_structure", {
  # Create structure vector with NA
  # Use valid IUPAC-condensed strings from existing structures
  glycans <- as_glycan_structure(c(
    "Gal(b1-3)GalNAc(a1-",
    NA,
    "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"
  ))

  result <- convert_to_generic(glycans)
  expect_equal(length(result), 3)
  expect_false(is.na(result[1]))
  expect_true(is.na(result[2]))
  expect_false(is.na(result[3]))

  # Verify non-NA elements are converted
  expect_true(all(grepl("Hex", as.character(result[!is.na(result)]))))
})

test_that("convert_to_generic with all NA structures", {
  # All NA elements
  glycans <- as_glycan_structure(c(NA, NA))
  result <- convert_to_generic(glycans)
  expect_equal(length(result), 2)
  expect_true(all(is.na(result)))
})

test_that("get_mono_type skips NA compositions", {
  concrete <- glycan_composition(NA, c(Gal = 1))
  generic <- glycan_composition(NA, c(Hex = 1))
  all_na <- glycan_composition(NA, NA)

  expect_equal(get_mono_type(concrete), "concrete")
  expect_equal(get_mono_type(generic), "generic")
  expect_true(is.na(get_mono_type(all_na)))
})

#' Get the Structure Resolution Levels
#'
#' @description
#' Glycan structures can have four possible levels of resolution:
#' - "intact": All monosaccharides are concrete (e.g. "Man", "GlcNAc"),
#'   and no linkage or anomer contains "?".
#' - "partial": All monosaccharides are concrete (e.g. "Man", "GlcNAc"),
#'   at least one linkage or anomer contains "?",
#'   and at least one linkage or anomer has a non-"?" annotation.
#' - "topological": All monosaccharides are concrete (e.g. "Man", "GlcNAc"),
#'   and all linkages and anomers are completely unknown ("??-?"/"??").
#' - "basic": All monosaccharides are generic (e.g. "Hex", "HexNAc").
#'
#' Note that in theory you can have a glycan with generic monosaccharides with all linkages determined.
#' For example, "Hex(b1-3)HexNAc(a1-" is a valid glycan structure.
#' But in reality, this is almost impossible,
#' because linkage information is far more difficult to acquire than monosaccharide information.
#' This kind of glycan structure is also assigned to "basic" level.
#'
#' @param x A [glycan_structure()] vector.
#'
#' @returns A character scalar containing the structure level for `x`.
#'   If `x` is empty or all structures in `x` are NA, returns NA_character_.
#'
#' @examples
#' glycan <- as_glycan_structure("Gal(b1-3)GalNAc(a1-")
#' get_structure_level(glycan)
#'
#' @seealso [has_linkages()], [get_mono_type()]
#' @export
get_structure_level <- function(x) {
  checkmate::assert_class(x, "glyrepr_structure")

  non_na <- !structure_na_mask(x)

  if (!any(non_na)) {
    return(NA_character_)
  }
  if (length(x) == 0) {
    return(NA_character_)
  }

  x_valid <- x[non_na]
  mono_type <- get_mono_type.glyrepr_structure(x_valid)
  has_linkages_strict <- has_linkages(x_valid, strict = TRUE)
  has_linkages_lenient <- has_linkages(x_valid, strict = FALSE)

  if (mono_type == "generic") {
    .warn_generic_linkage_structure_level(has_linkages_lenient)
    return("basic")
  }

  if (all(has_linkages_strict)) {
    return("intact")
  }

  if (any(has_linkages_lenient)) {
    return("partial")
  }

  "topological"
}

#' Warn About Generic Structures With Linkage Annotation
#'
#' Generic structures are always treated as basic resolution,
#' even if they contain linkage or anomer annotations.
#'
#' @param has_linkages_lenient A logical vector returned by [has_linkages()]
#'   with `strict = FALSE`.
#' @returns Nothing. Called for its warning side effect.
#' @noRd
.warn_generic_linkage_structure_level <- function(has_linkages_lenient) {
  if (any(has_linkages_lenient)) {
    cli::cli_warn(c(
      "Generic glycan structures with linkage annotations are treated as {.val basic}.",
      "i" = "Linkage information is ignored when residues are generic."
    ), class = "glyrepr_warning_generic_structure_linkages")
  }
}

#' Reduce a Glycan Structure to a Lower Resolution Level
#'
#' This function reduces a glycan structure from a higher resolution level to a lower resolution level
#' (see [get_structure_level()] for four possible levels of resolution).
#' For example, it can reduce an "intact" structure to a "topological" structure,
#' or a "partial" structure to a "basic" structure.
#' One exception is that you can never reduce an "intact" structure to "partial" level,
#' because the "partial" level is not deterministic.
#'
#' @details
#' The logic is as follows:
#' - If `to_level` is "topological", this function calls [remove_linkages()] to remove all linkages.
#' - If `to_level` is "basic", this function calls [remove_linkages()] to remove all linkages,
#'   and [convert_to_generic()] to convert all monosaccharides to generic.
#'
#' @param x A [glycan_structure()] vector.
#' @param to_level The resolution level to reduce to. Can be "basic" or "topological".
#'   Must be a lower resolution level than `x`
#'   ("intact" > "partial" > "topological" > "basic").
#'   If `to_level` is the same as the structure level of `x`,
#'   the result will be the same as the input.
#'   You can use [get_structure_level()] to check the structure level of `x`.
#'
#' @returns A [glycan_structure()] vector reduced to the given resolution level.
#' @examples
#' glycan <- as_glycan_structure("Gal(b1-3)GalNAc(a1-")
#' reduce_structure_level(glycan, to_level = "topological")
#'
#' @seealso [get_structure_level()]
#' @export
reduce_structure_level <- function(x, to_level) {
  checkmate::assert_class(x, "glyrepr_structure")
  checkmate::assert_choice(to_level, c("basic", "topological"))

  from_level <- get_structure_level(x)

  if (is.na(from_level)) {
    # Two situations can lead to NA structure level:
    #. 1. x is empty.
    #. 2. All structures in x are NA.
    # In both cases, we can just return x without any modification.
    return(x)
  }

  level_ranks <- c("basic" = 1, "topological" = 2, "partial" = 3, "intact" = 4)
  if (level_ranks[[from_level]] < level_ranks[[to_level]]) {
    cli::cli_abort(c(
      "Cannot reduce a structure to a higher resolution level.",
      "x" = "Structure level of {.arg x}: {.val {from_level}}.",
      "i" = "Target level: {.val {to_level}} (> {.val {from_level}}).",
      "i" = "You can use {.fn get_structure_level} to check the structure level of {.arg x}."
    ))
  }

  x <- remove_linkages(x)
  if (to_level == "basic") {
    x <- convert_to_generic(x)
  }
  x
}

test_that("get_structure_level works for each glycan separately", {
  glycan1 <- as_glycan_structure("Gal(b1-3)GalNAc(a1-")
  glycan2 <- as_glycan_structure("Gal(b1-?)GalNAc(a1-")
  glycan3 <- as_glycan_structure("Gal(??-?)GalNAc(??-")
  glycan4 <- as_glycan_structure("Hex(??-?)HexNAc(??-")
  glycan5 <- as_glycan_structure("Hex(b1-3)HexNAc(a1-")

  expect_equal(get_structure_level(glycan1), "intact")
  expect_equal(get_structure_level(glycan2), "partial")
  expect_equal(get_structure_level(glycan3), "topological")
  expect_equal(get_structure_level(glycan4), "basic")
  expect_warning(
    expect_equal(get_structure_level(glycan5), "basic"),
    class = "glyrepr_warning_generic_structure_linkages"
  )
})

test_that("get_structure_level works for an intact glycan vector", {
  glycans <- as_glycan_structure(c(
    "Gal(b1-3)GalNAc(a1-",
    "GalNAc(a1-"
  ))
  expect_equal(get_structure_level(glycans), "intact")
})

test_that("get_structure_level works for a partial glycan vector with only one unknown linkage", {
  glycans <- as_glycan_structure(c(
    "Gal(b1-?)GalNAc(a1-",
    "GalNAc(a1-"
  ))
  expect_equal(get_structure_level(glycans), "partial")
})

test_that("get_structure_level works for a partial glycan vector with only one known linkage", {
  glycans <- as_glycan_structure(c(
    "Gal(?1-?)GalNAc(??-",
    "GalNAc(??-"
  ))
  expect_equal(get_structure_level(glycans), "partial")
})

test_that("get_structure_level works for a partial glycan vector with only one known reducing end configuration", {
  glycans <- as_glycan_structure(c(
    "Gal(??-?)GalNAc(??-",
    "GalNAc(?1-"
  ))
  expect_equal(get_structure_level(glycans), "partial")
})

test_that("get_structure_level works for a topological glycan vector", {
  glycans <- as_glycan_structure(c(
    "Gal(??-?)GalNAc(??-",
    "GalNAc(??-"
  ))
  expect_equal(get_structure_level(glycans), "topological")
})

test_that("get_structure_level works for a basic glycan vector", {
  glycans <- as_glycan_structure(c(
    "Hex(??-?)HexNAc(??-",
    "HexNAc(??-"
  ))
  expect_equal(get_structure_level(glycans), "basic")
})

test_that("get_structure_level works for a basic glycan vector with linkages", {
  glycans <- as_glycan_structure(c(
    "Hex(b1-3)HexNAc(a1-",
    "HexNAc(a1-"
  ))
  expect_snapshot(res <- get_structure_level(glycans))
  # should warn about the rare case of a generic glycan with linkages,
  # and tell the user that it will be treated as basic

  expect_equal(res, "basic")
})

test_that("get_structure_level ignores NA", {
  glycans <- as_glycan_structure(c(
    "Hex(??-?)HexNAc(??-",
    "HexNAc(??-",
    NA
  ))
  expect_equal(get_structure_level(glycans), "basic")
})

test_that("get_structure_level return NA_character_ for all-NA vector", {
  glycans <- as_glycan_structure(c(NA, NA))
  expect_equal(get_structure_level(glycans), NA_character_)
})

test_that("get_structure_level returns NA_character_ for empty vector", {
  glycans <- as_glycan_structure(character())
  expect_equal(get_structure_level(glycans), NA_character_)
})

test_that("reduce_structure_level works for each glycan separately", {
  glycan_intact <- as_glycan_structure("Gal(b1-3)GalNAc(a1-")
  glycan_partial <- as_glycan_structure("Gal(b1-?)GalNAc(a1-")
  glycan_topological <- as_glycan_structure("Gal(??-?)GalNAc(??-")

  # intact to topological
  expect_equal(
    as.character(reduce_structure_level(
      glycan_intact,
      to_level = "topological"
    )),
    "Gal(??-?)GalNAc(??-"
  )
  # partial to topological
  expect_equal(
    as.character(reduce_structure_level(
      glycan_partial,
      to_level = "topological"
    )),
    "Gal(??-?)GalNAc(??-"
  )
  # topological to topological (identity)
  expect_equal(
    as.character(reduce_structure_level(
      glycan_topological,
      to_level = "topological"
    )),
    "Gal(??-?)GalNAc(??-"
  )

  # intact to basic
  expect_equal(
    as.character(reduce_structure_level(glycan_intact, to_level = "basic")),
    "Hex(??-?)HexNAc(??-"
  )
  # partial to basic
  expect_equal(
    as.character(reduce_structure_level(glycan_partial, to_level = "basic")),
    "Hex(??-?)HexNAc(??-"
  )
  # topological to basic
  expect_equal(
    as.character(reduce_structure_level(
      glycan_topological,
      to_level = "basic"
    )),
    "Hex(??-?)HexNAc(??-"
  )
})

test_that("reduce_structure_level rejects higher level", {
  glycan <- as_glycan_structure("Hex(??-?)HexNAc(??-")
  expect_snapshot(
    reduce_structure_level(glycan, to_level = "topological"),
    error = TRUE
  )
})

test_that("reduce_structure_level works for multiple glycans", {
  glycans <- as_glycan_structure(c(
    "Gal(b1-3)GalNAc(a1-",
    "Gal(b1-?)GalNAc(a1-",
    "Gal(??-?)GalNAc(??-"
  ))
  expect_equal(
    as.character(reduce_structure_level(glycans, to_level = "topological")),
    c("Gal(??-?)GalNAc(??-", "Gal(??-?)GalNAc(??-", "Gal(??-?)GalNAc(??-")
  )
})

test_that("reduce_structure_level works for all-NA vectors", {
  glycans <- as_glycan_structure(c(NA, NA))
  expect_equal(reduce_structure_level(glycans, to_level = "topological"), glycans)
})

test_that("reduce_structure_level works for empty vectors", {
  glycans <- as_glycan_structure(character())
  expect_equal(reduce_structure_level(glycans, to_level = "topological"), glycans)
})

#' Convert Monosaccharides to Generic Type
#'
#' This function converts monosaccharide types of monosaccharide characters,
#' glycan compositions, or glycan structures from concrete to generic type.
#' This is a simplified version that only supports conversion from "concrete"
#' to "generic" monosaccharides.
#'
#' @details
#'
#' # Two types of monosaccharides
#'
#' There are two types of monosaccharides:
#' - concrete: e.g. "Gal", "GlcNAc", "Glc", "Fuc", etc.
#' - generic: e.g. "Hex", "HexNAc", "HexA", "HexN", etc.
#'
#' For the full list of monosaccharides, use [available_monosaccharides()].
#'
#' @param x Either of these objects:
#'   - A character of monosaccharide;
#'   - A glycan composition vector ("glyrepr_composition" object);
#'   - A glycan structure vector ("glyrepr_structure" object).
#'
#' @returns A new object of the same class as `x`
#' with monosaccharides converted to generic type.
#'
#' @examples
#' # Convert character vectors
#' convert_to_generic(c("Gal", "GlcNAc"))
#'
#' # Convert glycan compositions
#' comps <- glycan_composition(
#'   c(Gal = 5, GlcNAc = 2),
#'   c(Glc = 5, GalNAc = 4, Fuc = 1)
#' )
#' convert_to_generic(comps)
#'
#' # Convert glycan structures
#' strucs <- c(n_glycan_core(), o_glycan_core_1())
#' convert_to_generic(strucs)
#'
#' @export
convert_to_generic <- function(x) {
  UseMethod("convert_to_generic")
}

#' @export
#' @rdname convert_to_generic
convert_to_generic.character <- function(x) {
  checkmate::assert_character(x)

  from <- get_mono_type(x)

  # Check if conversion is valid (no backward conversion)
  if (any(from == "generic")) {
    # Already generic, return as-is
    return(x)
  }

  convert_mono_type_impl(x)
}

#' @export
#' @rdname convert_to_generic
convert_to_generic.glyrepr_structure <- function(x) {
  if (!is_glycan_structure(x)) {
    cli::cli_abort(c(
      "Input must be a glyrepr_structure vector.",
      "i" = "Use `glycan_structure()` to create a glyrepr_structure from igraph objects."
    ))
  }

  # Get current mono types
  from <- get_mono_type(x)

  # Use spmap_structure with optimized implementation
  spmap_structure(list(x, from), function(graph, from) {
    if (from == "generic") {
      return(graph)
    }
    convert_glycan_mono_type_impl(graph, from, "generic")
  })
}

#' @export
#' @rdname convert_to_generic
convert_to_generic.glyrepr_composition <- function(x) {
  if (!is_glycan_composition(x)) {
    cli::cli_abort(c(
      "Input must be a glyrepr_composition vector.",
      "i" = "Use `glycan_composition()` to create a glyrepr_composition from named vectors."
    ))
  }

  if (length(x) == 0) {
    return(x)
  }

  # Get current mono types
  current_type <- get_mono_type.glyrepr_composition(x)
  if (is.na(current_type)) {
    return(x)
  }
  if (current_type == "generic") {
    return(x)
  }

  # Convert each composition
  compositions <- vctrs::field(x, "data")
  new_compositions <- purrr::map(compositions, function(comp) {
    # Handle NULL/empty compositions (from NA elements)
    if (is.null(comp) || length(comp) == 0) {
      return(NULL) # NA elements are stored as NULL
    }

    # Convert monosaccharide names
    old_names <- names(comp)
    new_names <- convert_mono_type_impl(old_names)

    # Create new composition with converted names, aggregating counts for duplicates
    result <- tapply(comp, new_names, sum, na.rm = TRUE)
    res_names <- names(result)
    result <- as.integer(result)
    names(result) <- res_names

    .reorder_composition_components(result)
  })

  # Create new composition vector
  new_glycan_composition(new_compositions)
}

#' Get Monosaccharide Types
#'
#' This function determines the type of monosaccharides in character vectors,
#' glycan compositions, or glycan structures.
#' Supported types: "concrete" and "generic" (see details below).
#'
#' @details
#'
#' # Two types of monosaccharides
#'
#' There are two types of monosaccharides:
#' - concrete: e.g. "Gal", "GlcNAc", "Glc", "Fuc", etc.
#' - generic: e.g. "Hex", "HexNAc", "HexA", "HexN", etc.
#'
#' For the full list of monosaccharides, use [available_monosaccharides()].
#'
#' # Special monosaccharides
#'
#' Some monosaccharides are special in that they have no generic names in database or literature.
#' For example, "Mur" is a rare monosaccharide that has no popular generic name.
#' In `glyrepr`, we assign a "g" prefix to these monosaccharides as their generic names.
#' This includes "gNeu", "gKdn", "gPse", "gLeg", "gAci", "g4eLeg", "gBac", "gKdo", "gMur".
#' These names might only be meaningful inside `glycoverse`.
#' Take care when you export results from `glycoverse` functions to other analysis tools.
#'
#' @param x Either of these objects:
#'   - A character vector of monosaccharide names;
#'   - A glycan composition vector ("glyrepr_composition" object);
#'   - A glycan structure vector ("glyrepr_structure" object).
#'
#' @returns
#'   - For character input, returns a character vector of the same length as `x`.
#'   - For `glyrepr_structure` and `glyrepr_composition` input, returns a character scalar.
#'
#' @examples
#' # Character vector
#' get_mono_type(c("Gal", "Hex"))
#'
#' # Glycan structures
#' get_mono_type(n_glycan_core(mono_type = "concrete"))
#' get_mono_type(n_glycan_core(mono_type = "generic"))
#'
#' # Glycan compositions
#' comp <- glycan_composition(c(Glc = 2, GalNAc = 1))
#' get_mono_type(comp)
#'
#' @seealso [convert_to_generic()]
#'
#' @export
get_mono_type <- function(x) {
  UseMethod("get_mono_type")
}

#' @export
#' @rdname get_mono_type
get_mono_type.character <- function(x) {
  checkmate::assert_character(x)
  result <- vector("character", length = length(x))
  is_concrete <- x %in% monosaccharides$concrete
  is_generic <- x %in% monosaccharides$generic
  is_unknown <- !(is_concrete | is_generic)
  if (any(is_unknown)) {
    cli::cli_abort("Unknown monosaccharide: {.val {x[is_unknown]}}.")
  }
  result[is_concrete] <- "concrete"
  result[is_generic] <- "generic"
  result
}

#' @export
#' @rdname get_mono_type
get_mono_type.glyrepr_structure <- function(x) {
  if (!is_glycan_structure(x)) {
    cli::cli_abort(c(
      "Input must be a glyrepr_structure vector.",
      "i" = "Use `glycan_structure()` to create a glyrepr_structure from igraph objects."
    ))
  }

  if (length(x) == 0) {
    return(character())
  }

  graphs <- attr(x, "graphs")

  # Handle all-NA vectors (empty graphs list)
  if (length(graphs) == 0) {
    return(NA_character_)
  }

  graphs1 <- graphs[[1]]
  monos1 <- igraph::V(graphs1)$mono
  get_mono_type_impl(monos1)
}

#' @export
#' @rdname get_mono_type
get_mono_type.glyrepr_composition <- function(x) {
  if (!is_glycan_composition(x)) {
    cli::cli_abort(c(
      "Input must be a glyrepr_composition vector.",
      "i" = "Use `glycan_composition()` to create a glyrepr_composition from named vectors."
    ))
  }

  if (length(x) == 0) {
    return(character())
  }

  # All non-missing compositions in a composition vector must be of the same type.
  # Therefore, we just need to check the first non-missing composition.
  # Filter out substituents before determining monosaccharide type.
  compositions <- vctrs::field(x, "data")
  compositions <- purrr::discard(compositions, .is_na_composition_elem)
  if (length(compositions) == 0) {
    return(NA_character_)
  }

  comp_names <- names(compositions[[1]])
  mono_names <- comp_names[!comp_names %in% available_substituents()]
  get_mono_type_impl(mono_names)
}

#' Decide mono type from a vector of monosaccharide names
#'
#' This function handles the special cases where some monosaccharides
#' have the same name for both generic and concrete types.
#' For example, "Mur" is both a generic and concrete monosaccharide.
#'
#' This function is used internally when creating [glycan_composition()] and [glycan_structure()].
#'
#' @param x A character vector of monosaccharide names.
#' @returns A character scalar of monosaccharide type. Can be "concrete", "generic", "mixed", or "unknown".
#' @noRd
get_mono_type_impl <- function(x) {
  types <- tryCatch(
    get_mono_type.character(x),
    error = function(e) "unknown"
  )

  unique_types <- unique(types)
  if (length(unique_types) > 1) {
    "mixed"
  } else {
    unique_types[[1]]
  }
}

get_graph_mono_type <- function(graph) {
  # This function is the implementation of get_mono_type for a single glycan graph.
  monos <- igraph::vertex_attr(graph, "mono")
  get_mono_type_impl(monos)
}

#' Convert monosaccharide names to generic type
#'
#' This function converts concrete monosaccharide names to generic type.
#' Generic monosaccharides are returned as is.
#' It assumes that all monosaccharides are of the same type.
#' Substituents are not converted.
#'
#' @param monos A character vector of monosaccharide names.
#' @returns A character vector of monosaccharide names in generic type.
#' @noRd
convert_mono_type_impl <- function(monos) {
  mono_type <- get_mono_type_impl(monos)
  if (mono_type == "generic") {
    return(monos)
  }
  if (mono_type == "mixed") {
    cli::cli_abort("Mixed monosaccharide types are not supported.")
  }
  from_ <- monosaccharides[["concrete"]]
  to_ <- monosaccharides[["generic"]]

  # Handle NA values - preserve them
  is_na <- is.na(monos)
  if (any(is_na)) {
    # If all values are NA, return as-is
    if (all(is_na)) {
      return(monos)
    }
    result <- monos
    # Convert non-NA values
    non_na <- !is_na
    result[non_na] <- convert_mono_type_impl(monos[non_na])
    return(result)
  }

  # Non-NA case: convert using base R vectorized operations
  is_substituent <- monos %in% available_substituents()
  result <- to_[match(monos, from_)]
  result[is_substituent] <- monos[is_substituent]
  result
}

convert_glycan_mono_type_impl <- function(glycan, from, to) {
  old_names <- igraph::V(glycan)$mono
  new_names <- convert_mono_type_impl(old_names)
  raise_error_for_na(old_names, new_names, to)
  igraph::set_vertex_attr(glycan, "mono", value = new_names)
}

raise_error_for_na <- function(old_names, new_names, to) {
  bad_names <- old_names[is.na(new_names)]
  if (length(bad_names) > 0) {
    cli::cli_abort(
      "Some monosaccharides cannot be converted to {.val {to}}: {.val {bad_names}}.",
      call = rlang::expr(convert_to_generic())
    )
  }
}

#' Get the Anomeric information
#'
#' @param x A glycan structure vector (glyrepr_structure).
#' @returns a character vector of the anomeric information.
#' @export
#'
#' @examples
#' x <- n_glycan_core()
#' get_anomer(x)
get_anomer <- function(x) {
  smap_chr(x, function(g) g$anomer)
}

#' Determine if a Glycan Structure has Linkages
#'
#' Unknown linkages in a glycan structure are represented by "??-?".
#' Also, a linkage can be partially known (e.g. "a?-?").
#' This function checks if a glycan structure has linkages,
#' in a strict or lenient way.
#'
#' @param glycan A [glycan_structure()] vector.
#' @param strict A logical value.
#'   * If `FALSE` (default), a glycan is considered to have linkages if any
#'     linkage is partially known (not "??-?").
#'   * If `TRUE`, a glycan is considered to have linkages only if all linkages
#'     are fully determined (no "?" in the linkage).
#'
#' @returns A logical vector indicating if each glycan structure has linkages.
#'
#' @examples
#' glycan <- o_glycan_core_1(linkage = TRUE)
#' has_linkages(glycan)
#' print(glycan)
#'
#' glycan <- remove_linkages(glycan)
#' has_linkages(glycan)
#' print(glycan)
#'
#' glycan <- as_glycan_structure("Gal(b1-?)GalNAc(a1-")
#' has_linkages(glycan)
#' has_linkages(glycan, strict = TRUE)
#'
#' @seealso [remove_linkages()], [possible_linkages()]
#'
#' @export
has_linkages <- function(glycan, strict = FALSE) {
  checkmate::assert_class(glycan, "glyrepr_structure")
  checkmate::assert_flag(strict)

  smap_lgl(glycan, .has_linkages_single, strict = strict)
}

# Internal function to check linkages in a single igraph
.has_linkages_single <- function(glycan, strict) {
  if (strict) {
    linkages <- igraph::E(glycan)$linkage
    anomer <- glycan$anomer
    all(!stringr::str_detect(c(linkages, anomer), stringr::fixed("?")))
  } else {
    any(igraph::E(glycan)$linkage != "??-?") | glycan$anomer != "??"
  }
}

#' Generate Possible Linkages
#'
#' @description
#' Given an obscure linkage format (having "?", e.g. "a2-?"),
#' this function generates all possible linkages based on the format.
#' See [valid_linkages()] for details.
#'
#' The ranges of possible anomers, first positions, and second positions
#' can be specified using `anomer_range`, `pos1_range`, and `pos2_range`.
#'
#' @param linkage A linkage string.
#' @param anomer_range A character vector of possible anomers.
#' Default is `c("a", "b")`.
#' @param pos1_range A numeric vector of possible first positions.
#' Default is `1:2`.
#' @param pos2_range A numeric vector of possible second positions.
#' Default is `1:9`.
#' @param include_unknown A logical value. If `TRUE`, "?" will be included.
#' Default is `FALSE`.
#'
#' @returns A character vector of possible linkages.
#'
#' @examples
#' possible_linkages("a2-?")
#' possible_linkages("??-2")
#' possible_linkages("a1-3")
#' possible_linkages("a?-?", pos1_range = 2, pos2_range = c(2, 3))
#' possible_linkages("?1-6", include_unknown = TRUE)
#'
#' @seealso [has_linkages()], [remove_linkages()], [valid_linkages()]
#'
#' @export
possible_linkages <- function(
  linkage,
  anomer_range = c("a", "b"),
  pos1_range = 1:2,
  pos2_range = 1:9,
  include_unknown = FALSE
) {
  # Input checks
  checkmate::assert_character(linkage, len = 1)
  checkmate::assert_character(anomer_range, pattern = "^[ab]$", unique = TRUE)
  checkmate::assert_numeric(pos1_range, lower = 1, upper = 2, unique = TRUE)
  checkmate::assert_numeric(pos2_range, lower = 1, upper = 9, unique = TRUE)
  checkmate::assert_flag(include_unknown)

  # Check if the linkage is valid
  if (!valid_linkages(linkage)) {
    cli::cli_abort("Invalid linkage format.")
  }

  # Add unknown elements
  if (include_unknown) {
    anomer_range <- c(anomer_range, "?")
    pos1_range <- c(pos1_range, "?")
    pos2_range <- c(pos2_range, "?")
  }

  # Possible linkage elements
  current_anomer <- stringr::str_sub(linkage, 1, 1)
  current_pos1 <- stringr::str_sub(linkage, 2, 2)
  current_pos2 <- stringr::str_sub(linkage, 4, -1)

  anomers <- if (current_anomer == "?") anomer_range else current_anomer
  first_positions <- if (current_pos1 == "?") pos1_range else current_pos1
  if (current_pos2 == "?") {
    second_positions <- pos2_range
  } else if (stringr::str_detect(current_pos2, "/")) {
    second_positions <- stringr::str_split(current_pos2, "/")[[1]]
  } else {
    second_positions <- current_pos2
  }

  # Generate possible linkages
  purrr::pmap_chr(
    expand.grid(anomers, first_positions, second_positions),
    ~ paste0(..1, ..2, "-", ..3)
  )
}


#' Remove All Linkages from a Glycan
#'
#' This function replaces all linkages in a glycan structure with "??-?",
#' as well as the reducing end anomer with "??-".
#'
#' @param glycan A glyrepr_structure vector.
#'
#' @returns A glyrepr_structure vector with all linkages removed.
#'
#' @examples
#' glycan <- o_glycan_core_1(linkage = TRUE)
#' glycan
#' remove_linkages(glycan)
#'
#' @export
remove_linkages <- function(glycan) {
  if (!is_glycan_structure(glycan)) {
    cli::cli_abort(c(
      "Input must be a glyrepr_structure vector.",
      "i" = "Use `glycan_structure()` to create a glyrepr_structure from igraph objects."
    ))
  }

  smap_structure(glycan, .remove_linkages_single)
}

# Internal function to remove linkages from a single igraph
.remove_linkages_single <- function(glycan) {
  res <- igraph::set_edge_attr(glycan, "linkage", value = "??-?")
  res <- igraph::set_graph_attr(res, "anomer", value = "??")
  res
}

#' Convert Glycan Structure to IUPAC-like Sequence
#'
#' @description
#' Convert a glycan structure to a sequence representation in the form of
#' mono(linkage)mono, with branches represented by square brackets [].
#' The backbone is chosen as the longest path, and for branches, linkages are
#' ordered lexicographically with smaller linkages on the backbone.
#'
#' @details
#' # Sequence Format
#'
#' The sequence follows the format mono(linkage)mono, where:
#' - mono: monosaccharide name with optional substituents (e.g., Glc, GlcNAc, Glc3Me)
#' - linkage: glycosidic linkage (e.g., b1-4, a1-3)
#' - Branches are enclosed in square brackets []
#' - Substituents are appended directly to monosaccharide names (e.g., Glc3Me for Glc with 3Me substituent)
#'
#' # Backbone Selection
#'
#' The backbone is selected as the longest path in the tree. For branches,
#' the same rule applies recursively.
#'
#' # Linkage Comparison
#'
#' Linkages are compared lexicographically:
#' 1. First by anomeric configuration: ? > b > a
#' 2. Then by first position: ? > numbers (numerically)
#' 3. Finally by second position: ? > numbers (numerically)
#'
#' Smaller linkages are placed on the backbone, larger ones in branches.
#'
#' @param glycan A glyrepr_structure vector.
#'
#' @returns A character vector representing the IUPAC sequences.
#'
#' @examples
#' # Simple linear structure
#' structure_to_iupac(o_glycan_core_1())
#'
#' # Branched structure
#' structure_to_iupac(n_glycan_core())
#'
#' # Structure with substituents
#' graph <- igraph::make_graph(~ 1-+2)
#' igraph::V(graph)$mono <- c("Glc", "GlcNAc")
#' igraph::V(graph)$sub <- c("3Me", "6Ac")
#' igraph::E(graph)$linkage <- "b1-4"
#' graph$anomer <- "a1"
#' glycan <- glycan_structure(graph)
#' structure_to_iupac(glycan)  # Returns "GlcNAc6Ac(b1-4)Glc3Me(a1-"
#'
#' # Vectorized structures
#' structs <- c(o_glycan_core_1(), n_glycan_core())
#' structure_to_iupac(structs)
#'
#' @export
structure_to_iupac <- function(glycan) {
  if (!is_glycan_structure(glycan)) {
    cli::cli_abort(c(
      "Input must be a glyrepr_structure vector.",
      "i" = "Use `glycan_structure()` to create a glyrepr_structure from igraph objects."
    ))
  }

  vctrs::vec_data(glycan)
}

# Internal function to convert a single igraph to IUPAC
.structure_to_iupac_single <- function(glycan) {
  root <- which(igraph::degree(glycan, mode = "in") == 0)
  seq_cache <- build_seq_cache(glycan, root)

  # Step 1: Generate pseudo-IUPAC sequence starting from root
  # `pseudo_seq` is a string like "V1E1[V2E2]V3E3".
  # V1, V2, V3 are the vertex indices, E1, E2, E3 are the edge indices.
  pseudo_seq <- seq_glycan(root, seq_cache)

  # Step 2: Replace vertex and edge indices with actual monosaccharides and linkages
  real_seq <- replace_mono_and_link(pseudo_seq, glycan)

  anomer <- glycan$anomer
  paste0(real_seq, "(", anomer, "-")
}

#' Calculate linkage rank used for ordering linkages
#'
#' This is the reciprocal of the second position of the linkage.
#' "?" always rank the highest, so assigned 1.
#' For example, "b1-4" has rank 1/4, "a2-3" has rank 1/3, "a?-?" has rank 1.
#'
#' @param linkages Character vector in format "xy-z" (e.g., "b1-4", "a2-3")
#' @returns Numeric rank of the linkage.
#' @noRd
calculate_linkage_rank <- function(linkages) {
  pos2 <- stringr::str_split_i(linkages, "-", 2) # second position
  suppressWarnings(
    dplyr::if_else(
      pos2 == "?" | stringr::str_detect(pos2, "/"),
      1,
      1 / as.numeric(pos2)
    )
  )
}

#' Calculate depth (longest path to leaf) for each node
#'
#' @param glycan An igraph object representing a glycan structure
#' @param root Root vertex
#' @returns Numeric vector of depths for each node (indexed by vertex id)
#' @noRd
calculate_depths <- function(glycan, root) {
  vcount <- igraph::vcount(glycan)
  depths <- rep(NA_real_, vcount)

  # Use a local function to calculate depths with memoization
  calculate_single_depth <- function(node) {
    node_index <- as.integer(node)

    if (!is.na(depths[node_index])) {
      return(depths[node_index])
    }

    # Get child nodes
    children <- as.integer(igraph::neighbors(glycan, node_index, mode = "out"))

    if (length(children) == 0) {
      # Leaf node
      depths[node_index] <<- 0
    } else {
      # Calculate depth for all children first
      child_depths <- vapply(children, calculate_single_depth, numeric(1))

      # This node's depth is 1 + max child depth
      depths[node_index] <<- 1 + max(child_depths)
    }

    depths[node_index]
  }

  # Calculate depths for all nodes starting from root
  calculate_single_depth(root)

  # Fill in any remaining nodes (shouldn't be needed for connected graph)
  for (v in seq_len(vcount)) {
    if (is.na(depths[v])) {
      calculate_single_depth(v)
    }
  }

  depths
}

#' Build adjacency cache for sequence generation
#'
#' It caches the following information:
#' - children: `children[[i]]` is all the children ids of vertex i.
#' - edge_ids: `edge_ids[[i]]` is all the edge ids directed from vertex i.
#' - linkages: `linkages[[i]]` is all the linkage attributes (e.g. "b1-4") of the edges directed from vertex i.
#' - depths: `depths[i]` is the maximum depth of the subtree rooted at vertex i.
#' - signatures: `signatures[i]` is the signature of the subtree rooted at vertex i.
#'   The signature of a node is a string used for branch ordering in ties breaking.
#'
#' @param glycan An igraph object representing a glycan structure
#' @param root Root vertex index used for depth calculation
#' @returns List containing child vertices, edge ids, linkages per parent, and node depths
#' @noRd
build_seq_cache <- function(glycan, root) {
  vcount <- igraph::vcount(glycan)
  edge_ids <- seq_len(igraph::ecount(glycan))
  edge_vertices <- igraph::ends(glycan, igraph::E(glycan), names = FALSE)
  edge_linkages <- igraph::edge_attr(glycan, "linkage")

  children <- vector("list", vcount)
  parent_edge_ids <- vector("list", vcount)
  parent_linkages <- vector("list", vcount)

  for (edge_id in edge_ids) {
    parent <- edge_vertices[edge_id, 1]
    child <- edge_vertices[edge_id, 2]

    children[[parent]] <- c(children[[parent]], child)
    parent_edge_ids[[parent]] <- c(parent_edge_ids[[parent]], edge_id)
    parent_linkages[[parent]] <- c(
      parent_linkages[[parent]],
      edge_linkages[edge_id]
    )
  }

  # ===== Calculate node signatures =====
  mono_vec <- igraph::vertex_attr(glycan, "mono")
  sub_vec <- igraph::vertex_attr(glycan, "sub")
  mono_sub <- ifelse(
    is.na(sub_vec) | sub_vec == "",
    mono_vec,
    paste0(mono_vec, stringr::str_remove_all(sub_vec, ","))
  )

  signature <- rep(NA_character_, vcount)

  compute_sig <- function(node) {
    if (!is.na(signature[node])) {
      return(signature[node])
    }

    kids <- children[[node]]

    if (is.null(kids) || length(kids) == 0) {
      signature[node] <<- mono_sub[node]
    } else {
      toks <- character(length(kids))
      for (i in seq_along(kids)) {
        kid <- kids[i]
        link_i <- parent_linkages[[node]][i]
        toks[i] <- paste0(link_i, "->", compute_sig(kid))
      }

      toks <- sort(toks)

      signature[node] <<- paste0(
        mono_sub[node],
        "{",
        paste0(toks, collapse = ","),
        "}"
      )
    }

    signature[node]
  }

  # trigger calculation of all node signatures
  for (n in seq_len(vcount)) {
    compute_sig(n)
  }
  # ===== End of calculating node signatures =====

  list(
    children = children,
    edge_ids = parent_edge_ids,
    linkages = parent_linkages,
    depths = calculate_depths(glycan, root),
    signatures = signature
  )
}

#' Generate glycan sequence recursively using pseudo-IUPAC format
#'
#' @param node Current node
#' @param cache Precomputed adjacency and edge metadata
#' @returns Character string representing the pseudo-IUPAC sequence
#' @noRd
seq_glycan <- function(node, cache) {
  children <- cache$children[[node]]
  edge_ids <- cache$edge_ids[[node]]

  # Base case: leaf node
  if (is.null(children) || length(children) == 0) {
    # Return vertex index with V prefix
    return(paste0("V", as.character(node)))
  }

  if (length(children) > 1) {
    children_order <- order_branches(node, cache)
    backbone_child <- children[[children_order$backbone]]
    backbone_edge_id <- edge_ids[[children_order$backbone]]
    backbone_seq <- seq_glycan(backbone_child, cache)
    branch_children <- children[children_order$branches]
    branch_edge_ids <- edge_ids[children_order$branches]
    branch_seqs <- purrr::map_chr(branch_children, ~ seq_glycan(.x, cache))
    branch_seqs <- paste0("[", branch_seqs, "E", branch_edge_ids, "]")
  } else {
    backbone_child <- children[[1]]
    backbone_edge_id <- edge_ids[[1]]
    backbone_seq <- seq_glycan(backbone_child, cache)
    branch_seqs <- ""
  }

  # Combine: backbone + E<edge_index> + branches + V<node_index>
  branches_str <- paste0(branch_seqs, collapse = "")
  paste0(
    backbone_seq,
    "E",
    backbone_edge_id,
    branches_str,
    "V",
    as.character(node)
  )
}

#' Order branches
#'
#' This function orders all branches of a node by depth, then by linkages.
#'
#' @param node A node index.
#' @param cache Precomputed adjacency and edge metadata
#' @returns A list of two elements:
#'   - "backbone": the index of the backbone child.
#'   - "branches": the indices of the branches in order.
#'   All indices can only be used on `cache$children[[node]]`.
#'   For example, when `backbone` is 2, it means `cache$children[[node]][[2]]` is the backbone child.
#' @noRd
order_branches <- function(node, cache) {
  # Find the backbone child
  children <- cache$children[[node]]
  child_depths <- cache$depths[children]
  child_linkages <- cache$linkages[[node]]
  linkage_ranks <- calculate_linkage_rank(child_linkages)
  child_sigs <- cache$signatures[children]
  backbone_child_index <- order(
    child_depths,
    linkage_ranks,
    child_sigs,
    decreasing = TRUE
  )[[1]]

  # Order the rest of the children
  if (length(children) > 1) {
    branch_order <- order(linkage_ranks, child_sigs, decreasing = TRUE)
    branch_order <- branch_order[branch_order != backbone_child_index]
    list(backbone = backbone_child_index, branches = branch_order)
  } else {
    list(backbone = backbone_child_index, branches = integer(0))
  }
}

#' Replace vertex and edge indices with actual monosaccharides and linkages
#'
#' @param pseudo_seq Character string containing pseudo-IUPAC sequence with V and E prefixes
#' @param glycan An igraph object representing a glycan structure
#' @returns Character string with V<index> replaced by monosaccharides and E<index> replaced by linkages
#' @noRd
replace_mono_and_link <- function(pseudo_seq, glycan) {
  # Replace vertex indices (V<index>) with monosaccharides and substituents
  vertex_pattern <- "V(\\d+)"
  vertex_matches <- stringr::str_match_all(pseudo_seq, vertex_pattern)[[1]]

  if (nrow(vertex_matches) > 0) {
    for (i in seq_len(nrow(vertex_matches))) {
      vertex_index <- as.numeric(vertex_matches[i, 2])
      mono <- igraph::vertex_attr(glycan, "mono", vertex_index)
      sub <- igraph::vertex_attr(glycan, "sub", vertex_index)

      # Combine monosaccharide with substituent if present
      mono_with_sub <- if (sub == "") {
        mono
      } else {
        # Remove commas to get IUPAC format
        iupac_sub <- stringr::str_remove_all(sub, ",")
        paste0(mono, iupac_sub)
      }

      # Replace V<index> with actual monosaccharide
      pseudo_seq <- stringr::str_replace(
        pseudo_seq,
        paste0("V", vertex_index),
        mono_with_sub
      )
    }
  }

  # Replace edge indices (E<index>) with linkages
  edge_pattern <- "E(\\d+)"
  edge_matches <- stringr::str_match_all(pseudo_seq, edge_pattern)[[1]]

  if (nrow(edge_matches) > 0) {
    for (i in seq_len(nrow(edge_matches))) {
      edge_index <- as.numeric(edge_matches[i, 2])
      linkage <- igraph::edge_attr(glycan, "linkage", edge_index)
      # Replace E<index> with actual linkage wrapped in parentheses
      pseudo_seq <- stringr::str_replace(
        pseudo_seq,
        paste0("E", edge_index),
        paste0("(", linkage, ")")
      )
    }
  }

  pseudo_seq
}

# Is the graph directed?
is_directed_graph <- function(graph) {
  igraph::is_directed(graph)
}


# Is the graph an outward tree?
is_out_tree <- function(graph) {
  igraph::is_tree(graph, mode = "out")
}


# Does the graph have these vertex attributes?
has_vertex_attrs <- function(graph, attrs) {
  all(attrs %in% igraph::vertex_attr_names(graph))
}


# Does the graph have these edge attributes?
has_edge_attrs <- function(graph, attrs) {
  if (
    getRversion() < "4.4.0" &&
      .Platform$OS.type == "windows" &&
      igraph::vcount(graph) == 1
  ) {
    # It seems like when the graph has no edges,
    # setting edge attributes will not work on Windows with R < 4.4.0.
    # So we skip this check to circumvent R CMD check.
    return(TRUE)
  }
  all(attrs %in% igraph::edge_attr_names(graph))
}


.unique_no_na <- function(x) unique(x[!is.na(x)])


# Are all monosaaccharides known?
is_known_mono <- function(monos) {
  known_monos <- c(
    .unique_no_na(monosaccharides$generic),
    monosaccharides$concrete
  )
  monos %in% known_monos
}


# Are generic and concrete monosaccharides mixed?
mix_generic_concrete <- function(monos) {
  has_generic <- any(monos %in% .unique_no_na(monosaccharides$generic))
  has_concrete <- any(monos %in% monosaccharides$concrete)
  sum(as.integer(c(has_generic, has_concrete))) > 1
}


# Is a valid subtituent?
valid_substituent <- function(sub) {
  # Apply to each element if input is a vector
  purrr::map_lgl(sub, function(single_sub) {
    # Empty substituent is always valid
    if (single_sub == "") {
      return(TRUE)
    }

    # Split by commas to handle multiple substituents
    individual_subs <- stringr::str_split(single_sub, ",")[[1]]

    # Check if each individual substituent is valid
    subs_pattern <- stringr::str_c(available_substituents(), collapse = "|")
    pattern <- stringr::str_glue("^[\\d\\?]({subs_pattern})$")

    individual_valid <- purrr::map_lgl(
      individual_subs,
      ~ stringr::str_detect(.x, pattern)
    )

    # All individual substituents must be valid
    if (!all(individual_valid)) {
      return(FALSE)
    }

    # Check if substituents are properly sorted by position
    # Extract positions from individual substituents
    positions <- purrr::map_chr(
      individual_subs,
      ~ stringr::str_extract(.x, "^[\\d\\?]")
    )

    # For sorting, convert ? to a high numeric value
    numeric_positions <- purrr::map_dbl(positions, function(pos) {
      if (pos == "?") {
        return(Inf)
      } else {
        return(as.numeric(pos))
      }
    })

    # Check if positions are sorted in ascending order
    is_sorted <- all(numeric_positions == sort(numeric_positions))

    # Check for duplicate positions (not allowed)
    has_duplicates <- any(duplicated(positions))

    return(is_sorted && !has_duplicates)
  })
}


# Is a valid anomer?
valid_anomer <- function(anomer) {
  stringr::str_detect(anomer, "^[ab\\?][\\d\\?]$")
}


#' Check if Linkages are Valid
#'
#' Valid linkages are in the form of "a1-2", "b1-4", "a?-1", etc.
#' Specifically, the pattern is `xy-z`:
#' - `x`: the anomer, either "a", "b", or "?".
#' - `y`: the first position, either "1", "2" or "?".
#' - `z`: the second position, either a 1-9 digit or "?".
#' Can also be multiple positions separated by "/", e.g. "1/2/3".
#' "?" could not be used with "/".
#'
#' @param linkages A character vector of linkages.
#'
#' @returns A logical vector.
#'
#' @examples
#' # Valid linkages
#' valid_linkages(c("a1-2", "?1-4", "a?-1", "b?-?", "??-?", "a1/2-3"))
#'
#' # Invalid linkages
#' valid_linkages(c("a1-2/?", "1-4", "a/b1-2", "c1-2", "a9-1"))
#'
#' @export
valid_linkages <- function(linkages) {
  checkmate::assert_character(linkages)
  anomer_p <- "[ab\\?]"
  pos1_p <- "([12]|\\?)"
  pos2_p <- "([1-9](/[1-9])*|\\?)"
  linkage_pattern <- stringr::str_glue("^{anomer_p}{pos1_p}-{pos2_p}$")
  stringr::str_detect(linkages, linkage_pattern)
}


# Check if any duplicated linkage positions exist.
# This means the same position of one residue can not be connected to multiple other residues.
any_dup_linkage_pos <- function(glycan) {
  for (v in igraph::V(glycan)) {
    links <- igraph::incident(glycan, v, mode = "out")$linkage
    pos2 <- stringr::str_split_i(links, stringr::fixed("-"), 2)
    pos2 <- pos2[pos2 != "?"]
    pos2 <- pos2[!stringr::str_detect(pos2, stringr::fixed("/"))]
    if (any(duplicated(pos2))) {
      return(TRUE)
    }
  }
  return(FALSE)
}

patrick::with_parameters_test_that(
  "`valid_linkages()` works for good linkages",
  {
    expect_true(valid_linkages(char))
  },
  char = c("b1-4", "a1-3")
)


patrick::with_parameters_test_that(
  "`valid_linkages()` works for unknown linkages",
  {
    expect_true(valid_linkages(char))
  },
  char = c("b1-?", "a?-3", "?1-4", "??-?", "a2-3/6")
)


patrick::with_parameters_test_that(
  "`valid_linkages()` fails for bad linkages",
  {
    expect_false(valid_linkages(char))
  },
  char = c("1-4", "c1-4", "a3-1", "a1-10", "b1", "abc", "")
)

# This table referred to https://www.ncbi.nlm.nih.gov/glycans/snfg.html
monosaccharides <- tibble::tribble(
  ~generic  , ~concrete  ,
  # Hexose
  "Hex"     , "Glc"      ,
  "Hex"     , "Man"      ,
  "Hex"     , "Gal"      ,
  "Hex"     , "Gul"      ,
  "Hex"     , "Alt"      ,
  "Hex"     , "All"      ,
  "Hex"     , "Tal"      ,
  "Hex"     , "Ido"      ,
  # HexNAc
  "HexNAc"  , "GlcNAc"   ,
  "HexNAc"  , "GalNAc"   ,
  "HexNAc"  , "ManNAc"   ,
  "HexNAc"  , "GulNAc"   ,
  "HexNAc"  , "AltNAc"   ,
  "HexNAc"  , "AllNAc"   ,
  "HexNAc"  , "TalNAc"   ,
  "HexNAc"  , "IdoNAc"   ,
  # Hexosamine
  "HexN"    , "GlcN"     ,
  "HexN"    , "ManN"     ,
  "HexN"    , "GalN"     ,
  "HexN"    , "GulN"     ,
  "HexN"    , "AltN"     ,
  "HexN"    , "AllN"     ,
  "HexN"    , "TalN"     ,
  "HexN"    , "IdoN"     ,
  # Hexuronate
  "HexA"    , "GlcA"     ,
  "HexA"    , "ManA"     ,
  "HexA"    , "GalA"     ,
  "HexA"    , "GulA"     ,
  "HexA"    , "AltA"     ,
  "HexA"    , "AllA"     ,
  "HexA"    , "TalA"     ,
  "HexA"    , "IdoA"     ,
  # Deoxyhexose
  "dHex"    , "Fuc"      ,
  "dHex"    , "Qui"      ,
  "dHex"    , "Rha"      ,
  "dHex"    , "6dGul"    ,
  "dHex"    , "6dAlt"    ,
  "dHex"    , "6dTal"    ,
  # DeoxyhexNAc
  "dHexNAc" , "QuiNAc"   ,
  "dHexNAc" , "RhaNAc"   ,
  "dHexNAc" , "6dAltNAc" ,
  "dHexNAc" , "6dTalNAc" ,
  "dHexNAc" , "FucNAc"   ,
  # Di-deoxyhexose
  "ddHex"   , "Oli"      ,
  "ddHex"   , "Tyv"      ,
  "ddHex"   , "Abe"      ,
  "ddHex"   , "Par"      ,
  "ddHex"   , "Dig"      ,
  "ddHex"   , "Col"      ,
  # Pentose
  "Pen"     , "Ara"      ,
  "Pen"     , "Lyx"      ,
  "Pen"     , "Xyl"      ,
  "Pen"     , "Rib"      ,
  # 3-deoxy-nonulosonic acids
  "NeuAc"   , "Neu5Ac"   ,
  "NeuGc"   , "Neu5Gc"   ,
  "gNeu"    , "Neu"      ,
  "gKdn"    , "Kdn"      ,
  # 3,9-dideoxy-nonulosonic acids
  "gPse"    , "Pse"      ,
  "gLeg"    , "Leg"      ,
  "gAci"    , "Aci"      ,
  "g4eLeg"  , "4eLeg"    ,
  # Unknown
  "gBac"    , "Bac"      ,
  "Hep"     , "LDmanHep" ,
  "gKdo"    , "Kdo"      ,
  "HepA"    , "Dha"      ,
  "Hep"     , "DDmanHep" ,
  "MurAc"   , "MurNAc"   ,
  "MurGc"   , "MurNGc"   ,
  "gMur"    , "Mur"      ,
  # Assigned
  "Pen"     , "Api"      ,
  "Hex"     , "Fru"      ,
  "Hex"     , "Tag"      ,
  "Hex"     , "Sor"      ,
  "Hex"     , "Psi"
)


#' Get Available Monosaacharides
#'
#' This function returns a character vector of monosaccharide names of
#' the given type. See [get_mono_type()] for monosaacharide types.
#'
#' @param mono_type A character string specifying the type of monosaccharides.
#'  Can be "all", "generic", or "concrete". Default is "all".
#'
#' @returns A character vector of monosaccharide names.
#'
#' @examples
#' available_monosaccharides()
#'
#' @export
available_monosaccharides <- function(mono_type = "all") {
  checkmate::assert_choice(mono_type, c("all", "generic", "concrete"))
  if (mono_type == "all") {
    unique(purrr::discard(unlist(monosaccharides, use.names = FALSE), is.na))
  } else {
    unique(purrr::discard(monosaccharides[[mono_type]], is.na))
  }
}


#' Check if a Monosaccharide is Known
#'
#' This function checks if a vector of monosaccharide names are known.
#'
#' @param mono A character vector of monosaccharide names.
#'
#' @returns A logical vector.
#'
#' @examples
#' is_known_monosaccharide(c("Gal", "Hex"))
#' is_known_monosaccharide(c("X", "Hx", "Nac"))
#'
#' @export
is_known_monosaccharide <- function(mono) {
  checkmate::assert_character(mono)
  (mono %in% monosaccharides$generic | mono %in% monosaccharides$concrete)
}

patrick::with_parameters_test_that(
  "known monosaccharides",
  {
    expect_true(is_known_monosaccharide(mono))
  },
  mono = c("Gal", "Hex")
)


patrick::with_parameters_test_that(
  "unknown monosaccharides",
  {
    expect_false(is_known_monosaccharide(mono))
  },
  mono = c("X", "Hx", "Nac")
)


test_that("get all generic monosaccharides", {
  res <- available_monosaccharides("generic")
  expect_contains(res, c("Hex", "HexNAc", "dHex"))
  expect_true(length(unique(res)) == length(res))
})


test_that("get all concrete monosaccharides", {
  res <- available_monosaccharides("concrete")
  expect_contains(res, c("Gal", "Man", "GlcNAc"))
  expect_true(length(unique(res)) == length(res))
})


test_that("get all monosaccharides", {
  res <- available_monosaccharides()
  expect_contains(res, c("Hex", "Man"))
  expect_true(length(unique(res)) == length(res))
})

#' Map Functions Over Glycan Structure Vectors
#'
#' @description
#' These functions apply a function to each unique structure in a glycan structure vector,
#' taking advantage of hash-based deduplication to avoid redundant computation.
#' Similar to purrr mapping functions, but optimized for glycan structure vectors.
#'
#' @param .x A glycan structure vector (glyrepr_structure).
#' @param .f A function that takes an igraph object and returns a result.
#'   Can be a function, purrr-style lambda (`~ .x$attr`), or a character string naming a function.
#' @param ... Additional arguments passed to `.f`.
#' @param .ptype A prototype for the return type (for `smap_vec`).
#' @param .parallel Logical; whether to use parallel processing. If `FALSE` (default),
#'   parallel processing is disabled. Set to `TRUE` to enable parallel processing.
#'
#' @details
#' These functions only compute `.f` once for each unique structure, then map
#' the results back to the original vector positions. This is much more efficient
#' than applying `.f` to each element individually when there are duplicate structures.
#'
#'
#' **Return Types:**
#' - `smap()`: Returns a list with the same length as `.x`
#' - `smap_vec()`: Returns an atomic vector with the same length as `.x`
#' - `smap_lgl()`: Returns a logical vector
#' - `smap_int()`: Returns an integer vector
#' - `smap_dbl()`: Returns a double vector
#' - `smap_chr()`: Returns a character vector
#' - `smap_structure()`: Returns a new glycan structure vector (`.f` must return igraph objects)
#'
#' @returns
#' - `smap()`: A list
#' - `smap_vec()`: An atomic vector of type specified by `.ptype`
#' - `smap_lgl/int/dbl/chr()`: Atomic vectors of the corresponding type
#' - `smap_structure()`: A new glyrepr_structure object
#'
#' @examples
#' # Create a structure vector with duplicates
#' core1 <- o_glycan_core_1()
#' core2 <- n_glycan_core()
#' structures <- c(core1, core2, core1)  # core1 appears twice
#'
#' # Map a function that counts vertices - only computed twice, not three times
#' smap_int(structures, igraph::vcount)
#'
#' # Map a function that returns logical
#' smap_lgl(structures, function(g) igraph::vcount(g) > 5)
#'
#' # Use purrr-style lambda functions
#' smap_int(structures, ~ igraph::vcount(.x))
#' smap_lgl(structures, ~ igraph::vcount(.x) > 5)
#'
#' # Map a function that modifies structure (must return igraph)
#' add_vertex_names <- function(g) {
#'   if (!("name" %in% igraph::vertex_attr_names(g))) {
#'     igraph::set_vertex_attr(g, "name", value = paste0("v", seq_len(igraph::vcount(g))))
#'   } else {
#'     g
#'   }
#' }
#' smap_structure(structures, add_vertex_names)
#'
#' @name smap
NULL

# Helper function to rebuild glycan_structure with proper deduplication
# after modifications that may create identical graphs
.rebuild_structure_with_dedup <- function(modified_graphs, idx_mapping) {
  # Get new IUPACs for all modified graphs
  new_unique_iupacs <- purrr::map_chr(
    modified_graphs,
    .structure_to_iupac_single
  )
  new_iupacs <- new_unique_iupacs[idx_mapping]

  # Re-deduplicate graphs based on new IUPACs to handle cases where
  # modifications create identical graphs
  unique_new_indices <- which(!duplicated(new_unique_iupacs))
  final_unique_graphs <- modified_graphs[unique_new_indices]
  final_unique_iupacs <- new_unique_iupacs[unique_new_indices]
  names(final_unique_graphs) <- final_unique_iupacs

  # Create result glycan_structure
  new_glycan_structure(new_iupacs, final_unique_graphs)
}

# Helper function for parallel processing
.smap_apply <- function(
  data_list,
  func,
  use_parallel = FALSE,
  auto_threshold = 100,
  ...
) {
  n_tasks <- length(data_list)

  # Handle NULL values for backward compatibility
  if (is.null(use_parallel)) {
    use_parallel <- FALSE
  }

  if (use_parallel && n_tasks > 1) {
    # Check if future backend is set up
    if (!future::nbrOfWorkers() > 1) {
      cli::cli_inform(
        "No parallel backend detected. Using sequential processing."
      )
      use_parallel <- FALSE
    }
  } else if (use_parallel && n_tasks <= 1) {
    use_parallel <- FALSE
  }

  if (use_parallel) {
    furrr::future_map(data_list, func, ...)
  } else {
    purrr::map(data_list, func, ...)
  }
}

#' Split a glycan structure vector into valid and missing positions
#'
#' @param x A glycan structure vector.
#' @returns A list with vector data, graph data, names, and NA masks.
#' @noRd
.structure_map_input <- function(x) {
  codes <- vctrs::vec_data(x)
  na_mask <- is.na(codes)

  list(
    codes = codes,
    graphs = attr(x, "graphs"),
    input_names = names(x),
    na_mask = na_mask,
    na_count = sum(na_mask),
    has_na = any(na_mask),
    all_na = length(codes) > 0 && all(na_mask),
    valid_codes = codes[!na_mask],
    valid_names = names(x)[!na_mask]
  )
}

#' Build a deduplicated-combination table for structure mapping
#'
#' @param valid_codes Non-missing structure codes.
#' @param args Additional recycled argument vectors aligned with `valid_codes`.
#' @returns A tibble with a `code`, one `arg*` column per argument, and `combo_key`.
#' @noRd
.structure_combo_table <- function(valid_codes, args = list()) {
  combinations_df <- tibble::tibble(code = valid_codes)
  key_components <- list(valid_codes)

  for (i in seq_along(args)) {
    arg_name <- paste0("arg", i)
    combinations_df[[arg_name]] <- args[[i]]
    key_components <- append(
      key_components,
      list(purrr::map_chr(args[[i]], .generate_value_key))
    )
  }

  combinations_df$combo_key <- do.call(paste, c(key_components, sep = "|||"))
  combinations_df
}

#' Return unique rows from a structure combination table
#'
#' @param combinations_df A table returned by `.structure_combo_table()`.
#' @returns A tibble containing unique combination keys.
#' @noRd
.unique_structure_combos <- function(combinations_df) {
  combinations_df[!duplicated(combinations_df$combo_key), ]
}

#' Extract mapped arguments from one combination-table row
#'
#' @param row A one-row tibble from `.structure_combo_table()`.
#' @param n_args Number of argument columns to extract.
#' @returns A list of mapped argument values.
#' @noRd
.extract_combo_args <- function(row, n_args) {
  purrr::map(seq_len(n_args), function(i) {
    .extract_from_list_column(row[[paste0("arg", i)]])
  })
}

#' Restore a list result to the original structure vector shape
#'
#' @param valid_results Results for non-missing positions.
#' @param map_input Metadata from `.structure_map_input()`.
#' @param na_value Value to place at missing positions.
#' @returns A list with original length and names.
#' @noRd
.restore_list_with_na <- function(valid_results, map_input, na_value = NA) {
  if (!map_input$has_na) {
    names(valid_results) <- map_input$input_names
    return(valid_results)
  }

  result <- vector("list", length(map_input$codes))
  result[!map_input$na_mask] <- valid_results
  result[map_input$na_mask] <- list(na_value)
  names(result) <- map_input$input_names
  result
}

#' Restore a structure result to the original structure vector shape
#'
#' @param valid_result Result for non-missing positions.
#' @param map_input Metadata from `.structure_map_input()`.
#' @returns A glycan structure vector with original length and names.
#' @noRd
.restore_structure_with_na <- function(valid_result, map_input) {
  if (map_input$all_na) {
    result <- new_na_glycan_structure(length(map_input$codes))
    names(result) <- map_input$input_names
    return(result)
  }

  if (map_input$has_na) {
    result_iupacs <- rep(NA_character_, length(map_input$codes))
    result_iupacs[!map_input$na_mask] <- vctrs::vec_data(valid_result)

    result <- new_glycan_structure(result_iupacs, attr(valid_result, "graphs"))
    names(result) <- map_input$input_names
    return(result)
  }

  names(valid_result) <- map_input$input_names
  valid_result
}

# Helper function for common smap logic
.smap_base <- function(.x, .f, ..., .parallel = FALSE, .convert_fn = NULL) {
  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input must be a glycan_structure vector.")
  }

  .f <- rlang::as_function(.f)

  map_input <- .structure_map_input(.x)

  # If all elements are NA, return NA results
  if (map_input$all_na) {
    if (is.null(.convert_fn)) {
      result <- vector("list", length(map_input$codes))
    } else {
      # Get typed NA by running .convert_fn on a properly-typed NA
      # Use NA_character_ for character output (as.character(NA) gives "NA" string)
      # For other types, NA coerces correctly
      test_na <- NA_character_ # Default to character type
      typed_na <- .convert_fn(list(test_na))
      result <- rep(typed_na, map_input$na_count)
    }
    names(result) <- map_input$input_names
    return(result)
  }

  # Extract dots without .parallel
  dots <- list(...)

  # Apply function only to unique graphs (NA already filtered out)
  unique_codes <- names(map_input$graphs)
  unique_results <- .smap_apply(
    unique_codes,
    function(code) {
      do.call(.f, c(list(map_input$graphs[[code]]), dots))
    },
    use_parallel = .parallel
  )
  names(unique_results) <- unique_codes

  # If no conversion function provided, return list (for smap)
  if (is.null(.convert_fn)) {
    # Optimized mapping: use match() instead of individual lookups
    idx <- match(map_input$valid_codes, unique_codes)
    valid_results <- unique_results[idx]
    names(valid_results) <- map_input$valid_names

    return(.restore_list_with_na(valid_results, map_input, na_value = NULL))
  }

  # Convert to target type and map back to original positions
  unique_converted <- .convert_fn(unique_results)
  names(unique_converted) <- unique_codes

  # Optimized mapping: use match() instead of individual lookups
  idx <- match(map_input$valid_codes, unique_codes)
  valid_result <- unique_converted[idx]

  # Initialize result with NA and assign valid results
  result <- rep(NA, length(map_input$codes))
  result[!map_input$na_mask] <- valid_result
  names(result) <- map_input$input_names
  return(result)
}

# Helper function to generate a hash-based key for complex objects
# Used for deduplication when dealing with nested lists and other complex structures
.generate_object_hash <- function(obj) {
  serialized_data <- serialize(obj, connection = NULL)
  hash_input <- sum(as.integer(serialized_data))
  hex_hash <- as.hexmode(hash_input)
  format(hex_hash, width = 8)
}

# Helper function to extract values from tibble list-columns
# Tibble stores single-element lists as list(data) in list-columns, requiring unwrapping
.extract_from_list_column <- function(value) {
  if (is.list(value) && length(value) == 1) {
    value[[1]] # Extract from list-column wrapper
  } else {
    value # Use as-is for non-list or multi-element cases
  }
}

# Helper function to generate deduplication keys for arbitrary values
# Handles both simple values (convert to character) and complex objects (hash-based keys)
.generate_value_key <- function(value, prefix = "list_") {
  if (is.list(value)) {
    paste0(prefix, .generate_object_hash(value))
  } else {
    as.character(value)
  }
}

#' @rdname smap
#' @export
smap <- function(.x, .f, ..., .parallel = FALSE) {
  .smap_base(.x, .f, ..., .parallel = .parallel, .convert_fn = NULL)
}

#' @rdname smap
#' @export
smap_vec <- function(.x, .f, ..., .ptype = NULL, .parallel = FALSE) {
  .smap_base(
    .x,
    .f,
    ...,
    .parallel = .parallel,
    .convert_fn = function(results) vctrs::vec_c(!!!results, .ptype = .ptype)
  )
}

#' @rdname smap
#' @export
smap_lgl <- function(.x, .f, ..., .parallel = FALSE) {
  .smap_base(
    .x,
    .f,
    ...,
    .parallel = .parallel,
    .convert_fn = function(results) {
      as.logical(unlist(results, use.names = FALSE))
    }
  )
}

#' @rdname smap
#' @export
smap_int <- function(.x, .f, ..., .parallel = FALSE) {
  .smap_base(
    .x,
    .f,
    ...,
    .parallel = .parallel,
    .convert_fn = function(results) {
      as.integer(unlist(results, use.names = FALSE))
    }
  )
}

#' @rdname smap
#' @export
smap_dbl <- function(.x, .f, ..., .parallel = FALSE) {
  .smap_base(
    .x,
    .f,
    ...,
    .parallel = .parallel,
    .convert_fn = function(results) {
      as.double(unlist(results, use.names = FALSE))
    }
  )
}

#' @rdname smap
#' @export
smap_chr <- function(.x, .f, ..., .parallel = FALSE) {
  .smap_base(
    .x,
    .f,
    ...,
    .parallel = .parallel,
    .convert_fn = function(results) {
      as.character(unlist(results, use.names = FALSE))
    }
  )
}

#' @rdname smap
#' @export
smap_structure <- function(.x, .f, ..., .parallel = FALSE) {
  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input must be a glycan_structure vector.")
  }

  .f <- rlang::as_function(.f)
  map_input <- .structure_map_input(.x)

  # Extract dots without .parallel
  dots <- list(...)

  if (map_input$all_na) {
    return(.restore_structure_with_na(NULL, map_input))
  }

  # Apply function only to unique graphs
  unique_iupacs <- names(map_input$graphs)
  new_graphs <- .smap_apply(
    unique_iupacs,
    function(iupac) {
      result <- do.call(.f, c(list(map_input$graphs[[iupac]]), dots))
      if (!inherits(result, "igraph")) {
        cli::cli_abort(
          "Function `.f` must return an igraph object when using `smap_structure()`."
        )
      }
      result
    },
    use_parallel = .parallel
  )

  # Rebuild glycan_structure with proper deduplication
  idx <- match(map_input$valid_codes, unique_iupacs)
  valid_result <- .rebuild_structure_with_dedup(new_graphs, idx)
  names(valid_result) <- map_input$valid_names

  .restore_structure_with_na(valid_result, map_input)
}

#' Apply Function to Unique Structures Only
#'
#' @description
#' Apply a function only to the unique structures in a glycan structure vector,
#' returning results in the same order as the unique structures appear.
#' This is useful when you need to perform expensive computations but only
#' care about unique results.
#'
#'
#' @param .x A glycan structure vector (glyrepr_structure).
#' @param .f A function that takes an igraph object and returns a result.
#'   Can be a function, purrr-style lambda (`~ .x$attr`), or a character string naming a function.
#' @param ... Additional arguments passed to `.f`.
#' @param .parallel Logical; whether to use parallel processing. If `FALSE` (default),
#'   parallel processing is disabled. Set to `TRUE` to enable parallel processing.
#'   See examples in \code{\link{smap}} for how to set up and use parallel processing.
#'
#' @return A list with results for each unique structure, named by their hash codes.
#'
#' @examples
#' # Create a structure vector with duplicates
#' core1 <- o_glycan_core_1()
#' structures <- c(core1, core1, core1)  # same structure 3 times
#'
#' # Only compute once for the unique structure
#' unique_results <- smap_unique(structures, igraph::vcount)
#' length(unique_results)  # 1, not 3
#'
#' # Use purrr-style lambda
#' unique_results2 <- smap_unique(structures, ~ igraph::vcount(.x))
#' length(unique_results2)  # 1, not 3
#'
#' @export
smap_unique <- function(.x, .f, ..., .parallel = FALSE) {
  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input must be a glycan_structure vector.")
  }

  .f <- rlang::as_function(.f)

  graphs <- attr(.x, "graphs")

  # Extract dots without .parallel
  dots <- list(...)

  # Apply function only to unique graphs
  results <- .smap_apply(
    graphs,
    function(g) {
      do.call(.f, c(list(g), dots))
    },
    use_parallel = .parallel
  )
  results
}

#' Test Predicates on Glycan Structure Vectors
#'
#' @description
#' These functions test predicates on unique structures in a glycan structure vector,
#' taking advantage of hash-based deduplication to avoid redundant computation.
#' Similar to purrr predicate functions, but optimized for glycan structure vectors.
#'
#' @param .x A glycan structure vector (glyrepr_structure).
#' @param .p A predicate function that takes an igraph object and returns a logical value.
#'   Can be a function, purrr-style lambda (`~ .x$attr`), or a character string naming a function.
#' @param ... Additional arguments passed to `.p`.
#'
#' @details
#' These functions only evaluate `.p` once for each unique structure, making them
#' much more efficient than applying `.p` to each element individually when there
#' are duplicate structures.
#'
#' **Return Values:**
#' - `ssome()`: Returns `TRUE` if at least one unique structure satisfies the predicate
#' - `severy()`: Returns `TRUE` if all unique structures satisfy the predicate
#' - `snone()`: Returns `TRUE` if no unique structures satisfy the predicate
#'
#' @return A single logical value.
#'
#' @examples
#' # Create a structure vector with duplicates
#' core1 <- o_glycan_core_1()
#' core2 <- n_glycan_core()
#' structures <- c(core1, core2, core1)  # core1 appears twice
#'
#' # Test if some structures have more than 5 vertices
#' ssome(structures, function(g) igraph::vcount(g) > 5)
#'
#' # Test if all structures have at least 3 vertices
#' severy(structures, function(g) igraph::vcount(g) >= 3)
#'
#' # Test if no structures have more than 20 vertices
#' snone(structures, function(g) igraph::vcount(g) > 20)
#'
#' # Use purrr-style lambda functions
#' ssome(structures, ~ igraph::vcount(.x) > 5)
#' severy(structures, ~ igraph::vcount(.x) >= 3)
#' snone(structures, ~ igraph::vcount(.x) > 20)
#'
#' @name smap_predicates
NULL

#' @rdname smap_predicates
#' @export
ssome <- function(.x, .p, ...) {
  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input must be a glycan_structure vector.")
  }

  .p <- rlang::as_function(.p)

  graphs <- attr(.x, "graphs")

  # Apply predicate only to unique graphs using purrr::some
  purrr::some(graphs, .p, ...)
}

#' @rdname smap_predicates
#' @export
severy <- function(.x, .p, ...) {
  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input must be a glycan_structure vector.")
  }

  .p <- rlang::as_function(.p)

  graphs <- attr(.x, "graphs")

  # Apply predicate only to unique graphs using purrr::every
  purrr::every(graphs, .p, ...)
}

#' @rdname smap_predicates
#' @export
snone <- function(.x, .p, ...) {
  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input must be a glycan_structure vector.")
  }

  .p <- rlang::as_function(.p)

  graphs <- attr(.x, "graphs")

  # Apply predicate only to unique graphs using purrr::none
  purrr::none(graphs, .p, ...)
}

#' Map Functions Over Two Glycan Structure Vectors
#'
#' @description
#' These functions apply a function to each unique structure combination in two glycan structure vectors,
#' taking advantage of hash-based deduplication to avoid redundant computation.
#' Similar to purrr map2 functions, but optimized for glycan structure vectors.
#'
#' @param .x A glycan structure vector (glyrepr_structure).
#' @param .y A vector of the same length as `.x`, or length 1 (will be recycled).
#' @param .f A function that takes an igraph object (from `.x`) and a value (from `.y`) and returns a result.
#'   Can be a function, purrr-style lambda (`~ .x + .y`), or a character string naming a function.
#' @param ... Additional arguments passed to `.f`.
#' @param .ptype A prototype for the return type (for `smap2_vec`).
#' @param .parallel Logical; whether to use parallel processing. If `FALSE` (default),
#'   parallel processing is disabled. Set to `TRUE` to enable parallel processing.
#'   See examples in \code{\link{smap}} for how to set up and use parallel processing.
#'
#' @details
#' These functions only compute `.f` once for each unique combination of structure and corresponding
#' `.y` value, then map the results back to the original vector positions. This is much more efficient
#' than applying `.f` to each element pair individually when there are duplicate structure-value combinations.
#'
#' **NA Handling:**
#' NA elements in `.x` are preserved in the output - the function is not applied to NA positions,
#' and the corresponding results are set to NA.
#'
#'
#' **Return Types:**
#' - `smap2()`: Returns a list with the same length as `.x`
#' - `smap2_vec()`: Returns an atomic vector with the same length as `.x`
#' - `smap2_lgl()`: Returns a logical vector
#' - `smap2_int()`: Returns an integer vector
#' - `smap2_dbl()`: Returns a double vector
#' - `smap2_chr()`: Returns a character vector
#' - `smap2_structure()`: Returns a new glycan structure vector (`.f` must return igraph objects)
#'
#' @return
#' - `smap2()`: A list
#' - `smap2_vec()`: An atomic vector of type specified by `.ptype`
#' - `smap2_lgl/int/dbl/chr()`: Atomic vectors of the corresponding type
#' - `smap2_structure()`: A new glyrepr_structure object
#'
#' @examples
#' # Create structure vectors with duplicates
#' core1 <- o_glycan_core_1()
#' core2 <- n_glycan_core()
#' structures <- c(core1, core2, core1)  # core1 appears twice
#' weights <- c(1.0, 2.0, 1.0)  # corresponding weights
#'
#' # Map a function that uses both structure and weight
#' smap2_dbl(structures, weights, function(g, w) igraph::vcount(g) * w)
#'
#' # Use purrr-style lambda functions
#' smap2_dbl(structures, weights, ~ igraph::vcount(.x) * .y)
#'
#' # Test with recycling (single weight for all structures)
#' smap2_dbl(structures, 2.5, ~ igraph::vcount(.x) * .y)
#'
#' # Map a function that modifies structure based on second argument
#' # This example adds a graph attribute instead of modifying topology
#' add_weight_attr <- function(g, weight) {
#'   igraph::set_graph_attr(g, "weight", weight)
#' }
#' weights_to_add <- c(1.5, 2.5, 1.5)
#' smap2_structure(structures, weights_to_add, add_weight_attr)
#'
#' @name smap2
NULL

#' @rdname smap2
#' @export
smap2 <- function(.x, .y, .f, ..., .parallel = FALSE) {
  # FUNCTION-LEVEL BUG FIX DOCUMENTATION:
  #
  # This function previously had a critical bug when handling nested list objects
  # in the .y parameter. The issue occurred in two places:
  #
  # 1. DATA FRAME EXPANSION BUG:
  #    Using data.frame() with nested lists caused unwanted row expansion.
  #    Example: .y = list(list(c(6,5,4,3,2,1))) would create 6 rows instead of 1,
  #    breaking the 1:1 correspondence with glycan structures.
  #
  # 2. LIST-COLUMN EXTRACTION BUG:
  #    When extracting values from tibble list-columns, an extra layer of list
  #    wrapping was not properly handled.
  #
  # SOLUTION:
  #    - Replace data.frame() with tibble::tibble() for proper list-column support
  #    - Implement hash-based keys for complex list objects
  #    - Add proper unwrapping logic when extracting from list-columns
  #
  # This fix ensures smap2 works correctly with nested data structures commonly
  # used in glycan analysis workflows.

  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input `.x` must be a glycan_structure vector.")
  }

  # Handle empty input
  if (length(.x) == 0) {
    return(list())
  }

  # Recycle .y to match length of .x
  .y <- vctrs::vec_recycle(.y, length(.x))

  .f <- rlang::as_function(.f)

  map_input <- .structure_map_input(.x)

  # If all elements are NA, return NA results
  if (map_input$all_na) {
    result <- vector("list", length(map_input$codes))
    result[map_input$na_mask] <- list(NA)
    names(result) <- map_input$input_names
    return(result)
  }

  valid_y <- .y[!map_input$na_mask]

  # Create unique combinations data frame for proper handling
  #
  # BUG FIX NOTES:
  # Prior to this fix, using data.frame() with nested list objects in the y_val column
  # caused unwanted expansion where the inner vectors would be unwrapped into separate rows.
  # For example, if .y = list(list(c(6,5,4,3,2,1))), data.frame() would create 6 rows
  # instead of 1, breaking the length correspondence with the glycan structures.
  #
  # The fix involves two changes:
  # 1. Use tibble::tibble() instead of data.frame() to properly handle list-columns
  # 2. Create hash-based keys for complex list objects to enable proper deduplication

  combinations_df <- .structure_combo_table(
    map_input$valid_codes,
    list(valid_y)
  )
  unique_combinations_df <- .unique_structure_combos(combinations_df)

  # Apply function only to unique combinations
  unique_results <- .smap_apply(
    seq_len(nrow(unique_combinations_df)),
    function(i) {
      row <- unique_combinations_df[i, ]
      y_val <- .extract_combo_args(row, 1)[[1]]

      .f(map_input$graphs[[row$code]], y_val, ...)
    },
    use_parallel = .parallel
  )
  names(unique_results) <- unique_combinations_df$combo_key

  # Map results back to original vector positions
  valid_results <- purrr::map(combinations_df$combo_key, ~ unique_results[[.x]])
  names(valid_results) <- map_input$valid_names

  .restore_list_with_na(valid_results, map_input, na_value = NA)
}

#' @rdname smap2
#' @export
smap2_vec <- function(.x, .y, .f, ..., .ptype = NULL, .parallel = FALSE) {
  results <- smap2(.x, .y, .f, ..., .parallel = .parallel)
  vctrs::vec_c(!!!results, .ptype = .ptype)
}

#' @rdname smap2
#' @export
smap2_lgl <- function(.x, .y, .f, ..., .parallel = FALSE) {
  smap2_vec(.x, .y, .f, ..., .ptype = logical(), .parallel = .parallel)
}

#' @rdname smap2
#' @export
smap2_int <- function(.x, .y, .f, ..., .parallel = FALSE) {
  smap2_vec(.x, .y, .f, ..., .ptype = integer(), .parallel = .parallel)
}

#' @rdname smap2
#' @export
smap2_dbl <- function(.x, .y, .f, ..., .parallel = FALSE) {
  smap2_vec(.x, .y, .f, ..., .ptype = double(), .parallel = .parallel)
}

#' @rdname smap2
#' @export
smap2_chr <- function(.x, .y, .f, ..., .parallel = FALSE) {
  smap2_vec(.x, .y, .f, ..., .ptype = character(), .parallel = .parallel)
}

#' @rdname smap2
#' @export
smap2_structure <- function(.x, .y, .f, ..., .parallel = FALSE) {
  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input `.x` must be a glycan_structure vector.")
  }

  # Handle empty input
  if (length(.x) == 0) {
    return(glycan_structure())
  }

  # Recycle .y to match length of .x
  .y <- vctrs::vec_recycle(.y, length(.x))

  .f <- rlang::as_function(.f)

  map_input <- .structure_map_input(.x)

  # If all elements are NA, return NA structure
  if (map_input$all_na) {
    return(.restore_structure_with_na(NULL, map_input))
  }

  valid_y <- .y[!map_input$na_mask]

  # Create unique combinations data frame for proper handling
  # BUG FIX: Use tibble to prevent unwanted expansion of nested lists (same fix as smap2)
  combinations_df <- .structure_combo_table(
    map_input$valid_codes,
    list(valid_y)
  )
  unique_combinations_df <- .unique_structure_combos(combinations_df)

  # Apply function only to unique combinations
  new_graphs <- .smap_apply(
    seq_len(nrow(unique_combinations_df)),
    function(i) {
      row <- unique_combinations_df[i, ]
      y_val <- .extract_combo_args(row, 1)[[1]]

      result <- .f(map_input$graphs[[row$code]], y_val, ...)
      if (!inherits(result, "igraph")) {
        cli::cli_abort(
          "Function `.f` must return an igraph object when using `smap2_structure()`."
        )
      }
      result
    },
    use_parallel = .parallel
  )

  # Rebuild glycan_structure with proper deduplication
  idx <- match(combinations_df$combo_key, unique_combinations_df$combo_key)
  valid_result <- .rebuild_structure_with_dedup(new_graphs, idx)

  # Restore names for valid results
  names(valid_result) <- map_input$valid_names

  .restore_structure_with_na(valid_result, map_input)
}

#' Map Functions Over Glycan Structure Vectors and Multiple Arguments
#'
#' @description
#' These functions apply a function to each unique structure in a glycan structure vector
#' along with corresponding elements from multiple other vectors,
#' taking advantage of hash-based deduplication to avoid redundant computation.
#' Similar to purrr pmap functions, but optimized for glycan structure vectors.
#'
#' @param .l A list where the first element is a glycan structure vector (glyrepr_structure)
#'   and the remaining elements are vectors of the same length or length 1 (will be recycled).
#' @param .f A function that takes an igraph object (from first element of `.l`) and
#'   values from other elements, returning a result.
#'   Can be a function, purrr-style lambda (`~ .x + .y + .z`), or a character string naming a function.
#' @param ... Additional arguments passed to `.f`.
#' @param .ptype A prototype for the return type (for `spmap_vec`).
#' @param .parallel Logical; whether to use parallel processing. If `FALSE` (default),
#'   parallel processing is disabled. Set to `TRUE` to enable parallel processing.
#'   See examples in \code{\link{smap}} for how to set up and use parallel processing.
#'
#' @details
#' These functions only compute `.f` once for each unique combination of structure and corresponding
#' values from other vectors, then map the results back to the original vector positions.
#'
#' **NA Handling:**
#' NA elements in the first argument (glycan structure vector) are preserved in the output.
#'
#' **Time Complexity Performance:**
#'
#' Performance scales with unique combinations of all arguments rather than total vector length.
#' When argument vectors are highly redundant, performance approaches O(unique_structures).
#' Scaling factor shows time increase when vector size increases 20x.
#'
#' **Return Types:**
#' - `spmap()`: Returns a list with the same length as the input vectors
#' - `spmap_vec()`: Returns an atomic vector with the same length as the input vectors
#' - `spmap_lgl()`: Returns a logical vector
#' - `spmap_int()`: Returns an integer vector
#' - `spmap_dbl()`: Returns a double vector
#' - `spmap_chr()`: Returns a character vector
#' - `spmap_structure()`: Returns a new glycan structure vector (`.f` must return igraph objects)
#'
#' @return
#' - `spmap()`: A list
#' - `spmap_vec()`: An atomic vector of type specified by `.ptype`
#' - `spmap_lgl/int/dbl/chr()`: Atomic vectors of the corresponding type
#' - `spmap_structure()`: A new glyrepr_structure object
#'
#' @examples
#' # Create structure vectors with duplicates
#' core1 <- o_glycan_core_1()
#' core2 <- n_glycan_core()
#' structures <- c(core1, core2, core1)  # core1 appears twice
#' weights <- c(1.0, 2.0, 1.0)  # corresponding weights
#' factors <- c(2, 3, 2)  # corresponding factors
#'
#' # Map a function that uses structure, weight, and factor
#' spmap_dbl(list(structures, weights, factors),
#'           function(g, w, f) igraph::vcount(g) * w * f)
#'
#' # Use purrr-style lambda functions
#' spmap_dbl(list(structures, weights, factors), ~ igraph::vcount(..1) * ..2 * ..3)
#'
#' # Test with recycling
#' spmap_dbl(list(structures, 2.0, 3), ~ igraph::vcount(..1) * ..2 * ..3)
#'
#' @name spmap
NULL

#' @rdname spmap
#' @export
spmap <- function(.l, .f, ..., .parallel = FALSE) {
  # FUNCTION-LEVEL BUG FIX DOCUMENTATION:
  # This function had a similar bug to smap2 where data.frame() would unwrap
  # nested list arguments, causing incorrect row expansion. Fixed by using tibble
  # and proper hash-based key generation for complex objects.
  # Also fixed NA handling to skip NA elements in the structure vector.

  # Check if it's actually a plain list, not a vctrs object that looks like a list
  if (!inherits(.l, "list") || inherits(.l, "vctrs_vctr") || length(.l) == 0) {
    cli::cli_abort("Input `.l` must be a non-empty list.")
  }

  if (!is_glycan_structure(.l[[1]])) {
    cli::cli_abort("First element of `.l` must be a glycan_structure vector.")
  }

  # Handle empty input
  if (length(.l[[1]]) == 0) {
    return(list())
  }

  # Recycle all vectors to match length of first vector
  target_length <- length(.l[[1]])
  .l <- purrr::map(.l, ~ vctrs::vec_recycle(.x, target_length))

  .f <- rlang::as_function(.f)

  map_input <- .structure_map_input(.l[[1]])

  # If all elements are NA, return NA results
  if (map_input$all_na) {
    result <- vector("list", length(map_input$codes))
    result[map_input$na_mask] <- list(NA)
    names(result) <- map_input$input_names
    return(result)
  }

  valid_args <- purrr::map(.l[-1], ~ .x[!map_input$na_mask])

  # Create unique combinations data frame for proper handling
  # BUG FIX: Use tibble to prevent unwanted expansion of nested lists
  combinations_df <- .structure_combo_table(map_input$valid_codes, valid_args)
  unique_combinations_df <- .unique_structure_combos(combinations_df)

  # Apply function only to unique combinations
  unique_results <- .smap_apply(
    seq_len(nrow(unique_combinations_df)),
    function(i) {
      row <- unique_combinations_df[i, ]
      # Build full argument list: first is structure, then other args
      args <- c(
        list(map_input$graphs[[row$code]]),
        .extract_combo_args(row, length(valid_args))
      )
      do.call(.f, c(args, list(...)))
    },
    use_parallel = .parallel
  )
  names(unique_results) <- unique_combinations_df$combo_key

  # Map results back to original positions
  valid_results <- purrr::map(combinations_df$combo_key, ~ unique_results[[.x]])
  names(valid_results) <- map_input$valid_names

  .restore_list_with_na(valid_results, map_input, na_value = NA)
}

#' @rdname spmap
#' @export
spmap_vec <- function(.l, .f, ..., .ptype = NULL, .parallel = FALSE) {
  results <- spmap(.l, .f, ..., .parallel = .parallel)
  vctrs::vec_c(!!!results, .ptype = .ptype)
}

#' @rdname spmap
#' @export
spmap_lgl <- function(.l, .f, ..., .parallel = FALSE) {
  spmap_vec(.l, .f, ..., .ptype = logical(), .parallel = .parallel)
}

#' @rdname spmap
#' @export
spmap_int <- function(.l, .f, ..., .parallel = FALSE) {
  spmap_vec(.l, .f, ..., .ptype = integer(), .parallel = .parallel)
}

#' @rdname spmap
#' @export
spmap_dbl <- function(.l, .f, ..., .parallel = FALSE) {
  spmap_vec(.l, .f, ..., .ptype = double(), .parallel = .parallel)
}

#' @rdname spmap
#' @export
spmap_chr <- function(.l, .f, ..., .parallel = FALSE) {
  spmap_vec(.l, .f, ..., .ptype = character(), .parallel = .parallel)
}

#' @rdname spmap
#' @export
spmap_structure <- function(.l, .f, ..., .parallel = FALSE) {
  # FUNCTION-LEVEL BUG FIX DOCUMENTATION:
  # This function had the same nested list expansion bug as spmap and smap2.
  # Fixed by using tibble and proper list-column handling.
  # Also fixed NA handling to skip NA elements in the structure vector.

  if (!is.list(.l) || length(.l) == 0) {
    cli::cli_abort("Input `.l` must be a non-empty list.")
  }

  if (!is_glycan_structure(.l[[1]])) {
    cli::cli_abort("First element of `.l` must be a glycan_structure vector.")
  }

  # Handle empty input
  if (length(.l[[1]]) == 0) {
    return(glycan_structure())
  }

  # Recycle all vectors to match length of first vector
  target_length <- length(.l[[1]])
  .l <- purrr::map(.l, ~ vctrs::vec_recycle(.x, target_length))

  .f <- rlang::as_function(.f)

  map_input <- .structure_map_input(.l[[1]])

  # If all elements are NA, return NA structure
  if (map_input$all_na) {
    return(.restore_structure_with_na(NULL, map_input))
  }

  valid_args <- purrr::map(.l[-1], ~ .x[!map_input$na_mask])

  # Create unique combinations data frame for proper handling
  # BUG FIX: Use tibble to prevent unwanted expansion of nested lists
  combinations_df <- .structure_combo_table(map_input$valid_codes, valid_args)
  unique_combinations_df <- .unique_structure_combos(combinations_df)

  # Apply function only to unique combinations
  new_graphs <- .smap_apply(
    seq_len(nrow(unique_combinations_df)),
    function(i) {
      row <- unique_combinations_df[i, ]
      # Build full argument list: first is structure, then other args
      args <- c(
        list(map_input$graphs[[row$code]]),
        .extract_combo_args(row, length(valid_args))
      )
      result <- do.call(.f, c(args, list(...)))
      if (!inherits(result, "igraph")) {
        cli::cli_abort(
          "Function `.f` must return an igraph object when using `spmap_structure()`."
        )
      }
      result
    },
    use_parallel = .parallel
  )

  # Rebuild glycan_structure with proper deduplication
  idx <- match(combinations_df$combo_key, unique_combinations_df$combo_key)
  valid_result <- .rebuild_structure_with_dedup(new_graphs, idx)
  names(valid_result) <- map_input$valid_names

  .restore_structure_with_na(valid_result, map_input)
}

#' Map Functions Over Glycan Structure Vectors with Indices
#'
#' @description
#' These functions apply a function to each unique structure in a glycan structure vector
#' along with their corresponding indices,
#' taking advantage of hash-based deduplication to avoid redundant computation.
#' Similar to purrr imap functions, but optimized for glycan structure vectors.
#'
#' @param .x A glycan structure vector (glyrepr_structure).
#' @param .f A function that takes an igraph object (from `.x`) and an index/name,
#'   returning a result.
#'   Can be a function, purrr-style lambda (`~ paste(.x, .y)`), or a character string naming a function.
#' @param ... Additional arguments passed to `.f`.
#' @param .ptype A prototype for the return type (for `simap_vec`).
#'
#' @details
#' These functions only compute `.f` once for each unique combination of structure and corresponding
#' index/name, then map the results back to the original vector positions. This is much more efficient
#' than applying `.f` to each element individually when there are duplicate structures.
#'
#' **IMPORTANT PERFORMANCE NOTE:**
#' Due to the inclusion of position indices, `simap` functions have **O(total_structures)**
#' time complexity because each position creates a unique combination, even with identical structures.
#'
#' **Alternative:** Consider `smap()` functions if position information is not required.
#'
#' The index passed to `.f` is the position in the original vector (1-based).
#' If the vector has names, the names are passed instead of indices.
#'
#' **Return Types:**
#' - `simap()`: Returns a list with the same length as `.x`
#' - `simap_vec()`: Returns an atomic vector with the same length as `.x`
#' - `simap_lgl()`: Returns a logical vector
#' - `simap_int()`: Returns an integer vector
#' - `simap_dbl()`: Returns a double vector
#' - `simap_chr()`: Returns a character vector
#' - `simap_structure()`: Returns a new glycan structure vector (`.f` must return igraph objects)
#'
#' @return
#' - `simap()`: A list
#' - `simap_vec()`: An atomic vector of type specified by `.ptype`
#' - `simap_lgl()`: Returns a logical vector
#' - `simap_int()`: Returns an integer vector
#' - `simap_dbl()`: Returns a double vector
#' - `simap_chr()`: Returns a character vector
#' - `simap_structure()`: A new glyrepr_structure object
#'
#' @examples
#' # Create structure vectors with duplicates
#' core1 <- o_glycan_core_1()
#' core2 <- n_glycan_core()
#' structures <- c(core1, core2, core1)  # core1 appears twice
#'
#' # Map a function that uses both structure and index
#' simap_chr(structures, function(g, i) paste0("Structure_", i, "_vcount_", igraph::vcount(g)))
#'
#' # Use purrr-style lambda functions
#' simap_chr(structures, ~ paste0("Pos", .y, "_vertices", igraph::vcount(.x)))
#'
#' @name simap
NULL

#' @rdname simap
#' @export
simap <- function(.x, .f, ...) {
  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input `.x` must be a glycan_structure vector.")
  }

  # Handle empty input
  if (length(.x) == 0) {
    return(list())
  }

  .f <- rlang::as_function(.f)

  map_input <- .structure_map_input(.x)

  if (map_input$all_na) {
    result <- vector("list", length(map_input$codes))
    result[map_input$na_mask] <- list(NA)
    names(result) <- map_input$input_names
    return(result)
  }

  # Get indices or names
  if (!is.null(names(.x))) {
    indices <- names(.x)
  } else {
    indices <- seq_along(.x)
  }

  valid_indices <- indices[!map_input$na_mask]

  # Create unique combinations data frame for proper handling
  combinations_df <- .structure_combo_table(
    map_input$valid_codes,
    list(valid_indices)
  )
  unique_combinations_df <- .unique_structure_combos(combinations_df)

  # Apply function only to unique combinations
  unique_results <- purrr::map(
    seq_len(nrow(unique_combinations_df)),
    function(i) {
      row <- unique_combinations_df[i, ]
      index <- .extract_combo_args(row, 1)[[1]]
      .f(map_input$graphs[[row$code]], index, ...)
    }
  )
  names(unique_results) <- unique_combinations_df$combo_key

  # Map results back to original vector positions
  result <- purrr::map(combinations_df$combo_key, ~ unique_results[[.x]])
  names(result) <- valid_indices

  .restore_list_with_na(result, map_input, na_value = NA)
}

#' @rdname simap
#' @export
simap_vec <- function(.x, .f, ..., .ptype = NULL) {
  results <- simap(.x, .f, ...)
  vctrs::vec_c(!!!results, .ptype = .ptype)
}

#' @rdname simap
#' @export
simap_lgl <- function(.x, .f, ...) {
  simap_vec(.x, .f, ..., .ptype = logical())
}

#' @rdname simap
#' @export
simap_int <- function(.x, .f, ...) {
  simap_vec(.x, .f, ..., .ptype = integer())
}

#' @rdname simap
#' @export
simap_dbl <- function(.x, .f, ...) {
  simap_vec(.x, .f, ..., .ptype = double())
}

#' @rdname simap
#' @export
simap_chr <- function(.x, .f, ...) {
  simap_vec(.x, .f, ..., .ptype = character())
}

#' @rdname simap
#' @export
simap_structure <- function(.x, .f, ...) {
  if (!is_glycan_structure(.x)) {
    cli::cli_abort("Input `.x` must be a glycan_structure vector.")
  }

  # Handle empty input
  if (length(.x) == 0) {
    return(glycan_structure())
  }

  .f <- rlang::as_function(.f)

  map_input <- .structure_map_input(.x)

  if (map_input$all_na) {
    return(.restore_structure_with_na(NULL, map_input))
  }

  # Get indices or names
  if (!is.null(names(.x))) {
    indices <- names(.x)
  } else {
    indices <- seq_along(.x)
  }

  valid_indices <- indices[!map_input$na_mask]

  # Create unique combinations data frame for proper handling
  combinations_df <- .structure_combo_table(
    map_input$valid_codes,
    list(valid_indices)
  )
  unique_combinations_df <- .unique_structure_combos(combinations_df)

  # Apply function only to unique combinations
  new_graphs <- purrr::map(seq_len(nrow(unique_combinations_df)), function(i) {
    row <- unique_combinations_df[i, ]
    index <- .extract_combo_args(row, 1)[[1]]
    result <- .f(map_input$graphs[[row$code]], index, ...)
    if (!inherits(result, "igraph")) {
      cli::cli_abort(
        "Function `.f` must return an igraph object when using `simap_structure()`."
      )
    }
    result
  })

  # Rebuild glycan_structure with proper deduplication
  idx <- match(combinations_df$combo_key, unique_combinations_df$combo_key)
  valid_result <- .rebuild_structure_with_dedup(new_graphs, idx)
  names(valid_result) <- valid_indices

  .restore_structure_with_na(valid_result, map_input)
}

test_that("smap functions preserve names in output", {
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)
  names(structures) <- c("A", "B", "C")

  # smap_int should preserve names
  result <- smap_int(structures, igraph::vcount)
  expect_equal(names(result), c("A", "B", "C"))

  # smap_chr should preserve names
  result_chr <- smap_chr(structures, ~ .x$anomer)
  expect_equal(names(result_chr), c("A", "B", "C"))

  # smap should preserve names in list
  result_list <- smap(structures, ~ igraph::vcount(.x))
  expect_equal(names(result_list), c("A", "B", "C"))
})

test_that("smap functions work with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test smap_int with regular function
  result <- smap_int(structures, igraph::vcount)
  expect_equal(length(result), 3)
  expect_type(result, "integer")

  # Test smap_chr with regular function
  result_chr <- smap_chr(structures, function(g) g$anomer)
  expect_equal(length(result_chr), 3)
  expect_type(result_chr, "character")

  # Test smap_lgl with regular function
  result_lgl <- smap_lgl(structures, function(g) igraph::vcount(g) > 5)
  expect_equal(length(result_lgl), 3)
  expect_type(result_lgl, "logical")
})

test_that("smap functions work with purrr-style lambda functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test smap_int with purrr lambda
  result_lambda <- smap_int(structures, ~ igraph::vcount(.x))
  result_regular <- smap_int(structures, igraph::vcount)
  expect_equal(result_lambda, result_regular)

  # Test smap_chr with purrr lambda
  result_chr_lambda <- smap_chr(structures, ~ .x$anomer)
  result_chr_regular <- smap_chr(structures, function(g) g$anomer)
  expect_equal(result_chr_lambda, result_chr_regular)

  # Test smap_lgl with purrr lambda
  result_lgl_lambda <- smap_lgl(structures, ~ igraph::vcount(.x) > 5)
  result_lgl_regular <- smap_lgl(structures, function(g) igraph::vcount(g) > 5)
  expect_equal(result_lgl_lambda, result_lgl_regular)
})

test_that("smap_unique works with purrr-style lambda functions", {
  # Create test structures with duplicates
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1, core1)

  # Test with regular function
  result_regular <- smap_unique(structures, igraph::vcount)
  expect_equal(length(result_regular), 1) # Only one unique structure

  # Test with purrr lambda
  result_lambda <- smap_unique(structures, ~ igraph::vcount(.x))
  expect_equal(result_lambda, result_regular)
})

test_that("smap_structure works with purrr-style lambda functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1)

  # Function that adds vertex names if not present
  add_names_regular <- function(g) {
    if (!("name" %in% igraph::vertex_attr_names(g))) {
      igraph::set_vertex_attr(
        g,
        "name",
        value = paste0("v", seq_len(igraph::vcount(g)))
      )
    } else {
      g
    }
  }

  # Test with regular function
  result_regular <- smap_structure(structures, add_names_regular)

  # Test with purrr lambda
  result_lambda <- smap_structure(
    structures,
    ~ {
      if (!("name" %in% igraph::vertex_attr_names(.x))) {
        igraph::set_vertex_attr(
          .x,
          "name",
          value = paste0("v", seq_len(igraph::vcount(.x)))
        )
      } else {
        .x
      }
    }
  )

  # Both results should be equivalent
  expect_s3_class(result_lambda, "glyrepr_structure")
  expect_equal(length(result_lambda), length(result_regular))
})

test_that("get_anomer function works correctly", {
  # Test the specific case from the user's example
  x <- n_glycan_core()
  result <- get_anomer(x)
  expect_type(result, "character")
  expect_equal(length(result), 1)
  expect_equal(result, "b1")

  # Test that it's equivalent to purrr lambda style
  result_lambda <- smap_chr(x, ~ .x$anomer)
  expect_equal(result, result_lambda)
})

test_that("ssome works with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1() # smaller structure
  core2 <- n_glycan_core() # larger structure
  structures <- c(core1, core2, core1)

  # Test if some structures have more than 5 vertices
  result <- ssome(structures, function(g) igraph::vcount(g) > 5)
  expect_type(result, "logical")
  expect_equal(length(result), 1)
  expect_true(is.logical(result))

  # Test if some structures have more than 20 vertices (should be FALSE)
  result_false <- ssome(structures, function(g) igraph::vcount(g) > 20)
  expect_false(result_false)

  # Test if some structures have at least 1 vertex (should be TRUE)
  result_true <- ssome(structures, function(g) igraph::vcount(g) >= 1)
  expect_true(result_true)
})

test_that("ssome works with purrr-style lambda functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test with purrr lambda
  result_lambda <- ssome(structures, ~ igraph::vcount(.x) > 5)
  result_regular <- ssome(structures, function(g) igraph::vcount(g) > 5)
  expect_equal(result_lambda, result_regular)

  # Test with another lambda
  result_lambda2 <- ssome(structures, ~ igraph::vcount(.x) >= 1)
  expect_true(result_lambda2)
})

test_that("severy works with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test if all structures have at least 1 vertex (should be TRUE)
  result_true <- severy(structures, function(g) igraph::vcount(g) >= 1)
  expect_true(result_true)

  # Test if all structures have more than 10 vertices (should be FALSE)
  result_false <- severy(structures, function(g) igraph::vcount(g) > 10)
  expect_false(result_false)

  # Test if all structures have at least 3 vertices
  result <- severy(structures, function(g) igraph::vcount(g) >= 3)
  expect_type(result, "logical")
  expect_equal(length(result), 1)
})

test_that("severy works with purrr-style lambda functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test with purrr lambda
  result_lambda <- severy(structures, ~ igraph::vcount(.x) >= 1)
  result_regular <- severy(structures, function(g) igraph::vcount(g) >= 1)
  expect_equal(result_lambda, result_regular)

  # Test with another lambda
  result_lambda2 <- severy(structures, ~ igraph::vcount(.x) >= 3)
  expect_type(result_lambda2, "logical")
})

test_that("snone works with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test if no structures have more than 50 vertices (should be TRUE)
  result_true <- snone(structures, function(g) igraph::vcount(g) > 50)
  expect_true(result_true)

  # Test if no structures have at least 1 vertex (should be FALSE)
  result_false <- snone(structures, function(g) igraph::vcount(g) >= 1)
  expect_false(result_false)

  # Test return type
  result <- snone(structures, function(g) igraph::vcount(g) > 20)
  expect_type(result, "logical")
  expect_equal(length(result), 1)
})

test_that("snone works with purrr-style lambda functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test with purrr lambda
  result_lambda <- snone(structures, ~ igraph::vcount(.x) > 50)
  result_regular <- snone(structures, function(g) igraph::vcount(g) > 50)
  expect_equal(result_lambda, result_regular)

  # Test with another lambda
  result_lambda2 <- snone(structures, ~ igraph::vcount(.x) >= 1)
  expect_false(result_lambda2)
})

test_that("structure predicate functions work with duplicate structures efficiently", {
  # Create structures with many duplicates to test efficiency
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1, core1, core1, core1) # same structure 5 times

  # These should only evaluate the predicate once for the unique structure
  result_some <- ssome(structures, function(g) igraph::vcount(g) > 3)
  result_every <- severy(structures, function(g) igraph::vcount(g) > 3)
  result_none <- snone(structures, function(g) igraph::vcount(g) > 50)

  expect_type(result_some, "logical")
  expect_type(result_every, "logical")
  expect_type(result_none, "logical")

  expect_equal(length(result_some), 1)
  expect_equal(length(result_every), 1)
  expect_equal(length(result_none), 1)
})

test_that("structure predicate functions handle edge cases", {
  # Test with single structure
  single <- o_glycan_core_1()
  expect_type(ssome(single, ~ igraph::vcount(.x) > 0), "logical")
  expect_type(severy(single, ~ igraph::vcount(.x) > 0), "logical")
  expect_type(snone(single, ~ igraph::vcount(.x) > 50), "logical")
})

# Tests for smap2 functions -----------------------------------------

test_that("smap2 functions work with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)
  weights <- c(1.0, 2.0, 1.5)

  # Test smap2_dbl with regular function
  result <- smap2_dbl(structures, weights, function(g, w) igraph::vcount(g) * w)
  expect_equal(length(result), 3)
  expect_type(result, "double")

  # Test smap2_int with regular function
  n_vertices <- c(1, 2, 1)
  result_int <- smap2_int(structures, n_vertices, function(g, n) {
    igraph::vcount(g) + n
  })
  expect_equal(length(result_int), 3)
  expect_type(result_int, "integer")

  # Test smap2_lgl with regular function
  thresholds <- c(5, 6, 5)
  result_lgl <- smap2_lgl(structures, thresholds, function(g, t) {
    igraph::vcount(g) > t
  })
  expect_equal(length(result_lgl), 3)
  expect_type(result_lgl, "logical")
})

test_that("smap2 functions work with purrr-style lambda functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)
  weights <- c(1.0, 2.0, 1.5)

  # Test smap2_dbl with purrr lambda
  result_lambda <- smap2_dbl(structures, weights, ~ igraph::vcount(.x) * .y)
  result_regular <- smap2_dbl(structures, weights, function(g, w) {
    igraph::vcount(g) * w
  })
  expect_equal(result_lambda, result_regular)

  # Test smap2_lgl with purrr lambda
  thresholds <- c(5, 6, 5)
  result_lgl_lambda <- smap2_lgl(
    structures,
    thresholds,
    ~ igraph::vcount(.x) > .y
  )
  result_lgl_regular <- smap2_lgl(structures, thresholds, function(g, t) {
    igraph::vcount(g) > t
  })
  expect_equal(result_lgl_lambda, result_lgl_regular)
})

test_that("smap2 functions work with recycling", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test with single weight (should be recycled)
  single_weight <- 2.5
  result <- smap2_dbl(structures, single_weight, ~ igraph::vcount(.x) * .y)
  expect_equal(length(result), 3)
  expect_type(result, "double")

  # All results should use the same weight
  expected <- smap_dbl(structures, ~ igraph::vcount(.x) * 2.5)
  expect_equal(result, expected)
})

test_that("smap2_structure works with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2) # different structures

  # Function that adds a graph attribute (doesn't change structure topology)
  add_attribute <- function(g, attr_val) {
    igraph::set_graph_attr(g, "test_attr", attr_val)
  }

  attr_values <- c("value1", "value2")
  result <- smap2_structure(structures, attr_values, add_attribute)

  expect_s3_class(result, "glyrepr_structure")
  expect_equal(length(result), 2)

  # Check that the attributes were added correctly
  result_graphs <- get_structure_graphs(result)
  expect_equal(igraph::graph_attr(result_graphs[[1]], "test_attr"), "value1")
  expect_equal(igraph::graph_attr(result_graphs[[2]], "test_attr"), "value2")
})

test_that("smap2_structure works with purrr-style lambda functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2) # different structures

  # Test with purrr lambda that adds vertex attribute based on second argument
  labels <- c("label1", "label2")
  result <- smap2_structure(
    structures,
    labels,
    ~ {
      igraph::set_graph_attr(.x, "custom_label", .y)
    }
  )

  expect_s3_class(result, "glyrepr_structure")
  expect_equal(length(result), 2)

  # Check that the custom attributes were added
  result_graphs <- get_structure_graphs(result)
  expect_equal(igraph::graph_attr(result_graphs[[1]], "custom_label"), "label1")
  expect_equal(igraph::graph_attr(result_graphs[[2]], "custom_label"), "label2")
})

test_that("smap2 functions handle duplicate structures efficiently", {
  # Create structures with duplicates to test efficiency
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1, core1) # same structure 3 times
  weights <- c(1.0, 2.0, 1.0) # first and third are same

  # This should only compute twice: once for (core1, 1.0) and once for (core1, 2.0)
  result <- smap2_dbl(structures, weights, function(g, w) igraph::vcount(g) * w)

  expect_equal(length(result), 3)
  expect_type(result, "double")

  # First and third results should be equal since they have same structure and weight
  expect_equal(result[1], result[3])
  # Second result should be different
  expect_true(result[2] != result[1])
})

test_that("smap2 functions validate inputs", {
  core1 <- o_glycan_core_1()
  structures <- core1

  # Test that first argument must be glycan_structure
  expect_error(smap2_dbl(c(1, 2, 3), c(1, 2, 3), ~ .x * .y), "glycan_structure")
  expect_error(
    smap2_structure("not_a_structure", c(1), ~.x),
    "glycan_structure"
  )

  # Test that smap2_structure validates return type
  expect_error(
    smap2_structure(structures, 1, ~"not_an_igraph"),
    "igraph object"
  )
})

test_that("smap2 functions handle edge cases", {
  # Test with single structure
  single <- o_glycan_core_1()
  result <- smap2_dbl(single, 3.0, ~ igraph::vcount(.x) * .y)
  expect_equal(length(result), 1)
  expect_type(result, "double")

  # Test with empty structures
  empty <- glycan_structure()
  empty_result <- smap2_dbl(empty, numeric(0), ~ igraph::vcount(.x) * .y)
  expect_equal(length(empty_result), 0)
  expect_type(empty_result, "double")
})

test_that("smap2 basic functionality works", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)
  values <- c("a", "b", "c")

  # Test smap2 (returns list)
  result <- smap2(structures, values, function(g, v) {
    list(vcount = igraph::vcount(g), value = v)
  })

  expect_type(result, "list")
  expect_equal(length(result), 3)
  expect_equal(result[[1]]$value, "a")
  expect_equal(result[[2]]$value, "b")
  expect_equal(result[[3]]$value, "c")
})

test_that("smap2_chr works correctly", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)
  prefixes <- c("prefix1_", "prefix2_", "prefix3_")

  # Test smap2_chr
  result <- smap2_chr(structures, prefixes, function(g, p) {
    paste0(p, igraph::vcount(g))
  })

  expect_type(result, "character")
  expect_equal(length(result), 3)
  expect_true(all(grepl("^prefix[123]_\\d+$", result)))
})

# Tests for spmap functions -----------------------------------------

test_that("spmap functions work with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2)
  weights <- c(1.0, 2.0)
  factors <- c(2, 3)

  # Test spmap_dbl with regular function
  result <- spmap_dbl(list(structures, weights, factors), function(g, w, f) {
    igraph::vcount(g) * w * f
  })
  expect_equal(length(result), 2)
  expect_type(result, "double")

  # Test spmap_int with regular function
  add_values <- c(1, 2)
  result_int <- spmap_int(list(structures, add_values), function(g, a) {
    igraph::vcount(g) + a
  })
  expect_equal(length(result_int), 2)
  expect_type(result_int, "integer")

  # Test spmap_lgl with regular function
  thresholds <- c(5, 6)
  result_lgl <- spmap_lgl(list(structures, thresholds), function(g, t) {
    igraph::vcount(g) > t
  })
  expect_equal(length(result_lgl), 2)
  expect_type(result_lgl, "logical")
})

test_that("spmap functions work with purrr-style lambda functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2)
  weights <- c(1.0, 2.0)
  factors <- c(2, 3)

  # Test spmap_dbl with purrr lambda
  result_lambda <- spmap_dbl(
    list(structures, weights, factors),
    ~ igraph::vcount(..1) * ..2 * ..3
  )
  result_regular <- spmap_dbl(
    list(structures, weights, factors),
    function(g, w, f) igraph::vcount(g) * w * f
  )
  expect_equal(result_lambda, result_regular)

  # Test spmap_chr with purrr lambda
  prefixes <- c("pre1_", "pre2_")
  result_chr_lambda <- spmap_chr(
    list(structures, prefixes),
    ~ paste0(..2, igraph::vcount(..1))
  )
  expect_equal(length(result_chr_lambda), 2)
  expect_type(result_chr_lambda, "character")
})

test_that("spmap functions work with recycling", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test with single weight (should be recycled)
  single_weight <- 2.0
  single_factor <- 3
  result <- spmap_dbl(
    list(structures, single_weight, single_factor),
    ~ igraph::vcount(..1) * ..2 * ..3
  )
  expect_equal(length(result), 3)
  expect_type(result, "double")

  # All results should use the same weight and factor
  expected <- smap_dbl(structures, ~ igraph::vcount(.x) * 2.0 * 3)
  expect_equal(result, expected)
})

test_that("spmap_structure works with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2)

  # Function that adds graph attribute based on multiple arguments
  add_attributes <- function(g, label, value) {
    g <- igraph::set_graph_attr(g, "custom_label", label)
    g <- igraph::set_graph_attr(g, "custom_value", value)
    g
  }

  labels <- c("label1", "label2")
  values <- c(10, 20)
  result <- spmap_structure(list(structures, labels, values), add_attributes)

  expect_s3_class(result, "glyrepr_structure")
  expect_equal(length(result), 2)

  # Check that attributes were added
  graphs <- attr(result, "graphs")
  first_structure <- graphs[[vctrs::vec_data(result)[1]]]
  expect_equal(igraph::graph_attr(first_structure, "custom_label"), "label1")
  expect_equal(igraph::graph_attr(first_structure, "custom_value"), 10)
})

test_that("spmap functions handle duplicate structures efficiently", {
  # Create structures with duplicates to test efficiency
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1, core1)
  weights <- c(1.0, 2.0, 1.0)
  factors <- c(2, 3, 2)

  # This should only compute twice: once for (core1, 1.0, 2) and once for (core1, 2.0, 3)
  result <- spmap_dbl(list(structures, weights, factors), function(g, w, f) {
    igraph::vcount(g) * w * f
  })

  expect_equal(length(result), 3)
  expect_type(result, "double")

  # First and third should be equal (same combination)
  expect_equal(result[1], result[3])
  expect_false(result[1] == result[2]) # Different combination
})

test_that("spmap functions validate inputs", {
  core1 <- o_glycan_core_1()
  structures <- core1

  # Test that .l must be a list
  expect_error(spmap_dbl(structures, ~ igraph::vcount(.x)), "non-empty list")
  expect_error(spmap_dbl(list(), ~ igraph::vcount(.x)), "non-empty list")

  # Test that first element must be glycan_structure
  expect_error(
    spmap_dbl(list(c(1, 2, 3), c(1, 2, 3)), ~ .x * .y),
    "glycan_structure"
  )
  expect_error(
    spmap_structure(list("not_a_structure"), ~.x),
    "glycan_structure"
  )

  # Test that spmap_structure validates return type
  expect_error(
    spmap_structure(list(structures, 1), ~"not_an_igraph"),
    "igraph object"
  )
})

test_that("spmap functions handle edge cases", {
  # Test with single structure
  single <- o_glycan_core_1()
  result <- spmap_dbl(list(single, 3.0, 2), ~ igraph::vcount(..1) * ..2 * ..3)
  expect_equal(length(result), 1)
  expect_type(result, "double")

  # Test with empty structures
  empty <- glycan_structure()
  empty_result <- spmap_dbl(
    list(empty, numeric(0), numeric(0)),
    ~ igraph::vcount(..1) * ..2 * ..3
  )
  expect_equal(length(empty_result), 0)
  expect_type(empty_result, "double")
})

test_that("spmap basic functionality works", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2)
  values1 <- c("a", "b")
  values2 <- c(1, 2)

  # Test spmap (returns list)
  result <- spmap(list(structures, values1, values2), function(g, v1, v2) {
    list(vcount = igraph::vcount(g), value1 = v1, value2 = v2)
  })

  expect_type(result, "list")
  expect_equal(length(result), 2)
  expect_equal(result[[1]]$value1, "a")
  expect_equal(result[[1]]$value2, 1)
  expect_equal(result[[2]]$value1, "b")
  expect_equal(result[[2]]$value2, 2)
})

# Tests for simap functions -----------------------------------------

test_that("simap functions work with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test simap_chr with regular function (using index)
  result <- simap_chr(structures, function(g, i) {
    paste0("Structure_", i, "_vcount_", igraph::vcount(g))
  })
  expect_equal(length(result), 3)
  expect_type(result, "character")
  expect_true(grepl("Structure_1_", result[1]))
  expect_true(grepl("Structure_2_", result[2]))
  expect_true(grepl("Structure_3_", result[3]))

  # Test simap_int with regular function (using index)
  result_int <- simap_int(structures, function(g, i) igraph::vcount(g) + i)
  expect_equal(length(result_int), 3)
  expect_type(result_int, "integer")

  # Test simap_lgl with regular function (using index)
  result_lgl <- simap_lgl(structures, function(g, i) i > 1)
  expect_equal(length(result_lgl), 3)
  expect_type(result_lgl, "logical")
  expect_false(result_lgl[1]) # index 1
  expect_true(result_lgl[2]) # index 2
  expect_true(result_lgl[3]) # index 3
})

test_that("simap functions work with purrr-style lambda functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Test simap_chr with purrr lambda
  result_lambda <- simap_chr(
    structures,
    ~ paste0("Pos", .y, "_vertices", igraph::vcount(.x))
  )
  result_regular <- simap_chr(structures, function(g, i) {
    paste0("Pos", i, "_vertices", igraph::vcount(g))
  })
  expect_equal(result_lambda, result_regular)

  # Test simap_dbl with purrr lambda
  result_dbl <- simap_dbl(structures, ~ igraph::vcount(.x) * .y)
  expect_equal(length(result_dbl), 3)
  expect_type(result_dbl, "double")
})

test_that("simap_structure works with regular functions", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2)

  # Function that adds index/name as graph attribute
  add_index_attr <- function(g, idx) {
    igraph::set_graph_attr(g, "position", idx)
  }

  result <- simap_structure(structures, add_index_attr)

  expect_s3_class(result, "glyrepr_structure")
  expect_equal(length(result), 2)

  # Check that attributes were added
  graphs <- attr(result, "graphs")
  first_structure <- graphs[[vctrs::vec_data(result)[1]]]
  expect_equal(igraph::graph_attr(first_structure, "position"), 1)
})

test_that("simap functions handle duplicate structures efficiently", {
  # Create structures with duplicates to test efficiency
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1, core1)

  # This should compute three times since indices are different: (core1, 1), (core1, 2), (core1, 3)
  result <- simap_chr(structures, function(g, i) {
    paste0("Structure_", i, "_vcount_", igraph::vcount(g))
  })

  expect_equal(length(result), 3)
  expect_type(result, "character")

  # All should be different because indices are different
  expect_false(result[1] == result[2])
  expect_false(result[1] == result[3])
  expect_false(result[2] == result[3])

  # Test that each result contains the expected structure and index information
  expect_true(grepl("Structure_1_", result[1]))
  expect_true(grepl("Structure_2_", result[2]))
  expect_true(grepl("Structure_3_", result[3]))
})

test_that("simap functions validate inputs", {
  core1 <- o_glycan_core_1()
  structures <- core1

  # Test that first argument must be glycan_structure
  expect_error(simap_chr(c(1, 2, 3), ~ paste(.x, .y)), "glycan_structure")
  expect_error(simap_structure("not_a_structure", ~.x), "glycan_structure")

  # Test that simap_structure validates return type
  expect_error(
    simap_structure(structures, ~"not_an_igraph"),
    "igraph object"
  )
})

test_that("simap functions handle edge cases", {
  # Test with single structure
  single <- o_glycan_core_1()
  result <- simap_chr(
    single,
    ~ paste0("Index", .y, "_vertices", igraph::vcount(.x))
  )
  expect_equal(length(result), 1)
  expect_type(result, "character")
  expect_true(grepl("Index1_", result[1]))

  # Test with empty structures
  empty <- glycan_structure()
  empty_result <- simap_chr(empty, ~ paste0(.y, "_", igraph::vcount(.x)))
  expect_equal(length(empty_result), 0)
  expect_type(empty_result, "character")
})

test_that("simap basic functionality works", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2)

  # Test simap (returns list)
  result <- simap(structures, function(g, i) {
    list(vcount = igraph::vcount(g), index = i)
  })

  expect_type(result, "list")
  expect_equal(length(result), 2)
  expect_equal(result[[1]]$index, 1)
  expect_equal(result[[2]]$index, 2)
  expect_true(is.numeric(result[[1]]$vcount))
  expect_true(is.numeric(result[[2]]$vcount))
})

# Tests for parallel functionality -----------------------------------------

test_that("smap parallel parameter works correctly", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)

  # Simple test function
  test_func <- function(g) igraph::vcount(g)

  # Test sequential processing (explicit)
  result_seq <- smap_int(structures, test_func, .parallel = FALSE)
  expect_equal(length(result_seq), 3)
  expect_type(result_seq, "integer")

  # Test default behavior (should be sequential for small dataset)
  result_default <- smap_int(structures, test_func)
  expect_equal(result_seq, result_default)

  # Test that results are identical regardless of parallel setting
  # Note: We don't test .parallel = TRUE here to avoid requiring parallel backend setup
  expect_equal(result_seq, result_default)
})

test_that("smap parallel parameter is not passed to user function", {
  # Create test structures
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1)

  # Function that would fail if .parallel is passed as argument
  strict_func <- function(g) {
    # This function only accepts one argument
    if (length(as.list(match.call())) > 2) {
      stop("Too many arguments passed to user function")
    }
    igraph::vcount(g)
  }

  # This should work without error - .parallel should not be passed to user function
  expect_no_error(smap_int(structures, strict_func, .parallel = FALSE))
  expect_no_error(smap_int(structures, strict_func, .parallel = NULL))
})

test_that("smap handles additional arguments correctly with parallel parameter", {
  # Create test structures
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2)

  # Function that takes additional arguments
  func_with_args <- function(g, multiplier = 1, offset = 0) {
    igraph::vcount(g) * multiplier + offset
  }

  # Test with additional arguments and parallel parameter
  result1 <- smap_int(
    structures,
    func_with_args,
    multiplier = 2,
    offset = 1,
    .parallel = FALSE
  )
  result2 <- smap_int(
    structures,
    func_with_args,
    multiplier = 2,
    offset = 1,
    .parallel = NULL
  )

  expect_equal(length(result1), 2)
  expect_equal(result1, result2)
  expect_type(result1, "integer")

  # Verify the calculation is correct by computing expected values
  # Get the actual structures from the glycan_structure object
  graphs <- attr(structures, "graphs")
  codes <- vctrs::vec_data(structures)

  expected1 <- igraph::vcount(graphs[[codes[1]]]) * 2 + 1
  expected2 <- igraph::vcount(graphs[[codes[2]]]) * 2 + 1
  expect_equal(result1[1], expected1)
  expect_equal(result1[2], expected2)
})

test_that("smap auto-parallel threshold behavior", {
  # Create a small dataset (should not trigger auto-parallel)
  core1 <- o_glycan_core_1()
  small_structures <- c(core1, core1, core1) # 3 total, 1 unique

  # Simple function
  simple_func <- function(g) igraph::vcount(g)

  # With small dataset, auto-parallel should behave like sequential
  result_auto <- smap_int(small_structures, simple_func, .parallel = NULL)
  result_seq <- smap_int(small_structures, simple_func, .parallel = FALSE)

  expect_equal(result_auto, result_seq)
  expect_equal(length(result_auto), 3)
  expect_type(result_auto, "integer")
})

test_that("smap parallel parameter validation", {
  # Create test structures
  core1 <- o_glycan_core_1()
  structures <- core1

  # Test that .parallel accepts valid values
  expect_no_error(smap_int(structures, igraph::vcount, .parallel = TRUE))
  expect_no_error(smap_int(structures, igraph::vcount, .parallel = FALSE))
  expect_no_error(smap_int(structures, igraph::vcount, .parallel = NULL))

  # Test that invalid .parallel values are handled gracefully
  # The function should either work or give a meaningful error, not crash
  expect_error(
    smap_int(structures, igraph::vcount, .parallel = "invalid"),
    "invalid 'x' type"
  )
})

test_that("smap_structure correctly updates unique structures count when modifications create duplicates", {
  # Create structures that will become identical after modification
  glycans <- c("Gal(a1-3)GalNAc(a1-", "Gal(a1-4)GalNAc(a1-")
  structures <- as_glycan_structure(glycans)

  # Before modification: should have 2 unique structures
  expect_equal(length(attr(structures, "graphs")), 2)

  # Remove linkages - both structures should become identical
  result <- remove_linkages(structures)

  # After modification: should have only 1 unique structure
  expect_equal(length(attr(result, "graphs")), 1)

  # Both elements should have the same IUPAC code
  expect_equal(as.character(result)[1], as.character(result)[2])
  expect_equal(as.character(result)[1], "Gal(??-?)GalNAc(??-")
})

test_that("smap2_structure correctly updates unique structures count when modifications create duplicates", {
  # Create structures that will become identical after modification
  glycans <- c("Gal(a1-3)GalNAc(a1-", "Gal(a1-4)GalNAc(a1-")
  structures <- as_glycan_structure(glycans)
  values <- c("test1", "test2")

  # Before modification: should have 2 unique structures
  expect_equal(length(attr(structures, "graphs")), 2)

  # Function that removes linkages regardless of second argument
  remove_linkages_func <- function(g, val) {
    igraph::set_edge_attr(g, "linkage", value = "??-?") |>
      igraph::set_graph_attr("anomer", value = "??")
  }

  result <- smap2_structure(structures, values, remove_linkages_func)

  # After modification: should have only 1 unique structure
  expect_equal(length(attr(result, "graphs")), 1)

  # Both elements should have the same IUPAC code
  expect_equal(as.character(result)[1], as.character(result)[2])
  expect_equal(as.character(result)[1], "Gal(??-?)GalNAc(??-")
})

test_that("spmap_structure correctly updates unique structures count when modifications create duplicates", {
  # Create structures that will become identical after modification
  glycans <- c("Gal(a1-3)GalNAc(a1-", "Gal(a1-4)GalNAc(a1-")
  structures <- as_glycan_structure(glycans)
  values1 <- c("test1", "test2")
  values2 <- c(1, 2)

  # Before modification: should have 2 unique structures
  expect_equal(length(attr(structures, "graphs")), 2)

  # Function that removes linkages regardless of other arguments
  remove_linkages_func <- function(g, val1, val2) {
    igraph::set_edge_attr(g, "linkage", value = "??-?") |>
      igraph::set_graph_attr("anomer", value = "??")
  }

  result <- spmap_structure(
    list(structures, values1, values2),
    remove_linkages_func
  )

  # After modification: should have only 1 unique structure
  expect_equal(length(attr(result, "graphs")), 1)

  # Both elements should have the same IUPAC code
  expect_equal(as.character(result)[1], as.character(result)[2])
  expect_equal(as.character(result)[1], "Gal(??-?)GalNAc(??-")
})

# Additional regression tests for the smap2 nested list fix
test_that("smap2 handles real glycan structures with nested match results correctly", {
  # Create a realistic glycan structure (simulating glyenzy use case)
  glycan_code <- "GlcNAc(b1-2)Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"
  structures <- as_glycan_structure(glycan_code)

  # Create nested list simulating enzyme rule match results
  # This is the exact structure that was causing the original bug
  rule_matches <- list(list(c(6, 5, 4, 3, 2, 1)))

  # Test that smap2 processes this correctly
  call_count <- 0
  result <- smap2(structures, rule_matches, function(graph, match_data) {
    call_count <<- call_count + 1

    # Verify the graph is correct
    expect_true(igraph::is_igraph(graph))
    expect_equal(igraph::vcount(graph), 6)

    # Verify the match data structure is preserved
    expect_true(is.list(match_data))
    expect_length(match_data, 1)
    expect_equal(match_data[[1]], c(6, 5, 4, 3, 2, 1))

    return("processed")
  })

  # Should be called exactly once (not 6 times due to expansion bug)
  expect_equal(call_count, 1)
  expect_length(result, 1)
  expect_equal(result[[1]], "processed")
})

test_that("smap2 tibble fix handles multiple glycans with different nested structures", {
  # Create multiple structures
  glycan_codes <- c(
    "GlcNAc(b1-2)Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-",
    "Gal(a1-3)GalNAc(a1-"
  )
  structures <- as_glycan_structure(glycan_codes)

  # Create different nested structures for each glycan
  nested_data <- list(
    list(c(6, 5, 4, 3, 2, 1)), # First glycan: complex match
    list(c(2, 1)) # Second glycan: simple match
  )

  results <- smap2(structures, nested_data, function(graph, match_data) {
    return(list(
      vertex_count = igraph::vcount(graph),
      match_length = length(match_data[[1]])
    ))
  })

  # Verify results correspond correctly to inputs
  expect_length(results, 2)
  expect_equal(results[[1]]$vertex_count, 6)
  expect_equal(results[[1]]$match_length, 6)
  expect_equal(results[[2]]$vertex_count, 2)
  expect_equal(results[[2]]$match_length, 2)
})

test_that("smap2 hash generation works correctly for different list structures", {
  # Test the hash-based key generation with various list structures
  structures <- as_glycan_structure("Gal(a1-3)GalNAc(a1-")[1]

  # Different nested list structures that should produce different keys
  list_variants <- list(
    list(c(1, 2, 3)),
    list(c(3, 2, 1)), # Same elements, different order
    list(c(1, 2, 3, 4)), # Different length
    list(list(a = 1, b = 2)) # Named list
  )

  # Each should be handled correctly without expansion
  for (i in seq_along(list_variants)) {
    result <- smap2(structures, list_variants[i], function(graph, data) {
      expect_true(is.list(data))
      return(paste0("variant_", i))
    })

    expect_length(result, 1)
    expect_equal(result[[1]], paste0("variant_", i))
  }
})

test_that("smap2 refactored code maintains performance with complex inputs", {
  # Performance regression test for the refactored hash generation
  glycan_codes <- c(
    "Gal(a1-3)GalNAc(a1-",
    "GlcNAc(b1-2)Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-",
    "Man(a1-2)Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"
  )
  structures <- as_glycan_structure(glycan_codes)

  # Create moderately complex nested structures
  complex_nested <- list(
    list(c(1:10)),
    list(list(x = 1:5, y = letters[1:5])),
    list(matrix(1:6, nrow = 2))
  )

  # Should complete without errors and reasonable time
  start_time <- Sys.time()
  result <- smap2(structures, complex_nested, function(graph, data) {
    return("completed")
  })
  end_time <- Sys.time()

  # Basic correctness checks
  expect_length(result, 3)
  expect_true(all(unlist(result) == "completed"))

  # Performance should be reasonable (less than 1 second for this simple case)
  expect_true(as.numeric(end_time - start_time) < 1)
})

test_that("smap2 edge case: empty lists and mixed types", {
  glycan_codes <- c(
    "Gal(a1-3)GalNAc(a1-",
    "GlcNAc(b1-2)Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"
  )
  structures <- as_glycan_structure(glycan_codes)

  # Test with empty lists and mixed simple/complex types
  mixed_data <- list(
    list(), # Empty list
    "simple_string" # Non-list data
  )

  result <- smap2(structures, mixed_data, function(graph, data) {
    if (is.list(data)) {
      return(paste0("list_", length(data)))
    } else {
      return(paste0("non_list_", data))
    }
  })

  expect_length(result, 2)
  expect_equal(result[[1]], "list_0")
  expect_equal(result[[2]], "non_list_simple_string")
})

# Tests for names preservation -----------------------------------------

test_that("smap_structure preserves names", {
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1)
  names(structures) <- c("X", "Y")

  add_attr <- function(g) {
    igraph::set_graph_attr(g, "test", "value")
  }

  result <- smap_structure(structures, add_attr)
  expect_equal(names(result), c("X", "Y"))
})

test_that("smap2 functions preserve names", {
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)
  names(structures) <- c("A", "B", "C")
  weights <- c(1.0, 2.0, 1.0)

  # smap2_dbl should preserve names
  result <- smap2_dbl(structures, weights, function(g, w) igraph::vcount(g) * w)
  expect_equal(names(result), c("A", "B", "C"))

  # smap2 should preserve names in list
  result_list <- smap2(structures, weights, function(g, w) {
    list(count = igraph::vcount(g), weight = w)
  })
  expect_equal(names(result_list), c("A", "B", "C"))
})

test_that("smap2_structure preserves names", {
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1)
  names(structures) <- c("X", "Y")
  values <- c(1, 2)

  add_attr <- function(g, v) {
    igraph::set_graph_attr(g, "value", v)
  }

  result <- smap2_structure(structures, values, add_attr)
  expect_equal(names(result), c("X", "Y"))
})

# Tests for spmap names preservation -----------------------------------------

test_that("spmap functions preserve names", {
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2)
  names(structures) <- c("A", "B")
  weights <- c(1.0, 2.0)
  factors <- c(2, 3)

  # spmap_dbl should preserve names
  result <- spmap_dbl(list(structures, weights, factors), function(g, w, f) {
    igraph::vcount(g) * w * f
  })
  expect_equal(names(result), c("A", "B"))

  # spmap should preserve names in list
  result_list <- spmap(list(structures, weights, factors), function(g, w, f) {
    list(count = igraph::vcount(g), weight = w)
  })
  expect_equal(names(result_list), c("A", "B"))
})

test_that("spmap_structure preserves names", {
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1)
  names(structures) <- c("X", "Y")
  values1 <- c("a", "b")
  values2 <- c(1, 2)

  add_attrs <- function(g, v1, v2) {
    g <- igraph::set_graph_attr(g, "label", v1)
    igraph::set_graph_attr(g, "num", v2)
  }

  result <- spmap_structure(list(structures, values1, values2), add_attrs)
  expect_equal(names(result), c("X", "Y"))
})

# Tests for simap names preservation -----------------------------------------

test_that("simap functions preserve names", {
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)
  names(structures) <- c("A", "B", "C")

  # simap_chr should preserve names
  result <- simap_chr(structures, function(g, i) paste0("Structure_", i))
  expect_equal(names(result), c("A", "B", "C"))

  # simap should preserve names in list
  result_list <- simap(structures, function(g, i) list(index = i))
  expect_equal(names(result_list), c("A", "B", "C"))
})

test_that("simap_structure preserves names", {
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1)
  names(structures) <- c("X", "Y")

  add_index_attr <- function(g, idx) {
    igraph::set_graph_attr(g, "index", idx)
  }

  result <- simap_structure(structures, add_index_attr)
  expect_equal(names(result), c("X", "Y"))
})

test_that("smap_unique documents behavior with named input", {
  core1 <- o_glycan_core_1()
  structures <- c(core1, core1, core1)
  names(structures) <- c("A", "B", "C") # All same structure, different names

  # smap_unique operates on unique structures only
  result <- smap_unique(structures, igraph::vcount)

  # Result is named by structure hash (IUPAC), not input names
  expect_true(is.null(names(result)) || startsWith(names(result), "Gal"))
})

test_that("get_structure_level returns one unnamed level for a named vector", {
  core1 <- o_glycan_core_1()
  core2 <- n_glycan_core()
  structures <- c(core1, core2, core1)
  names(structures) <- c("A", "B", "C")

  result <- get_structure_level(structures)
  expect_equal(result, "intact")
  expect_null(names(result))
})

# Tests for NA handling in smap functions -----------------------------------

test_that("smap skips NA elements", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA), n_glycan_core())
  counts <- smap_int(structs, igraph::vcount)
  expect_equal(length(counts), 3)
  expect_false(is.na(counts[1]))
  expect_true(is.na(counts[2]))
  expect_false(is.na(counts[3]))
})

test_that("smap_structure skips NA elements", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA), n_glycan_core())
  result <- smap_structure(structs, identity)
  expect_equal(length(result), 3)
  expect_false(is.na(result[1]))
  expect_true(is.na(result[2]))
  expect_false(is.na(result[3]))
})

test_that("ssome skips NA elements", {
  structs <- c(glycan_structure(NA), o_glycan_core_1())
  expect_true(ssome(structs, function(g) igraph::vcount(g) > 1))
})

test_that("severy skips NA elements", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA))
  expect_true(severy(structs, function(g) igraph::vcount(g) > 1))
})

test_that("snone skips NA elements", {
  structs <- glycan_structure(NA, NA)
  expect_true(snone(structs, function(g) igraph::vcount(g) > 100))
})

test_that(".smap_base handles NA elements correctly", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA), n_glycan_core())

  # smap_int should return NA for position 2
  result <- smap_int(structs, igraph::vcount)
  expect_equal(length(result), 3)
  expect_false(is.na(result[1]))
  expect_true(is.na(result[2]))
  expect_false(is.na(result[3]))

  # smap_chr should return NA for position 2
  result_chr <- smap_chr(structs, ~ .x$anomer)
  expect_equal(length(result_chr), 3)
  expect_false(is.na(result_chr[1]))
  expect_true(is.na(result_chr[2]))
  expect_false(is.na(result_chr[3]))

  # smap should return NULL for position 2 (for composition compatibility)
  result_list <- smap(structs, igraph::vcount)
  expect_equal(length(result_list), 3)
  expect_false(is.na(result_list[[1]]))
  expect_true(is.null(result_list[[2]]))
  expect_false(is.na(result_list[[3]]))
})

test_that(".smap_base handles all-NA vector with correct NA types", {
  # All-NA vector
  all_na_structs <- glycan_structure(NA, NA, NA)

  # smap_chr should return NA_character_, not "NA" string
  result_chr <- smap_chr(all_na_structs, ~ .x$anomer)
  expect_equal(length(result_chr), 3)
  expect_type(result_chr, "character")
  expect_true(all(is.na(result_chr)))
  # Verify it's true NA_character_, not the string "NA"
  expect_identical(result_chr[1], NA_character_)

  # smap_int should return NA_integer_
  result_int <- smap_int(all_na_structs, igraph::vcount)
  expect_equal(length(result_int), 3)
  expect_type(result_int, "integer")
  expect_true(all(is.na(result_int)))

  # smap_dbl should return NA_real_
  result_dbl <- smap_dbl(all_na_structs, igraph::vcount)
  expect_equal(length(result_dbl), 3)
  expect_type(result_dbl, "double")
  expect_true(all(is.na(result_dbl)))

  # smap_lgl should return NA (logical)
  result_lgl <- smap_lgl(all_na_structs, ~ igraph::vcount(.x) > 0)
  expect_equal(length(result_lgl), 3)
  expect_type(result_lgl, "logical")
  expect_true(all(is.na(result_lgl)))

  # smap returns list with NULL for NA elements
  result_list <- smap(all_na_structs, igraph::vcount)
  expect_equal(length(result_list), 3)
  expect_true(all(purrr::map_lgl(result_list, is.null)))
})

# Tests for NA handling in smap2 functions -----------------------------------

test_that("smap2 handles NA elements", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA), n_glycan_core())

  # smap2_dbl should return NA for position 2
  result <- smap2_dbl(structs, c(1, 2, 3), function(g, n) igraph::vcount(g) + n)
  expect_equal(length(result), 3)
  expect_false(is.na(result[1]))
  expect_true(is.na(result[2]))
  expect_false(is.na(result[3]))

  # smap2 with lambda should work
  result_lambda <- smap2_dbl(structs, c(1, 2, 3), ~ igraph::vcount(.x) + .y)
  expect_equal(result, result_lambda)
})

test_that("smap2 handles NA at different positions", {
  # NA at beginning
  structs1 <- c(glycan_structure(NA), o_glycan_core_1(), n_glycan_core())
  result1 <- smap2_dbl(structs1, c(1, 2, 3), ~ igraph::vcount(.x) + .y)
  expect_true(is.na(result1[1]))
  expect_false(is.na(result1[2]))
  expect_false(is.na(result1[3]))

  # NA at end
  structs3 <- c(o_glycan_core_1(), n_glycan_core(), glycan_structure(NA))
  result3 <- smap2_dbl(structs3, c(1, 2, 3), ~ igraph::vcount(.x) + .y)
  expect_false(is.na(result3[1]))
  expect_false(is.na(result3[2]))
  expect_true(is.na(result3[3]))
})

test_that("smap2_structure handles NA elements", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA), n_glycan_core())
  names(structs) <- c("A", "B", "C")
  values <- c(1, 2, 3)

  add_attr <- function(g, v) {
    igraph::set_graph_attr(g, "value", v)
  }

  result <- smap2_structure(structs, values, add_attr)

  expect_equal(length(result), 3)
  expect_equal(names(result), c("A", "B", "C"))

  # Check the graphs directly - there should be 2 unique graphs
  result_graphs <- attr(result, "graphs")
  expect_equal(length(result_graphs), 2)

  # Check that attributes were set correctly on the graphs
  expect_equal(igraph::graph_attr(result_graphs[[1]], "value"), 1)
  expect_equal(igraph::graph_attr(result_graphs[[2]], "value"), 3)

  # Verify result has correct codes (2 actual codes + 1 NA)
  result_codes <- vctrs::vec_data(result)
  expect_equal(sum(is.na(result_codes)), 1)
  expect_equal(sum(!is.na(result_codes)), 2)
})

# Tests for NA handling in spmap functions -----------------------------------

test_that("spmap handles NA elements", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA), n_glycan_core())

  # spmap_dbl should return NA for position 2
  result <- spmap_dbl(list(structs, c(1, 2, 3)), ~ igraph::vcount(..1) + ..2)
  expect_equal(length(result), 3)
  expect_false(is.na(result[1]))
  expect_true(is.na(result[2]))
  expect_false(is.na(result[3]))

  # spmap with lambda should work
  result_lambda <- spmap_dbl(list(structs, c(1, 2, 3)), function(g, n) {
    igraph::vcount(g) + n
  })
  expect_equal(result, result_lambda)
})

test_that("spmap_structure handles NA elements", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA), n_glycan_core())

  result <- spmap_structure(list(structs, c(1, 2, 3)), function(g, n) g)
  expect_equal(length(result), 3)
  expect_false(is.na(result[1]))
  expect_true(is.na(result[2]))
  expect_false(is.na(result[3]))
})

# Tests for NA handling in simap functions -----------------------------------

test_that("simap handles NA elements", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA), n_glycan_core())

  result <- simap_int(structs, function(g, i) igraph::vcount(g) + i)

  expect_equal(length(result), 3)
  expect_false(is.na(result[1]))
  expect_true(is.na(result[2]))
  expect_false(is.na(result[3]))
})

test_that("simap_structure handles NA elements", {
  structs <- c(o_glycan_core_1(), glycan_structure(NA), n_glycan_core())

  result <- simap_structure(structs, function(g, i) g)

  expect_equal(length(result), 3)
  expect_false(is.na(result[1]))
  expect_true(is.na(result[2]))
  expect_false(is.na(result[3]))
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(glyrepr)

## -----------------------------------------------------------------------------
# Our test data: some common glycan structures
iupacs <- c(
  "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-",  # N-glycan core
  "Gal(b1-3)GalNAc(a1-",                                    # O-glycan core 1
  "Gal(b1-3)[GlcNAc(b1-6)]GalNAc(a1-",                     # O-glycan core 2
  "Man(a1-3)[Man(a1-6)]Man(a1-3)[Man(a1-6)]Man(a1-",          # Branched mannose
  "GlcNAc6Ac(b1-4)Glc3Me(a1-"                              # With decorations
)

struc <- as_glycan_structure(iupacs)

# Now create a realistic dataset with lots of repetition.
large_struc <- rep(struc, 1000)  # 5,000 total structures
large_struc

## -----------------------------------------------------------------------------
# Only 5 unique graphs are stored internally
length(attr(large_struc, "structures"))

# But we have 5,000 total elements
length(large_struc)

## -----------------------------------------------------------------------------
library(lobstr)
obj_sizes(struc, large_struc)

## -----------------------------------------------------------------------------
# This will not work and will raise an error.
tryCatch(
  purrr::map_int(large_struc, ~ igraph::vcount(.x)),
  error = function(e) cat("Error:", rlang::cnd_message(e))
)

## -----------------------------------------------------------------------------
vertex_counts <- smap_int(large_struc, ~ igraph::vcount(.x))
vertex_counts[1:10]

## -----------------------------------------------------------------------------
vertex_counts <- smap_int(large_struc, igraph::vcount)
summary(vertex_counts)

## -----------------------------------------------------------------------------
has_many_vertices <- smap_lgl(large_struc, ~ igraph::vcount(.x) > 4)
sum(has_many_vertices)

## -----------------------------------------------------------------------------
degree_sequences <- smap(large_struc, ~ igraph::degree(.x))
degree_sequences[1:3]

## -----------------------------------------------------------------------------
ssome(large_struc, ~ any(igraph::degree(.x) == 0))

## -----------------------------------------------------------------------------
severy(large_struc, ~ igraph::is_connected(.x))

## -----------------------------------------------------------------------------
# smap2: Apply function with additional parameters
thresholds <- c(3, 4, 5)
large_enough <- smap2_lgl(struc[1:3], thresholds, function(g, threshold) {
  igraph::vcount(g) >= threshold
})
large_enough

## -----------------------------------------------------------------------------
# simap: Include position information
indexed_report <- simap_chr(large_struc[1:3], function(g, i) {
  paste0("#", i, ": ", igraph::vcount(g), " vertices")
})
indexed_report

## -----------------------------------------------------------------------------
# Create a large dataset with high redundancy
huge_struc <- rep(struc, 5000)  # 25,000 structures, only 5 unique

cat("Dataset size:", length(huge_struc), "structures\n")
cat("Unique structures:", length(attr(huge_struc, "structures")), "\n")
cat("Redundancy factor:", length(huge_struc) / length(attr(huge_struc, "structures")), "x\n")

library(tictoc)

# Optimized approach: smap only processes 5 unique structures
tic("smap_int (optimized)")
vertex_counts_optimized <- smap_int(huge_struc, igraph::vcount)
toc()

# Naive approach: extract all graphs and process each one
tic("Naive approach (all graphs)")
all_graphs <- get_structure_graphs(huge_struc)  # Extracts all 25,000 graphs
vertex_counts_naive <- purrr::map_int(all_graphs, igraph::vcount)
toc()

# Verify results are equivalent (though data types may differ)
all.equal(vertex_counts_optimized, vertex_counts_naive)

## -----------------------------------------------------------------------------
# Calculate clustering coefficient for each structure
clustering_coeffs <- smap_dbl(large_struc, ~ igraph::transitivity(.x, type = "global"))
summary(clustering_coeffs)

## -----------------------------------------------------------------------------
# Create a compact structure summary.
structure_analysis <- smap(large_struc, function(g) {
  list(
    vertices = igraph::vcount(g),
    edges = igraph::ecount(g),
    diameter = ifelse(igraph::is_connected(g), igraph::diameter(g), NA),
    clustering = igraph::transitivity(g, type = "global")
  )
})

# Convert to a more usable format
analysis_df <- do.call(rbind, lapply(structure_analysis, data.frame))
head(analysis_df)

## -----------------------------------------------------------------------------
# Find only structures with exactly 5 vertices
has_five_vertices <- smap_lgl(large_struc, ~ igraph::vcount(.x) == 5)
five_vertex_structures <- large_struc[has_five_vertices]

cat("Found", sum(has_five_vertices), "structures with exactly 5 vertices\n")

## -----------------------------------------------------------------------------
# Custom motif detector
detect_branching <- function(g) {
  degrees <- igraph::degree(g)
  any(degrees >= 3)
}

# Apply to a large dataset using unique structure optimization.
has_branching <- smap_lgl(large_struc, detect_branching)
cat("Structures with branching:", sum(has_branching), "out of", length(large_struc), "\n")

# Use smap2 to check structures against complexity thresholds
complexity_thresholds <- rep(c(3, 4, 5, 2, 4), 1000)  # Thresholds for each structure
meets_threshold <- smap2_lgl(large_struc, complexity_thresholds, function(g, threshold) {
  igraph::vcount(g) >= threshold
})
cat("Structures meeting complexity threshold:", sum(meets_threshold), "out of", length(large_struc), "\n")

## -----------------------------------------------------------------------------
sessionInfo()

#' Get Color for Concrete Monosaccharides
#'
#' @param mono A monosaccharide name (character), potentially with substituents
#' @return A color code (character)
#' @keywords internal
get_mono_color <- function(mono) {
  # Extract base monosaccharide name (without substituents)
  base_mono <- .extract_base_mono(mono)

  dplyr::recode_values(
    base_mono,
    c(
      "Glc",
      "GlcNAc",
      "GlcN",
      "GlcA",
      "Qui",
      "QuiNAc",
      "Oli",
      "Bac",
      "Api"
    ) ~ "#0072BC",
    c(
      "Man",
      "ManNAc",
      "ManN",
      "ManA",
      "Rha",
      "RhaNAc",
      "Tyv",
      "Ara",
      "Kdn",
      "Pse",
      "LDmanHep",
      "Fru"
    ) ~ "#00A651",
    c("Gal", "GalNAc", "GalN", "GalA", "Lyx", "Leg", "Kdo", "Tag") ~ "#FFD400",
    c(
      "Gul",
      "GulNAc",
      "GulN",
      "GulA",
      "6dGul",
      "Abe",
      "Xyl",
      "Dha",
      "Sor"
    ) ~ "#F47920",
    c(
      "Alt",
      "AltNAc",
      "AltN",
      "AltA",
      "6dAlt",
      "6dAltNAc",
      "Par",
      "Rib",
      "Aci",
      "DDmanHep",
      "Psi"
    ) ~ "#F69EA1",
    c("All", "AllNAc", "AllN", "AllA", "Dig", "Neu5Ac", "MurNAc") ~ "#A54399",
    c(
      "Tal",
      "TalNAc",
      "TalN",
      "TalA",
      "6dTal",
      "6dTalNAc",
      "Col",
      "Neu5Gc",
      "4eLeg",
      "MurNGc"
    ) ~ "#8FCCE9",
    c("Ido", "IdoNAc", "IdoN", "IdoA", "Neu", "Mur") ~ "#A17A4D",
    c("Fuc", "FucNAc", "Sia") ~ "#ED1C24",
    default = "black"
  )
}

#' Add Colors to Monosaccharides
#'
#' @param monos A character vector of monosaccharide names
#' @param colored A logical value indicating whether to add colors
#' @return A character vector with ANSI color codes
#' @keywords internal
add_colors <- function(monos, colored = TRUE) {
  if (!colored) {
    return(monos)
  }

  colors <- purrr::map_chr(monos, get_mono_color)
  color_styles <- purrr::map(colors, cli::make_ansi_style)
  purrr::map2_chr(monos, color_styles, ~ .y(.x))
}


#' Replace Monosaccharides in String with Colored Versions
#'
#' @param text Character string containing monosaccharide names
#' @param mono_names Character vector of monosaccharide names to replace
#' @return Character string with monosaccharides replaced by colored versions
#' @keywords internal
replace_monos_with_colored <- function(text, mono_names) {
  # Get unique monosaccharides and their colors
  unique_monos <- unique(mono_names)
  colored_monos <- add_colors(unique_monos, colored = TRUE)
  names(colored_monos) <- unique_monos

  # Replace each monosaccharide in the text with its colored version
  result <- text
  for (i in seq_along(unique_monos)) {
    mono <- unique_monos[i]
    colored_mono <- colored_monos[i]

    # Special handling for monosaccharides that might have substituents
    # For example, "Neu5Ac" should match in "Neu5Ac9Ac"
    if (mono %in% c("Neu5Ac", "Neu5Gc", "Neu")) {
      # For sialic acids, match the base name followed by optional substituents
      pattern <- paste0("\\b", mono, "(?=[0-9]|\\(|$)")
      result <- stringr::str_replace_all(result, pattern, colored_mono)
    } else {
      # For other monosaccharides, use more flexible matching
      # Match the mono name followed by optional substituents (digits + letters)
      pattern <- paste0("\\b", mono, "(?=[0-9]|\\(|$)")
      result <- stringr::str_replace_all(result, pattern, colored_mono)
    }
  }

  result
}

#' Add Gray Color to Linkages in IUPAC String
#'
#' @param iupac_text Character string of IUPAC notation
#' @return Character string with linkages colored gray
#' @keywords internal
add_gray_linkages <- function(iupac_text) {
  # Gray color style
  gray_style <- cli::make_ansi_style("gray")

  result <- iupac_text

  # Pattern 1: Complete linkages like (b1-3), (a1-6)
  complete_pattern <- "\\(([ab?]\\d*-\\d*)\\)"
  result <- stringr::str_replace_all(result, complete_pattern, function(match) {
    linkage <- stringr::str_sub(match, 2, -2) # Remove parentheses
    paste0("(", gray_style(linkage), ")")
  })

  # Pattern 2: Incomplete linkages at end like (a1-, (?1-
  incomplete_pattern <- "\\(([ab?]\\d*-)$"
  result <- stringr::str_replace_all(
    result,
    incomplete_pattern,
    function(match) {
      linkage <- stringr::str_sub(match, 2, -1) # Remove opening parenthesis only
      paste0("(", gray_style(linkage))
    }
  )

  result
}

#' Apply Colors to IUPAC String (Monosaccharides + Gray Linkages)
#'
#' @param iupac_text Character string of IUPAC notation
#' @param mono_names Character vector of monosaccharide names to color
#' @return Character string with colored monosaccharides and gray linkages
#' @keywords internal
colorize_iupac_string <- function(iupac_text, mono_names) {
  # First, color the monosaccharides
  result <- replace_monos_with_colored(iupac_text, mono_names)

  # Then, make linkages gray
  result <- add_gray_linkages(result)

  result
}

#' Extract Base Monosaccharide Name (Without Substituents)
#'
#' @param mono A monosaccharide name (character), potentially with substituents
#' @return The base monosaccharide name without substituents
#' @keywords internal
.extract_base_mono <- function(mono) {
  # Use a simplified version of the substituent extraction logic
  # This avoids circular dependency with iupac-to-structure.R

  # Handle special cases first
  if (mono == "Neu5Ac") {
    return("Neu5Ac")
  } else if (stringr::str_starts(mono, "Neu5Ac") && nchar(mono) > 6) {
    return("Neu5Ac")
  } else if (mono == "Neu4Ac5Ac") {
    return("Neu5Ac")
  } else if (mono == "Neu4Ac5Gc") {
    return("Neu5Gc")
  }

  # For other monosaccharides, remove substituents
  subs_pattern <- stringr::str_c(available_substituents(), collapse = "|")
  single_sub_pattern <- stringr::str_glue("[1-9\\?]({subs_pattern})")

  # Find all substituents
  all_subs <- stringr::str_extract_all(mono, single_sub_pattern)[[1]]

  if (length(all_subs) > 0) {
    # Remove all substituents to get base monosaccharide
    clean_mono <- mono
    for (sub in all_subs) {
      clean_mono <- stringr::str_remove(clean_mono, stringr::fixed(sub))
    }
    return(clean_mono)
  } else {
    return(mono)
  }
}

test_that("monosaccharides with substituents are colored correctly", {
  expect_equal(get_mono_color("Neu5Ac9Ac"), "#A54399") # Purple for Neu5Ac
  expect_equal(get_mono_color("Gal6S"), "#FFD400") # Yellow for Gal
  expect_equal(get_mono_color("Man3Me"), "#00A651") # Green for Man
  expect_equal(get_mono_color("Glc"), "#0072BC") # Blue for Glc
})

test_that("special sialic acid cases are handled correctly", {
  # Test color assignment for special cases
  expect_equal(get_mono_color("Neu4Ac5Ac"), "#A54399") # Purple for Neu5Ac
  expect_equal(get_mono_color("Neu4Ac5Gc"), "#8FCCE9") # Light blue for Neu5Gc
})

#' Get the Number of Monosaccharides
#'
#' When `mono` is:
#' - `NULL` (default), returns the total number of monosaccharides and substituents.
#' - A string, returns the number of the specified monosaccharide or substituent.
#'
#' @details
#' When `mono` is "generic" (e.g. "Hex", "HexNAc"),
#' it counts all "concrete" monosaccharides that match.
#' For example, "Hex" will count all Glc, Man, Gal, etc.
#' When `mono` is "concrete" (e.g. "Gal", "GalNAc"),
#' NA is returned when the composition is "generic".
#'
#' @param x A glycan composition (`glyrepr_composition`) or
#'   a glycan structure (`glyrepr_structure`) vector.
#' @param mono The monosaccharide or substituent to count. A character scalar.
#'   If `NULL` (default), return the total number of monosaccharides.
#' @param include_subs Whether to include substituents when `mono` is `NULL`.
#'   Default is `FALSE`.
#'
#' @returns A numeric vector of the same length as `x`.
#'
#' @examples
#' comp <- glycan_composition(c(Gal = 1, Man = 1, GalNAc = 1))
#' count_mono(comp, "Hex")
#' count_mono(comp, "Gal")
#'
#' struct <- as_glycan_structure("Gal(b1-3)GlcNAc(b1-4)Glc(a1-")
#' count_mono(struct, "Gal")
#'
#' # Total number of monosaccharides
#' count_mono(comp)
#'
#' @export
count_mono <- function(x, mono = NULL, include_subs = FALSE) {
  UseMethod("count_mono")
}

.check_count_mono_args <- function(mono, include_subs) {
  checkmate::assert_flag(include_subs)
  if (is.null(mono)) {
    return()
  }
  checkmate::assert_string(mono)
  if (!mono %in% c(available_monosaccharides(), available_substituents())) {
    cli::cli_abort("{.arg mono} must be a known monosaccharide or substituent.")
  }
}

#' @rdname count_mono
#' @export
count_mono.glyrepr_composition <- function(
  x,
  mono = NULL,
  include_subs = FALSE
) {
  .check_count_mono_args(mono, include_subs)

  # Special behavior when `mono` is NULL: count all monosaccharides and substituents
  if (is.null(mono)) {
    data <- vctrs::field(vctrs::vec_data(x), "data")
    if (!include_subs) {
      data <- purrr::map(data, ~ .x[names(.x) %in% available_monosaccharides()])
    }
    # Handle NULL elements (NA compositions) - return NA for them
    return(purrr::map_int(data, function(x) {
      if (is.null(x) || length(x) == 0) NA_integer_ else sum(x)
    }))
  }

  # Get the type of the monosaccharide
  if (mono %in% available_substituents()) {
    mono_type <- "substituent" # the value is not used anywhere, just for readability
  } else {
    mono_type <- get_mono_type(mono)
  }
  # special monosaccharides are those having the same name for both generic and concrete types
  if (is.na(mono_type)) {
    mono_type <- "special" # the value is not used anywhere, just for readability
  }

  # Convert to generic if needed
  if (mono_type == "generic") {
    x <- convert_to_generic(x)
  }
  data <- vctrs::field(vctrs::vec_data(x), "data")

  # Count the number of monosaccharides or substituents
  count_one <- function(one_mono, mono) {
    # Handle NULL elements (NA compositions) - return NA for them
    if (is.null(one_mono) || length(one_mono) == 0) {
      return(NA_integer_)
    }
    if (mono %in% names(one_mono)) {
      n <- one_mono[[mono]]
      if (is.na(n)) {
        n <- 0L
      }
    } else {
      n <- 0L
    }
    n
  }
  res <- purrr::map_int(data, count_one, mono = mono)
  if (mono_type == "concrete") {
    x_mono_type <- get_mono_type(x)
    res[x_mono_type == "generic"] <- NA_integer_
  }
  res
}

#' @rdname count_mono
#' @export
count_mono.glyrepr_structure <- function(x, mono = NULL, include_subs = FALSE) {
  .check_count_mono_args(mono, include_subs)
  comps <- as_glycan_composition(x)
  count_mono.glyrepr_composition(comps, mono, include_subs)
}

---
title: "Getting Started with glyrepr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with glyrepr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

R gives us many data structures for different jobs.
For example, `c(1L, 2L, 3L)` represents an integer vector,
`c("a", "b", "c")` represents a character vector,
and data frames represent tabular data.
Before we can inspect or manipulate a kind of data well,
we first need a suitable structure for representing it.

What about glycans?

When we talk about glycans, we usually talk about two related things:
their compositions and their structures.
A composition can be represented as a named vector,
for example `c(Hex = 3, HexNAc = 2, Neu5Ac = 1)`.
A structure can be represented as a graph,
for example with the `igraph` package.
Those representations are useful, but managing the details by hand quickly becomes cumbersome,
especially for complex glycans or large datasets.
This is where the `glyrepr` package comes in.

`glyrepr` provides two main vector types:
`glycan_composition()` and `glycan_structure()`.
These vectors are designed to feel like ordinary R vectors:
you can subset them, concatenate them, sort them, put them in tibbles,
and use them in vectorized workflows.

These two representations are the foundation of `glycoverse`.
Most higher-level packages in the ecosystem build on them,
so it is worth taking a little time to get comfortable here.

```{r setup}
library(glyrepr)
```

## Glycan Composition Vectors

Let's start with glycan composition vectors.

### Creating Glycan Composition Vectors

A glycan composition vector can be created with either `glycan_composition()` or `as_glycan_composition()`.

`glycan_composition()` is the direct constructor.
It takes one or more named vectors, where the names are monosaccharides and the values are counts.

```{r}
comps <- glycan_composition(
  c(Man = 5, GlcNAc = 2),
  c(Man = 3, Gal = 2, GlcNAc = 4),
  c(Man = 3, Gal = 2, GlcNAc = 4, Neu5Ac = 1, Fuc = 1)
)
comps
```

This creates a glycan composition vector called `comps` with three glycan compositions.

`as_glycan_composition()` is more flexible and can convert several input types.

From a list of named vectors:

```{r}
as_glycan_composition(list(
  c(Man = 5, GlcNAc = 2),
  c(Man = 3, Gal = 2, GlcNAc = 4),
  c(Man = 3, Gal = 2, GlcNAc = 4, Neu5Ac = 1, Fuc = 1)
))
```

This looks similar to `glycan_composition()`,
but it is more convenient when your compositions are already stored in a list,
for example after reading or generating them programmatically.

```{r}
comp_list <- list(
  c(Man = 5, GlcNAc = 2),
  c(Man = 3, Gal = 2, GlcNAc = 4),
  c(Man = 3, Gal = 2, GlcNAc = 4, Neu5Ac = 1, Fuc = 1)
)
as_glycan_composition(comp_list)
```

From a character vector:

```{r}
as_glycan_composition(c("H5N2", "Hex(3)HexNAc(2)"))
```

Both compact notation (`"H5N2"`) and Byonic-style notation (`"Hex(3)HexNAc(2)"`) are supported.

From a glycan structure vector, which we will discuss below:

```{r}
strucs <- c(o_glycan_core_1(), o_glycan_core_2())
as_glycan_composition(strucs)
```

### Two Types of Monosaccharide Names

Before we move on, let's briefly discuss the two types of monosaccharide names used in `glyrepr`:

1. **Generic names**: Hex, HexNAc, dHex, etc.
2. **Specific names**: Man, GlcNAc, Fuc, etc.

Generic names are common in mass spectrometry data,
where it is often difficult to distinguish isomers from MS evidence alone.
For example, one `Hex` residue could be Man, Gal, Glc, or another hexose.
Specific names carry more biological detail and are common in glycan databases and literature.

`glyrepr` supports both types of names, but you cannot mix them in the same vector.

```{r}
# This raises an error because the monosaccharide names are mixed.
try(as_glycan_composition(c("Hex(5)HexNAc(2)", "Man(5)GlcNAc(2)")), silent = TRUE)
```

The same rule applies to glycan structure vectors.

### Inspecting Glycan Composition Vectors

The main function for inspecting glycan compositions is `count_mono()`.
Let's demonstrate it using the `comps` vector we created earlier.

```{r}
comps
```

The first argument is a glycan composition vector, and the second argument is the monosaccharide you want to count.

```{r}
count_mono(comps, "Man")
```

```{r}
count_mono(comps, "Neu5Ac")
```

`count_mono()` works with both generic and specific monosaccharide names.
It understands the relationship between them, so it can count at the level you ask for.

```{r}
count_mono(comps, "Hex")
```

Note that both "Man" and "Gal" are counted as "Hex" here.

You can also omit the second argument to get the total monosaccharide count for each composition.

```{r}
count_mono(comps)
```

### Manipulating Glycan Composition Vectors

One useful mental model for glycan composition vectors
(and glycan structure vectors) is that they behave like atomic vectors.
This means they support familiar operations like subsetting, concatenation, and sorting.

**Concatenation**:

```{r}
c(comps, comps)
```

**Subsetting**:

```{r}
comps[1:2]
```

```{r}
comps[integer()]
```

**Length**:

```{r}
length(comps)
```

**Unique**:

```{r}
dup_comps <- c(comps, comps)
dup_comps
```

```{r}
unique(dup_comps)
```

**Repeated**:

```{r}
rep(comps, times = 2)
```

**Sorting**:

```{r}
sort(comps)
```

```{r}
sort(comps, decreasing = TRUE)
```

### Working with Tibbles

One of the most useful features of glycan composition vectors is that they work smoothly with tibbles and data frames.

```{r}
library(tibble)

tb <- tibble(
  id = c("glycan1", "glycan2", "glycan3"),
  composition = comps
)
tb
```

You can use `tidyverse` functions to perform operations on the glycan composition column.

```{r}
library(dplyr)

tb |>
  mutate(n_sia = count_mono(composition, "Neu5Ac")) |>
  filter(n_sia > 0)
```

### Missing Values and Names

Glycan composition vectors can also handle missing values.

```{r}
comps_with_na <- glycan_composition(c(Man = 5, GlcNAc = 2), NA)
comps_with_na
```

```{r}
count_mono(comps_with_na, "Man")
```

For technical reasons, glycan composition vectors cannot have element names right now.
In practice, this is usually fine:
the composition itself remains the data value,
and identifiers can live in a separate tibble column.

## Glycan Structure Vectors

Now let's move on to the core of glycan representation:
glycan structure vectors.
Many of the same ideas apply,
including the atomic vector nature, vectorized operations, and seamless integration with tibbles.
Because structures are more complex than compositions,
there are a few additional features and considerations to keep in mind.

### Creating Glycan Structure Vectors

As with glycan composition vectors,
you can create glycan structure vectors with either `glycan_structure()` or `as_glycan_structure()`.

Under the hood, glycan structures are represented as `igraph` objects,
because glycans are naturally graph-like.
Most of the time, you do not need to work with those graphs directly:
`glyrepr` gives you a higher-level interface for common structure operations.

`glycan_structure()` is the direct constructor and takes one or more `igraph` objects.
Creating those graphs manually can be tedious,
so you will usually start with `as_glycan_structure()` instead.
It can parse IUPAC-condensed strings into glycan structure vectors.

If you are not familiar with IUPAC-condensed notation,
check out this [article](https://glycoverse.github.io/glyrepr/articles/iupac.html) for a quick introduction.
We recommend getting familiar with this notation,
because it is the main text language for communicating glycan structures in `glycoverse`.

Parsing IUPAC-condensed strings is straightforward with `as_glycan_structure()`.
(For parsing other formats like GlycoCT, use the [glyparse](https://github.com/glycoverse/glyparse) package.)

```{r}
strucs <- as_glycan_structure(c(
  "Gal(b1-3)GalNAc(a1-",
  "Gal(b1-3)[GlcNAc(b1-6)]GalNAc(a1-"
))
strucs
```

### Inspecting Glycan Structure Vectors

`strucs` prints a bit like a character vector with colors,
but under the hood the glycan structures are stored as `igraph` objects.
You can access the underlying `igraph` objects using `get_structure_graphs()`.

```{r}
get_structure_graphs(strucs)
```

Details of how a glycan structure is modeled as a graph are covered in the [glycan graph vignette](https://glycoverse.github.io/glyrepr/articles/glycan-graph.html).
You do not need all of those details for everyday use,
but a quick skim can make the structure functions easier to understand.

The `count_mono()` function also works with glycan structure vectors.

```{r}
count_mono(strucs, "Gal")
```

Other functions, including `has_linkages()`, `get_mono_type()`, and `get_structure_level()`,
inspect specific aspects of the structures.

```{r}
# This function works element-wise
has_linkages(strucs)
```

```{r}
get_mono_type(strucs)
```

```{r}
get_structure_level(strucs)
```

### Structure Levels

The structures we have seen so far are "intact" structures.
They contain specific monosaccharides and complete linkage information.
In real datasets, though, glycan structures often have missing information.
For example, we might know the topology but not the linkages,
or we might only know generic monosaccharide classes.

To accommodate these scenarios, `glyrepr` defines four levels of glycan structures:

- **Intact**: specific monosaccharides and complete linkage information.
- **Partial**: specific monosaccharides, with at least one missing linkage annotation.
- **Topological**: specific monosaccharides, but all linkage information is missing.
- **Basic**: generic monosaccharides, with linkage information treated as missing.

Structure levels are defined at the vector level,
so one glycan structure vector has one level.

Let's see some examples.

**Intact structures**:

```{r}
as_glycan_structure(c(
  "Gal(b1-3)GalNAc(a1-",
  "Gal(b1-3)[GlcNAc(b1-6)]GalNAc(a1-"
)) |> get_structure_level()
```

**Partial structures**:

```{r}
as_glycan_structure(c(
  "Gal(b1-?)GalNAc(a1-",
  "Gal(b1-?)[GlcNAc(b1-6)]GalNAc(a1-"
)) |> get_structure_level()
```

**Topological structures**:

```{r}
as_glycan_structure(c(
  "Gal(??-?)GalNAc(??-",
  "Gal(??-?)[GlcNAc(??-?)]GalNAc(??-"
)) |> get_structure_level()
```

**Basic structures**:

```{r}
as_glycan_structure(c(
  "Hex(??-?)HexNAc(??-",
  "Hex(??-?)[HexNAc(??-?)]HexNAc(??-"
)) |> get_structure_level()
```

In theory, you can have something like `"Hex(b1-3)HexNAc(a1-"`,
with generic monosaccharides but all linkages intact.
In practice, linkage information is usually harder to obtain than monosaccharide identities,
so this situation is rare.

If you create such a vector, `glyrepr` classifies it as `"basic"` and warns you.

```{r}
as_glycan_structure(c(
  "Hex(a1-3)HexNAc(a1-",
  "Hex(a1-3)[HexNAc(b1-6)]HexNAc(a1-"
)) |> get_structure_level()
```

### Manipulating Glycan Structure Vectors

Glycan structure vectors also support vectorized operations like subsetting and concatenation.
They also have structure-specific helpers for reducing resolution,
removing linkages, and removing substituents.

```{r}
strucs
```

```{r}
reduce_structure_level(strucs, to_level = "basic")
```

```{r}
remove_linkages(strucs)  # same as reduce_structure_level(strucs, to_level = "topological")
```

```{r}
strucs_with_subs <- as_glycan_structure(c(
  "Gal6S(b1-3)GalNAc(a1-",
  "Gal6S(b1-3)[GlcNAc(b1-6)]GalNAc(a1-"
))
remove_substituents(strucs_with_subs)
```

### Missing Values and Names

Like glycan composition vectors, glycan structure vectors can also handle missing values.

Different from glycan composition vectors, though, glycan structure vectors can have names.
This is a very useful feature we introduced in version 0.10.0.

```{r}
strings <- c(
  glycan1 = "Gal(b1-3)GalNAc(a1-",
  glycan2 = "Gal(b1-3)[GlcNAc(b1-6)]GalNAc(a1-"
)
as_glycan_structure(strings)
```

You will find the names useful when working with the `glymotif` package.

## What's Next?

This vignette sets the foundation for working with glycan representations in `glycoverse`.
From here, you can explore the packages that build on these representations:

- [glyparse](https://github.com/glycoverse/glyparse): parsing text nomenclatures into glycan representations.
- [glymotif](https://github.com/glycoverse/glymotif): identifying and analyzing glycan motifs.
- [glydet](https://github.com/glycoverse/glydet): calculating derived traits and quantifying motifs.
- [glydraw](https://github.com/glycoverse/glydraw): visualizing glycan structures with SNFG notation.
- [glyenzy](https://github.com/glycoverse/glyenzy): inspecting and modeling glycan biosynthesis.

Happy exploring.

---
title: "Efficient Glycan Manipulation with smap"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Efficient Glycan Manipulation with smap}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

This vignette introduces the `smap` family of functions.
These functions are useful when you want to apply custom `igraph`-based operations to glycan structure vectors.

This guide assumes you are comfortable with R programming and have some familiarity with graph concepts.
If you are just getting started,
read the "Getting Started with glyrepr" vignette first.

```{r setup}
library(glyrepr)
```

## Unique Structure Optimization

Before using `smap`, it helps to understand why these functions exist.

### The Problem

Working with glycan structures means working with graphs, 
and graph operations are computationally expensive. 
When you are analyzing thousands of glycans from a large-scale study,
this becomes a real bottleneck.

### The Solution

`glyrepr` implements an optimization called **unique structure storage**.
Instead of storing thousands of identical graphs, 
it stores only the unique ones and keeps track of which original positions they belong to.

Let's see this in action:

```{r}
# Our test data: some common glycan structures
iupacs <- c(
  "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-",  # N-glycan core
  "Gal(b1-3)GalNAc(a1-",                                    # O-glycan core 1
  "Gal(b1-3)[GlcNAc(b1-6)]GalNAc(a1-",                     # O-glycan core 2
  "Man(a1-3)[Man(a1-6)]Man(a1-3)[Man(a1-6)]Man(a1-",          # Branched mannose
  "GlcNAc6Ac(b1-4)Glc3Me(a1-"                              # With decorations
)

struc <- as_glycan_structure(iupacs)

# Now create a realistic dataset with lots of repetition.
large_struc <- rep(struc, 1000)  # 5,000 total structures
large_struc
```

Notice that the object reports only 5 unique structures.
The vector has 5,000 elements,
but only 5 unique graphs are stored internally.

We can verify that directly:

```{r}
# Only 5 unique graphs are stored internally
length(attr(large_struc, "structures"))

# But we have 5,000 total elements
length(large_struc)
```

### Memory Savings

```{r}
library(lobstr)
obj_sizes(struc, large_struc)
```

The memory difference can be substantial.
For repeated structures,
the optimized representation can be much smaller than storing every graph independently.

## The `smap` Family

There is one important consequence of this internal representation:
regular `lapply()` or `purrr::map()` functions do not operate directly on a glycan structure vector as if it were a list of graphs.

```{r}
# This will not work and will raise an error.
tryCatch(
  purrr::map_int(large_struc, ~ igraph::vcount(.x)),
  error = function(e) cat("Error:", rlang::cnd_message(e))
)
```

**Why does this fail?** 
Because `purrr` functions don't understand the internal structure optimization of `glycan_structure` objects.

### Structure-Aware Mapping

The `smap` functions are structure-aware alternatives to `purrr` mapping functions.
They understand the unique structure optimization and work directly with the underlying graph objects.

```{r}
vertex_counts <- smap_int(large_struc, ~ igraph::vcount(.x))
vertex_counts[1:10]
```

The "s" stands for "structure":
these functions operate on the underlying `igraph` objects that represent glycan structures.

## The `smap` Toolkit

The `smap` family provides glycan-aware equivalents for virtually all `purrr` functions:

| purrr | smap | purrr | smap |
|-------|------|-------|------|
| `map()` | `smap()` | `map2()` | `smap2()` |
| `map_lgl()` | `smap_lgl()` | `map2_lgl()` | `smap2_lgl()` |
| `map_int()` | `smap_int()` | `map2_int()` | `smap2_int()` |
| `map_dbl()` | `smap_dbl()` | `map2_dbl()` | `smap2_dbl()` |
| `map_chr()` | `smap_chr()` | `map2_chr()` | `smap2_chr()` |
| `some()` | `ssome()` | `pmap()` | `spmap()` |
| `every()` | `severy()` | `pmap_*()` | `spmap_*()` |
| `none()` | `snone()` | `imap()` | `simap()` |
| | | `imap_*()` | `simap_*()` |

As a simple rule, replace `map` with `smap`,
`pmap` with `spmap`, 
and `imap` with `simap`. 
The function signatures are designed to feel familiar if you already use `purrr`.

### Basic Examples

**Count vertices in each structure:**
```{r}
vertex_counts <- smap_int(large_struc, igraph::vcount)
summary(vertex_counts)
```

**Find structures with more than 4 vertices:**
```{r}
has_many_vertices <- smap_lgl(large_struc, ~ igraph::vcount(.x) > 4)
sum(has_many_vertices)
```

**Get the degree sequence of each structure:**
```{r}
degree_sequences <- smap(large_struc, ~ igraph::degree(.x))
degree_sequences[1:3]
```

**Check if any structure has isolated vertices:**
```{r}
ssome(large_struc, ~ any(igraph::degree(.x) == 0))
```

**Verify all structures are connected:**
```{r}
severy(large_struc, ~ igraph::is_connected(.x))
```

### Beyond Basic `smap()`

Quick examples of the extended family:

```{r}
# smap2: Apply function with additional parameters
thresholds <- c(3, 4, 5)
large_enough <- smap2_lgl(struc[1:3], thresholds, function(g, threshold) {
  igraph::vcount(g) >= threshold
})
large_enough
```

```{r}
# simap: Include position information
indexed_report <- simap_chr(large_struc[1:3], function(g, i) {
  paste0("#", i, ": ", igraph::vcount(g), " vertices")
})
indexed_report
```

**Performance note:**
`simap` functions do not benefit from the unique structure optimization.
Since each element has a different index,
the combination of `(structure, index)` is always unique, 
breaking the deduplication that makes other `smap` functions fast. 
Use `simap` only when you truly need position information.

## Performance

The main performance benefit of `smap` functions comes from automatic deduplication:

```{r}
# Create a large dataset with high redundancy
huge_struc <- rep(struc, 5000)  # 25,000 structures, only 5 unique

cat("Dataset size:", length(huge_struc), "structures\n")
cat("Unique structures:", length(attr(huge_struc, "structures")), "\n")
cat("Redundancy factor:", length(huge_struc) / length(attr(huge_struc, "structures")), "x\n")

library(tictoc)

# Optimized approach: smap only processes 5 unique structures
tic("smap_int (optimized)")
vertex_counts_optimized <- smap_int(huge_struc, igraph::vcount)
toc()

# Naive approach: extract all graphs and process each one
tic("Naive approach (all graphs)")
all_graphs <- get_structure_graphs(huge_struc)  # Extracts all 25,000 graphs
vertex_counts_naive <- purrr::map_int(all_graphs, igraph::vcount)
toc()

# Verify results are equivalent (though data types may differ)
all.equal(vertex_counts_optimized, vertex_counts_naive)
```

The higher the redundancy, the larger the performance gain.
In real glycoproteomics datasets with repeated structures, 
this optimization can provide about 10x speedups.

## Additional Patterns

### Working with Complex Functions

The function you pass to `smap` must accept an `igraph` object as its first argument. 
You can use purrr-style lambda notation:

```{r}
# Calculate clustering coefficient for each structure
clustering_coeffs <- smap_dbl(large_struc, ~ igraph::transitivity(.x, type = "global"))
summary(clustering_coeffs)
```

### Combining Multiple Metrics

```{r}
# Create a compact structure summary.
structure_analysis <- smap(large_struc, function(g) {
  list(
    vertices = igraph::vcount(g),
    edges = igraph::ecount(g),
    diameter = ifelse(igraph::is_connected(g), igraph::diameter(g), NA),
    clustering = igraph::transitivity(g, type = "global")
  )
})

# Convert to a more usable format
analysis_df <- do.call(rbind, lapply(structure_analysis, data.frame))
head(analysis_df)
```

### Memory-Efficient Filtering

```{r}
# Find only structures with exactly 5 vertices
has_five_vertices <- smap_lgl(large_struc, ~ igraph::vcount(.x) == 5)
five_vertex_structures <- large_struc[has_five_vertices]

cat("Found", sum(has_five_vertices), "structures with exactly 5 vertices\n")
```

## When to Use `smap` Functions

**Use `smap` functions when:**

- You need to apply `igraph`-based functions to glycan structures.
- You want better performance with datasets containing repeated structures.
- You are building custom glycan analysis pipelines.

**Use regular R functions when:**

- You are working with compositions.
- You are operating on string representations.

**Special note on `simap`:**

While `simap` functions are convenient for position-aware operations, 
they do not provide performance benefits over regular `imap` functions.
The inclusion of index information breaks the unique structure optimization, 
making each `(structure, index)` pair unique even when structures are identical.

## Example: Custom Motif Detection

Here's how you might build a custom glycan analysis pipeline using `smap` functions:

```{r}
# Custom motif detector
detect_branching <- function(g) {
  degrees <- igraph::degree(g)
  any(degrees >= 3)
}

# Apply to a large dataset using unique structure optimization.
has_branching <- smap_lgl(large_struc, detect_branching)
cat("Structures with branching:", sum(has_branching), "out of", length(large_struc), "\n")

# Use smap2 to check structures against complexity thresholds
complexity_thresholds <- rep(c(3, 4, 5, 2, 4), 1000)  # Thresholds for each structure
meets_threshold <- smap2_lgl(large_struc, complexity_thresholds, function(g, threshold) {
  igraph::vcount(g) >= threshold
})
cat("Structures meeting complexity threshold:", sum(meets_threshold), "out of", length(large_struc), "\n")
```

## Summary

The `smap` family provides structure-aware mapping functions for glycan structure vectors.
It lets you write custom graph-based analyses while preserving the unique structure optimization used by `glyrepr`.

Key takeaways:

- Unique structure optimization stores repeated structures efficiently.
- `smap` functions are `purrr`-like tools that understand glycan structure vectors.
- Performance gains are strongest when datasets contain repeated structures.
- Use `smap` for structures, and use regular R or `purrr` functions for other data types.

## Session Information

```{r}
sessionInfo()
```

---
title: "Glycan Graphs: The Network Behind Glycan Structures"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Glycan Graphs: The Network Behind Glycan Structures}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette is for users who want to understand how `glyrepr` stores glycan structures internally.
Some familiarity with graph theory and the `igraph` package will help.
If those concepts are new to you,
the [igraph documentation](https://r.igraph.org) is a useful companion reference.

## Glycans as Graphs

Glycans are naturally represented as directed graphs.
In `glyrepr`, a glycan structure is stored as an outward-directed tree,
where each vertex represents a monosaccharide and each edge represents a glycosidic linkage.

Behind the scenes,
each `glycan_structure()` object is backed by an `igraph` object.
Most workflows can stay at the `glycan_structure()` level,
but the graph representation is useful when you need custom structure analysis.

```{r setup}
library(glyrepr)
```

## What Is Stored in Memory?

A glycan can carry many kinds of information:
linear oriented C-atoms,
basetype,
substituents,
configuration,
anomeric center,
ring size,
linkage positions,
and more.

Some packages, such as Python's `glypy`, store a very detailed glycan model.
That comprehensive strategy is useful for specialized tasks such as MS/MS spectra simulation,
but it can be overkill for everyday omics research.

`glyrepr` takes a more compact approach:
if a feature can be derived from an IUPAC-condensed text representation,
`glyrepr` stores it.
Details such as configuration and ring size are not stored directly,
because they are often predictable for common carbohydrates and are not needed for many glycomics workflows.

For a closer look at IUPAC-condensed notation,
see the [IUPAC-condensed vignette](https://glycoverse.github.io/glyrepr/articles/iupac.html).

## Extracting the Graph

You cannot pass a `glycan_structure()` object directly to most `igraph` functions.
First, extract the underlying graph with `get_structure_graphs()`:

```{r}
glycan <- n_glycan_core()
graph <- get_structure_graphs(glycan)
graph
```

The printed graph contains several pieces of information.

**First line:** Directed Named ("DN") graph with 5 vertices (sugar units) and 4 edges (bonds). 

**Graph-level attributes:**

- `anomer`: the anomeric configuration of the reducing end.

**Vertex attributes:**

- `name`: a unique ID for each monosaccharide.
- `mono`: the monosaccharide type, such as "Hex" or "HexNAc".
- `sub`: chemical decorations attached to the monosaccharide.

**Edge attributes:**

- `linkage`: how the monosaccharides are connected, including bond positions and configurations.

**Connection pattern:** "1->2" means vertex 1 connects to vertex 2. 
`glyrepr` treats bonds as arrows pointing from the core toward the branches.
The direction is a modeling choice that makes traversal and structure operations easier.

You can also plot the graph with `igraph`:

```{r}
plot(graph)
```

## Graph Components

### Vertices

Each vertex represents a monosaccharide with three key properties:

**Names:**
These are auto-generated identifiers, usually simple integers,
but they could be anything as long as they're unique:

```{r}
igraph::V(graph)$name
```

**Monosaccharides:**
These are IUPAC-condensed names like "Hex", 
"HexNAc", 
"Glc", 
"GlcNAc". 

```{r}
igraph::V(graph)$mono
```

For the full list of available monosaccharides,
check [SNFG notation](https://www.ncbi.nlm.nih.gov/glycans/snfg.html) or run `available_monosaccharides()`.

**Substituents:**
Chemical decorations like "Me" (methyl), 
"Ac" (acetyl), 
"S" (sulfate), 
etc. 
Position matters.
"3Me" = methyl at position 3, 
"?S" = sulfate at unknown position:

```{r}
igraph::V(graph)$sub
```

Multiple decorations are comma-separated and sorted by position:

```{r}
glycan2 <- as_glycan_structure("Glc3Me6S(a1-")
graph2 <- get_structure_graphs(glycan2)
igraph::V(graph2)$sub
```

### Edges

Edges represent glycosidic bonds with a simple but powerful format:

```
<target anomeric config><target position> - <source position>
```

Here is an example where "Gal" has an "a" anomeric configuration,
linking from position 3 of "GalNAc" to position 1 of "Gal":

```{r}
glycan3 <- as_glycan_structure("Gal(a1-3)GalNAc(b1-")
graph3 <- get_structure_graphs(glycan3)
igraph::E(graph3)$linkage
```

`glyrepr` stores anomer information in edges rather than vertices.
This follows the way linkages are written in IUPAC-condensed notation,
for example "Neu5Ac with an a2-3 linkage".

### Graph-Level Attributes

**Anomer:** The anomeric configuration of the reducing end.

```{r}
graph$anomer
```

## Working with the Graph

### Using `igraph`

Once you understand the graph structure, 
you can use `igraph` functions for custom structure analysis.

**Example 1:** Count branched structures (sugars with multiple children):

```{r}
sum(igraph::degree(graph, mode = "out") > 1)
```

**Example 2:** Explore the structure with breadth-first search:

```{r}
bfs_result <- igraph::bfs(graph, root = 1, mode = "out")
bfs_result$order
```

### Using `smap` Functions

Working with multiple glycans? You could use `purrr`:

```{r}
library(purrr)

glycans <- c(n_glycan_core(), o_glycan_core_1(), o_glycan_core_2())
graphs <- get_structure_graphs(glycans)  # Extract graphs first
map_int(graphs, ~ igraph::vcount(.x))    # Then analyze
```

For glycan structure vectors,
`glyrepr`'s `smap` functions are usually a better fit:

```{r}
smap_int(glycans, ~ igraph::vcount(.x))  # Direct analysis, no intermediate step.
```

The main advantage of `smap` functions is how they handle duplicates.
Real datasets often contain many repeated structures,
and `smap` optimizes by processing unique structures once, 
then efficiently expanding results back to the original dimensions.

The [smap vignette](https://glycoverse.github.io/glyrepr/articles/smap.html) covers this workflow in more detail.

### Motif Analysis with `glymotif`

One important application is identifying biologically meaningful motifs, or functional substructures.
The `glymotif` package, 
built on this graph foundation, 
specializes in exactly this task.

See the [`glymotif` introduction](https://glycoverse.github.io/glymotif/articles/glymotif.html) for examples.

## Summary

In this vignette, you saw:

- how glycan structures map to directed graphs.
- what information is stored and what is deliberately omitted.
- how to extract and inspect the underlying graphs.
- how `igraph`, `smap`, and `glymotif` can build on this representation.

The graph representation might seem complex at first, 
but it's this solid foundation that enables all the sophisticated glycan analysis capabilities in the `glycoverse`. 
Most users will not need to manipulate the graph directly,
but understanding the model makes it easier to extend `glyrepr` when custom analysis is needed.

# format.glyrepr_structure includes names with tab separation

    Code
      format(glycans)
    Output
      [1] "A\tGal(b1-3)GalNAc(a1-                                "
      [2] "B\tMan(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"

# format.glyrepr_structure without names works correctly

    Code
      format(glycans)
    Output
      [1] "Gal(b1-3)GalNAc(a1-                                "
      [2] "Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-"

# truncation works in tibble

    Code
      print(tibble, width = 30)
    Output
      # A tibble: 3 x 2
        struc                      a
        <struct>               <dbl>
      1 Man(a1-3)[Man(a1-6)]M~     1
      2 Man(a1-3)[Man(a1-6)]M~     1
      3 Man(a1-3)[Man(a1-6)]M~     1

# get_structure_level works for a basic glycan vector with linkages

    Code
      res <- get_structure_level(glycans)
    Condition
      Warning:
      Generic glycan structures with linkage annotations are treated as "basic".
      i Linkage information is ignored when residues are generic.

# reduce_structure_level rejects higher level

    Code
      reduce_structure_level(glycan, to_level = "topological")
    Condition
      Error in `reduce_structure_level()`:
      ! Cannot reduce a structure to a higher resolution level.
      x Structure level of `x`: "basic".
      i Target level: "topological" (> "basic").
      i You can use `get_structure_level()` to check the structure level of `x`.

# truncation works in tibble for compositions

    Code
      print(tibble, width = 30)
    Output
      # A tibble: 3 x 2
        comp                a
        <comp>          <dbl>
      1 Hex(1)HexNAc(1)     1
      2 Hex(2)HexNAc(1)     1
      3 Hex(3)dHex(1)       1

# as_glycan_composition works for E and L

    Code
      comps <- as_glycan_composition(chars)
    Condition
      Warning:
      Simple composition codes "E" and/or "L" are parsed as "NeuAc".
      i Linkage-specific Neu5Ac information is discarded in glycan compositions.

# N-glycan core structure

    Code
      print(glycan, verbose = TRUE)
    Output
      <glycan_structure[1]>
      [1] Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-
      # Unique structures: 1

# N-glycan core structure without linkages

    Code
      print(glycan, verbose = TRUE)
    Output
      <glycan_structure[1]>
      [1] Man(??-?)[Man(??-?)]Man(??-?)GlcNAc(??-?)GlcNAc(??-
      # Unique structures: 1

# N-glycan core structure with generic monosaccharides

    Code
      print(glycan, verbose = TRUE)
    Output
      <glycan_structure[1]>
      [1] Hex(a1-3)[Hex(a1-6)]Hex(b1-4)HexNAc(b1-4)HexNAc(b1-
      # Unique structures: 1

# O-glycan core 1

    Code
      print(glycan, verbose = TRUE)
    Output
      <glycan_structure[1]>
      [1] Gal(b1-3)GalNAc(a1-
      # Unique structures: 1

# O-glycan core 2

    Code
      print(glycan, verbose = TRUE)
    Output
      <glycan_structure[1]>
      [1] Gal(b1-3)[GlcNAc(b1-6)]GalNAc(a1-
      # Unique structures: 1

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/structure.R
\name{format_glycan_structure_subset}
\alias{format_glycan_structure_subset}
\title{Format a Subset of Glycan Structures with Optional Colors}
\usage{
format_glycan_structure_subset(x, indices, colored = TRUE)
}
\arguments{
\item{x}{A glyrepr_structure object}

\item{indices}{Indices of structures to format}

\item{colored}{A logical value indicating whether to add colors}
}
\value{
A character vector of formatted structures for the specified indices
}
\description{
Format a Subset of Glycan Structures with Optional Colors
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smap.R
\name{smap_predicates}
\alias{smap_predicates}
\alias{ssome}
\alias{severy}
\alias{snone}
\title{Test Predicates on Glycan Structure Vectors}
\usage{
ssome(.x, .p, ...)

severy(.x, .p, ...)

snone(.x, .p, ...)
}
\arguments{
\item{.x}{A glycan structure vector (glyrepr_structure).}

\item{.p}{A predicate function that takes an igraph object and returns a logical value.
Can be a function, purrr-style lambda (\code{~ .x$attr}), or a character string naming a function.}

\item{...}{Additional arguments passed to \code{.p}.}
}
\value{
A single logical value.
}
\description{
These functions test predicates on unique structures in a glycan structure vector,
taking advantage of hash-based deduplication to avoid redundant computation.
Similar to purrr predicate functions, but optimized for glycan structure vectors.
}
\details{
These functions only evaluate \code{.p} once for each unique structure, making them
much more efficient than applying \code{.p} to each element individually when there
are duplicate structures.

\strong{Return Values:}
\itemize{
\item \code{ssome()}: Returns \code{TRUE} if at least one unique structure satisfies the predicate
\item \code{severy()}: Returns \code{TRUE} if all unique structures satisfy the predicate
\item \code{snone()}: Returns \code{TRUE} if no unique structures satisfy the predicate
}
}
\examples{
# Create a structure vector with duplicates
core1 <- o_glycan_core_1()
core2 <- n_glycan_core()
structures <- c(core1, core2, core1)  # core1 appears twice

# Test if some structures have more than 5 vertices
ssome(structures, function(g) igraph::vcount(g) > 5)

# Test if all structures have at least 3 vertices
severy(structures, function(g) igraph::vcount(g) >= 3)

# Test if no structures have more than 20 vertices
snone(structures, function(g) igraph::vcount(g) > 20)

# Use purrr-style lambda functions
ssome(structures, ~ igraph::vcount(.x) > 5)
severy(structures, ~ igraph::vcount(.x) >= 3)
snone(structures, ~ igraph::vcount(.x) > 20)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glyrepr-package.R
\name{glyrepr-vctrs}
\alias{glyrepr-vctrs}
\title{Internal vctrs methods}
\description{
Internal vctrs methods
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glyrepr-package.R
\docType{package}
\name{glyrepr-package}
\alias{glyrepr}
\alias{glyrepr-package}
\title{glyrepr: Representation for Glycan Compositions and Structures}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Computational representations of glycan compositions and structures, including details such as linkages, anomers, and substituents. Supports varying levels of monosaccharide specificity (e.g., "Hex" or "Gal") and ambiguous linkages. Provides robust parsing and generation of IUPAC-condensed structure strings. Optimized for vectorized operations on glycan structures, with efficient handling of duplications. As the cornerstone of the glycoverse ecosystem, this package delivers the foundational data structures that power glycomics and glycoproteomics analysis workflows.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://glycoverse.github.io/glyrepr/}
  \item \url{https://github.com/glycoverse/glyrepr}
  \item Report bugs at \url{https://github.com/glycoverse/glyrepr/issues}
}

}
\author{
\strong{Maintainer}: Bin Fu \email{23110220018@m.fudan.edu.cn} (\href{https://orcid.org/0000-0001-8567-2997}{ORCID}) [copyright holder]

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color-utils.R
\name{add_colors}
\alias{add_colors}
\title{Add Colors to Monosaccharides}
\usage{
add_colors(monos, colored = TRUE)
}
\arguments{
\item{monos}{A character vector of monosaccharide names}

\item{colored}{A logical value indicating whether to add colors}
}
\value{
A character vector with ANSI color codes
}
\description{
Add Colors to Monosaccharides
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color-utils.R
\name{colorize_iupac_string}
\alias{colorize_iupac_string}
\title{Apply Colors to IUPAC String (Monosaccharides + Gray Linkages)}
\usage{
colorize_iupac_string(iupac_text, mono_names)
}
\arguments{
\item{iupac_text}{Character string of IUPAC notation}

\item{mono_names}{Character vector of monosaccharide names to color}
}
\value{
Character string with colored monosaccharides and gray linkages
}
\description{
Apply Colors to IUPAC String (Monosaccharides + Gray Linkages)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color-utils.R
\name{replace_monos_with_colored}
\alias{replace_monos_with_colored}
\title{Replace Monosaccharides in String with Colored Versions}
\usage{
replace_monos_with_colored(text, mono_names)
}
\arguments{
\item{text}{Character string containing monosaccharide names}

\item{mono_names}{Character vector of monosaccharide names to replace}
}
\value{
Character string with monosaccharides replaced by colored versions
}
\description{
Replace Monosaccharides in String with Colored Versions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/anomer.R
\name{get_anomer}
\alias{get_anomer}
\title{Get the Anomeric information}
\usage{
get_anomer(x)
}
\arguments{
\item{x}{A glycan structure vector (glyrepr_structure).}
}
\value{
a character vector of the anomeric information.
}
\description{
Get the Anomeric information
}
\examples{
x <- n_glycan_core()
get_anomer(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example-glycans.R
\name{n_glycan_core}
\alias{n_glycan_core}
\alias{o_glycan_core_1}
\alias{o_glycan_core_2}
\title{Example Glycan Structures}
\usage{
n_glycan_core(linkage = TRUE, mono_type = "concrete")

o_glycan_core_1(linkage = TRUE, mono_type = "concrete")

o_glycan_core_2(linkage = TRUE, mono_type = "concrete")
}
\arguments{
\item{linkage}{A logical indicating whether to include linkages (e.g. "b1-4").
Default is \code{TRUE}.}

\item{mono_type}{A character string specifying the type of monosaccharides.
Can be "generic" (Hex, HexNAc, dHex, NeuAc, etc.)
or "concrete" (Man, Gal, HexNAc, Fuc, etc.). Default is "concrete".}
}
\value{
A glycan structure (igraph) object.
}
\description{
Create example glycan structures for testing and demonstration.
Includes \strong{N-glycan core} and \strong{O-glycan core 1} and \strong{core 2}.
}
\section{N-Glycan Core}{
\strong{N-Glycans} are branched oligosaccharides that are bound, most commonly,
via GlcNAc to an Asn residue of the protein backbone.
A common motif of all N-glycans is the \strong{chitobiose core},
composed of three mannose and two GlcNAc moieties,
which is commonly attached to the protein backbone via GlcNAc.
The mannose residue is branched and connected via a1,3- and a1,6-glycosidic
linkages to the two other mannose building blocks.

\if{html}{\out{<div class="sourceCode">}}\preformatted{    Man
  a1-6 \\   b1-4      b1-4      b1-
        Man -- GlcNAc -- GlcNAc -
  a1-3 /
    Man
}\if{html}{\out{</div>}}
}

\section{O-Glycan Core}{
\strong{O-Glycans} are highly abundant in extracellular proteins.
Generally, O-glycans are extended following four major core structures:
\strong{core 1}, \strong{core 2}, core 3, and core 4.
The first two are by far the most common core structures in O-glycosylation
and are found throughout the body.

\strong{core 1}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{          a1-
    GalNAc -
   / b1-3
Gal
}\if{html}{\out{</div>}}

\strong{core 2}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{GlcNAc
      \\ b1-6 a1-
       GalNAc -
      / b1-3
   Gal
}\if{html}{\out{</div>}}
}

\examples{
print(n_glycan_core(), verbose = TRUE)
print(o_glycan_core_1(), verbose = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linkage.R
\name{remove_linkages}
\alias{remove_linkages}
\title{Remove All Linkages from a Glycan}
\usage{
remove_linkages(glycan)
}
\arguments{
\item{glycan}{A glyrepr_structure vector.}
}
\value{
A glyrepr_structure vector with all linkages removed.
}
\description{
This function replaces all linkages in a glycan structure with "??-?",
as well as the reducing end anomer with "??-".
}
\examples{
glycan <- o_glycan_core_1(linkage = TRUE)
glycan
remove_linkages(glycan)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/structure-to-iupac.R
\name{structure_to_iupac}
\alias{structure_to_iupac}
\title{Convert Glycan Structure to IUPAC-like Sequence}
\usage{
structure_to_iupac(glycan)
}
\arguments{
\item{glycan}{A glyrepr_structure vector.}
}
\value{
A character vector representing the IUPAC sequences.
}
\description{
Convert a glycan structure to a sequence representation in the form of
mono(linkage)mono, with branches represented by square brackets [].
The backbone is chosen as the longest path, and for branches, linkages are
ordered lexicographically with smaller linkages on the backbone.
}
\section{Sequence Format}{
The sequence follows the format mono(linkage)mono, where:
\itemize{
\item mono: monosaccharide name with optional substituents (e.g., Glc, GlcNAc, Glc3Me)
\item linkage: glycosidic linkage (e.g., b1-4, a1-3)
\item Branches are enclosed in square brackets []
\item Substituents are appended directly to monosaccharide names (e.g., Glc3Me for Glc with 3Me substituent)
}
}

\section{Backbone Selection}{
The backbone is selected as the longest path in the tree. For branches,
the same rule applies recursively.
}

\section{Linkage Comparison}{
Linkages are compared lexicographically:
\enumerate{
\item First by anomeric configuration: ? > b > a
\item Then by first position: ? > numbers (numerically)
\item Finally by second position: ? > numbers (numerically)
}

Smaller linkages are placed on the backbone, larger ones in branches.
}

\examples{
# Simple linear structure
structure_to_iupac(o_glycan_core_1())

# Branched structure
structure_to_iupac(n_glycan_core())

# Structure with substituents
graph <- igraph::make_graph(~ 1-+2)
igraph::V(graph)$mono <- c("Glc", "GlcNAc")
igraph::V(graph)$sub <- c("3Me", "6Ac")
igraph::E(graph)$linkage <- "b1-4"
graph$anomer <- "a1"
glycan <- glycan_structure(graph)
structure_to_iupac(glycan)  # Returns "GlcNAc6Ac(b1-4)Glc3Me(a1-"

# Vectorized structures
structs <- c(o_glycan_core_1(), n_glycan_core())
structure_to_iupac(structs)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/substituent.R
\name{remove_substituents}
\alias{remove_substituents}
\title{Remove All Substituents from a Glycan}
\usage{
remove_substituents(glycan)
}
\arguments{
\item{glycan}{A glyrepr_structure vector.}
}
\value{
A glyrepr_structure vector with all substituents removed.
}
\description{
This function replaces all substituents in a glycan structure with empty strings.
}
\examples{
(glycan <- o_glycan_core_1())
remove_substituents(glycan)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/structure-level.R
\name{get_structure_level}
\alias{get_structure_level}
\title{Get the Structure Resolution Levels}
\usage{
get_structure_level(x)
}
\arguments{
\item{x}{A \code{\link[=glycan_structure]{glycan_structure()}} vector.}
}
\value{
A character scalar containing the structure level for \code{x}.
If \code{x} is empty or all structures in \code{x} are NA, returns NA_character_.
}
\description{
Glycan structures can have four possible levels of resolution:
\itemize{
\item "intact": All monosaccharides are concrete (e.g. "Man", "GlcNAc"),
and no linkage or anomer contains "?".
\item "partial": All monosaccharides are concrete (e.g. "Man", "GlcNAc"),
at least one linkage or anomer contains "?",
and at least one linkage or anomer has a non-"?" annotation.
\item "topological": All monosaccharides are concrete (e.g. "Man", "GlcNAc"),
and all linkages and anomers are completely unknown ("??-?"/"??").
\item "basic": All monosaccharides are generic (e.g. "Hex", "HexNAc").
}

Note that in theory you can have a glycan with generic monosaccharides with all linkages determined.
For example, "Hex(b1-3)HexNAc(a1-" is a valid glycan structure.
But in reality, this is almost impossible,
because linkage information is far more difficult to acquire than monosaccharide information.
This kind of glycan structure is also assigned to "basic" level.
}
\examples{
glycan <- as_glycan_structure("Gal(b1-3)GalNAc(a1-")
get_structure_level(glycan)

}
\seealso{
\code{\link[=has_linkages]{has_linkages()}}, \code{\link[=get_mono_type]{get_mono_type()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/structure.R
\name{get_structure_graphs}
\alias{get_structure_graphs}
\title{Access Individual Glycan Structures}
\usage{
get_structure_graphs(x, return_list = NULL)
}
\arguments{
\item{x}{A glycan structure vector.}

\item{return_list}{If \code{TRUE}, always returns a list.
If \code{FALSE} and \code{x} has a length of 1, return the igraph object directly.
If not provided (default), \code{FALSE} when \code{x} has a length of 1 and \code{TRUE} otherwise.}
}
\value{
A list of igraph objects or an igraph object directly (see \code{return_list} parameter).
}
\description{
Extract individual glycan structure graphs from a glycan structure vector.
}
\examples{
structures <- c(o_glycan_core_1(), n_glycan_core())
get_structure_graphs(structures)
get_structure_graphs(structures)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smap.R
\name{spmap}
\alias{spmap}
\alias{spmap_vec}
\alias{spmap_lgl}
\alias{spmap_int}
\alias{spmap_dbl}
\alias{spmap_chr}
\alias{spmap_structure}
\title{Map Functions Over Glycan Structure Vectors and Multiple Arguments}
\usage{
spmap(.l, .f, ..., .parallel = FALSE)

spmap_vec(.l, .f, ..., .ptype = NULL, .parallel = FALSE)

spmap_lgl(.l, .f, ..., .parallel = FALSE)

spmap_int(.l, .f, ..., .parallel = FALSE)

spmap_dbl(.l, .f, ..., .parallel = FALSE)

spmap_chr(.l, .f, ..., .parallel = FALSE)

spmap_structure(.l, .f, ..., .parallel = FALSE)
}
\arguments{
\item{.l}{A list where the first element is a glycan structure vector (glyrepr_structure)
and the remaining elements are vectors of the same length or length 1 (will be recycled).}

\item{.f}{A function that takes an igraph object (from first element of \code{.l}) and
values from other elements, returning a result.
Can be a function, purrr-style lambda (\code{~ .x + .y + .z}), or a character string naming a function.}

\item{...}{Additional arguments passed to \code{.f}.}

\item{.parallel}{Logical; whether to use parallel processing. If \code{FALSE} (default),
parallel processing is disabled. Set to \code{TRUE} to enable parallel processing.
See examples in \code{\link{smap}} for how to set up and use parallel processing.}

\item{.ptype}{A prototype for the return type (for \code{spmap_vec}).}
}
\value{
\itemize{
\item \code{spmap()}: A list
\item \code{spmap_vec()}: An atomic vector of type specified by \code{.ptype}
\item \code{spmap_lgl/int/dbl/chr()}: Atomic vectors of the corresponding type
\item \code{spmap_structure()}: A new glyrepr_structure object
}
}
\description{
These functions apply a function to each unique structure in a glycan structure vector
along with corresponding elements from multiple other vectors,
taking advantage of hash-based deduplication to avoid redundant computation.
Similar to purrr pmap functions, but optimized for glycan structure vectors.
}
\details{
These functions only compute \code{.f} once for each unique combination of structure and corresponding
values from other vectors, then map the results back to the original vector positions.

\strong{NA Handling:}
NA elements in the first argument (glycan structure vector) are preserved in the output.

\strong{Time Complexity Performance:}

Performance scales with unique combinations of all arguments rather than total vector length.
When argument vectors are highly redundant, performance approaches O(unique_structures).
Scaling factor shows time increase when vector size increases 20x.

\strong{Return Types:}
\itemize{
\item \code{spmap()}: Returns a list with the same length as the input vectors
\item \code{spmap_vec()}: Returns an atomic vector with the same length as the input vectors
\item \code{spmap_lgl()}: Returns a logical vector
\item \code{spmap_int()}: Returns an integer vector
\item \code{spmap_dbl()}: Returns a double vector
\item \code{spmap_chr()}: Returns a character vector
\item \code{spmap_structure()}: Returns a new glycan structure vector (\code{.f} must return igraph objects)
}
}
\examples{
# Create structure vectors with duplicates
core1 <- o_glycan_core_1()
core2 <- n_glycan_core()
structures <- c(core1, core2, core1)  # core1 appears twice
weights <- c(1.0, 2.0, 1.0)  # corresponding weights
factors <- c(2, 3, 2)  # corresponding factors

# Map a function that uses structure, weight, and factor
spmap_dbl(list(structures, weights, factors),
          function(g, w, f) igraph::vcount(g) * w * f)

# Use purrr-style lambda functions
spmap_dbl(list(structures, weights, factors), ~ igraph::vcount(..1) * ..2 * ..3)

# Test with recycling
spmap_dbl(list(structures, 2.0, 3), ~ igraph::vcount(..1) * ..2 * ..3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/substituent.R
\name{available_substituents}
\alias{available_substituents}
\title{Available Substituents}
\usage{
available_substituents()
}
\value{
A character vector.
}
\description{
Get the available substituents for monosaccharides.
}
\examples{
available_substituents()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smap.R
\name{simap}
\alias{simap}
\alias{simap_vec}
\alias{simap_lgl}
\alias{simap_int}
\alias{simap_dbl}
\alias{simap_chr}
\alias{simap_structure}
\title{Map Functions Over Glycan Structure Vectors with Indices}
\usage{
simap(.x, .f, ...)

simap_vec(.x, .f, ..., .ptype = NULL)

simap_lgl(.x, .f, ...)

simap_int(.x, .f, ...)

simap_dbl(.x, .f, ...)

simap_chr(.x, .f, ...)

simap_structure(.x, .f, ...)
}
\arguments{
\item{.x}{A glycan structure vector (glyrepr_structure).}

\item{.f}{A function that takes an igraph object (from \code{.x}) and an index/name,
returning a result.
Can be a function, purrr-style lambda (\code{~ paste(.x, .y)}), or a character string naming a function.}

\item{...}{Additional arguments passed to \code{.f}.}

\item{.ptype}{A prototype for the return type (for \code{simap_vec}).}
}
\value{
\itemize{
\item \code{simap()}: A list
\item \code{simap_vec()}: An atomic vector of type specified by \code{.ptype}
\item \code{simap_lgl()}: Returns a logical vector
\item \code{simap_int()}: Returns an integer vector
\item \code{simap_dbl()}: Returns a double vector
\item \code{simap_chr()}: Returns a character vector
\item \code{simap_structure()}: A new glyrepr_structure object
}
}
\description{
These functions apply a function to each unique structure in a glycan structure vector
along with their corresponding indices,
taking advantage of hash-based deduplication to avoid redundant computation.
Similar to purrr imap functions, but optimized for glycan structure vectors.
}
\details{
These functions only compute \code{.f} once for each unique combination of structure and corresponding
index/name, then map the results back to the original vector positions. This is much more efficient
than applying \code{.f} to each element individually when there are duplicate structures.

\strong{IMPORTANT PERFORMANCE NOTE:}
Due to the inclusion of position indices, \code{simap} functions have \strong{O(total_structures)}
time complexity because each position creates a unique combination, even with identical structures.

\strong{Alternative:} Consider \code{smap()} functions if position information is not required.

The index passed to \code{.f} is the position in the original vector (1-based).
If the vector has names, the names are passed instead of indices.

\strong{Return Types:}
\itemize{
\item \code{simap()}: Returns a list with the same length as \code{.x}
\item \code{simap_vec()}: Returns an atomic vector with the same length as \code{.x}
\item \code{simap_lgl()}: Returns a logical vector
\item \code{simap_int()}: Returns an integer vector
\item \code{simap_dbl()}: Returns a double vector
\item \code{simap_chr()}: Returns a character vector
\item \code{simap_structure()}: Returns a new glycan structure vector (\code{.f} must return igraph objects)
}
}
\examples{
# Create structure vectors with duplicates
core1 <- o_glycan_core_1()
core2 <- n_glycan_core()
structures <- c(core1, core2, core1)  # core1 appears twice

# Map a function that uses both structure and index
simap_chr(structures, function(g, i) paste0("Structure_", i, "_vcount_", igraph::vcount(g)))

# Use purrr-style lambda functions
simap_chr(structures, ~ paste0("Pos", .y, "_vertices", igraph::vcount(.x)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/structure.R
\name{as_glycan_structure}
\alias{as_glycan_structure}
\title{Convert to Glycan Structure Vector}
\usage{
as_glycan_structure(x)
}
\arguments{
\item{x}{An object to convert to a glycan structure vector.
Can be an igraph object, a list of igraph objects,
a character vector of IUPAC-condensed strings,
or an existing glyrepr_structure object.}
}
\value{
A glyrepr_structure object.
}
\description{
Convert an object to a glycan structure vector.
}
\examples{
library(igraph)

# Convert a single igraph
graph <- make_graph(~ 1-+2)
V(graph)$mono <- c("GlcNAc", "GlcNAc")
V(graph)$sub <- ""
E(graph)$linkage <- "b1-4"
graph$anomer <- "a1"
as_glycan_structure(graph)

# Convert a list of igraphs
o_glycan_vec <- o_glycan_core_1()
o_glycan_graph <- get_structure_graphs(o_glycan_vec)
as_glycan_structure(list(graph, o_glycan_graph))

# Convert a character vector of IUPAC-condensed strings
as_glycan_structure(c("GlcNAc(b1-4)GlcNAc(b1-", "Man(a1-2)GlcNAc(b1-"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count-mono.R
\name{count_mono}
\alias{count_mono}
\alias{count_mono.glyrepr_composition}
\alias{count_mono.glyrepr_structure}
\title{Get the Number of Monosaccharides}
\usage{
count_mono(x, mono = NULL, include_subs = FALSE)

\method{count_mono}{glyrepr_composition}(x, mono = NULL, include_subs = FALSE)

\method{count_mono}{glyrepr_structure}(x, mono = NULL, include_subs = FALSE)
}
\arguments{
\item{x}{A glycan composition (\code{glyrepr_composition}) or
a glycan structure (\code{glyrepr_structure}) vector.}

\item{mono}{The monosaccharide or substituent to count. A character scalar.
If \code{NULL} (default), return the total number of monosaccharides.}

\item{include_subs}{Whether to include substituents when \code{mono} is \code{NULL}.
Default is \code{FALSE}.}
}
\value{
A numeric vector of the same length as \code{x}.
}
\description{
When \code{mono} is:
\itemize{
\item \code{NULL} (default), returns the total number of monosaccharides and substituents.
\item A string, returns the number of the specified monosaccharide or substituent.
}
}
\details{
When \code{mono} is "generic" (e.g. "Hex", "HexNAc"),
it counts all "concrete" monosaccharides that match.
For example, "Hex" will count all Glc, Man, Gal, etc.
When \code{mono} is "concrete" (e.g. "Gal", "GalNAc"),
NA is returned when the composition is "generic".
}
\examples{
comp <- glycan_composition(c(Gal = 1, Man = 1, GalNAc = 1))
count_mono(comp, "Hex")
count_mono(comp, "Gal")

struct <- as_glycan_structure("Gal(b1-3)GlcNAc(b1-4)Glc(a1-")
count_mono(struct, "Gal")

# Total number of monosaccharides
count_mono(comp)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/composition.R
\name{as_glycan_composition}
\alias{as_glycan_composition}
\title{Convert to Glycan Composition}
\usage{
as_glycan_composition(x)
}
\arguments{
\item{x}{An object to convert to a glycan composition.
Supported inputs include:
\itemize{
\item Named integer vectors or lists of named integer vectors
\item Character vectors with composition strings (e.g., "Hex(5)HexNAc(2)")
\item \code{glyrepr_structure} objects (counts both monosaccharides and substituents)
\item Existing \code{glyrepr_composition} objects (returned as-is)
}}
}
\value{
A \code{glyrepr_composition} object.
}
\description{
Converts an object to a glycan composition using vctrs casting framework.
This function provides a convenient way to convert various input types
to \code{\link[=glycan_composition]{glycan_composition()}}.
}
\details{
This function uses the vctrs casting framework for type conversion.
When converting from glycan structures, both monosaccharides and substituents
are counted. Substituents are extracted from the \code{sub} attribute of each
vertex in the structure. For example, a vertex with \code{sub = "3Me"}
contributes one "Me" substituent to the composition.

Simple composition strings use one-letter residue codes: "H" for "Hex",
"N" for "HexNAc", "F" for "dHex", "S"/"A" for "NeuAc", and "G" for
"NeuGc". "E" and "L" are also accepted as linkage-specific Neu5Ac codes;
they are converted to "NeuAc" with a warning because composition objects do
not preserve linkage information.
}
\examples{
# From a single named vector
as_glycan_composition(c(Hex = 5, HexNAc = 2))

# From a list of named vectors
as_glycan_composition(list(c(Hex = 5, HexNAc = 2), c(Hex = 3, HexNAc = 1)))

# From a character vector of Byonic composition strings
as_glycan_composition(c("Hex(5)HexNAc(2)", "Hex(3)HexNAc(1)"))

# From a character vector of simple composition strings
as_glycan_composition(c("H5N2", "H5N4S1F1"))

# From an existing composition (returns as-is)
comp <- glycan_composition(c(Hex = 5, HexNAc = 2))
as_glycan_composition(comp)

# From a glycan structure vector
strucs <- c(n_glycan_core(), o_glycan_core_1())
as_glycan_composition(strucs)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linkage.R
\name{has_linkages}
\alias{has_linkages}
\title{Determine if a Glycan Structure has Linkages}
\usage{
has_linkages(glycan, strict = FALSE)
}
\arguments{
\item{glycan}{A \code{\link[=glycan_structure]{glycan_structure()}} vector.}

\item{strict}{A logical value.
\itemize{
\item If \code{FALSE} (default), a glycan is considered to have linkages if any
linkage is partially known (not "??-?").
\item If \code{TRUE}, a glycan is considered to have linkages only if all linkages
are fully determined (no "?" in the linkage).
}}
}
\value{
A logical vector indicating if each glycan structure has linkages.
}
\description{
Unknown linkages in a glycan structure are represented by "??-?".
Also, a linkage can be partially known (e.g. "a?-?").
This function checks if a glycan structure has linkages,
in a strict or lenient way.
}
\examples{
glycan <- o_glycan_core_1(linkage = TRUE)
has_linkages(glycan)
print(glycan)

glycan <- remove_linkages(glycan)
has_linkages(glycan)
print(glycan)

glycan <- as_glycan_structure("Gal(b1-?)GalNAc(a1-")
has_linkages(glycan)
has_linkages(glycan, strict = TRUE)

}
\seealso{
\code{\link[=remove_linkages]{remove_linkages()}}, \code{\link[=possible_linkages]{possible_linkages()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/monosaccharide.R
\name{is_known_monosaccharide}
\alias{is_known_monosaccharide}
\title{Check if a Monosaccharide is Known}
\usage{
is_known_monosaccharide(mono)
}
\arguments{
\item{mono}{A character vector of monosaccharide names.}
}
\value{
A logical vector.
}
\description{
This function checks if a vector of monosaccharide names are known.
}
\examples{
is_known_monosaccharide(c("Gal", "Hex"))
is_known_monosaccharide(c("X", "Hx", "Nac"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/structure.R
\name{glycan_structure}
\alias{glycan_structure}
\alias{is_glycan_structure}
\title{Create a Glycan Structure Vector}
\usage{
glycan_structure(...)

is_glycan_structure(x)
}
\arguments{
\item{...}{igraph graph objects to be converted to glycan structures, or existing
glycan structure vectors. Supports mixed input of multiple objects.}

\item{x}{An object to check or convert.}
}
\value{
A \code{glyrepr_structure} class glycan structure vector object.
}
\description{
\code{glycan_structure()} creates an efficient glycan structure vector for storing and
processing glycan molecular structures. The function employs hash-based deduplication
mechanisms, making it suitable for glycoproteomics, glycomics analysis, and glycan
structure comparison studies.
}
\section{Data Structure Overview}{
A glycan structure vector is a vctrs record with an additional S3 class
\code{glyrepr_structure}. Therefore, \code{sloop::s3_class()} returns the class hierarchy
\code{c("glyrepr_structure", "vctrs_rcrd")}.

Each glycan structure must satisfy the following constraints:
\subsection{Graph Structure Requirements}{
\itemize{
\item Must be a directed graph with an outward tree structure (reducing end as root)
\item Must have a graph attribute \code{anomer} in the format "a1" or "b1"
\itemize{
\item Unknown parts can be represented with "?", e.g., "?1", "a?", "??"
}
}
}

\subsection{Node Attributes}{
\itemize{
\item \code{mono}: Monosaccharide names, must be known monosaccharide types
\itemize{
\item Generic names: Hex, HexNAc, dHex, NeuAc, etc.
\item Concrete names: Glc, Gal, Man, GlcNAc, etc.
\item Cannot mix generic and concrete names
\item NA values are not allowed
}
\item \code{sub}: Substituent information
\itemize{
\item Single substituent format: "xY" (x = position, Y = substituent name),
e.g., "2Ac", "3S"
\item Multiple substituents separated by commas and ordered by position,
e.g., "3Me,4Ac", "2S,6P"
\item No substituents represented by empty string ""
}
}
}

\subsection{Edge Attributes}{
\itemize{
\item \code{linkage}: Glycosidic linkage information in format "a/bX-Y"
\itemize{
\item Standard format: e.g., "b1-4", "a2-3"
\item Unknown positions allowed: "a1-?", "b?-3", "??-?"
\item Partially unknown positions: "a1-3/6", "a1-3/6/9"
\item NA values are not allowed
}
}
}
}

\section{Node and Edge Order}{
The indices of vertices and linkages in a glycan correspond directly to their
order in the IUPAC-condensed string, which is printed when you print a
\code{\link[=glycan_structure]{glycan_structure()}}.
For example, for the glycan \verb{Man(a1-3)[Man(a1-6)]Man(b1-4)GlcNAc(b1-4)GlcNAc(b1-},
the vertices are "Man", "Man", "Man", "GlcNAc", "GlcNAc",
and the linkages are "a1-3", "a1-6", "b1-4", "b1-4".
}

\section{NA Support}{
Glycan structure vectors support NA values for representing missing or
unknown structures:
\itemize{
\item Create with \code{glycan_structure(NA)} or \code{glycan_structure(NULL)}
\item Combine with valid structures: \code{c(struct1, NA, struct2)}
\item Convert from character: \code{as_glycan_structure(c("Glc(a1-", NA))}
\item \code{smap} functions skip NA elements gracefully
\item \code{is.na()} returns \code{TRUE} for NA elements
}
}

\section{Naming Support}{
Glycan structure vectors can have names, which are preserved during operations.
This is particularly useful when working with the \code{glymotif} package.
}

\section{As characters}{
One side-effect of the current implementation is that you can treat a glycan structure vector
as a pure character vector of IUPAC-condensed strings.
In fact, \code{is.character()} returns \code{TRUE} for a glycan structure vector,
and all \code{stringr} functions work directly on the vector.

However, we still recommend using \code{as.character()} to explicitly convert to character when needed,
to avoid confusion and ensure that the intended behavior is clear.
}

\examples{
library(igraph)

# Example 1: Create a simple glycan structure GlcNAc(b1-4)GlcNAc
graph <- make_graph(~ 1-+2)  # Create graph with two monosaccharides
V(graph)$mono <- c("GlcNAc", "GlcNAc")  # Set monosaccharide types
V(graph)$sub <- ""  # No substituents
E(graph)$linkage <- "b1-4"  # b1-4 glycosidic linkage
graph$anomer <- "a1"  # a anomeric carbon

# Create glycan structure vector
simple_struct <- glycan_structure(graph)
print(simple_struct)

# Example 2: Use predefined glycan core structures
n_core <- n_glycan_core()  # N-glycan core structure
o_core1 <- o_glycan_core_1()  # O-glycan Core 1 structure

# Example 3: Create complex structure with substituents
complex_graph <- make_graph(~ 1-+2-+3)
V(complex_graph)$mono <- c("GlcNAc", "Gal", "Neu5Ac")
V(complex_graph)$sub <- c("", "", "")  # Add substituents as needed
E(complex_graph)$linkage <- c("b1-4", "a2-3")
complex_graph$anomer <- "b1"

complex_struct <- glycan_structure(complex_graph)
print(complex_struct)

# Example 4: Check if object is a glycan structure
is_glycan_structure(simple_struct)  # TRUE
is_glycan_structure(graph)          # FALSE

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/monosaccharide.R
\name{available_monosaccharides}
\alias{available_monosaccharides}
\title{Get Available Monosaacharides}
\usage{
available_monosaccharides(mono_type = "all")
}
\arguments{
\item{mono_type}{A character string specifying the type of monosaccharides.
Can be "all", "generic", or "concrete". Default is "all".}
}
\value{
A character vector of monosaccharide names.
}
\description{
This function returns a character vector of monosaccharide names of
the given type. See \code{\link[=get_mono_type]{get_mono_type()}} for monosaacharide types.
}
\examples{
available_monosaccharides()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mono-type.R
\name{convert_to_generic}
\alias{convert_to_generic}
\alias{convert_to_generic.character}
\alias{convert_to_generic.glyrepr_structure}
\alias{convert_to_generic.glyrepr_composition}
\title{Convert Monosaccharides to Generic Type}
\usage{
convert_to_generic(x)

\method{convert_to_generic}{character}(x)

\method{convert_to_generic}{glyrepr_structure}(x)

\method{convert_to_generic}{glyrepr_composition}(x)
}
\arguments{
\item{x}{Either of these objects:
\itemize{
\item A character of monosaccharide;
\item A glycan composition vector ("glyrepr_composition" object);
\item A glycan structure vector ("glyrepr_structure" object).
}}
}
\value{
A new object of the same class as \code{x}
with monosaccharides converted to generic type.
}
\description{
This function converts monosaccharide types of monosaccharide characters,
glycan compositions, or glycan structures from concrete to generic type.
This is a simplified version that only supports conversion from "concrete"
to "generic" monosaccharides.
}
\section{Two types of monosaccharides}{
There are two types of monosaccharides:
\itemize{
\item concrete: e.g. "Gal", "GlcNAc", "Glc", "Fuc", etc.
\item generic: e.g. "Hex", "HexNAc", "HexA", "HexN", etc.
}

For the full list of monosaccharides, use \code{\link[=available_monosaccharides]{available_monosaccharides()}}.
}

\examples{
# Convert character vectors
convert_to_generic(c("Gal", "GlcNAc"))

# Convert glycan compositions
comps <- glycan_composition(
  c(Gal = 5, GlcNAc = 2),
  c(Glc = 5, GalNAc = 4, Fuc = 1)
)
convert_to_generic(comps)

# Convert glycan structures
strucs <- c(n_glycan_core(), o_glycan_core_1())
convert_to_generic(strucs)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate-helper.R
\name{valid_linkages}
\alias{valid_linkages}
\title{Check if Linkages are Valid}
\usage{
valid_linkages(linkages)
}
\arguments{
\item{linkages}{A character vector of linkages.}
}
\value{
A logical vector.
}
\description{
Valid linkages are in the form of "a1-2", "b1-4", "a?-1", etc.
Specifically, the pattern is \code{xy-z}:
\itemize{
\item \code{x}: the anomer, either "a", "b", or "?".
\item \code{y}: the first position, either "1", "2" or "?".
\item \code{z}: the second position, either a 1-9 digit or "?".
Can also be multiple positions separated by "/", e.g. "1/2/3".
"?" could not be used with "/".
}
}
\examples{
# Valid linkages
valid_linkages(c("a1-2", "?1-4", "a?-1", "b?-?", "??-?", "a1/2-3"))

# Invalid linkages
valid_linkages(c("a1-2/?", "1-4", "a/b1-2", "c1-2", "a9-1"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color-utils.R
\name{add_gray_linkages}
\alias{add_gray_linkages}
\title{Add Gray Color to Linkages in IUPAC String}
\usage{
add_gray_linkages(iupac_text)
}
\arguments{
\item{iupac_text}{Character string of IUPAC notation}
}
\value{
Character string with linkages colored gray
}
\description{
Add Gray Color to Linkages in IUPAC String
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/structure-level.R
\name{reduce_structure_level}
\alias{reduce_structure_level}
\title{Reduce a Glycan Structure to a Lower Resolution Level}
\usage{
reduce_structure_level(x, to_level)
}
\arguments{
\item{x}{A \code{\link[=glycan_structure]{glycan_structure()}} vector.}

\item{to_level}{The resolution level to reduce to. Can be "basic" or "topological".
Must be a lower resolution level than \code{x}
("intact" > "partial" > "topological" > "basic").
If \code{to_level} is the same as the structure level of \code{x},
the result will be the same as the input.
You can use \code{\link[=get_structure_level]{get_structure_level()}} to check the structure level of \code{x}.}
}
\value{
A \code{\link[=glycan_structure]{glycan_structure()}} vector reduced to the given resolution level.
}
\description{
This function reduces a glycan structure from a higher resolution level to a lower resolution level
(see \code{\link[=get_structure_level]{get_structure_level()}} for four possible levels of resolution).
For example, it can reduce an "intact" structure to a "topological" structure,
or a "partial" structure to a "basic" structure.
One exception is that you can never reduce an "intact" structure to "partial" level,
because the "partial" level is not deterministic.
}
\details{
The logic is as follows:
\itemize{
\item If \code{to_level} is "topological", this function calls \code{\link[=remove_linkages]{remove_linkages()}} to remove all linkages.
\item If \code{to_level} is "basic", this function calls \code{\link[=remove_linkages]{remove_linkages()}} to remove all linkages,
and \code{\link[=convert_to_generic]{convert_to_generic()}} to convert all monosaccharides to generic.
}
}
\examples{
glycan <- as_glycan_structure("Gal(b1-3)GalNAc(a1-")
reduce_structure_level(glycan, to_level = "topological")

}
\seealso{
\code{\link[=get_structure_level]{get_structure_level()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linkage.R
\name{possible_linkages}
\alias{possible_linkages}
\title{Generate Possible Linkages}
\usage{
possible_linkages(
  linkage,
  anomer_range = c("a", "b"),
  pos1_range = 1:2,
  pos2_range = 1:9,
  include_unknown = FALSE
)
}
\arguments{
\item{linkage}{A linkage string.}

\item{anomer_range}{A character vector of possible anomers.
Default is \code{c("a", "b")}.}

\item{pos1_range}{A numeric vector of possible first positions.
Default is \code{1:2}.}

\item{pos2_range}{A numeric vector of possible second positions.
Default is \code{1:9}.}

\item{include_unknown}{A logical value. If \code{TRUE}, "?" will be included.
Default is \code{FALSE}.}
}
\value{
A character vector of possible linkages.
}
\description{
Given an obscure linkage format (having "?", e.g. "a2-?"),
this function generates all possible linkages based on the format.
See \code{\link[=valid_linkages]{valid_linkages()}} for details.

The ranges of possible anomers, first positions, and second positions
can be specified using \code{anomer_range}, \code{pos1_range}, and \code{pos2_range}.
}
\examples{
possible_linkages("a2-?")
possible_linkages("??-2")
possible_linkages("a1-3")
possible_linkages("a?-?", pos1_range = 2, pos2_range = c(2, 3))
possible_linkages("?1-6", include_unknown = TRUE)

}
\seealso{
\code{\link[=has_linkages]{has_linkages()}}, \code{\link[=remove_linkages]{remove_linkages()}}, \code{\link[=valid_linkages]{valid_linkages()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color-utils.R
\name{.extract_base_mono}
\alias{.extract_base_mono}
\title{Extract Base Monosaccharide Name (Without Substituents)}
\usage{
.extract_base_mono(mono)
}
\arguments{
\item{mono}{A monosaccharide name (character), potentially with substituents}
}
\value{
The base monosaccharide name without substituents
}
\description{
Extract Base Monosaccharide Name (Without Substituents)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smap.R
\name{smap_unique}
\alias{smap_unique}
\title{Apply Function to Unique Structures Only}
\usage{
smap_unique(.x, .f, ..., .parallel = FALSE)
}
\arguments{
\item{.x}{A glycan structure vector (glyrepr_structure).}

\item{.f}{A function that takes an igraph object and returns a result.
Can be a function, purrr-style lambda (\code{~ .x$attr}), or a character string naming a function.}

\item{...}{Additional arguments passed to \code{.f}.}

\item{.parallel}{Logical; whether to use parallel processing. If \code{FALSE} (default),
parallel processing is disabled. Set to \code{TRUE} to enable parallel processing.
See examples in \code{\link{smap}} for how to set up and use parallel processing.}
}
\value{
A list with results for each unique structure, named by their hash codes.
}
\description{
Apply a function only to the unique structures in a glycan structure vector,
returning results in the same order as the unique structures appear.
This is useful when you need to perform expensive computations but only
care about unique results.
}
\examples{
# Create a structure vector with duplicates
core1 <- o_glycan_core_1()
structures <- c(core1, core1, core1)  # same structure 3 times

# Only compute once for the unique structure
unique_results <- smap_unique(structures, igraph::vcount)
length(unique_results)  # 1, not 3

# Use purrr-style lambda
unique_results2 <- smap_unique(structures, ~ igraph::vcount(.x))
length(unique_results2)  # 1, not 3

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smap.R
\name{smap2}
\alias{smap2}
\alias{smap2_vec}
\alias{smap2_lgl}
\alias{smap2_int}
\alias{smap2_dbl}
\alias{smap2_chr}
\alias{smap2_structure}
\title{Map Functions Over Two Glycan Structure Vectors}
\usage{
smap2(.x, .y, .f, ..., .parallel = FALSE)

smap2_vec(.x, .y, .f, ..., .ptype = NULL, .parallel = FALSE)

smap2_lgl(.x, .y, .f, ..., .parallel = FALSE)

smap2_int(.x, .y, .f, ..., .parallel = FALSE)

smap2_dbl(.x, .y, .f, ..., .parallel = FALSE)

smap2_chr(.x, .y, .f, ..., .parallel = FALSE)

smap2_structure(.x, .y, .f, ..., .parallel = FALSE)
}
\arguments{
\item{.x}{A glycan structure vector (glyrepr_structure).}

\item{.y}{A vector of the same length as \code{.x}, or length 1 (will be recycled).}

\item{.f}{A function that takes an igraph object (from \code{.x}) and a value (from \code{.y}) and returns a result.
Can be a function, purrr-style lambda (\code{~ .x + .y}), or a character string naming a function.}

\item{...}{Additional arguments passed to \code{.f}.}

\item{.parallel}{Logical; whether to use parallel processing. If \code{FALSE} (default),
parallel processing is disabled. Set to \code{TRUE} to enable parallel processing.
See examples in \code{\link{smap}} for how to set up and use parallel processing.}

\item{.ptype}{A prototype for the return type (for \code{smap2_vec}).}
}
\value{
\itemize{
\item \code{smap2()}: A list
\item \code{smap2_vec()}: An atomic vector of type specified by \code{.ptype}
\item \code{smap2_lgl/int/dbl/chr()}: Atomic vectors of the corresponding type
\item \code{smap2_structure()}: A new glyrepr_structure object
}
}
\description{
These functions apply a function to each unique structure combination in two glycan structure vectors,
taking advantage of hash-based deduplication to avoid redundant computation.
Similar to purrr map2 functions, but optimized for glycan structure vectors.
}
\details{
These functions only compute \code{.f} once for each unique combination of structure and corresponding
\code{.y} value, then map the results back to the original vector positions. This is much more efficient
than applying \code{.f} to each element pair individually when there are duplicate structure-value combinations.

\strong{NA Handling:}
NA elements in \code{.x} are preserved in the output - the function is not applied to NA positions,
and the corresponding results are set to NA.

\strong{Return Types:}
\itemize{
\item \code{smap2()}: Returns a list with the same length as \code{.x}
\item \code{smap2_vec()}: Returns an atomic vector with the same length as \code{.x}
\item \code{smap2_lgl()}: Returns a logical vector
\item \code{smap2_int()}: Returns an integer vector
\item \code{smap2_dbl()}: Returns a double vector
\item \code{smap2_chr()}: Returns a character vector
\item \code{smap2_structure()}: Returns a new glycan structure vector (\code{.f} must return igraph objects)
}
}
\examples{
# Create structure vectors with duplicates
core1 <- o_glycan_core_1()
core2 <- n_glycan_core()
structures <- c(core1, core2, core1)  # core1 appears twice
weights <- c(1.0, 2.0, 1.0)  # corresponding weights

# Map a function that uses both structure and weight
smap2_dbl(structures, weights, function(g, w) igraph::vcount(g) * w)

# Use purrr-style lambda functions
smap2_dbl(structures, weights, ~ igraph::vcount(.x) * .y)

# Test with recycling (single weight for all structures)
smap2_dbl(structures, 2.5, ~ igraph::vcount(.x) * .y)

# Map a function that modifies structure based on second argument
# This example adds a graph attribute instead of modifying topology
add_weight_attr <- function(g, weight) {
  igraph::set_graph_attr(g, "weight", weight)
}
weights_to_add <- c(1.5, 2.5, 1.5)
smap2_structure(structures, weights_to_add, add_weight_attr)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iupac-to-structure.R
\name{.parse_iupac_condensed_single}
\alias{.parse_iupac_condensed_single}
\title{Parse IUPAC-condensed string to glycan structure}
\usage{
.parse_iupac_condensed_single(x)
}
\arguments{
\item{x}{A single IUPAC-condensed string}
}
\value{
An igraph object representing the glycan structure
}
\description{
Internal functions for parsing IUPAC-condensed strings into igraph objects.
This supports the as_glycan_structure.character method.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mono-type.R
\name{get_mono_type}
\alias{get_mono_type}
\alias{get_mono_type.character}
\alias{get_mono_type.glyrepr_structure}
\alias{get_mono_type.glyrepr_composition}
\title{Get Monosaccharide Types}
\usage{
get_mono_type(x)

\method{get_mono_type}{character}(x)

\method{get_mono_type}{glyrepr_structure}(x)

\method{get_mono_type}{glyrepr_composition}(x)
}
\arguments{
\item{x}{Either of these objects:
\itemize{
\item A character vector of monosaccharide names;
\item A glycan composition vector ("glyrepr_composition" object);
\item A glycan structure vector ("glyrepr_structure" object).
}}
}
\value{
\itemize{
\item For character input, returns a character vector of the same length as \code{x}.
\item For \code{glyrepr_structure} and \code{glyrepr_composition} input, returns a character scalar.
}
}
\description{
This function determines the type of monosaccharides in character vectors,
glycan compositions, or glycan structures.
Supported types: "concrete" and "generic" (see details below).
}
\section{Two types of monosaccharides}{
There are two types of monosaccharides:
\itemize{
\item concrete: e.g. "Gal", "GlcNAc", "Glc", "Fuc", etc.
\item generic: e.g. "Hex", "HexNAc", "HexA", "HexN", etc.
}

For the full list of monosaccharides, use \code{\link[=available_monosaccharides]{available_monosaccharides()}}.
}

\section{Special monosaccharides}{
Some monosaccharides are special in that they have no generic names in database or literature.
For example, "Mur" is a rare monosaccharide that has no popular generic name.
In \code{glyrepr}, we assign a "g" prefix to these monosaccharides as their generic names.
This includes "gNeu", "gKdn", "gPse", "gLeg", "gAci", "g4eLeg", "gBac", "gKdo", "gMur".
These names might only be meaningful inside \code{glycoverse}.
Take care when you export results from \code{glycoverse} functions to other analysis tools.
}

\examples{
# Character vector
get_mono_type(c("Gal", "Hex"))

# Glycan structures
get_mono_type(n_glycan_core(mono_type = "concrete"))
get_mono_type(n_glycan_core(mono_type = "generic"))

# Glycan compositions
comp <- glycan_composition(c(Glc = 2, GalNAc = 1))
get_mono_type(comp)

}
\seealso{
\code{\link[=convert_to_generic]{convert_to_generic()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smap.R
\name{smap}
\alias{smap}
\alias{smap_vec}
\alias{smap_lgl}
\alias{smap_int}
\alias{smap_dbl}
\alias{smap_chr}
\alias{smap_structure}
\title{Map Functions Over Glycan Structure Vectors}
\usage{
smap(.x, .f, ..., .parallel = FALSE)

smap_vec(.x, .f, ..., .ptype = NULL, .parallel = FALSE)

smap_lgl(.x, .f, ..., .parallel = FALSE)

smap_int(.x, .f, ..., .parallel = FALSE)

smap_dbl(.x, .f, ..., .parallel = FALSE)

smap_chr(.x, .f, ..., .parallel = FALSE)

smap_structure(.x, .f, ..., .parallel = FALSE)
}
\arguments{
\item{.x}{A glycan structure vector (glyrepr_structure).}

\item{.f}{A function that takes an igraph object and returns a result.
Can be a function, purrr-style lambda (\code{~ .x$attr}), or a character string naming a function.}

\item{...}{Additional arguments passed to \code{.f}.}

\item{.parallel}{Logical; whether to use parallel processing. If \code{FALSE} (default),
parallel processing is disabled. Set to \code{TRUE} to enable parallel processing.}

\item{.ptype}{A prototype for the return type (for \code{smap_vec}).}
}
\value{
\itemize{
\item \code{smap()}: A list
\item \code{smap_vec()}: An atomic vector of type specified by \code{.ptype}
\item \code{smap_lgl/int/dbl/chr()}: Atomic vectors of the corresponding type
\item \code{smap_structure()}: A new glyrepr_structure object
}
}
\description{
These functions apply a function to each unique structure in a glycan structure vector,
taking advantage of hash-based deduplication to avoid redundant computation.
Similar to purrr mapping functions, but optimized for glycan structure vectors.
}
\details{
These functions only compute \code{.f} once for each unique structure, then map
the results back to the original vector positions. This is much more efficient
than applying \code{.f} to each element individually when there are duplicate structures.

\strong{Return Types:}
\itemize{
\item \code{smap()}: Returns a list with the same length as \code{.x}
\item \code{smap_vec()}: Returns an atomic vector with the same length as \code{.x}
\item \code{smap_lgl()}: Returns a logical vector
\item \code{smap_int()}: Returns an integer vector
\item \code{smap_dbl()}: Returns a double vector
\item \code{smap_chr()}: Returns a character vector
\item \code{smap_structure()}: Returns a new glycan structure vector (\code{.f} must return igraph objects)
}
}
\examples{
# Create a structure vector with duplicates
core1 <- o_glycan_core_1()
core2 <- n_glycan_core()
structures <- c(core1, core2, core1)  # core1 appears twice

# Map a function that counts vertices - only computed twice, not three times
smap_int(structures, igraph::vcount)

# Map a function that returns logical
smap_lgl(structures, function(g) igraph::vcount(g) > 5)

# Use purrr-style lambda functions
smap_int(structures, ~ igraph::vcount(.x))
smap_lgl(structures, ~ igraph::vcount(.x) > 5)

# Map a function that modifies structure (must return igraph)
add_vertex_names <- function(g) {
  if (!("name" \%in\% igraph::vertex_attr_names(g))) {
    igraph::set_vertex_attr(g, "name", value = paste0("v", seq_len(igraph::vcount(g))))
  } else {
    g
  }
}
smap_structure(structures, add_vertex_names)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color-utils.R
\name{get_mono_color}
\alias{get_mono_color}
\title{Get Color for Concrete Monosaccharides}
\usage{
get_mono_color(mono)
}
\arguments{
\item{mono}{A monosaccharide name (character), potentially with substituents}
}
\value{
A color code (character)
}
\description{
Get Color for Concrete Monosaccharides
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/composition.R
\name{glycan_composition}
\alias{glycan_composition}
\alias{is_glycan_composition}
\title{Create a Glycan Composition}
\usage{
glycan_composition(...)

is_glycan_composition(x)
}
\arguments{
\item{...}{Named integer vectors.
Names are monosaccharides or substituents, values are numbers of residues.
Monosaccharides and substituents can be mixed in the same composition.}

\item{x}{A list of named integer vectors.}
}
\value{
A glyrepr_composition object.
}
\description{
Create a glycan composition from a list of named integer vectors.
Compositions can contain both monosaccharides and substituents.
}
\details{
Compositions can contain:
\itemize{
\item Monosaccharides: either generic (e.g., "Hex", "HexNAc") or concrete (e.g., "Glc", "Gal").
All monosaccharides in a composition vector must be of the same type.
\item Substituents: e.g., "Me", "Ac", "S". These can be mixed with either
generic or concrete monosaccharides.
}

Components are automatically sorted with monosaccharides first (according to
their order in the monosaccharides table), followed by substituents (according
to their order in \code{available_substituents()}). Duplicate components are
automatically summed.
}
\examples{
# A vector with one composition (generic monosaccharides)
glycan_composition(c(Hex = 5, HexNAc = 2))
# A vector with multiple compositions
glycan_composition(c(Hex = 5, HexNAc = 2), c(Hex = 5, HexNAc = 4, dHex = 2))
# Residues are reordered automatically
glycan_composition(c(HexNAc = 1, Hex = 2))
# An example for generic monosaccharides
glycan_composition(c(Hex = 2, HexNAc = 1))
# An example for concrete monosaccharides
glycan_composition(c(Glc = 2, Gal = 1))
# Compositions with substituents
glycan_composition(c(Glc = 1, S = 1))
glycan_composition(c(Hex = 3, HexNAc = 2, Me = 1, Ac = 1))
# Substituents are sorted after monosaccharides
glycan_composition(c(S = 1, Gal = 1, Ac = 1, Glc = 1))

}
\seealso{
\code{available_monosaccharides()}, \code{available_substituents()}
}
